
// ----------------------------------------------------------------------

void adjustField(int zFieldTgt) {
  // adjusts the slider field value for zFld
  FieldTgt = zFieldTgt;
  int zVal;  // limits
  if (sliderDwn < 1) {
    // just been clicked so get the current value
    switch(FieldTgt) {
      case FldT1maxLft: FieldVal = T1maxAngLft; FldLL = 0; FldUL = 160; break;
      case FldT1maxRht: FieldVal = T1maxAngRht; FldLL = 0; FldUL = 160; break;
      case FldT2maxLft: FieldVal = T2maxAngLft; FldLL = 10; FldUL = 140; break;
      case FldT2minLft: FieldVal = T2minAngLft; FldLL = 0; FldUL = 130; break;
      case FldT2maxRht: FieldVal = T2maxAngRht; FldLL = 10; FldUL = 140; break;
      case FldT2minRht: FieldVal = T2minAngRht; FldLL = 0; FldUL = 130; break;
      case FldT1WlkLft: FieldVal = WalkLift; FldLL = 10; FldUL = 60; break;
    }
  }
  // calculate new value and limit
  zVal = FieldVal + (mX - mXD)/3;
  if (zVal > FldUL) {zVal = FldUL;}
  if (zVal < FldLL) {zVal = FldLL;}
  switch(FieldTgt) {
    case FldT1maxLft: T1maxAngLft = zVal; calcRideHeights(); break;
    case FldT1maxRht: T1maxAngRht = zVal; calcRideHeights(); break;
    case FldT2maxLft: T2maxAngLft = zVal; calcMidSteps(); break;
    case FldT2minLft: T2minAngLft = zVal; calcMidSteps(); break;
    case FldT2maxRht: T2maxAngRht = zVal; calcMidSteps(); break;
    case FldT2minRht: T2minAngRht = zVal; calcMidSteps(); break;
    case FldT1WlkLft: WalkLift = zVal; break;
  } 
  checkDepFields();
  rateFlag = true; drawFlag = 1;
  sliderDwn = 1; // set slider down flag to prevent inc/dec from false triggering
}

// ----------------------------------------------------------------------

void adjustLinkVal(int zT) {
  // called from the linking process, adjust an angle by AngDif
  AngVal[zT] = getAngVal(AngVal[zT]) + AngDif;
  if (AngVal[zT] < AngVLL[zT]) {
    AngVal[zT] = AngVLL[zT];
  }
  else if (AngVal[zT] > AngVUL[zT]) {
    AngVal[zT] = AngVUL[zT];
  }
  ArrayVal[(rowPnt * rowWidth) + zT] = AngVal[zT];
  setSliderX();

  // send value to usb port
  msgTx += "ST" + str(Ch[zT]) + "." +"SA" + str(AngVal[zT]) + ".";
}


// ######################################################################


int Calc_AngLA(int zAng2,float zR_RH) {
  // calculates intermediate tier 1 angles from the parameters passed
  float zUB = (float)zAng2;
  float zUH = 75.22 * sin(zUB/57.2958);
  float zLHT = zR_RH - zUH;
  //print("zUH "); print(zUH); print("\tzLHT "); println(zLHT);
  float zLB = 90.0 - degrees(asin(zLHT/103.5));
  //print("zUB "); print(zUB); print("\tzLB "); println(zLB);
  return (int)(90.0 + zUB - zLB);
}

// ----------------------------------------------------------------------

void calcMidSteps() {
  // simulator angles have been adjusted so recalculate mid points
  T2midAngLft = T2minAngLft + ((T2maxAngLft - T2minAngLft)/2);
  T2midAngRht = T2minAngRht + ((T2maxAngRht - T2minAngRht)/2);
  CalcNomStepsLft = T2maxAngLft - T2minAngLft + 1;
  CalcNomStepsRht = T2maxAngRht - T2minAngRht + 1;
}

// ----------------------------------------------------------------------

void calcRates() {
  // work through the list and calculate rate angle values
  // we determine the start angle, end angle and number of intermediate steps to do this
  HelpOFF++;  // turn off help system if not already off
  //println("calcRates()...");
  //t0 = millis();  // used to time function duration
  rateFlag = false;
  int zFA = 0;  // finish angle
  int zSA = 0;  // start angle
  int zSR = 0;  // start angle row number
  for (int zC = 0; zC < (rowWidth - 1); zC++) {
    // we work down each column, one angle at a time
    // only the first column can contain commnd lines
    int zI = 0;  // counts the number of steps associated with the rate angles and the end fixed angle
    int zIT = 0; // temporary step value
    int zS = 0;  // state machine pointer
    for (int zP = 0; zP < rowNum; zP++) {
      int zR = zP * rowWidth;  // initialise the row pointer for each column
      //if ((zC > 0) || (ArrayVal[zR + zC] >= 0)) {
      if (ArrayVal[zR] >= 0) {
        // the row is not a command line so use the data
        //println("col: " + str(zC) + "  row: " + str(zP) + "  zS=" + str(zS) + "  Ang: " + str(ArrayVal[zR + zC]));
        switch (zS) {
          case 0:  // looking for first angle(s)
            // in this state if we find a rate angle before a valid angle we clear the rate flag
            if (ArrayVal[zR + zC] >= 400) {
              // this is the error condition, this rate angle does not have a fixed angle before it
              // so we remove the rate flag and set it as a fixed angle
              ArrayVal[zR + zC] -= 400;  // remove the rate flag from this angle
            } 
            zSA = getAngVal(ArrayVal[zR + zC]);  // store the fixed angle as the start angle
            zS++; break;  // move onto the next task for the next angle
          case 1:  // tracking fixed angles until we find a rate angle
            if (ArrayVal[zR + zC] >= 400) {
              // this is the first rate angle, so we now change the task to look for the next
              // fixed angle before calculating the rate(s)
              zSR = zP;  // store the row reference for this angle
              zIT = ArrayVal[zR + rowWidth -1];  // collect the number of steps
              while (zIT > 200) {zIT -= 200;}  // strip off any rate profile
              zI = zIT; zS++;
            } else {
              // this is a fixed angle so record it as a valid start angle
              zSA = getAngVal(ArrayVal[zR + zC]);  // store the fixed angle as the start angle
            } break;
          case 2:  // look for a fixed angle to set a target range for the rate angles
            if (ArrayVal[zR + zC] >= 400) {
              // we've encountered another rate angle so collect the number of steps
              zIT = ArrayVal[zR + rowWidth -1]; // add on the number of steps
              while (zIT > 200) {zIT -= 200;}    // strip off any rate profile
              zI += zIT;
            } else {
              // this is the terminating angle we are looking for so we can no calculate the rate angles
              zIT = ArrayVal[zR + rowWidth -1];  // add on the number of steps to get the total
              while (zIT > 200) {zIT -= 200;}    // strip off any rate profile
              zI += zIT;
              // now recalculate each rate angle based on its associate number of steps
              int zIC = 0;  // initiatlise the cumulative step value
              zFA = getAngVal(ArrayVal[zR + zC]);  // note finish angle
              
              // now set the rates, within the range, ignoring Cmd lines
              for (int zN = zSR; zN < zP; zN++) {
                //print("Adjusting rate in col: " + str(zC) + "  row: " + str(zN));
                int zRN = zN * rowWidth;
                if (ArrayVal[zRN] >= 0) {
                  // this is not a command line
                  zIT =  ArrayVal[zRN + rowWidth -1]; // get the cumulative steps
                  while (zIT > 200) {zIT -= 200;}    // strip off any rate profile
                  zIC += zIT;
                  int zAng = zSA + ((zIC * (zFA - zSA))/zI);
                  //println("  to: " + zAng);
                  //println("zSA: " + str(zSA) + 
                  //println("zIC: " + str(zIC) + "  zI: " + str(zI));
                  ArrayVal[zRN + zC] = zAng + 400;
                }
              }
              // now restart the tasks from this fixed angle
              zI = 0;    // reset the number of steps associated with the rate angles and the end fixed angle
              zSA = zFA; // store the fixed finish angle as the new start angle
              zSR = zP;  // store the row reference for this angle
              zS = 1;    // reset the task pointer to look for more rate angles
            }
            break;
        }
      }
    }
  }
  //println("calcRates() completed!");
  //println("t1-t0 = " + str(millis() - t0));
  HelpOFF--;  // turn on help system if held off by this function
}

// ----------------------------------------------------------------------

void calcRideHeights() {
  // calculate the ride heights due to a T1 max angle change
  // initial calcs
  float zR_UC = T2maxAngRht - 90.0;           // forward angles
  float zL_UC = T2maxAngLft - 90.0;
  float zR_LB = 180.0 - T1maxAngRht + zR_UC;  // T1 vertical angle
  float zL_LB = 180.0 - T1maxAngLft + zL_UC;
  // conversion factors:
  //   90° == 1.57079557 rads
  // 1 rad == 57.2958°
  //float zR_UH = 75.22 * sin(1.57079557 - (zR_UC/57.2958));  // height contribution of upper leg
  float zR_UH = 75.22 * sin(radians(90.0 - zR_UC));  // height contribution of upper leg
  float zL_UH = 75.22 * sin(radians(90.0 - zL_UC));
  float zR_LH = 103.5 * sin(radians(90.0 - zR_LB));  // height contribution of lower leg
  float zL_LH = 103.5 * sin(radians(90.0 - zL_LB));
  float zR_RH = zR_UH + zR_LH;  // combined 'ride' height
  float zL_RH = zL_UH + zL_LH;
  CalcMaxHgtRht = (int)zR_RH;  // display the ride height
  CalcMaxHgtLft = (int)zL_RH;
}

// ----------------------------------------------------------------------

void calcSelSteps() {
  // calculate steps for selected rows using difference with previous
  // row, providing it is not a command line
  if (SelCnt < 1) {return;}
   drawFlag = 1;
  if (rowNum < 2) {msgTx = "Error - < 2 lines"; return;}
  
  int zA = 0;       // any Temp[] angle
  int zB = 0;       // any row angle
  int zD = 0;       // differential angle range
  int zM;           // flag recording rate profile for this line
  int zN = 0;       // number of lines set/modified
  boolean zNS = false;  // == true if a new step value is determined
  int zP = 0;       // servo angle pointer
  int zRP;          // array memory pointer
  int zSteps =  1;  // min num steps
  int zT = 0;       // flag > 0 if Temp[] holds a reference line

  for (int zR = 0; zR < rowNum; zR++) {
    zRP = zR * rowWidth;        // current row data pointer
    if (!Select[zR]) {
      // this line is not selected so store it as a previous reference if not a command line
      if (ArrayVal[zRP] >= 0) {grabRowInTemp(zR); zT = 1;}
    } else {
      if (ArrayVal[zRP] >= 0) {
        // not a command row so perform the analysis
        if (zT < 1) {
          // either the first row or the first row with angle data, so use it as reference
          grabRowInTemp(zR); zT = 1;
        } else {
          // this row is selected so, and we have a reference line, so do the calculation
          // first grab the rate profile, to re-apply later
          zSteps = ArrayVal[zRP + rowWidth - 1];
          zM = 0;
          if (zSteps > 600) {zM = 3;}
          else if (zSteps > 400) {zM = 2;}
          else if (zSteps > 200) {zM = 1;}
          zSteps =  1;  // reset the temp step value for each row
          zNS = false;  // clear the new step flag
          for (zP = 0; zP < (rowWidth - 1); zP++) {
            // compare each angle with the Temp[] reference row, ignoring rate angles
            zA = getAngVal(Temp[zP]);  // strip off rate and Pwr flags
            if (!AngRate) {
              // the Temp[] reference is not a rate angle
              zB = getAngVal(ArrayVal[zP + zRP]);
              if (!AngRate) {
                // the current angle is not a rate angle too, so do the comparison
                zD = abs(zB - zA);
                zD /= DegStep;  // divide the angle range by DegStep
                if (zD > zSteps) {zSteps = zD; zNS = true;}
              }
            }
          }
          if (keySHIFT) {zSteps = 30;}  // using default value over-ride
          if (keySHIFT || zNS) {
            if (zM == 3) {zSteps += 600;}
            else if (zM == 2) {zSteps += 400;}
            else if (zM == 1) {zSteps += 200;}
            ArrayVal[zRP + rowWidth - 1] = zSteps; zN++;
          }
          // now grab this modified row as the reference for the next selected row
          grabRowInTemp(zR);
        }
      }
    }
  } 
  if (zT < 1) {msgTx = "Error - no ref line"; return;}
  if (zN < 1) {msgTx = "Error - no lines set"; return;}

  msgTx = str(zN) + " line steps set";
  rateFlag = true;
}

// ----------------------------------------------------------------------

void Calc_Walk() {
  // calculates a walk sequence based on the Scrn1 settings
  WalkSim = 0;  // ensure simulator is stopped
  cursor(WAIT); ClearArray(0); ClearSelect(); mDwn = 0;
  int zFR1,zFR2,zFL1,zFL2,zRR1,zRR2,zRL1,zRL2;
  int zRP = 0;  // rowPnt
  
  // calculate nominal number of steps, but Steps is actually used
  CalcNomStepsLft = T2maxAngLft - T2minAngLft + 1;
  CalcNomStepsRht = T2maxAngRht - T2minAngRht + 1;
  
  // initial calcs
  float zR_UC = T2maxAngRht - 90.0;           // forward angles
  float zL_UC = T2maxAngLft - 90.0;
  float zR_LB = 180.0 - T1maxAngRht + zR_UC;  // T1 vertical angle
  float zL_LB = 180.0 - T1maxAngLft + zL_UC;
  // conversion factors:
  //   90° == 1.57079557 rads
  // 1 rad == 57.2958°
  //float zR_UH = 75.22 * sin(1.57079557 - (zR_UC/57.2958));  // height contribution of upper leg
  float zR_UH = 75.22 * sin(radians(90.0 - zR_UC));  // height contribution of upper leg
  float zL_UH = 75.22 * sin(radians(90.0 - zL_UC));
  float zR_LH = 103.5 * sin(radians(90.0 - zR_LB));  // height contribution of lower leg
  float zL_LH = 103.5 * sin(radians(90.0 - zL_LB));
  float zR_RH = zR_UH + zR_LH;  // combined 'ride' height
  float zL_RH = zL_UH + zL_LH;
  CalcMaxHgtRht = (int)zR_RH;  // display the ride height
  CalcMaxHgtLft = (int)zL_RH;
  if (WalkHeight > CalcMaxHgtRht) {WalkHeight = CalcMaxHgtRht;}
  if (WalkHeight > CalcMaxHgtLft) {WalkHeight = CalcMaxHgtLft;}
  
  // create the move to ready line
  int zMR = T2minAngRht + ((T2maxAngRht - T2minAngRht)/2);  // calc tier 2 mid-angle for ready position
  zFR2 = zMR; zRR2 = zMR;
  int zML = T2minAngLft + ((T2maxAngLft - T2minAngLft)/2);  // calc tier 2 mid-angle for ready position
  zFL2 = zML; zRL2 = zML;
  zFR1 = Calc_AngLA(zFR2,zR_RH); zRR1 = zFR1;
  zFL1 = Calc_AngLA(zFL2,zL_RH); zRL1 = zFL1;
  //         Row,zFR1,zFR2,zFR3,zFL1,zFL2,zFL3,zRR1,zRR2,zRR3,zRL1,zRL2,zRL3,zStp
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,  30);  // overwrite first row
  
  // Phase A - create the left up for fly step
  // zFR2, zFL2, zRR2 & zRL2 do not move in this first step
  zFL1 -= WalkLift;    // zFL2 lifts ready to fly forward
  zRR1 -= WalkLift;    // zRR2 lifts ready to fly forward
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  
  // Phase B - create the left up fly step
  zFR2 --; zFR1 = Calc_AngLA(zFR2,zR_RH);                  // FR leg pulling back
  zFL2 = T2maxAngLft; zFL1 = Calc_AngLA(zFL2,zL_RH) - WalkLift;  // FL fly to angles
  zRR2 = T2maxAngRht; zRR1 = Calc_AngLA(zRR2,zR_RH) - WalkLift;  // RR fly to angles
  zRL2 --; zRL1 = Calc_AngLA(zRL2,zL_RH);                  // RL leg pulling back
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row

  // Phase C - create the left landing step
  zFR2 --; zFR1 = Calc_AngLA(zFR2,zR_RH);            // FR leg pulling back
  zFL2 = T2maxAngLft; zFL1 = Calc_AngLA(zFL2,zL_RH); // FL landing at max fwd angle
  zRR2 = T2maxAngRht; zRR1 = Calc_AngLA(zRR2,zR_RH); // FL landing at max fwd angle
  zRL2 --; zRR1 = Calc_AngLA(zRR2,zL_RH);            // RL leg pulling back
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  
  // Phase D - create the 1st phase of pull back
  while (zFR2 >= T2minAngRht) {
    zFR2 --; zFR1 = Calc_AngLA(zFR2,zR_RH);   // FR leg pulling back
    zFL2 --; zFL1 = Calc_AngLA(zFL2,zL_RH);   // FL leg pulling back
    zRR2 --; zRR1 = Calc_AngLA(zRR2,zR_RH);   // RR leg pulling back
    zRL2 --; zRL1 = Calc_AngLA(zRL2,zL_RH);   // RL leg pulling back
    zRP++;
    setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  }
  
  // Phase E - create the right up fly step
  zFR1 -= WalkLift;                                // FR fly to angles
  zFL2 --; zFL1 = Calc_AngLA(zFL2,zL_RH);    // FL leg pulling back
  zRR2 --; zRR1 = Calc_AngLA(zRR2,zR_RH);    // RR leg pulling back
  zRL1 -= WalkLift;                                // RL fly to angles
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  
  // Phase F1 - create intermediate point
  zFR2 = zMR; zFR1 = Calc_AngLA(zFR2,zR_RH) - WalkLift;  // FR mid fly to angles
  zFL2 --; zFL1 = Calc_AngLA(zFL2,zL_RH);          // FL leg pulling back
  zRR2 --; zRR1 = Calc_AngLA(zRR2,zR_RH);          // RR leg pulling back
  zRL2 = zMR; zRL1 = Calc_AngLA(zRL2,zL_RH) - WalkLift;  // RL mid fly to angles
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row

  // Phase F2 - create max reach step
  zFR2 = T2maxAngRht; zFR1 = Calc_AngLA(zFR2,zR_RH) - WalkLift;  // FR max fwd angle
  zFL2 --; zFL1 = Calc_AngLA(zFL2,zL_RH);                  // FL leg pulling back
  zRR2 --; zRR1 = Calc_AngLA(zRR2,zR_RH);                  // RR leg pulling back
  zRL2 = T2maxAngLft; zRL1 = Calc_AngLA(zRL2,zL_RH) - WalkLift;  // RL max fwd angle
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row

  // Phase G - create the right landing step
  zFR2 = T2maxAngRht; zFR1 = Calc_AngLA(zFR2,zR_RH); // FR landing at max fwd angle
  zFL2 --; zFL1 = Calc_AngLA(zFL2,zL_RH);            // FL leg pulling back
  zRR2 --; zRR1 = Calc_AngLA(zRR2,zR_RH);            // RR leg pulling back
  zRL2 = T2maxAngLft; zRL1 = Calc_AngLA(zRL2,zL_RH); // RL landing at max fwd angle
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  
  // Phase H - create the 2nd phase of pull back
  while (zFR2 >= (zMR - 2)) {
    zFR2 --; zFR1 = Calc_AngLA(zFR2,zR_RH);   // FR leg pulling back
    zFL2 --; zFL1 = Calc_AngLA(zFL2,zL_RH);   // FL leg pulling back
    zRR2 --; zRR1 = Calc_AngLA(zRR2,zR_RH);   // RR leg pulling back
    zRL2 --; zRL1 = Calc_AngLA(zRL2,zL_RH);   // RL leg pulling back
    zRP++;
    setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  }
  
  // Phase J - create the left up fly step
  zFR2 --; zFR1 = Calc_AngLA(zFR2,zR_RH);    // FR leg pulling back
  zFL1 -= WalkLift;                                // FL fly to angles
  zRR1 -= WalkLift;                                // RR fly to angles
  zRL2 --; zRL1 = Calc_AngLA(zRL2,zL_RH);    // RL leg pulling back
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  
  // Phase K - create intermediate point
  zFR2 --; zFR1 = Calc_AngLA(zFR2,zR_RH);          // FR leg pulling back
  zFL2 = zML; zFL1 = Calc_AngLA(zFL2,zL_RH) - WalkLift;  // FL mid fly to angles
  zRR2 = zMR; zRR1 = Calc_AngLA(zRR2,zR_RH) - WalkLift;  // RL mid fly to angles
  zRL2 --; zRL1 = Calc_AngLA(zRL2,zL_RH);          // RR leg pulling back
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  
  // now branch to the second row
  zRP++;
  setRowData(zRP,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0); rowNum++;  // add new row
  
  setAngValFromRow();  // set servo values and sliders
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void changeAllSelVal(int zCol,int zDif) {
  // changes all associated selected row column values by zDif
  int zA = 0;   // any angle
  int zC0 = 0;  // any colum ref
  int zC1 = 0;  // any colum ref
  int zC2 = 0;  // any colum ref
  int zC3 = 0;  // any colum ref
  for (int zRp = 0; zRp < rowNum; zRp++) {
    if (Select[zRp]) {
      // this is a selected row so check for command line
      int zMp = zRp * rowWidth;
      if (ArrayVal[zMp] >= 0) {
        // not a command line so modify column values
        switch (zCol) {
          case  0: zC0 = 0; zC1 = 3; zC2 = 6; zC3 =  9; break;  // FR1
          case  1: zC0 = 1; zC1 = 4; zC2 = 7; zC3 = 10; break;  // FR2
          case  2: zC0 = 2; zC1 = 5; zC2 = 8; zC3 = 11; break;  // FR3
          case  3: zC0 = 0; zC1 = 3; zC2 = 6; zC3 =  9; break;  // FL1
          case  4: zC0 = 1; zC1 = 4; zC2 = 7; zC3 = 10; break;  // FL2
          case  5: zC0 = 2; zC1 = 5; zC2 = 8; zC3 = 11; break;  // FL3
          case  6: zC0 = 0; zC1 = 3; zC2 = 6; zC3 =  9; break;  // RR1
          case  7: zC0 = 1; zC1 = 4; zC2 = 7; zC3 = 10; break;  // RR2
          case  8: zC0 = 2; zC1 = 5; zC2 = 8; zC3 = 11; break;  // RR3
          case  9: zC0 = 0; zC1 = 3; zC2 = 6; zC3 =  9; break;  // RL1
          case 10: zC0 = 1; zC1 = 4; zC2 = 7; zC3 = 10; break;  // RL2
          case 11: zC0 = 2; zC1 = 5; zC2 = 8; zC3 = 11; break;  // RL3
        }
        zA = getAngVal(ArrayVal[zMp + zC0]);
        if (zA < AngVUL[zC0]) {ArrayVal[zMp + zC0] += zDif;}
        zA = getAngVal(ArrayVal[zMp + zC1]);
        if (zA < AngVUL[zC1]) {ArrayVal[zMp + zC1] += zDif;}
        zA = getAngVal(ArrayVal[zMp + zC2]);
        if (zA < AngVUL[zC2]) {ArrayVal[zMp + zC2] += zDif;}
        zA = getAngVal(ArrayVal[zMp + zC3]);
        if (zA < AngVUL[zC3]) {ArrayVal[zMp + zC3] += zDif;}
      }
    }
  }
}

// ----------------------------------------------------------------------

void changeAngleVal() {
  // user has clicked on an angle value in the list window so change it
  
  int zR = getRowMYD();  // row pointer +/-, == 0 if current row
  int zRP = rowPnt + zR;
  if ((zRP < 0) || (zRP >= rowNum)) {return;}  // row is before/after list
  if (ArrayVal[zRP * rowWidth] < 0) {changeCmdVal(); return;}  // command line so exit
  
  int zA = getAngleMXD();  // array column pointer based on mouse position in list
  int zAV = zA + (zRP * rowWidth);  // pointer to the angle in ArrayVal[]
  boolean zC = false;  // change flag, == true if a change has occured
  if (keySHIFT) {
    // special 'quick' copy/paste function
    if (mB == LEFT) {
      // paste quick angle value, which has previously been set
      // this angle will not have any Pwr or Rate attributes
      if (QuickAng >= 0) {
        if ((QuickAng >= AngVLL[zA]) && (QuickAng <= AngVUL[zA])) {
          ArrayVal[zAV] = QuickAng; zC = true; rateFlag = true;
        }
      }
    } else {
      // record 'quick' angle value
      // only record angle, not Pwr or Rate attributes
      QuickAng = getAngVal(ArrayVal[zAV]);
      msgTx = "QuickAng = " + str(QuickAng);
    }
  } else {
    // normal increment/decrement process
    // the angle may be 'rated' if so this will remove the rate flag
    // the ON/OFF power factor will not be affected
    if (mB == LEFT) {
      // increment angle value
      AngLast = getAngVal(ArrayVal[zAV]);
      if (AngLast < AngVUL[zA]) {
        if (Select[zRP]) {
          // this row is selected so change all column values by the same
          // amount as if they were all linked together
          changeAllSelVal(zA,1);
          if (Select[rowPnt]) {zR = 0;}  // current row is affected, so send an update
        } else {
          // normal row
          ArrayVal[zAV] = 1 + AngLast;
        }
        zC = true; AngDif = 1;
      }
    } else {
      // decrement angle value
      AngLast = getAngVal(ArrayVal[zAV]);
      if (AngLast > AngVLL[zA]) {
        if (Select[zRP]) {
          // this row is selected so change all column values by the same
          // amount as if they were all linked together
          changeAllSelVal(zA,-1);
          if (Select[rowPnt]) {zR = 0;}  // current row is affected, so send an update
        } else {
          // normal row
          ArrayVal[zAV] = AngLast - 1;
        }
        zC = true; AngDif = 1;
      }
    }
    if (!AngPwr) {ArrayVal[zAV] += 200;}  // assign the power setting
  }
  
  if (zC) {
    if (zR == 0) {
      // current row changed
      setAngValFromRow();     // change sliders
      msgTx = "ST" + str(Ch[zA]) + "." +"SA" + str(AngVal[zA]) + ".";
      if (Select[zRP]) {setSliderX(); sendRowData(rowPnt,rowPnt,1);}
      else if (Link[zA] > 0) {moveLinkedVal(zA);}
      else {setSliderX(); usbPortWrite(msgTx);}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
    }
    rateFlag = true; drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

void changeCmdType() {
  // it is possible to change the command 'Type' for executeable commands
  int zR = getRowMYD();                // row pointer +/-, == 0 if current row
  int zCP = (rowPnt + zR) * rowWidth;  // memory pointer
  int zXM = 7;                         // max Exe code value
  CmdCode = ArrayVal[zCP + 1];         // get command type code
  if (CmdCode < 6) {return;}           // not an executaable code
  
  if (mB == LEFT) {
    CmdCode++; if (CmdCode > zXM) {CmdCode = 6;}
  } else {
    CmdCode--; if (CmdCode < 6) {CmdCode = zXM;}
  }
  ArrayVal[zCP + 1] = CmdCode;
  // now set defaults for new Cmd line
  ArrayVal[zCP + 2] = 0; ArrayVal[zCP + 3] = 0; ArrayVal[zCP + 4] = 0; ArrayVal[zCP + 5] = 0; ArrayVal[zCP + 6] = 0;  // clear values
  switch (CmdCode) {
    case 6:  // For... next loop
      ArrayVal[zCP + 2] = 1; break;
    case 7:  // ...Next in For... loop
      break;
  }
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void changeCmdVal() {
  // user has clicked in list line region on a command line
  int zR = getRowMYD();  // row pointer +/-, == 0 if current row
  int zCP = (rowPnt + zR) * rowWidth;
  int zC2 = 0;  // changed 1st value flag
  AngField = getAngleMXD();
  CmdCode = ArrayVal[zCP + 1] ;  // get command type code
  CmdVal0 = ArrayVal[zCP + 2] ;  // get command 1st value
  if (CmdCode == 0) {return;}    // blank line
  else if (CmdCode == 1) {
     // Goto command
     if (mB == LEFT) {if (CmdVal0 < (rowNum -1)) {CmdVal0++; zC2++;}}
     else {if (CmdVal0 > 0) {CmdVal0--; zC2++;}}
  }
  else if (CmdCode == 2) {
     // Speed command
     if (mB == LEFT) {if (CmdVal0 < 250) {CmdVal0++; zC2++;}}
     else {if (CmdVal0 > 1) {CmdVal0--; zC2++;}}
  }
  else if (CmdCode == 3) {
     // Pause command
     if (mB == LEFT) {
       if (CmdVal0 < 1000) {
         CmdVal0+= Inc; zC2++; IncCnt++;
         if (IncCnt > 10) {IncCnt = 0; Inc += 5;}
         if (CmdVal0 > 1000) {CmdVal0 = 1000;}  // don't overshoot
       }
     } else {
       if (CmdVal0 > 1) {
         CmdVal0-= Inc; zC2++; IncCnt++;
         if (IncCnt > 10) {IncCnt = 0; Inc += 5;}
         if (CmdVal0 < 1) {CmdVal0 = 1;}  // don't undershoot
       }
     }
     Pause = CmdVal0;  // set the default value to match this
  }
  else if (CmdCode == 4) {
     // Flag command
     if (mB == LEFT) {if (CmdVal0 < 180) {CmdVal0++; zC2++;}}
     else {if (CmdVal0 > 1) {CmdVal0--; zC2++;}}
     Flag = CmdVal0;  // set the default value to match this
  }
  else if (CmdCode == 5) {
     // Balance command
          if ((mB == LEFT) && (CmdVal0 < 3)) {CmdVal0++; zC2++;}
     else if ((mB == RIGHT) && (CmdVal0 > 0)) {CmdVal0--; zC2++;}
  }
  else if (CmdCode == 6) {
     // For... before Next command
          if (AngField == 1) {changeCmdType();}
     else if (AngField >= 2) {
       if (mB == LEFT) {if (CmdVal0 < 1000) {CmdVal0++; zC2++;}}
       else {if (CmdVal0 > 1) {CmdVal0--; zC2++;}}
     } AngField = -1;
  }
  else if (CmdCode == 7) {
     // Next in ...For loop command
     if (AngField == 1) {changeCmdType();} 
  }

  if (zC2 > 0) {ArrayVal[zCP + 2] = CmdVal0; drawFlag = 1;}
}

// ----------------------------------------------------------------------

void changeSteps() {
  // user clicked on steps field so inc/dec value
  // check for command row
  boolean zC = false;  // something changed flag
  int zSteps =  1;
  int zD = 0;
  int zP = 0;
  int zM = 0;          // rate mode flag
  int zR = getRowMYD();
  zR = rowPnt + zR;
  if ((zR < 0) || (zR >= rowNum)) {return;}  // row is before list
  if (ArrayVal[zR * rowWidth] < 0) {return;}  // command line so exit
  
  // strip off step profile zM, to add it later
  zSteps =  ArrayVal[(zR * rowWidth) + 12];
       if (zSteps > 600) {zSteps -= 600; zM = 3;}
  else if (zSteps > 400) {zSteps -= 400; zM = 2;}
  else if (zSteps > 200) {zSteps -= 200; zM = 1;}
  if (keySHIFT) {
    // special 'quick' copy/paste function
    if (mB == LEFT) {
      // paste quick step value, which has previously been set
      // this angle will not have any Pwr or Rate attributes
      if (QuickStep >= 0) {
        zSteps = QuickStep; zC = true;
             if (zM > 2) {zSteps += 600;}
        else if (zM > 1) {zSteps += 400;}
        else if (zM > 0) {zSteps += 200;}
        ArrayVal[(zR * rowWidth) + 12] = zSteps;
      }
    } else {
      // record 'quick' angle value
      // only record angle, not Pwr or Rate attributes
      QuickStep = zSteps;
      msgTx = "QuickStep = " + str(QuickStep); drawFlag = 1;
    }
  } else {
    // normal inc/dec process
    if (mB == LEFT) {
      // left mouse button increments the value
      if (zSteps < 180) { // max steps value is 180
        zSteps+= Inc; zC = true; IncCnt++;
        if (IncCnt > 10) {IncCnt = 0; Inc++;}
        if (zSteps > 180) {zSteps = 180;}  // don't overshoot
      }
    } else {
      // right mouse button decrements the value
      if (zSteps > 1) { // min steps value is 1
        zSteps-= Inc; zC = true; IncCnt++;
        if (IncCnt > 10) {IncCnt = 0; Inc++;}
        if (zSteps < 1) {zSteps = 1;}  // don't overshoot
      }
    }
    if (zM > 2) {zSteps += 600;}
    else if (zM > 1) {zSteps += 400;}
    else if (zM > 0) {zSteps += 200;}
    ArrayVal[(zR * rowWidth) + 12] = zSteps;
  }
  if (zC) {rateFlag = true; drawFlag = 1;}
}

// ----------------------------------------------------------------------

void checkDepFields() {
  // simulator field values have changed, ensure that they don't overlap
  switch(FieldTgt) {
    case FldT2maxLft:
      if (T2maxAngLft < (T2minAngLft + 10)) {T2minAngLft = T2maxAngLft - 10;}
      break;
    case FldT2minLft:
      if (T2minAngLft > (T2maxAngLft - 10)) {T2maxAngLft = T2minAngLft + 10;}
      break;
    case FldT2maxRht:
      if (T2maxAngRht < (T2minAngRht + 10)) {T2minAngRht = T2maxAngRht - 10;}
      break;
    case FldT2minRht:
      if (T2minAngRht > (T2maxAngRht - 10)) {T2maxAngRht = T2minAngRht + 10;}
      break;
  }
}

// ----------------------------------------------------------------------

void checkGotoCmdDel(int zD,int zCnt) {
  // called when row zD has zCnt deleted to correct any goto commands
  int zP;
  for (int zR = 0; zR < rowNum; zR++) {
    // look for goto commands
    zP = zR * rowWidth;
    if (ArrayVal[zP] < 0) {
      // this is a commands line, but is it a goto?
      if (ArrayVal[zP + 1] == 1) {
        // this is a Goto command so check that zD is not less than branched row
        if (ArrayVal[zP + 2] > zD) {
          ArrayVal[zP + 2]-= zCnt;
          println("Goto decremented");
        }
      }
    }
  }
}

// ----------------------------------------------------------------------

void checkGotoCmdIns(int zI,int zCnt) {
  // called when row zI has zCnt inserted so correct any goto commands
  int zP;
  for (int zR = 0; zR < rowNum; zR++) {
    // look for goto commands
    zP = zR * rowWidth;
    if (ArrayVal[zP] < 0) {
      // this is a command line, but is it a goto?
      if (ArrayVal[zP + 1] == 1) {
        // this is a Goto command so check that zI is not less than branched row
        if (ArrayVal[zP + 2] >= zI) {
          ArrayVal[zP + 2]+= zCnt;
          println("Goto incremented");
        }
      }
    }
  }
}

// ----------------------------------------------------------------------

void ClearArray(int zM) {
  // clear the values in an array to 90 degrees
  // zM = 0  just reset the array data only
  // zM = 1  update angles and send instruction over usb port
  cursor(WAIT);
  int zS = 0; int zW = rowWidth;
  for (int zC = 0; zC < arrayMax; zC++) {
    zS++; 
    if (zS < zW) {
           if (zS ==  1) {ArrayVal[zC] =  0;}
      else if (zS ==  2) {ArrayVal[zC] =  0;}
      else if (zS ==  3) {ArrayVal[zC] = 95;}
      else if (zS ==  4) {ArrayVal[zC] =  0;}
      else if (zS ==  5) {ArrayVal[zC] =  0;}
      else if (zS ==  6) {ArrayVal[zC] = 85;}
      else if (zS ==  7) {ArrayVal[zC] =  0;}
      else if (zS ==  8) {ArrayVal[zC] =  0;}
      else if (zS ==  9) {ArrayVal[zC] = 95;}
      else if (zS == 10) {ArrayVal[zC] =  0;}
      else if (zS == 11) {ArrayVal[zC] =  0;}
      else if (zS == 12) {ArrayVal[zC] = 85;}
    } else {ArrayVal[zC] = Steps; zS = 0;}
  }
  // reset pointers
  rowPnt = 0; rowNum = 1; drawFlag = 1;
  if (zM > 0) {return;}
  
  setAngValFromRow();  // set servo values and sliders
  // update Arduino
  usbEn = true; msgTx = "MC."; usbPortWrite(msgTx);
  usbEn = false; cursor(ARROW);
}

// ----------------------------------------------------------------------

void ClearLinks() {
  // resets the link array contents
  for (int zL = 0; zL < rowWidth;zL++) {Link[zL] = 0;}
}

// ----------------------------------------------------------------------

void ClearRow(int zR) {
  // sets row zR to default angle values and steps
  zR = zR * rowWidth;
  for (int zP = zR; zP < (zR + rowWidth - 1); zP++) {
         if (zP ==  0) {ArrayVal[zP] =  0;}
    else if (zP ==  1) {ArrayVal[zP] =  0;}
    else if (zP ==  2) {ArrayVal[zP] = 95;}
    else if (zP ==  3) {ArrayVal[zP] =  0;}
    else if (zP ==  4) {ArrayVal[zP] =  0;}
    else if (zP ==  5) {ArrayVal[zP] = 85;}
    else if (zP ==  6) {ArrayVal[zP] =  0;}
    else if (zP ==  7) {ArrayVal[zP] =  0;}
    else if (zP ==  8) {ArrayVal[zP] = 95;}
    else if (zP ==  9) {ArrayVal[zP] =  0;}
    else if (zP == 10) {ArrayVal[zP] =  0;}
    else if (zP == 11) {ArrayVal[zP] = 85;}
} ArrayVal[zR + rowWidth -1] = Steps;
}

// ----------------------------------------------------------------------

void ClearRowBlockFT(int zF, int zT) {
  // clears a region from row zF to row zT inclusive
  for (int zR = zF; zR <= zT; zR++) {ClearRow(zR);}
}

// ----------------------------------------------------------------------

void ClearSelect() {
  // clear the contents of the Select[] array
  cursor(WAIT);
  for (int zS = 0; zS < arrayMax; zS++) {Select[zS] = false;}
  SelCnt = 0; SelLast = -1;
  cursor(ARROW);
}

// ----------------------------------------------------------------------

void closeCOMPort() {
  // Close the COM port and empty the serial buffer
  if (comFlag > 0) {
    usbPort.clear(); usbPort.stop();
    comFlag = 0; msgCOM = "----"; msgRx = ""; msgTx = ""; drawFlag = 1;
    println("COM Port Closed");
    servoOE = 1;  // turn OFF OE enabled lamp
  }
  usbEn = false; drawFlag = 1;
}

// ----------------------------------------------------------------------

void copyCopyTo(int zR) {
  // copies the contents of the copy memory into row memory starting row zR
  // ensure that we don't exceed the array
  int zT = CopyCnt; if ((zT + zR) >= rowDepth) {zT = rowDepth - zR;}
  zR = zR * rowWidth;
  for (int zP = 0; zP < (zT * rowWidth); zP++) {
    ArrayVal[zR + zP] = CopyMem[zP];
  }
}

// ----------------------------------------------------------------------

void copyLftToRgt(int zF,int zT) {
  // copy the left servo value into the right-hand side servo value
  ArrayVal[(rowPnt * rowWidth) + zT] = ArrayVal[(rowPnt * rowWidth) + zF];
  // update angle array
  AngVal[zT] = AngVal[zF];
  // update sliders
  setSliderX();
  // send the angle
  msgTx = "ST" + str(Ch[zT]) + "." +"SA" + str(AngVal[zT]) + ".";
  usbPortWrite(msgTx); rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void copyLinkedVal(int zS) {
  // servos are linked so copy target to linked pair
  switch(zS) {
    case 0: copyRgtToLft(FR1,FL1); break;  // FR1
    case 1: copyRgtToLft(FR2,FL2); break;  // FR2
    case 2: copyRgtToLft(FR3,FL3); break;  // FR3
    
    case 3: copyLftToRgt(FL1,FR1); break;  // FL1
    case 4: copyLftToRgt(FL2,FR2); break;  // FL2
    case 5: copyLftToRgt(FL3,FR3); break;  // FL3

    case 6: copyRgtToLft(RR1,RL1); break;  // RR1
    case 7: copyRgtToLft(RR2,RL2); break;  // RR2
    case 8: copyRgtToLft(RR3,RL3); break;  // RR3

    case 9: copyLftToRgt(RL1,RR1); break;  // RL1
    case 10: copyLftToRgt(RL2,RR2); break;  // RL2
    case 11: copyLftToRgt(RL3,RR3); break;  // RL3
  }
}

// ----------------------------------------------------------------------

void copyRgtToLft(int zF,int zT) {
  // copy the left servo value into the right-hand side servo value
  ArrayVal[(rowPnt * rowWidth) + zT] = ArrayVal[(rowPnt * rowWidth) + zF];
  // update angle array
  AngVal[zT] = AngVal[zF];
  // update sliders
  setSliderX();
  // send the angle
  msgTx = "ST" + str(Ch[zT]) + "." +"SA" + str(AngVal[zT]) + ".";
  usbPortWrite(msgTx); rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void copyRows() {
  // copy one or more rows into the CopyMem[] store
    int zP = 0; int zRP = rowPnt * rowWidth; int zCP; CopyCnt = 0;
  if (SelCnt < 1) {
    // copy current row
    for (zP = 0; zP < rowWidth; zP++) {
      CopyMem[zP] = ArrayVal[zP + zRP];
    }
    CopyCnt = 1;  // copied count
  } else {
    // copy selected row(s)
    cursor(WAIT);
    for (int zR = 0; zR < rowNum; zR++) {
      // work through each selected row
      if (Select[zR]) {
        // row is selected
        zRP = zR * rowWidth; zCP = CopyCnt * rowWidth;
        // grab the selected row values
        for (zP = 0; zP < rowWidth; zP++) {
          CopyMem[zP + zCP] = ArrayVal[zP + zRP];
        } CopyCnt++;  // count the copied rows
     }
    }
  }
  msgTx = str(CopyCnt) + " rows copied"; drawFlag = 1;
  cursor(ARROW);
}

// ----------------------------------------------------------------------

void copyRowBlockFT(int zF, int zT, int zN) {
  // copy a block of rows zF to row zT, zN depth
  if (zT < zF) {
    // copying up so it's simple
    if ((zF + zN) >= rowDepth) {zN = rowDepth - zF -1;}
    for (int zP = 0; zP < zN; zP++) {
      copyRowFT(zF + zP, zT + zP);
    }
  } else {
    // copying down so start at the bottom
    if ((zT + zN) >= rowDepth) {zN = rowDepth - zT -1;}
    for (int zP = zN -1; zP >= 0; zP--) {
      copyRowFT(zF + zP, zT + zP);
    }
  }
}

// ----------------------------------------------------------------------

void copyRowFT(int zF, int zT) {
  // copy a single row zF to row zT
  zF = zF * rowWidth; zT = zT * rowWidth;
  for (int zP = 0; zP < rowWidth; zP++) {
    ArrayVal[zP + zT] = ArrayVal[zP + zF];
  }
}

// ----------------------------------------------------------------------

void copyToClipboard() {
  // called when user clicks on 'Copy' button
  // if lines are selected it will only copy them
  cursor(WAIT);
  int zC = 0;  // column reference
  CopyData = "  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp " + CR;
  for (int zRow = 0; zRow < rowNum; zRow++) {
    if ((SelCnt == 0) || Select[zRow]) {
      int zP = zRow * rowWidth; zC = zP;
      CopyData += "  MemSet(" + Num3Dig(ArrayVal[zP] ) + ","; zP++;  // FR1
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // FR2
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // FR3
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // FL1
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // FL2
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // FL3
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // RR1
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // RR2
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // RR3
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // RL1
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // RL2
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // RL3
      CopyData += Num3Dig(ArrayVal[zP]) + ");";  // Stp
      
      // append CMD text?
      if (ArrayVal[zC] < 0) {
        // yes this is a command line
        CopyData += "  // ";  // C++ comment
        switch (ArrayVal[zC + 1]) {
          case 0: CopyData += "---"; break;
          case 1: CopyData += "Goto row " + str(ArrayVal[zC + 2]); break;
          case 2: CopyData += "Speed =  " + str(ArrayVal[zC + 2]); break;
          case 3: CopyData += "Pause for " + str(ArrayVal[zC + 2]); break;
          case 4: CopyData += "Flag = " + str(ArrayVal[zC + 2]); break;
          case 5: CopyData += "Balance = ";
                 if (ArrayVal[zC + 2] == 0) {CopyData += "OFF";}
            else if (ArrayVal[zC + 2] == 1) {CopyData += "topple only";} 
            else if (ArrayVal[zC + 2] == 2) {CopyData += "balance only";} 
            else if (ArrayVal[zC + 2] == 3) {CopyData += "topple and balance";} 
            break;
          case 6: CopyData += "For...1 to ..." + str(ArrayVal[zC + 2]) + " loop"; break;
          case 7: CopyData += "Next after a For... loop"; break;
        }
      }
      CopyData += CR;  // end of line CRLF
    }
  }
  msgTx = "Memory copied"; drawTxMsg();

  // now copy it
  StringSelection data = new StringSelection(CopyData);
  Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
  clipboard.setContents(data, data);
  drawFlag = 1; cursor(ARROW);
}

// ----------------------------------------------------------------------

void cutRows() {
  // copy and cut current or selected rows
  if (SelCnt < 1) {
    // no rows selected so use current row
    Select[rowPnt] = true;  // fake current row selected
  }
  
  cursor(WAIT);
  // first copy the selected rows
  copyRows();
  
  if (CopyCnt == rowNum) {
    // all of the rows were selected, so reset them as default rows
    // then reset the flags and row count
    for (int zR = 0; zR < rowNum; zR++) {ClearRow(zR);}
    rowPnt = 0; rowNum = 1;
  } else {
    // copy unselected rows up then reset the rest
    int zN = 0;   // counts the unselected rows
    int zT = -1;  // row target pointer for copying
    for (int zR = 0; zR < rowNum; zR++) {
      // work through each selected row
      if (Select[zR]) {
        // row is selected so reset it
        ClearRow(zR);
        // move the target pointer to the first selected row
        if (zT < 0) {zT = zR;}
     } else {
       zN++;  // count the number of unselected rows
       if ((zR > zT) && (zT >= 0)) {
         // move the unselected row to the previously selected one
         copyRowFT(zR,zT); zT++;
         ClearRow(zR);
       }
     }
    }
    // reset flags and numbers  
    rowNum = zN; if (rowPnt >= rowNum) {rowPnt = rowNum - 1;}
  }
  ClearSelect();
  sendRowData(rowPnt,rowPnt,1);  // send data
  setAngValFromRow();     // set servo values and sliders
  msgTx = str(CopyCnt) + " rows copied and cut"; drawFlag = 1;
  cursor(ARROW);
}


// ######################################################################


void decDegSteps() {
  // degrements the DegStep value
  if (DegStep > 1) {DegStep--; rateFlag = true; rateFlag = true; drawFlag = 1;}
}

// ----------------------------------------------------------------------

void decField(int zFld) {
  // decrements the value of field zFld
  switch(zFld) {
    case FldT1maxLft:
      if (T1maxAngLft > 0) {T1maxAngLft--; calcRideHeights();} break;
    case FldT1maxRht:
      if (T1maxAngRht > 0) {T1maxAngRht--; calcRideHeights();} break;
    case FldT2maxLft:
      if (T2maxAngLft > 10) {T2maxAngLft--;
        if (T2maxAngLft < (T2minAngLft + 10)) {T2minAngLft = T2maxAngLft - 10;}
      calcMidSteps();} break;
    case FldT2minLft:
      if (T2minAngLft > 0) {T2minAngLft--; calcMidSteps();} break;
    case FldT2maxRht:
      if (T2maxAngRht > 10) {T2maxAngRht--;;
        if (T2maxAngRht < (T2minAngRht + 10)) {T2minAngRht = T2maxAngRht - 10;}
      calcMidSteps();} break;
    case FldT2minRht:
      if (T2minAngRht > 0) {T2minAngRht--; calcMidSteps();} break;
    case FldT1WlkLft:
      if (WalkLift > 10) {WalkLift--; drawFlag = 1;} break;
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void decRowPnt() {
  // decrement the memory row point if it is greater than zero
  if (rowPnt > 0) {
    rowPnt--; sendRowData(rowPnt,rowPnt,1);
    setAngValFromRow();  // set servo values and sliders
    drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

void decSpeed() {
  // decrement Speed value towards minimum
  if (Speed > SpeedLL) {
    Speed-= Inc; IncCnt++;  if (keySHIFT) {Speed -= 9;}
    if (IncCnt > 10) {IncCnt = 0; Inc++;}
    if (Speed < SpeedLL) {Speed = SpeedLL;}
    sendSpeed(); drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

void decXSlider(int zS) {
  // decrement target slider value
  servoTgt = zS;
  if (getAngVal(AngVal[zS]) > AngVLL[zS]) {
    AngLast = getAngVal(AngVal[zS]);  // get angle and strip off rate and pwr
    AngVal[zS] = AngLast - 1;
    AngDif = AngVal[zS] - AngLast;
    // update value in memory array
    ArrayVal[(rowPnt * rowWidth) + zS] = AngVal[zS];
    // send value to usb port
    msgTx = "ST" + str(Ch[zS]) + "." +"SA" + str(AngVal[zS]) + ".";
    if (Link[zS] > 0) {moveLinkedVal(zS);}
    else {setSliderX(); usbPortWrite(msgTx); rateFlag = true; drawFlag = 1;}
    if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
  }
}

// ----------------------------------------------------------------------

void deleteRows() {
  // delete the current row until last, then reset row to default
  // if some selected then delete all of them
  // if the USB port is active send the revised row to the robot
  cursor(WAIT);
  if (rowNum == 1) {
    // only one row, so set it to default
    ClearRow(rowPnt);
  } else if (SelCnt > 0) {
    // one or more rows are selected
    if (SelCnt == rowNum) {
      // all of the rows are selected so clear memory and insert new row
      ClearSelect();
      ClearArray(1);
    } else {
      // only some rows are selected so just delete them
      int zNum = rowNum; int zS = 0;
      while (zS < rowNum) {
        // work through rows and delete selected ones
        if (Select[zS]) {
          // this is a selected row, copy data up to overwrite it, then
          // clear the last line
          moveRowBlockFT(zS + 1, zS, zNum - zS - 1);
          moveSelArrayUp();       // shift selects to match row deletion
          checkGotoCmdDel(zS,1);  // check for any goto commands and correct them if needed
          zNum--;
        } else {
          // not a selected row so simply increment the pointer
          zS++;
        }
      } rowNum = rowNum - SelCnt; if (rowPnt >= rowNum) {rowPnt = rowNum - 1;}
    }
  } else {
    // no rows are selected
    if (rowPnt == (rowNum - 1)) {
      // looking at last row so clear it and reduce the pointer and number of rows
      ClearRow(rowPnt);
      rowPnt--; rowNum--;
    } else {
      // delete a row somewhere within the list of rows by copying values up
      // and clearing the last line before reducing pointers
      moveRowBlockFT(rowPnt + 1, rowPnt, rowNum - rowPnt - 1);
      checkGotoCmdDel(rowPnt,1);  // check for any goto commands and correct them if needed
      // now reset the last line before adjusting pointers
      rowNum--;
    }
  } 
  ClearSelect(); rateFlag = true; drawFlag = 1;
  if (usbEn) {sendRowData(rowPnt,rowPnt,1);}
}


// ######################################################################


void ESC() {
  // user has pressed the ESC key. If in the middle of a click-n-drag
  // this will be terminated and the droid returned to its original position.
  if ((sliderDwn < 1) && (MultiPnt < 1)) {return;}  // not a click-n-drag
  
  // original position is stored in Temp[] on mouswe click, so return to 
  // that position after killing all comm events
  usbBuff = "***";  // send ann abort '*' to arduinio, several to make sure
  RetRowFromTemp(); // recall the original values
  if (MultiPnt > 0) {
    // multipoint move so send full row of data
    sendRowData(rowPnt,rowPnt,1);  // send current row data to Arduino
  } else {
    // slider move so send original servo value
    // send value to usb port
    if (Link[servoTgt] > 0) {sendRowData(rowPnt,rowPnt,1);}  // more than one change
    else {
      // only one slider active so reset 6that value
      msgTx = "ST" + str(Ch[servoTgt]) + "." +"SA" + str(ArrayVal[(rowPnt * rowWidth) + servoTgt]) + ".";
      usbPortWrite(msgTx);
    }
  }
  setAngValFromRow();  // set servo values and sliders
  mDwn = 0; sliderDwn = 0; MultiPnt = 0; Help_Message();
  rateFlag = true; drawFlag = 1;
  cursor(ARROW);
}

// ----------------------------------------------------------------------

void exit() {
  // program is closing so reset Arduino...
  usbPortWrite("XX.");
  delay(20); // allow 20ms for data to be sent
  super.exit();
}


// ######################################################################


void getAngFillCol(int zAng,int zCP) {
  // sets background colours based on the column, angle and min/max values
  switch(zCP) {
    case 0:  // FR1
      AR = 255;
      AG = mapInt(zAng,MinA[0],MaxA[0],210,67);
      AB = mapInt(zAng,MinA[0],MaxA[0],210,67);
      break;
    case 1:  //FR2
      AR = mapInt(zAng,MinA[1],MaxA[1],210,27);
      AG = 255;
      AB = mapInt(zAng,MinA[1],MaxA[1],210,76);
      break;
    case 2:  //FR3
      AR = mapInt(zAng,MinA[2],MaxA[2],200,20);
      AG = mapInt(zAng,MinA[2],MaxA[2],220,120);
      AB = 255;
      break;
      
    case 3:  // FL1
      AR = 255;
      AG = mapInt(zAng,MinA[3],MaxA[3],210,67);
      AB = mapInt(zAng,MinA[3],MaxA[3],210,67);
      break;
    case 4:  // FL2
      AR = mapInt(zAng,MinA[4],MaxA[4],210,27);
      AG = 255;
      AB = mapInt(zAng,MinA[4],MaxA[4],210,76);
      break;
    case 5:  //FL3
      AR = mapInt(zAng,MinA[5],MaxA[5],200,20);
      AG = mapInt(zAng,MinA[5],MaxA[5],220,120);
      AB = 255;
      break;
      
    case 6:  // RR1
      AR = 255;
      AG = mapInt(zAng,MinA[6],MaxA[6],210,67);
      AB = mapInt(zAng,MinA[6],MaxA[6],210,67);
      break;
    case 7:  // RR2
      AR = mapInt(zAng,MinA[7],MaxA[7],210,27);
      AG = 255;
      AB = mapInt(zAng,MinA[7],MaxA[7],210,76);
      break;
    case 8:  //RR3
      AR = mapInt(zAng,MinA[8],MaxA[8],200,20);
      AG = mapInt(zAng,MinA[8],MaxA[8],220,120);
      AB = 255;
      break;
      
    case 9:  // RL1
      AR = 255;
      AG = mapInt(zAng,MinA[9],MaxA[9],210,67);
      AB = mapInt(zAng,MinA[9],MaxA[9],210,67);
      break;
    case 10:  //RL2
      AR = mapInt(zAng,MinA[10],MaxA[10],210,27);
      AG = 255;
      AB = mapInt(zAng,MinA[10],MaxA[10],210,76);
      break;
    case 11:  //RL3
      AR = mapInt(zAng,MinA[11],MaxA[11],200,20);
      AG = mapInt(zAng,MinA[11],MaxA[11],220,120);
      AB = 255;
      break;
  }
}

// ----------------------------------------------------------------------

int getAngleMXD() {
  // returns an array pointer based on the mXD value
  int zA;
  zA = (mXD - 154) * 12/(495 - 154);
  zA = min(zA,11); zA = max(zA,0);
  //print("zA = "); println(zA);
  return zA;
}

// ----------------------------------------------------------------------

int getAngVal(int zAng) {
  // strips off any auto-rate offset and sets the temp AngRate flag
  // strips off any Pwr offset and sets the temp AngPwr flag
  if (zAng >= 400) {zAng -= 400; AngRate = true;} else {AngRate = false;}
  if (zAng >= 200) {zAng -= 200; AngPwr = false;} else {AngPwr = true;}
  return zAng;
}

// ----------------------------------------------------------------------

void getCOMPort() {
  // initialise serial comms if a port exists
  if (comFlag > 0) {
    // stops an existing connection if already made
    usbPort.clear(); usbPort.stop();
    comFlag = 0; msgCOM = "----"; delay(100);
  }
  comListLength = Serial.list().length;
  if (comListLength > 0) {
    comPnt++; if (comPnt >= comListLength) {comPnt = 0;}
    // test available serial port
    comName = Serial.list()[comPnt];
    println(comName);
    try {
      usbPort = new Serial(this, comName, 115200);
      //if (pingON > 0) {tick = 100;} // don't send ticks too early or board wont reset!
      //usbPort.bufferUntil(LF);
      comFlag = 1; msgCOM = comName;  msgTx = "USB port connected";  // USB port connected
    } catch(Exception e) {
      comFlag = 0; msgCOM = "-Error-"; msgTx = "Failed to connect";
    }
  } else {
    // no serial port available
    comFlag = 0; msgCOM = "-NA-"; msgRx = ""; msgTx = "No port available";
  }
  // set conditions
  usbEn = false;  // always disable usb port messaging
  drawFlag = 1;
}

// ----------------------------------------------------------------------

Object getFromClipboard (DataFlavor flavor) {

  Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard(); 
  Transferable contents = clipboard.getContents(null);
  Object object = null;

  if (contents != null && contents.isDataFlavorSupported(flavor))
  {
    try
    {
      object = contents.getTransferData(flavor);
      println("Clipboard.getFromClipboard() >> Object transferred from clipboard.");
    }

    catch (UnsupportedFlavorException e1) // Unlikely but we must catch it
    {
      println("Clipboard.getFromClipboard() >> Unsupported flavor: " + e1);
      //~  e1.printStackTrace();
    }

    catch (java.io.IOException e2)
    {
      println("Clipboard.getFromClipboard() >> Unavailable data: " + e2);
      //~  e2.printStackTrace();
    }
  }

  return object;
} 

// ----------------------------------------------------------------------

void getMinMaxAng() {
  // load the min max array from the current list
  // set all min values to max angle
  MinA[ 0] = 180; MinA[ 1] = 180; MinA[ 2] = 180; MinA[ 3] = 180;
  MinA[ 4] = 180; MinA[ 5] = 180; MinA[ 6] = 180; MinA[ 7] = 180;
  MinA[ 8] = 180; MinA[ 9] = 180; MinA[10] = 180; MinA[11] = 180;
  // set all max values to 1
  MaxA[ 0] = 0; MaxA[ 1] = 0; MaxA[ 2] = 0; MaxA[ 3] = 0;
  MaxA[ 4] = 0; MaxA[ 5] = 0; MaxA[ 6] = 0; MaxA[ 7] = 0;
  MaxA[ 8] = 0; MaxA[ 9] = 0; MaxA[10] = 0; MaxA[11] = 0;
  
  // now get the min/max values
  int zA;  // temp angle value
  for (int zR = 0; zR < rowNum; zR++) {
    int zP = zR * rowWidth;  // set the row memory pointer
    if (ArrayVal[zP] >= 0) {
      // not a Cmd line
      for (int zI = 0; zI < (rowWidth - 1); zI++) {
        zA = getAngVal(ArrayVal[zP+zI]);  
        if (zA > MaxA[zI]) {MaxA[zI] = zA;}
        else if (zA < MinA[zI]) {MinA[zI] = zA;}
      }
    }
  }
}

// ----------------------------------------------------------------------

boolean getRated() {
  // scans the array to see if any angles are set to rate if so return true
  for (int zL = 0; zL < rowNum; zL++) {
    int zR = zL * rowWidth;
    if (ArrayVal[zR] >= 0) {
      // not a command line so check for rate value (400) added to angle?
      // as soon as we find one we return with true
      for (int zP = 0; zP < (rowWidth - 1); zP++) {
        if (ArrayVal[zP + zR] >= 400) {return true;}
      }
    }
  }
  return false;
}

// ----------------------------------------------------------------------

int getProfile(int zR) {
  // returns the rate profile as a number for the row zR
  zR*= rowWidth; zR = ArrayVal[zR + 12];
       if (zR > 600) {zR = 3;}  // switch acceleration/deceleration
  else if (zR > 400) {zR = 2;}  // deceleration
  else if (zR > 200) {zR = 1;}  // acceleration
  else {zR = 0;}                // constant rate
  return zR;
}

// ----------------------------------------------------------------------

int getRowMYD() {
  // returns the row offset based on mYD
  // add this to rowPnt to get the actual row pointer under the mouse
  int zR;
  if (Scrn != 1) {
    // SCrn0 or Scrn2
    zR = -2 + (6 * (mYD - 548)/(688 - 548));
    zR = max(zR,-2); zR = min(zR,3);
  } else {
    // Scrn1
    zR = -13 + (27 * (mYD - 42)/(686 - 42));
    zR = max(zR,-13); zR = min(zR,26);
  }
  //print("Row mYD = "); println(zR);
  return zR;
}

// ----------------------------------------------------------------------

String getRowString(int zPnt) {
  // returns a string representing the row of data starting at zPnt
  String zR; Blank = false; Sep$ = ","; Slope$ = "C";
  zR = getStrArrayVal(zPnt) + Sep$; zPnt++;       // FR1
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // FR2
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // FR3
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // FL1
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // FL2
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // FL3
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // RR1
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // RR2
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // RR3
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // RL1
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // RL2
  zR = zR + getStrArrayVal(zPnt); zPnt++;         // RL3;
  return zR;
}

// ----------------------------------------------------------------------

String getStrArrayVal(int zPnt) {
  // returns the array value as a 3 character string, with leading spaces
  // if not a command line then also set value of Slope$
  // an angle value will be 0 - 180, plus an offset of 200 if Pwr is OFF
  // a step value will be 0 - 180, plus move profiles of 200,400 and 600
  String zS = "";
  if (Blank) {zS = "      "; return zS;}
  
  int zV = ArrayVal[zPnt];
  if (zV < 0) {
    zS = "Cmd"; TR = 255; TB = 0;
    if (ArrayVal[zPnt + 1] == 0) {
      Blank = true; zS = "   "; Sep$ = "  "; Slope$ = "   ";
    } return zS;
  }

  Slope$ = "C";
  if (zV >= 600) {zV -= 600; Slope$ = "S";}
  else if (zV >= 400) {zV -= 400; Slope$ = "D";}
  else if (zV >= 200) {zV -= 200; Slope$ = "A";}
  
  zS = str(zV);
  if (zV < 100) {zS = "  " + zS;}
  if (zV < 10) {zS = "  " + zS;}
  return zS;
}

// ----------------------------------------------------------------------

void getTextFromClipboard() {
  // gets text from clipboard and analyses it as an import
  String text = (String) getFromClipboard(DataFlavor.stringFlavor);
  if (text == null) {
    msgTx = "Clipboard is empty!"; drawTxMsg();
    return;
  }
  msgTx = "Reading Clipboard data"; drawTxMsg();
  text = trim(text);
  if (text.indexOf("MemSet") >= 0) {
    // we are going to load one or more lines of data so clear the array
    // and prepare pointers
    cursor(WAIT);
    int zS = 0; int zW = rowWidth;
    for (int zC = 0; zC < arrayMax; zC++) {
      zS++; if (zS < zW) {ArrayVal[zC] = 90;}
      else {ArrayVal[zC] = Steps; zS = 0;}
    }
    // reset pointers
    String zLft = ""; int zEOL = 0; int zRow = 0; int zNum = 0;
    // now parse the data
    while (text.indexOf("MemSet(") >= 0) {
      // remove any preceeding text
      text = text.substring(text.indexOf("MemSet("),text.length());
      // find the end of the row data
      zEOL = text.indexOf(";");
      if (zEOL < 0) {break;}  // end of data has been reached
      // grab the row of data
      zLft = text.substring(7,zEOL);
      println(zLft);
      // now extract the numbers
      int zP = zRow * rowWidth; zNum++;
      while(true) {
        zEOL = zLft.indexOf(",");
        if (zEOL >= 0) {
          ArrayVal[zP] = int(trim(zLft.substring(0,zEOL))); zP++;
          //print(zLft.substring(0,zEOL)); print("\t"); println(int(trim(zLft.substring(0,zEOL))));
          zLft = zLft.substring(zEOL+1,zLft.length());
        } else { 
          zEOL = zLft.indexOf(")");
          if (zEOL >= 0) {
            ArrayVal[zP] = int(trim(zLft.substring(0,zEOL))); zP++;  // SR1
            //print(zLft.substring(0,zEOL)); print("\t"); println(int(trim(zLft.substring(0,zEOL))));
            break;
          } else {break;}
        }
      }
      
      zRow++;
  
      if ((zEOL + 1) >= text.length()) {break;}
      // remove the left portion
      text = text.substring(zEOL + 1,text.length());
    }
    rowPnt = 0; rowNum = zNum; usbEn =false;
    setAngValFromRow();  // set servo values and sliders
    cursor(ARROW);
  } else {
    msgTx = "Clipboard data is invalid"; drawTxMsg();
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void getWalkDelay() {
  // sets the Walk timer based on the current rows step value
  // the calling code must check for Command lines
  int zP = simPnt * rowWidth;
  WaklDelay = 1000/Speed;
  if (WaklDelay < 4) {WaklDelay = 4;}  // limit to 4ms minimum
  
  // check for and remove any step profile shaping
  int zStep = ArrayVal[zP + 12];
  if (zStep > 600) {zStep -= 600;}
  else if (zStep > 400) {zStep -= 400;}
  else if (zStep > 200) {zStep -= 200;}
  WalkNext = millis() + (WaklDelay * zStep);
}

// ----------------------------------------------------------------------

void getXSliderOffset(int zS) {
  // called when mouse is clicked on a slider to determine the offset
  mXOff = JoyXval[zS] - mX;
}
// ----------------------------------------------------------------------

void gotoFirst() {
  // goto the first row
  if (rowPnt > 1) {usbEn = false;}
  rowPnt = 0; drawFlag = 1;
  sendRowData(rowPnt,rowPnt,1);  // send row data
  setAngValFromRow();     // set servo values and sliders
}

// ----------------------------------------------------------------------

void gotoLast() {
  // goto the last row
  if ((rowNum - rowPnt) > 2) {usbEn = false;}
  rowPnt = rowNum - 1; drawFlag = 1;
  sendRowData(rowPnt,rowPnt,1);  // send row data
  setAngValFromRow();     // set servo values and sliders
}

// ----------------------------------------------------------------------

void grabRowInTemp(int zRP) {
  // copy the row zRP values to Temp[] store
  zRP *= rowWidth;
  for (int zP = 0; zP < (rowWidth - 1); zP++) {
    Temp[zP] = ArrayVal[zP + zRP];
  }
}


// ######################################################################


void incDegSteps() {
  // increments the DegStep value
  if (DegStep < 30) {DegStep++; rateFlag = true; drawFlag = 1;}
}

// ----------------------------------------------------------------------

void incField(int zFld) {
  // increments the value of field zFld
  switch(zFld) {
    case FldT1maxLft:
      if (T1maxAngLft < 160) {T1maxAngLft++; calcRideHeights();} break;
    case FldT1maxRht:
      if (T1maxAngRht < 160) {T1maxAngRht++; calcRideHeights();} break;
    case FldT2maxLft:
      if (T2maxAngLft < 140) {T2maxAngLft++; calcMidSteps();} break;
    case FldT2minLft:
      if (T2minAngLft < 130) {T2minAngLft++;
        if (T2minAngLft > (T2maxAngLft - 10)) {T2maxAngLft = T2minAngLft + 10;}
      calcMidSteps();} break;
    case FldT2maxRht:
      if (T2maxAngRht < 140) {T2maxAngRht++; calcMidSteps();} break;
    case FldT2minRht:
      if (T2minAngRht < 130) {T2minAngRht++;
        if (T2minAngRht > (T2maxAngRht - 10)) {T2maxAngRht = T2minAngRht + 10;}
      calcMidSteps();} break;
    case FldT1WlkLft:
      if (WalkLift <60) {WalkLift++;} break;
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void incRowPnt() {
  // increment the memory row point if it is less than max rows
  if (rowPnt < (rowNum - 1)) {
    rowPnt++; sendRowData(rowPnt,rowPnt,1);
    setAngValFromRow();  // set servo values and sliders
    drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

void incSpeed() {
  // increment Speed value towards maximum
  if (Speed < SpeedUL) {
    Speed+= Inc; IncCnt++; if (keySHIFT) {Speed += 9;}
    if (IncCnt > 10) {IncCnt = 0; Inc++;}
    if (Speed > SpeedUL) {Speed = SpeedUL;}
    sendSpeed(); drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

void incXSlider(int zS) {
  // increment slider value
  servoTgt = zS;
  if (getAngVal(AngVal[zS]) < AngVUL[zS]) {
    AngLast = getAngVal(AngVal[zS]);
    AngVal[zS] = AngLast + 1;
    AngDif = AngVal[zS] - AngLast;
    // update value in memory array
    ArrayVal[(rowPnt * rowWidth) + zS] = AngVal[zS];
    // send value to usb port
    msgTx = "ST" + str(Ch[zS]) + "." +"SA" + str(AngVal[zS]) + ".";
    if (Link[zS] > 0) {moveLinkedVal(zS);}
    else {setSliderX(); usbPortWrite(msgTx); rateFlag = true; drawFlag = 1;}
    if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
  }
}

// ----------------------------------------------------------------------

void insertBalance() {
  // insert a Balance command either after or before the current row
  // default value is 1 until a flag is adjusted
  int zRP = rowPnt * rowWidth;
  newRow();  // create a new line, then overwrite it
  zRP = rowPnt * rowWidth;

  // now overwrite with the command data
  ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
  ArrayVal[zRP + 1] = 5;  // set the cmd = 4
  ArrayVal[zRP + 2] = 1;  // set the default 1 = ON, 0 = OFF value
  ArrayVal[zRP + 3] = 0;  // clear FL1
  ArrayVal[zRP + 4] = 0;  // clear FL2
  ArrayVal[zRP + 5] = 0;  // clear FL3
  ArrayVal[zRP + 6] = 0;  // clear RR1
  ArrayVal[zRP + 7] = 0;  // clear RR2
  ArrayVal[zRP + 8] = 0;  // clear RR3
  ArrayVal[zRP + 9] = 0;  // clear RL1
  ArrayVal[zRP + 10] = 0;  // clear RL2
  ArrayVal[zRP + 11] = 0;  // clear RL3
  ArrayVal[zRP + 12] = 0;  // clear steps
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void insertBlank() {
  // insert a 'Blank' command either after or before the current row
  int zRP = rowPnt * rowWidth;
  newRow();  // create a new line, then overwrite it
  zRP = rowPnt * rowWidth;
  
  // now overwrite with the command data
  ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
  ArrayVal[zRP + 1] = 0;  // set the Blank cmd = 0
  ArrayVal[zRP + 2] = 0;  // clear FR3
  ArrayVal[zRP + 3] = 0;  // clear FL1
  ArrayVal[zRP + 4] = 0;  // clear FL2
  ArrayVal[zRP + 5] = 0;  // clear FL3
  ArrayVal[zRP + 6] = 0;  // clear RR1
  ArrayVal[zRP + 7] = 0;  // clear RR2
  ArrayVal[zRP + 8] = 0;  // clear RR3
  ArrayVal[zRP + 9] = 0;  // clear RL1
  ArrayVal[zRP + 10] = 0;  // clear RL2
  ArrayVal[zRP + 11] = 0;  // clear RL3
  ArrayVal[zRP + 12] = 0;  // clear steps
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void InsertExCode() {
  // insert an executable code either after or before the current row
  // the default value is For... Next command, one loop
  //
  // 6  - For... command
  // 7  - Next command
  int zRP = rowPnt * rowWidth;
  newRow();  // create a new line, then overwrite it
  zRP = rowPnt * rowWidth;

  // now overwrite with the command data
  ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
  ArrayVal[zRP + 1] = 6;  // set cmd = 6, first excuteable command value, For...Next loop
  ArrayVal[zRP + 2] = 1;  // set the default 1 value
  ArrayVal[zRP + 3] = 0;  // clear FL1
  ArrayVal[zRP + 4] = 0;  // clear FL2
  ArrayVal[zRP + 5] = 0;  // clear FL3
  ArrayVal[zRP + 6] = 0;  // clear RR1
  ArrayVal[zRP + 7] = 0;  // clear RR2
  ArrayVal[zRP + 8] = 0;  // clear RR3
  ArrayVal[zRP + 9] = 0;  // clear RL1
  ArrayVal[zRP + 10] = 0;  // clear RL2
  ArrayVal[zRP + 11] = 0;  // clear RL3
  ArrayVal[zRP + 12] = 0;  // clear steps
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void InsertFlag() {
  // insert a Flag command either after or before the current row
  // default value is 1 until a flag is adjusted
  int zRP = rowPnt * rowWidth;
  newRow();  // create a new line, then overwrite it
  zRP = rowPnt * rowWidth;

  // now overwrite with the command data
  ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
  ArrayVal[zRP + 1] = 4;  // set the Flag cmd = 4
  ArrayVal[zRP + 2] = Flag;  // set the default Flag value
  ArrayVal[zRP + 3] = 0;  // clear FL1
  ArrayVal[zRP + 4] = 0;  // clear FL2
  ArrayVal[zRP + 5] = 0;  // clear FL3
  ArrayVal[zRP + 6] = 0;  // clear RR1
  ArrayVal[zRP + 7] = 0;  // clear RR2
  ArrayVal[zRP + 8] = 0;  // clear RR3
  ArrayVal[zRP + 9] = 0;  // clear RL1
  ArrayVal[zRP + 10] = 0;  // clear RL2
  ArrayVal[zRP + 11] = 0;  // clear RL3
  ArrayVal[zRP + 12] = 0;  // clear steps
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void insertGoto() {
  // insert a goto branch to a selected row, the 1st selected row in the list
  // find the selected line
  if (SelCnt < 1) {return;}   // abort if none selected. This should never happen
  usbEn = false;  // avoid any unintentional movement

  int zRP = rowPnt * rowWidth;
  if ((ArrayVal[zRP] < 0) && (ArrayVal[zRP + 1] == 1)) {
    // current line is a goto line, so simply update it
  } else {
    newRow();  // create a new line, then overwrite it
    zRP = rowPnt * rowWidth;
  }

  for (int zS = 0; zS < rowNum; zS++) {
    if (Select[zS]) {
      // 1st selected row, so set branch instruction in place of SR1
      ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
      ArrayVal[zRP + 1] = 1;  // set the goto cmd = 1
      ArrayVal[zRP + 2] = zS; // set the goto pointer
      ArrayVal[zRP + 3] = 0;  // clear FL1
      ArrayVal[zRP + 4] = 0;  // clear FL2
      ArrayVal[zRP + 5] = 0;  // clear FL3
      ArrayVal[zRP + 6] = 0;  // clear RR1
      ArrayVal[zRP + 7] = 0;  // clear RR2
      ArrayVal[zRP + 8] = 0;  // clear RR3
      ArrayVal[zRP + 9] = 0;  // clear RL1
      ArrayVal[zRP + 10] = 0;  // clear RL2
      ArrayVal[zRP + 11] = 0;  // clear RL3
      ArrayVal[zRP + 12] = 0;  // clear steps
      break;
    }
  } 
  ClearSelect();  // clear selsects as it may not be visible if at front of list
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void insertPause() {
  // insert a Pause command either after or before the current row
  // default value is 1 until a pause is adjusted

  int zRP = rowPnt * rowWidth;
  newRow();  // create a new line, then overwrite it
  zRP = rowPnt * rowWidth;

  // now overwrite with the command data
  ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
  ArrayVal[zRP + 1] = 3;  // set the pause cmd = 3
  ArrayVal[zRP + 2] = Pause; // set the current pause value
  ArrayVal[zRP + 3] = 0;  // clear FL1
  ArrayVal[zRP + 4] = 0;  // clear FL2
  ArrayVal[zRP + 5] = 0;  // clear FL3
  ArrayVal[zRP + 6] = 0;  // clear RR1
  ArrayVal[zRP + 7] = 0;  // clear RR2
  ArrayVal[zRP + 8] = 0;  // clear RR3
  ArrayVal[zRP + 9] = 0;  // clear RL1
  ArrayVal[zRP + 10] = 0;  // clear RL2
  ArrayVal[zRP + 11] = 0;  // clear RL3
  ArrayVal[zRP + 12] = 0;  // clear steps
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void insertReady() {
  // insert ready data either after or before the current row
  // if SHIFT key then only send data and execute
  if (keySHIFT) {
    msgTx  = "SP50.";
    msgTx += "ML0.130,70,95,130,70,85,130,70,95,130,70,85,640,MX0.";
    usbPortWrite(msgTx);
    msgTx = msgTx.substring(0,20) + "...";
  } else {
    int zRP = rowPnt * rowWidth;
    newRow();  // create a new line, then overwrite it
    zRP = rowPnt * rowWidth;
    // now overwrite with the command data
    ArrayVal[zRP] = 130;      // FR1
    ArrayVal[zRP + 1] = 70;  // FR2
    ArrayVal[zRP + 2] = 95;  // FR3
    
    ArrayVal[zRP + 3] = 130; // FL1
    ArrayVal[zRP + 4] = 70;  // FL2
    ArrayVal[zRP + 5] = 85;  // FL3
    
    ArrayVal[zRP + 6] = 130; // RR1
    ArrayVal[zRP + 7] = 70;  // RR2
    ArrayVal[zRP + 8] = 95;  // RR3
    
    ArrayVal[zRP + 9] = 130; // RL1
    ArrayVal[zRP + 10] = 70; // RL2
    ArrayVal[zRP + 11] = 85; // RL3
    
    ArrayVal[zRP + 12] = Steps + 600;
    sendRowData(rowPnt,rowPnt,1);   // send data
    setAngValFromRow();      // set servo values and sliders
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void insertRest() {
  // insert ready data either after or before the current row
  // if SHIFT key then only send data and execute
  if (keySHIFT) {
    // move to rest sitting, then turn power off
    msgTx  = "SP50.";
    msgTx += "ML0.85,60,95,85,60,85,11,0,95,11,0,85,640,";
    msgTx += "ML1.285,260,295,285,260,285,211,200,295,211,200,285,1,MO.";
    usbPortWrite(msgTx);
    msgTx = msgTx.substring(0,20) + "...";
  } else {
    int zRP = rowPnt * rowWidth;
    newRow();  // create a new line, then overwrite it
    zRP = rowPnt * rowWidth;
    // now overwrite with the command data
    ArrayVal[zRP] = 85;      // FR1
    ArrayVal[zRP + 1] = 60;  // FR2
    ArrayVal[zRP + 2] = 95;  // FR3
    
    ArrayVal[zRP + 3] = 85;  // FL1
    ArrayVal[zRP + 4] = 60;  // FL2
    ArrayVal[zRP + 5] = 85;  // FL3
    
    ArrayVal[zRP + 6] = 11;  // RR1
    ArrayVal[zRP + 7] = 0; // RR2
    ArrayVal[zRP + 8] = 95; // RR3
    
    ArrayVal[zRP + 9] = 11;  // RL1
    ArrayVal[zRP + 10] = 0; // RL2
    ArrayVal[zRP + 11] = 85; // RL3
    
    ArrayVal[zRP + 12] = Steps;
    sendRowData(rowPnt,rowPnt,1);   // send data
    setAngValFromRow();      // set servo values and sliders
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void insertSpeed() {
  // insert a speed command either after or before the current row
  // if current line is speed command then update

  int zRP = rowPnt * rowWidth;
  if ((ArrayVal[zRP] < 0) && (ArrayVal[zRP + 1] == 2)) {
    // current line is a speed line, so simply update it
  } else {
    newRow();  // create a new line, then overwrite it
    zRP = rowPnt * rowWidth;
  }
  // now overwrite with the command data
  ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
  ArrayVal[zRP + 1] = 2;  // set the speed cmd = 2
  ArrayVal[zRP + 2] = Speed; // set the current speed value
  ArrayVal[zRP + 3] = 0;  // clear FL1
  ArrayVal[zRP + 4] = 0;  // clear FL2
  ArrayVal[zRP + 5] = 0;  // clear FL3
  ArrayVal[zRP + 6] = 0;  // clear RR1
  ArrayVal[zRP + 7] = 0;  // clear RR2
  ArrayVal[zRP + 8] = 0;  // clear RR3
  ArrayVal[zRP + 9] = 0;  // clear RL1
  ArrayVal[zRP + 10] = 0;  // clear RL2
  ArrayVal[zRP + 11] = 0;  // clear RL3
  ArrayVal[zRP + 12] = 0;  // clear steps
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void insertStand() {
  // insert squat data either after or before the current row
  // if SHIFT key then only send data and execute
  if (keySHIFT) {
    // move to upright stand position
    msgTx  = "SP50.";
    msgTx += "ML0.30,0,95,30,0,85,30,0,95,30,0,85,220,";
    msgTx += "ML1.160,90,95,160,90,85,160,90,95,160,90,85,460,";
    msgTx += "ML2.360,290,95,360,290,85,360,290,95,360,290,85, 5,";
    msgTx += "MO.";
    usbPortWrite(msgTx);
    msgTx = msgTx.substring(0,20) + "...";
  } else {
    int zRP = rowPnt * rowWidth;
    newRow();  // create a new line, then overwrite it
    zRP = rowPnt * rowWidth;
    // now overwrite with the command data
    ArrayVal[zRP] = 160;      // FR1
    ArrayVal[zRP + 1] = 90;  // FR2
    ArrayVal[zRP + 2] = 95; // FR3
    
    ArrayVal[zRP + 3] = 160;  // FL1
    ArrayVal[zRP + 4] = 90;  // FL2
    ArrayVal[zRP + 5] = 85; // FL3
    
    ArrayVal[zRP + 6] = 160;  // RR1
    ArrayVal[zRP + 7] = 90;  // RR2
    ArrayVal[zRP + 8] = 95; // RR3
    
    ArrayVal[zRP + 9] = 160;  // RL1
    ArrayVal[zRP + 10] = 90; // RL2
    ArrayVal[zRP + 11] = 85; // RL3
    
    ArrayVal[zRP + 12] = Steps + 400;
    sendRowData(rowPnt,rowPnt,1);   // send data
    setAngValFromRow();      // set servo values and sliders
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void insertSquat() {
  // insert squat data either after or before the current row
  // if SHIFT key then only send data and execute
  if (keySHIFT) {
    // move to squat and turn power OFF
    msgTx  = "SP50.";
    msgTx += "ML0.0,0,95,0,0,85,0,0,95,0,0,85,705,";
    msgTx += "ML1.0,0,95,0,0,85,0,0,95,0,0,85,  5,";
    msgTx += "ML2.200,200,295,200,200,285,200,200,295,200,200,285,5,MO.";
    usbPortWrite(msgTx);
    msgTx = msgTx.substring(0,20) + "...";
  } else {
    int zRP = rowPnt * rowWidth;
    newRow();  // create a new line, then overwrite it
    zRP = rowPnt * rowWidth;
    // now overwrite with the command data
    ArrayVal[zRP] = 0;      // FR1
    ArrayVal[zRP + 1] = 0;  // FR2
    ArrayVal[zRP + 2] = 95; // FR3
    
    ArrayVal[zRP + 3] = 0;  // FL1
    ArrayVal[zRP + 4] = 0;  // FL2
    ArrayVal[zRP + 5] = 85; // FL3
    
    ArrayVal[zRP + 6] = 0;  // RR1
    ArrayVal[zRP + 7] = 0;  // RR2
    ArrayVal[zRP + 8] = 95; // RR3
    
    ArrayVal[zRP + 9] = 0;  // RL1
    ArrayVal[zRP + 10] = 0; // RL2
    ArrayVal[zRP + 11] = 85; // RL3
    
    ArrayVal[zRP + 12] = Steps + 600;
    sendRowData(rowPnt,rowPnt,1);   // send data
    setAngValFromRow();      // set servo values and sliders
  } rateFlag = true; drawFlag = 1;
}


// ######################################################################


void keyToggleRate() {
  // called by 'R' key pressed. Toggle angle rate flag if we are pointing
  // at a valid angle in the list.
  if (AngField < 0) {return;}  // not a valid field

  rateFlag = true; drawFlag = 1;
  int zRow = getRowMYD() + rowPnt;
  int zPnt = (zRow * rowWidth) + AngField;
  if (zRow == 0) {msgTx = "Error - first line"; return;}
  if ((zRow >= (rowNum - 1)) && (ArrayVal[zPnt] < 400)) {msgTx = "Error - last line"; return;}
  if (rowNum < 3) {msgTx = "Error - too few lines"; return;}
  if (ArrayVal[zRow * rowWidth] < 0) {msgTx = "Error - Command line"; return;}
  
  if (ArrayVal[zPnt] >= 400) {
    ArrayVal[zPnt] -= 400;   // remove the auto-rate +400 offset
    Rated = getRated();  // check to see if any more rate angles exist
  } else {
    ArrayVal[zPnt] += 400;  // add the auto-rate +400 offset
    Rated = true;
  }

  if (Rated) {calcRates();}  // update rate values
  
  if (zRow == rowPnt) {
    // changed a value in the current row
    AngVal[AngField] = ArrayVal[zPnt];
    // send updated value to usb port
    msgTx = "ST" + str(Ch[AngField]) + "." +"SA" + str(AngVal[AngField]) + ".";
    setAngValFromRow();
    usbPortWrite(msgTx);
    if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
  }
}


// ######################################################################


void makeIntermediate(int zR) {
  // sets the value of row zR to intermediate angles with previous and next row
  if (zR == 0) {return;}             // invalid intermediate row
  if (zR == (rowNum - 1)) {return;}  // invalid intermediate row
  if (ArrayVal[zR * rowWidth] < 0) {return;}  // command line
  
  // the STEP value will not be modified
  zR = zR * rowWidth;
  for (int zP = 0; zP < (rowWidth - 1); zP++) {
    ArrayVal[zR + zP] = ArrayVal[zR + zP - rowWidth] + ((ArrayVal[zR + zP + rowWidth] - ArrayVal[zR + zP - rowWidth]) / 2);
  }
  msgTx = "Intermediate row created";
}

// ----------------------------------------------------------------------

int mapInt(int x,int in_min,int in_max,int out_min,int out_max) {
  // integer version of Processing floating point map() function
  if ((in_max - in_min) == 0) {in_max += 1;}  // trap divide by zero error
  return ((x - in_min) * (out_max - out_min) / (in_max - in_min)) + out_min;
}

// ----------------------------------------------------------------------

void moveLinkedVal(int zS) {
  // move the linked slider by the same amount if possible
  if (AngDif == 0) {return;}  // no change so abort function
  
  int zT0 = 0;
  if (Link[zS] < 2) {
    // single linked pair so adjust associated target
    switch(zS) {
      case 0: zT0 = FL1; break;  // FR1 linked to FL1
      case 1: zT0 = FL2; break;  // FR2 linked to FL2
      case 2: zT0 = FL3; break;  // FR3 linked to FL3
      
      case 3: zT0 = FR1; break;  // FL1 linked to FR1
      case 4: zT0 = FR2; break;  // FL2 linked to FR2
      case 5: zT0 = FR3; break;  // FL3 linked to FR3
  
      case 6: zT0 = RL1; break;  // RR1 linked to RL1
      case 7: zT0 = RL2; break;  // RR2 linked to RL2
      case 8: zT0 = RL3; break;  // RR3 linked to RL3
  
      case 9: zT0 = RR1; break;  // RL1 linked to RR1
      case 10: zT0 = RR2; break;  // RL2 linked to RR2
      case 11: zT0 = RR3; break;  // RL3 linked to RR3
    }
    adjustLinkVal(zT0);
  } else {
    // multi-linked values, so adjust three other values
    int zT1 = 0; int zT2 = 0;
    switch(zS) {
      case 0: zT0 = FL1; zT1 = RR1; zT2 = RL1; break;  // FR1 linked to FL1, RR1, RL1
      case 1: zT0 = FL2; zT1 = RR2; zT2 = RL2; break;  // FR2 linked to FL2, RR2, RL2
      case 2: zT0 = FL3; zT1 = RR3; zT2 = RL3; break;  // FR3 linked to FL3, RR3, RL3
      
      case 3: zT0 = FR1; zT1 = RR1; zT2 = RL1; break;  // FL1 linked to FR1, RR1, RL1
      case 4: zT0 = FR2; zT1 = RR2; zT2 = RL2; break;  // FL2 linked to FR2, RR2, RL2
      case 5: zT0 = FR3; zT1 = RR3; zT2 = RL3; break;  // FL3 linked to FR3, RR3, RL3
  
      case 6: zT0 = RL1; zT1 = FR1; zT2 = FL1; break;  // RR1 linked to RL1, FR1, FL1
      case 7: zT0 = RL2; zT1 = FR2; zT2 = FL2; break;  // RR2 linked to RL2, FR2, FL2
      case 8: zT0 = RL3; zT1 = FR3; zT2 = FL3; break;  // RR3 linked to RL3, FR3, FL3
  
      case 9: zT0 = RR1; zT1 = FR1; zT2 = FL1; break;  // RL1 linked to RR1, FR1, FL1
      case 10: zT0 = RR2; zT1 = FR2; zT2 = FL2; break;  // RL2 linked to RR2, FR2, FL2
      case 11: zT0 = RR3; zT1 = FR3; zT2 = FL3; break;  // RL3 linked to RR3, FR3, FL3
    }
    adjustLinkVal(zT0); adjustLinkVal(zT1); adjustLinkVal(zT2);
  }
  usbPortWrite(msgTx);
}

// ----------------------------------------------------------------------

void moveRowBlockFT(int zF, int zT, int zN) {
  // moves row block from zF up to zT, block row size zN, then clears remainder
  copyRowBlockFT(zF,zT,zN);
  if (zT < zF) {
    // row block was copied up so clear region below copied area
    ClearRowBlockFT(zT + zN, zF + zN -1);
  } else {
    // row block was copied down so clear region above it
    ClearRowBlockFT(zF, zT -1);
  }
}

// ----------------------------------------------------------------------

void moveSelArrayUp() {
  // called when a selected row has been deleted
  int zT = 0;
  for (int zS = 1; zS < arrayMax; zS++) {
    Select[zT] = Select[zS]; zT++;
  } Select[arrayMax - 1] = false;
}

// ----------------------------------------------------------------------

void moveSelDwn() {
  // move selected rows towards rowNum
  if ((SelCnt < 1) || (rowNum < 2)) {return;}
  usbEn = false;  // avoid any unintentional movement
  
  cursor(WAIT);
  boolean zS = false;
  for (int zR = rowNum - 2; zR >= 0; zR--) {
    if (Select[zR] && !Select[zR + 1]) {
      // current line selected and next is not, so swap them
      int zF = zR * rowWidth; int zT = (zR + 1) * rowWidth; int zV;
      // swap the row angle and steps
      for (int zP = 0; zP < rowWidth; zP++) {
        zV = ArrayVal[zT + zP]; ArrayVal[zT + zP] = ArrayVal[zF + zP];
        ArrayVal[zF + zP] = zV; 
      } Select[zR + 1] = true; Select[zR] = false; zS = true;
    }
  } 
  if (zS && (rowPnt < (rowNum-1))) {rowPnt++;}
  rateFlag = true; drawFlag = 1;
  cursor(ARROW);
}

// ----------------------------------------------------------------------

void moveSelUp() {
  // move selected rows towards row zero
  if ((SelCnt < 1) || (rowNum < 2)) {return;}
  usbEn = false;  // avoid any unintentional movement
  
  cursor(WAIT);
  boolean zS = false;
  for (int zR = 1; zR < rowNum; zR++) {
    if (Select[zR] && !Select[zR - 1]) {
      // current line selected and previous not, so swap them
      int zF = zR * rowWidth; int zT = (zR - 1) * rowWidth; int zV;
      // swap the row angle and steps
      for (int zP = 0; zP < rowWidth; zP++) {
        zV = ArrayVal[zT + zP]; ArrayVal[zT + zP] = ArrayVal[zF + zP];
        ArrayVal[zF + zP] = zV; 
      } Select[zR-1] = true; Select[zR] = false; zS = true;
    }
  } 
  if (zS && (rowPnt > 0)) {rowPnt--;}
  rateFlag = true; drawFlag = 1;
  cursor(ARROW);
}

// ----------------------------------------------------------------------

void MultiDecX(int zS1,int zS2) {
  // multimove X decrement on servos zS1 and zS2
}

// ----------------------------------------------------------------------

void MultiDecY(int zS1,int zS2) {
  // multimove Y decrement on servos zS1 and zS2
  int zA = getAngVal(AngVal[zS1]);
  boolean zDF = false;  // drawFlag
  if (zA > AngVLL[zS1]) {
    // work on tier 1 angles first
    AngLast = zA; AngVal[zS1] = zA - 1; AngDif = -1;
    // update value in memory array
    ArrayVal[(rowPnt * rowWidth) + zS1] = AngVal[zS1];
    // send value to usb port
    msgTx = "ST" + str(Ch[zS1]) + "." +"SA" + str(AngVal[zS1]) + ".";
    if (Link[zS1] > 0) {moveLinkedVal(zS1);}
    else {setSliderX(); usbPortWrite(msgTx); rateFlag = true; zDF = false;}
    if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
  }

  // now do tier 2 angles if tier 1 is divisible by 2, ie even
  if (((zA % 2) != 0) && (zA != 0)) {
    // tier 1 angle is divisible by 2 so adjust tier 2 angle
    // using this method tier 2 wiull be adjusted twice as fast as tier 2
    zA = getAngVal(AngVal[zS2]);
    if (zA > AngVLL[zS2]) {
      // work on tier 2 angles
      AngLast = zA; AngVal[zS2] = zA - 1; AngDif = -1;
      // update value in memory array
      ArrayVal[(rowPnt * rowWidth) + zS2] = AngVal[zS2];
      // send value to usb port
      msgTx = "ST" + str(Ch[zS2]) + "." +"SA" + str(AngVal[zS2]) + ".";
      if (Link[zS2] > 0) {moveLinkedVal(zS2);}
      else {setSliderX(); usbPortWrite(msgTx); rateFlag = true; zDF = false;}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
    }
  }
  if (zDF) {drawFlag = 1;}
}

// ----------------------------------------------------------------------

void MultiIncX(int zS1,int zS2) {
  // multimove X increment on servos zS1 and zS2
}

// ----------------------------------------------------------------------

void MultiIncY(int zS1,int zS2) {
  // multimove Y increment on servos zS1 and zS2
  int zA = getAngVal(AngVal[zS1]);
  boolean zDF = false;  // drawFlag
  if (zA < AngVUL[zS1]) {
    // work on tier 1 angles first
    AngLast = zA; AngVal[zS1] = zA + 1; AngDif = 1;
    // update value in memory array
    ArrayVal[(rowPnt * rowWidth) + zS1] = AngVal[zS1];
    // send value to usb port
    msgTx = "ST" + str(Ch[zS1]) + "." +"SA" + str(AngVal[zS1]) + ".";
    if (Link[zS1] > 0) {moveLinkedVal(zS1);}
    else {setSliderX(); usbPortWrite(msgTx);}
    if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
    rateFlag = true; zDF = false;
  }

  // now do tier 2 angles if tier 1 is divisible by 2, ie even
  // note that tier 2 has a max value of 90 to straighten the leg
  if (((zA % 2) != 0) && (zA != 0)) {
    // tier 1 angle is divisible by 2 so adjust tier 2 angle
    // using this method tier 2 wiull be adjusted twice as fast as tier 2
    zA = getAngVal(AngVal[zS2]);
    if (zA != 90) {
      // work on tier 2 angles
      if (zA > 90) {AngDif = -1;} else {AngDif = 1;}
      AngLast = AngVal[zS2]; AngVal[zS2] = zA + AngDif;
      // update value in memory array
      ArrayVal[(rowPnt * rowWidth) + zS2] = AngVal[zS2];
      // send value to usb port
      msgTx = "ST" + str(Ch[zS2]) + "." +"SA" + str(AngVal[zS2]) + ".";
      if (Link[zS2] > 0) {moveLinkedVal(zS2);}
      else {setSliderX(); usbPortWrite(msgTx);}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
      rateFlag = true; zDF = false;
    }
  }
  if (zDF) {drawFlag = 1;}
}

// ----------------------------------------------------------------------

void MultiMoveXY(int zS1,int zS2) {
  // multimove in X/Y on servos zS1 and zS2
  // the movement is relative to the click mXD/mYD points
  // Temp[] holds the active row when the mouse was clicked
  
  int zA = getAngVal(Temp[zS1]);
  int zB = getAngVal(Temp[zS2]);
  int zD = 0;           // AngDif
  boolean zDF = false;  // drawFlag
  int zP = 0;  // new angle
  if (MultiPnt == 0) {AngLast = zA; AngLast0 = zB;}
  MultiPnt = 1;  Multi0 = zS1; Multi1 = zS2;  // prevent mouseDown() interaction
  // consider Y axis first
  if (mY > mYD) {
    // mouse is dragged down so increase angles
    zD = (mY - mYD)/3;  // mouse Y determines change
    zP = zA + zD;
    if ((zP >= AngVLL[zS1]) && (zP <= AngVUL[zS1])) {
      // work on tier 1 angles first
      AngVal[zS1] = zP;
      AngDif = zP - AngLast; AngLast = zP;
      // update value in memory array
      ArrayVal[(rowPnt * rowWidth) + zS1] = AngVal[zS1];
      // send value to usb port
      msgTx = "ST" + str(Ch[zS1]) + "." +"SA" + str(AngVal[zS1]) + ".";
      if (Link[zS1] > 0) {moveLinkedVal(zS1);}
      else {setSliderX(); usbPortWrite(msgTx);}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
      rateFlag = true; zDF = true;
    }
    // now do tier 2 angles at half the displacement
    // also add in the x-axis component
    zD /= 2; zD -= (mX - mXD)/5;
    zP = zB + zD;
    if ((zP >= AngVLL[zS2]) && (zP <= AngVUL[zS2])) {
      // work on tier 2 angles
      AngVal[zS2] = zP;
      AngDif = zP - AngLast0; AngLast0 = zP;
      // update value in memory array
      ArrayVal[(rowPnt * rowWidth) + zS2] = AngVal[zS2];
      // send value to usb port
      msgTx = "ST" + str(Ch[zS2]) + "." +"SA" + str(AngVal[zS2]) + ".";
      if (Link[zS2] > 0) {moveLinkedVal(zS2);}
      else {setSliderX(); usbPortWrite(msgTx);}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
      rateFlag = true; zDF = true;
    }
  } else {
    // mouse is dragged up so decrease angles
    zD = (mY - mYD)/3;  // mouse Y determines change
    zP = zA + zD;
    if ((zP >= AngVLL[zS1]) && (zP <= AngVUL[zS1])) {
      // work on tier 1 angles first
      AngVal[zS1] = zP;
      AngDif = zP - AngLast; AngLast = zP;
      // update value in memory array
      ArrayVal[(rowPnt * rowWidth) + zS1] = AngVal[zS1];
      // send value to usb port
      msgTx = "ST" + str(Ch[zS1]) + "." +"SA" + str(AngVal[zS1]) + ".";
      if (Link[zS1] > 0) {moveLinkedVal(zS1);}
      else {setSliderX(); usbPortWrite(msgTx);}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
      rateFlag = true; zDF = true;
    }
    // now do tier 2 angles at half the displacement
    zD /= 2; zD -= (mX - mXD)/5;
    zP = zB + zD;
    if ((zP >= AngVLL[zS2]) && (zP <= AngVUL[zS2])) {
      // work on tier 2 angles
      AngVal[zS2] = zP;
      AngDif = zP - AngLast0; AngLast0 = zP;
      // update value in memory array
      ArrayVal[(rowPnt * rowWidth) + zS2] = AngVal[zS2];
      // send value to usb port
      msgTx = "ST" + str(Ch[zS2]) + "." +"SA" + str(AngVal[zS2]) + ".";
      if (Link[zS2] > 0) {moveLinkedVal(zS2);}
      else {setSliderX(); usbPortWrite(msgTx);}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
      rateFlag = true; zDF = true;
    }
  }
  if (zDF) {drawFlag = 1;}
}


// ######################################################################


void newRow() {
  // include a new row either before or after the current line if memory permits
  // correct goto commands if they are any that are affected
  if (rowNum == rowDepth) {return;}  // array memory is exhausted
  
  if (mB == LEFT) {
    // add in a new row after the current one
    if (rowPnt != (rowNum-1)) {
      // current row is not the last one in the list so copy values down to create a new line
      for (int zR = rowNum; zR > rowPnt; zR--) {
        int zT = zR * rowWidth; int zF = (zR-1) * rowWidth;
        for (int zP = 0; zP < rowWidth; zP++) {
          ArrayVal[zT + zP] = ArrayVal[zF + zP];
        }
      }
    }
    // now create the new row by copying the current one
    // if that was a command line, create a new row
    if (ArrayVal[rowPnt * rowWidth] < 0) {
      // current row is a command line, so create a default row
      ClearRow(rowPnt + 1);
    } else {
      // duplicate line values
      int zT = (rowPnt + 1) * rowWidth; int zF = rowPnt * rowWidth;
      for (int zP = 0; zP < rowWidth; zP++) {
        ArrayVal[zT + zP] = ArrayVal[zF + zP];
      }
    }
    rowNum++;  // increase the number of rows in the list
    rowPnt++; 
    checkGotoCmdIns(rowPnt,1);  // check to correct Goto commands
  } else {
    // add in a new row before the current one by copying the current rows down one row
    // this process will naturally leave a copy from the original
    for (int zR = rowNum; zR > rowPnt; zR--) {
      int zT = zR * rowWidth; int zF = (zR-1) * rowWidth;
      for (int zP = 0; zP < rowWidth; zP++) {
        ArrayVal[zT + zP] = ArrayVal[zF + zP];
      }
    }
    if (ArrayVal[rowPnt * rowWidth] < 0) {
      // current row is a command line, so create a default row
      ClearRow(rowPnt);
    }
    rowNum++;  // increase the number of rows in the list
    checkGotoCmdIns(rowPnt,1);  // check to correct Goto commands
  }
  // if the SHIFT key is down then determine intermediate points
  if (keySHIFT) {makeIntermediate(rowPnt);}
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

String Num3Dig(int zV) {
  // returns a string number 3 characters wide
  String zNum = "";
  if (zV < 0) {zNum = " ";}  // command line
  else if (zV < 10) {zNum = "  ";}
  else if (zV < 100) {zNum = " ";}
  zNum = zNum + str(zV);
  return zNum;
}


// ######################################################################


void pasteRows() {
  // paste the contents of the copy buffer
  usbEn = false;  // avoid any unintentional movement
  msgTx = "No copied data"; rateFlag = true; rateFlag = true; drawFlag = 1;
  if (CopyCnt < 1) {return;}
  
  cursor(WAIT);
  int zF = -1; int zR = 0; int zS = 0; int zT = -1;
  if (SelCnt > 0) {
    // one or more rows selected so create correct gap and overwrite them
    // if there is several selected regions, only the first is overwritten
    if (SelCnt == rowNum) {
      // all of the rows are selected so reset rows and copy in data
      ClearRowBlockFT(0, rowNum - 1); copyCopyTo(0);
      rowPnt = 0; rowNum = CopyCnt;
    } else {
      // not all rows are selected, so find start of selection and count them
      for (zR = 0; zR < rowNum; zR++) {
        if (Select[zR]) {
          // a row is selected, is it the 1st? 
          zS++; if (zF < 0) {zF = zR;}  // zF points to the 1st selected row
        } else {
          // row is not selected, were previous selected rows detected?
          // if so, zT points at 1st non-selected row after selection
          if (zF >= 0) {if (zT < 0) {zT = zR; break;}}  // yes, so at end of selection
        }
      }
      // zF points to the 1st selected row
      // zT points to the 1st non-selected row after the selection
      // zS is the number of selected lines to be replaced
      if (zT >= 0) {
        // if zT != -1 then some data after selection
        // now create suitable gap size
        if (zS > CopyCnt) {
          // more selected rows than copy can fill, so close the gap
          moveRowBlockFT(zT, zF + CopyCnt, rowNum - zT);
        } else if (zS < CopyCnt) {
          // fewer selected rows than copy, so increase the gap
          moveRowBlockFT(zT, zT + CopyCnt - zS, rowNum - zT);
        }
      }
      // gap now matches copy count, so copy into it, starting at row zF
      zT = CopyCnt; if ((zT + zF) >= rowDepth) {zT = rowDepth - zF - 1;}
      copyCopyTo(zF);  // copy in starting at row zF
      if (zS > CopyCnt) {checkGotoCmdDel(zT,zS-CopyCnt);}
      if (zS < CopyCnt) {checkGotoCmdIns(zT,CopyCnt-zS);}
      rowNum = rowNum - SelCnt + CopyCnt;
    }
    ClearSelect();
  } else {
    // no selected rows, so create a gap after current row and copy in
    // ensure there are enough rows for the copy data
    int zCnt = CopyCnt;
    if (rowPnt == (rowNum - 1)) {
      // last line so simply copy onto end
      copyCopyTo(rowPnt + 1); rowNum = rowNum + CopyCnt;
      if (rowNum > rowDepth) {rowNum = rowDepth;}
      println("Add to end");
    } else if ((rowDepth - rowNum) <= CopyCnt) {
      // copy memory is larger than rows left so limit copy and simply
      // overwrite data area
      zCnt = rowDepth - rowNum;
      copyCopyTo(rowPnt + 1); rowNum = rowDepth;
      println("Too big!");
    } else {
      // copied data will fit so move data down to create a gap
      moveRowBlockFT(rowPnt + 1, rowPnt + CopyCnt +1, rowNum - rowPnt);
      copyCopyTo(rowPnt + 1); rowNum = rowNum + CopyCnt;
      if (rowNum > rowDepth) {rowNum = rowDepth;}
      checkGotoCmdIns(rowPnt,CopyCnt);
      println("It fits!");
    }
  }
  msgTx = "Copied data was pasted";
  if (rowNum == rowDepth) {msgTx = "Memory is consumed!";}
  cursor(ARROW);
}

// ----------------------------------------------------------------------

void playConst() {
  // send contents of memory and then execute it
  int zF = 0;           // first row of data
  int zL = rowNum - 1;  // last row of data
  drawFlag = 1;
  if (usbEn) {
    msgTx = "MC."; usbPortWrite(msgTx);
    sendSpeed();
    // check to see if rows are selected
    if (SelCnt > 0) {
      // rows are selected so find first selected block
      // if multiple blocks or rows are selected we only play the first
      zF = -1;
      for (int zR = 0; zR < rowNum; zR++) {
        if (Select[zR]) {
          if (zF < 0) {zF = zR;}  // first selected row
          zL = zR;                // track last selected row
        } else {if (zF >= 0) {break;}}  // exit at end of selected block
      }
    }
    // now export the data to RoboDog
    for (int zR = zF; zR <= zL; zR++) {
      sendRowData(zR,zR,0);
    }
    msgTx = "MF" + str(zF) + "."; usbPortWrite(msgTx);
    msgTx = "MR."; usbPortWrite(msgTx); playMode = 2;
    msgTx = "Play repeatedly";
  } else if (Scrn == 2) {
    // simulate the walking process by moving down through the list
    Walk_Init();
    WalkSim = 2;  // set Walk to continuous mode
    playMode = 2;
  }
}

// ----------------------------------------------------------------------

void playOnce() {
  // if usbEn then send contents of memory and execute it
  int zF = 0;           // first row of data
  int zL = rowNum - 1;  // last row of data
  drawFlag = 1;
  if (usbEn) {
    msgTx = "MC."; usbPortWrite(msgTx);
    sendSpeed();
    // check to see if rows are selected
    if (SelCnt > 0) {
      // rows are selected so find first selected block
      // if multiple blocks or rows are selected we only play the first
      zF = -1;
      for (int zR = 0; zR < rowNum; zR++) {
        if (Select[zR]) {
          if (zF < 0) {zF = zR;}  // first selected row
          zL = zR;                // track last selected row
        } else {if (zF >= 0) {break;}}  // exit at end of selected block
      }
    }
    // now export the data to RoboDog
    for (int zR = zF; zR <= zL; zR++) {
      sendRowData(zR,zR,0);
    }
    msgTx = "MF" + str(zF) + "."; usbPortWrite(msgTx);
    msgTx = "MO."; usbPortWrite(msgTx); playMode = 1;
    msgTx = "Play once";
  } else if (Scrn == 2) {
    // simulate the walking process by moving down through the list
    Walk_Init();
    WalkSim = 1;  // set Walk to once mode
    playMode = 1;
  }
}

// ----------------------------------------------------------------------

void playStop() {
  // send a command to stop any movement
  drawFlag = 1;
  if (usbEn) {
    msgTx = "MS.";
    usbPortWrite(msgTx);
    rowPnt = 0;  // return to the start row in the list
    sendRowData(rowPnt,rowPnt,1);  // move to first row angles
  } else if (Scrn == 2) {
    // stop Walk simulator
    WalkSim = 0; msgTx = "Sim Stopped";
  } playMode = 0;
}

// ----------------------------------------------------------------------

void primeTemp() {
  // prime the Temp[] array with the default angles
  Temp[ 0] =  0;  // FR1
  Temp[ 1] =  0;  // FR2
  Temp[ 2] = 95;  // FR3
  Temp[ 3] =  0;  // FL1
  Temp[ 4] =  0;  // FL2
  Temp[ 5] = 85;  // FL3
  Temp[ 6] =  0;  // RR1
  Temp[ 7] =  0;  // RR2
  Temp[ 8] = 95;  // RR3
  Temp[ 9] =  0;  // RL1
  Temp[10] =  0;  // RL2
  Temp[11] = 85;  // RL3
  Temp[12] = 30;  // Stp
}


// ######################################################################


void readRowSlider() {
  // determine row slider position and display accordingly
}

// ----------------------------------------------------------------------

void readXSlider(int zS) {
  // determine X-slider position and send value
  servoTgt = zS;        // record which slider was clicked
  AngLast = AngVal[zS]; // remember the current value
  if (sliderDwn < 1) {getXSliderOffset(zS);}  // just been clicked
  JoyXval[zS] = mX + mXOff; JoyXval[zS] = max(JoyXval[zS], JoyXLL[zS]);
  JoyXval[zS] = min(JoyXval[zS], JoyXUL[zS]);

  // remove servo rate & power offset settings
  AngVal[zS] = getAngVal(AngVal[zS]);
  switch(zS) {
    case 0:  // FR1 
      AngVal[zS] = AngVLL[zS] + (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 1: // FR2
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 2: // FR3
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 3: //FL1
      AngVal[zS] = AngVLL[zS] + (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 4: // FL2
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 5: // FL3
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 6: //RR1
      AngVal[zS] = AngVLL[zS] + (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 7: // RR2
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 8: // RR3
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 9: // RL1
      AngVal[zS] = AngVLL[zS] + (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 10: // RL2
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 11: // RL3
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
  } 
  
  if (AngLast != AngVal[zS]) {
    AngDif = AngVal[zS] - AngLast;  // note difference in case of linked sliders
    AngLast = AngVal[zS];
    // update value in memory array
    ArrayVal[(rowPnt * rowWidth) + zS] = AngVal[zS];
    // send value to usb port
    msgTx = "ST" + str(Ch[zS]) + "." +"SA" + str(AngVal[zS]) + ".";
    usbPortWrite(msgTx); rateFlag = true; drawFlag = 1;
    if (Link[zS] > 0) {moveLinkedVal(zS);}
  }
  sliderDwn = 1; // set slider down flag to prevent inc/dec from false triggering
}

// ----------------------------------------------------------------------

void recallM0() {
  // recall data stored in M0 if any
  msgTx = "Data store M0 is empty"; rateFlag = true; drawFlag = 1;
  if (M0 < 1) {return;}  // no data stored so exit
  
  rowNum = M0; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    ArrayVal[zP] = Mem0[zP];
  } 
  rowPnt = 0; usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  ClearSelect();
  msgTx = "Data loaded from M0";
}

// ----------------------------------------------------------------------

void recallM1() {
  // recall data stored in M1 if any
  msgTx = "Data store M1 is empty"; rateFlag = true; drawFlag = 1;
  if (M1 < 1) {return;}  // no data stored so exit

  rowNum = M1; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    ArrayVal[zP] = Mem1[zP];
  } 
  rowPnt = 0; usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  ClearSelect(); drawFlag = 1;
  msgTx = "Data loaded from M1";
}

// ----------------------------------------------------------------------

void recallM2() {
  // recall data stored in M2 if any
  msgTx = "Data store M2 is empty"; rateFlag = true; drawFlag = 1;
  if (M2 < 1) {return;}  // no data stored so exit

  rowNum = M2; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    ArrayVal[zP] = Mem2[zP];
  } 
  rowPnt = 0; usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  ClearSelect(); drawFlag = 1;
  msgTx = "Data loaded from M2";
}

// ----------------------------------------------------------------------

void recallM3() {
  // recall data stored in M0 if any
  msgTx = "Data store M3 is empty"; rateFlag = true; drawFlag = 1;
  if (M3 < 1) {return;}  // no data stored so exit

  rowNum = M3; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    ArrayVal[zP] = Mem3[zP];
  } 
  rowPnt = 0; usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  ClearSelect(); drawFlag = 1;
  msgTx = "Data loaded from M3";
}

// ----------------------------------------------------------------------

void resetDegSteps() {
  // degrements the DegStep value
  DegStep = 3; drawFlag = 1;
}

// ----------------------------------------------------------------------

void resetLftFlds() {
  // set the values in the left fields to their defaults
  T1maxAngLft = 140;  // Walk Trainer T1 maxAng left side
  T2maxAngLft = 90;   // Walk Trainer T2 maxAng left side
  T2midAngLft = 70;   // Walk Trainer T2 midAng left side
  T2minAngLft = 50;   // Walk Trainer T2 minAng left side
}

// ----------------------------------------------------------------------

void resetRhtFlds() {
  // set the values in the rightt fields to their defaults
  T1maxAngRht = 140;  // Walk Trainer T1 maxAng right side
  T2maxAngRht = 90;   // Walk Trainer T2 maxAng right side
  T2midAngRht = 70;   // Walk Trainer T2 midAng right side
  T2minAngRht = 50;   // Walk Trainer T2 minAng right side
}

// ----------------------------------------------------------------------

void resetSpeed() {
  // set the Speed to the default value
  Speed = SpeedDef; drawFlag = 1; sendSpeed();
}

// ----------------------------------------------------------------------

void resetStepRates() {
  // reset all start rate profiles
  for (int zR = 0; zR < rowNum; zR++) {
    int zP = (zR * rowWidth) + 12;
    int zS = ArrayVal[zP];
    // strip off rate values
    if (zS > 600) {zS = zS - 600;}
    else if (zS > 400) {zS = zS - 400;}
    else if (zS > 200) {zS = zS - 200;}
    ArrayVal[zP] = zS;
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void RetRowFromTemp() {
  // return the angle values in Temp[] store to the current row
  int zRP = rowPnt * rowWidth;
  for (int zP = 0; zP < (rowWidth - 1); zP++) {
    ArrayVal[zP + zRP] = Temp[zP];
  }
}

// ----------------------------------------------------------------------

void reverseSelRows() {
  // reverse the order of the selected rows
  // first find the range of selected rows, which need to be contiguous
  if ((SelCnt < 1) || (rowNum < 2)) {return;}
  usbEn = false;  // avoid any unintentional movement

  int zF = -1; int zL = -1; int zS;
  for (zS = 0; zS < rowNum; zS++) {
    if (Select[zS]) {
      // record the 1st selected row  
      if (zF < 0) {zF = zS;}
      // collect selected rows until break or end
      else {zL = zS;}
    } else {
      // exit if we find a n unselected row after finding one or more
      if (zF >= 0) {break;}
    }
  }
  rateFlag = true;drawFlag = 1;
  // did we only find one selected row?
  if (zL < 0) {msgTx = "Error. Only 1 row selected."; return;}
  
  // we only need to perform the operation on half the number of rows
  int zN = zL - zF + 1;
  msgTx = str(zN) + " rows reversed.";
  zN = zF + (zN/2);
  for (zS = zF; zS < zN; zS++) {
    //println(str(zS) + "   " + str(zL));
    int zRF = zS * rowWidth; int zRL = zL * rowWidth;
    for (int zP = 0; zP < rowWidth; zP++) {
      Temp[zP] = ArrayVal[zRF + zP];
      ArrayVal[zRF + zP] = ArrayVal[zRL + zP];
      ArrayVal[zRL + zP] = Temp[zP];
    }
    // progressively move the last row pointer towards the middle
    zL--;
  }
}


// ######################################################################


void sendButton(String zS) {
  // called when a button is clicked on the scrn
  msgTx = zS; usbPortWrite(msgTx);
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void sendEcho() {
  // we send Ech status regardless of usbEn status
  if (Echo) {
    msgTx = "SR1.";
  } else {
    msgTx = "SR0.";
  }
  if (usbEn) {usbPortWrite(msgTx);}
  else if (comFlag > 0) {usbPort.write(msgTx);}
}

// ----------------------------------------------------------------------

void sendRowData(int zR, int zA, int zEx) {
  // send the contents of a row as a group of individual angles, as:
  //
  // MLnn.n,n,. - start a memory load sequence, values separate by a comma, terminated by a period '.'
  //
  // called on demand or when clearing the array or switching rows
  if (zEx == 1) {sendSpeed();}
  msgTx = "ML" + str(zA) + ".";  // set RoboDog memory pointer
  int zP = zR * rowWidth;
  for (int zC = 0; zC < rowWidth; zC++){
    msgTx = msgTx + str(ArrayVal[zP + zC]) + ",";
  }
  if (zEx == 1) {
    msgTx += "MX" + str(zA) + ".";  // send execute command
  }
  usbPortWrite(msgTx);
  msgTx = msgTx.substring(0,20) + "...";
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void sendSpeed() {
  // called to send Speed value to arduinio
  msgTx = "SP" + str(Speed) + "."; usbPortWrite(msgTx);
}

// ----------------------------------------------------------------------

void setAngValFromRow() {
  // called when a row changes or memory is cleared, this uses memory row
  // data to update the angle values and sliders
  int zP = rowPnt * rowWidth;
  if (ArrayVal[zP] < 0) {return;}  // command row so ignore
  
  AngVal[FR1] = ArrayVal[zP]; zP++;
  AngVal[FR2] = ArrayVal[zP]; zP++;
  AngVal[FR3] = ArrayVal[zP]; zP++;
  AngVal[FL1] = ArrayVal[zP]; zP++;
  AngVal[FL2] = ArrayVal[zP]; zP++;
  AngVal[FL3] = ArrayVal[zP]; zP++;
  AngVal[RR1] = ArrayVal[zP]; zP++;
  AngVal[RR2] = ArrayVal[zP]; zP++;
  AngVal[RR3] = ArrayVal[zP]; zP++;
  AngVal[RL1] = ArrayVal[zP]; zP++;
  AngVal[RL2] = ArrayVal[zP]; zP++;
  AngVal[RL3] = ArrayVal[zP];
  setSliderX();
}

// ----------------------------------------------------------------------

void setGndY() {
  // set the GndY ground value based on the latest GndYM
  GndY = GndYM + 12;
}

// ----------------------------------------------------------------------

void setRowData(int zRow,int zFR1,int zFR2,int zFR3,int zFL1,int zFL2,int zFL3,int zRR1,int zRR2,int zRR3,int zRL1,int zRL2,int zRL3,int zStp) {
  // called to load data directly into row zRow
  zRow = zRow * rowWidth;
  ArrayVal[zRow] = zFR1; zRow++;
  ArrayVal[zRow] = zFR2; zRow++;
  ArrayVal[zRow] = zFR3; zRow++;
  ArrayVal[zRow] = zFL1; zRow++;
  ArrayVal[zRow] = zFL2; zRow++;
  ArrayVal[zRow] = zFL3; zRow++;
  ArrayVal[zRow] = zRR1; zRow++;
  ArrayVal[zRow] = zRR2; zRow++;
  ArrayVal[zRow] = zRR3; zRow++;
  ArrayVal[zRow] = zRL1; zRow++;
  ArrayVal[zRow] = zRL2; zRow++;
  ArrayVal[zRow] = zRL3; zRow++;
  ArrayVal[zRow] = zStp;
}

// ----------------------------------------------------------------------

void setSliderX() {
  // set the X-slider position for current AngVal[] values
  // note that some sliders work in reverse, as their corresponding servo is reversed
  // angles can also contain Pwr status, which has to be ignored
  // FR1
  JoyXval[0] = JoyXLL[0] + ((getAngVal(AngVal[0]) - AngVLL[0]) * (JoyXUL[0] - JoyXLL[0]))/(AngVUL[0] - AngVLL[0]);
  // FR2
  JoyXval[1] = JoyXUL[1] - ((getAngVal(AngVal[1]) - AngVLL[1]) * (JoyXUL[1] - JoyXLL[1]))/(AngVUL[1] - AngVLL[1]);
  // FR3
  JoyXval[2] = JoyXUL[2] - ((getAngVal(AngVal[2]) - AngVLL[2]) * (JoyXUL[2] - JoyXLL[2]))/(AngVUL[2] - AngVLL[2]);
  // FL1
  JoyXval[3] = JoyXLL[3] + ((getAngVal(AngVal[3]) - AngVLL[3]) * (JoyXUL[3] - JoyXLL[3]))/(AngVUL[3] - AngVLL[3]);
  // FL2
  JoyXval[4] = JoyXUL[4] - ((getAngVal(AngVal[4]) - AngVLL[4]) * (JoyXUL[4] - JoyXLL[4]))/(AngVUL[4] - AngVLL[4]);
  // FL3
  JoyXval[5] = JoyXUL[5] - ((getAngVal(AngVal[5]) - AngVLL[5]) * (JoyXUL[5] - JoyXLL[5]))/(AngVUL[5] - AngVLL[5]);
  // RR1
  JoyXval[6] = JoyXLL[6] + ((getAngVal(AngVal[6]) - AngVLL[6]) * (JoyXUL[6] - JoyXLL[6]))/(AngVUL[6] - AngVLL[6]);
  // RR2
  JoyXval[7] = JoyXUL[7] - ((getAngVal(AngVal[7]) - AngVLL[7]) * (JoyXUL[7] - JoyXLL[7]))/(AngVUL[7] - AngVLL[7]);
  // RR3
  JoyXval[8] = JoyXUL[8] - ((getAngVal(AngVal[8]) - AngVLL[8]) * (JoyXUL[8] - JoyXLL[8]))/(AngVUL[8] - AngVLL[8]);
  // RL1
  JoyXval[9] = JoyXLL[9] + ((getAngVal(AngVal[9]) - AngVLL[9]) * (JoyXUL[9] - JoyXLL[9]))/(AngVUL[9] - AngVLL[9]);
  // RL2
  JoyXval[10] = JoyXUL[10] - ((getAngVal(AngVal[10]) - AngVLL[10]) * (JoyXUL[10] - JoyXLL[10]))/(AngVUL[10] - AngVLL[10]);
  // RL3
  JoyXval[11] = JoyXUL[11] - ((getAngVal(AngVal[11]) - AngVLL[11]) * (JoyXUL[11] - JoyXLL[11]))/(AngVUL[11] - AngVLL[11]);
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void storeM0() {
  // transfer row data to memory store M0
  cursor(WAIT);
  M0 = rowNum; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    Mem0[zP] = ArrayVal[zP];
  } drawFlag = 1;
  msgTx = "Data stored in M0";
}

// ----------------------------------------------------------------------

void storeM1() {
  // transfer row data to memory store M1
  cursor(WAIT);
  M1 = rowNum; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    Mem1[zP] = ArrayVal[zP];
  } drawFlag = 1;
  msgTx = "Data stored in M1";
}

// ----------------------------------------------------------------------

void storeM2() {
  // transfer row data to memory store M2
  cursor(WAIT);
  M2 = rowNum; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    Mem2[zP] = ArrayVal[zP];
  } drawFlag = 1;
  msgTx = "Data stored in M2";
}

// ----------------------------------------------------------------------

void storeM3() {
  // transfer row data to memory store M0
  cursor(WAIT);
  M3 = rowNum; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    Mem3[zP] = ArrayVal[zP];
  } drawFlag = 1;
  msgTx = "Data stored in M3";
}


// ######################################################################


void toggleEcho() {
  // toggles the comms echo function
  Echo = !Echo; sendEcho();
  if (!Echo) {msgRx = "Arduino is muted";}
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void toggleLink(int zA,int zB) {
  // toggles the link between servo zA and servo zB
  if (Link[zA] == 0) {
    // set the links even if the values aren't the same
    Link[zA] = 1; Link[zB] = 1;
  } else {Link[zA] = 0; Link[zB] = 0;}
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void toggleMultiLink(int zA) {
  // toggles multiple links in servo rows 1, 2 and 3
  drawFlag = 1;
  if (zA == FR1) {
    if (Link[FR1] < 2) {
      // multilink is not active so set all four links by adding 2
      Link[FR1] += 2; Link[FL1] += 2; Link[RR1] += 2; Link[RL1] += 2;
    } else {
      // multilink is active so set all four links by subtracting 2
      Link[FR1] -= 2; Link[FL1] -= 2; Link[RR1] -= 2; Link[RL1] -= 2;
    } return;
  }
  if (zA == FR2) {
    if (Link[FR2] < 2) {
      // multilink is not active so set all four links by adding 2
      Link[FR2] += 2; Link[FL2] += 2; Link[RR2] += 2; Link[RL2] += 2;
    } else {
      // multilink is active so set all four links by subtracting 2
      Link[FR2] -= 2; Link[FL2] -= 2; Link[RR2] -= 2; Link[RL2] -= 2;
    } return;
  }
  if (zA == FR3) {
    if (Link[FR3] < 2) {
      // multilink is not active so set all four links by adding 2
      Link[FR3] += 2; Link[FL3] += 2; Link[RR3] += 2; Link[RL3] += 2;
    } else {
      // multilink is active so set all four links by subtracting 2
      Link[FR3] -= 2; Link[FL3] -= 2; Link[RR3] -= 2; Link[RL3] -= 2;
    } return;
  }
}

// ----------------------------------------------------------------------

void togglePlayPause() {
  // send command to toggle play pause
  drawFlag = 1;
  if (usbEn) {
    msgTx = "MP."; usbPortWrite(msgTx);
  } else if (Scrn == 2) {
    // pause/restart Walk simulator
    if (WalkSim != 0) {
      WalkSim = -WalkSim;  // toggle Walk mode +/-
      if (WalkSim < 0) {msgTx = "Sim Pause";}
    }
  }
}

// ----------------------------------------------------------------------

void togglePwr(int zS) {
  // toggle the servo PWM pulses ON/OFF by modifying the angle value
  // all angles are within 0 - 180, so we can add 200 for the OFF state
  // if a rate offset of 400 is applied we need to sustain this
  int zOff = 0; if (AngVal[zS] >= 400) {zOff = 400; AngVal[zS] -= 400;}
  
  if (AngVal[zS] >= 200) {
    AngVal[zS] -= 200;  // remove the Pwr offset
    togPwr = 1;         // set all mode to sustain power
  } else {
    AngVal[zS] += 200;  // add the Pwr offset
    togPwr = 2;         // set all mode to remove power
  }
  AngVal[zS] += zOff;
  ArrayVal[(rowPnt * rowWidth) + zS] = AngVal[zS];
  // send updated value to usb port
  msgTx = "ST" + str(Ch[zS]) + "." +"SA" + str(AngVal[zS]) + ".";
  setAngValFromRow();
  usbPortWrite(msgTx); drawFlag = 1;
  if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
}

// ----------------------------------------------------------------------

void togglePwrAll() {
  // sets the state of all servos to the current selection when the mouse
  // button is held down. togPwr indicates the target state
  if (togPwr > 1) {togPwr = 200;} else {togPwr = 0;}
  AngVal[FR1] = getAngVal(AngVal[FR1]) + togPwr; if (AngRate) {AngVal[FR1] += 400;}
  ArrayVal[(rowPnt * rowWidth) + FR1] = AngVal[FR1];
  
  AngVal[FR2] = getAngVal(AngVal[FR2]) + togPwr; if (AngRate) {AngVal[FR2] += 400;}
  ArrayVal[(rowPnt * rowWidth) + FR2] = AngVal[FR2];
  
  AngVal[FR3] = getAngVal(AngVal[FR3]) + togPwr; if (AngRate) {AngVal[FR3] += 400;}
  ArrayVal[(rowPnt * rowWidth) + FR3] = AngVal[FR3];
  
  AngVal[FL1] = getAngVal(AngVal[FL1]) + togPwr; if (AngRate) {AngVal[FL1] += 400;}
  ArrayVal[(rowPnt * rowWidth) + FL1] = AngVal[FL1];
  
  AngVal[FL2] = getAngVal(AngVal[FL2]) + togPwr; if (AngRate) {AngVal[FL2] += 400;}
  ArrayVal[(rowPnt * rowWidth) + FL2] = AngVal[FL2];
  
  AngVal[FL3] = getAngVal(AngVal[FL3]) + togPwr; if (AngRate) {AngVal[FL3] += 400;}
  ArrayVal[(rowPnt * rowWidth) + FL3] = AngVal[FL3];
  
  AngVal[RR1] = getAngVal(AngVal[RR1]) + togPwr; if (AngRate) {AngVal[RR1] += 400;}
  ArrayVal[(rowPnt * rowWidth) + RR1] = AngVal[RR1];
  
  AngVal[RR2] = getAngVal(AngVal[RR2]) + togPwr; if (AngRate) {AngVal[RR2] += 400;}
  ArrayVal[(rowPnt * rowWidth) + RR2] = AngVal[RR2];
  
  AngVal[RR3] = getAngVal(AngVal[RR3]) + togPwr; if (AngRate) {AngVal[RR3] += 400;}
  ArrayVal[(rowPnt * rowWidth) + RR3] = AngVal[RR3];
  
  AngVal[RL1] = getAngVal(AngVal[RL1]) + togPwr; if (AngRate) {AngVal[RL1] += 400;}
  ArrayVal[(rowPnt * rowWidth) + RL1] = AngVal[RL1];
  
  AngVal[RL2] = getAngVal(AngVal[RL2]) + togPwr; if (AngRate) {AngVal[RL2] += 400;}
  ArrayVal[(rowPnt * rowWidth) + RL2] = AngVal[RL2];
  
  AngVal[RL3] = getAngVal(AngVal[RL3]) + togPwr; if (AngRate) {AngVal[RL3] += 400;}
  ArrayVal[(rowPnt * rowWidth) + RL3] = AngVal[RL3];
  
  setAngValFromRow();
  sendRowData(rowPnt,rowPnt,1);
  mDwn = 0;        // clear mouse down timer
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void toggleRate(int zS) {
  // toggle the servo 'auto-rate' mode ON/OFF by modifying the angle value
  // all angles are within 0 - 180, +200 for Pwr OFF, so we can add 400 for the rate ON state
  rateFlag = true; drawFlag = 1;
  if (rowPnt == 0) {msgTx = "Error - first line"; return;}
  if ((rowPnt >= (rowNum - 1)) && (AngVal[zS] < 400)) {msgTx = "Error - last line"; return;}
  if (rowNum < 3) {msgTx = "Error - too few lines"; return;}
  
  if (AngVal[zS] >= 400) {
    AngVal[zS] -= 400;   // remove the auto-rate +400 offset
    Rated = getRated();  // check to see if any more rate angles exist
  } else {
    AngVal[zS] += 400;  // add the auto-rate +400 offset
    Rated = true;
  }
  ArrayVal[(rowPnt * rowWidth) + zS] = AngVal[zS];
  if (Rated) {calcRates();}  // update rate values
  
  // send updated value to usb port
  msgTx = "ST" + str(Ch[zS]) + "." +"SA" + str(AngVal[zS]) + ".";
  setAngValFromRow();
  usbPortWrite(msgTx);
  if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
}

// ----------------------------------------------------------------------

void toggleRowSelect() {
  // toggle the selection of a given row
  int zF; int zT;
  int zR = getRowMYD();
  zR = rowPnt + zR;
  if ((zR >= 0) && (zR < rowNum)) {
    // valid row pointer
    if (SelLast < 0) {
      // first time selection
      Select[zR] = true; SelCnt++;
    } else {
      if (zR > SelLast) {zF = SelLast; zT = zR;} else {zF = zR; zT = SelLast;}
      if (Select[zR]) {
        if (keySHIFT) {
          // multi-line deselect
          for (zR = zF; zR < zT + 1; zR++) {
            if (Select[zR]) {Select[zR] = false; SelCnt--;}
          }
        } else {
          // single line deselect
          Select[zR] = false; SelCnt--;
        }
      } else {
        if (keySHIFT) {
          // multi-line select
          for (zR = zF; zR < zT + 1; zR++) {
            if (!Select[zR]) {Select[zR] = true; SelCnt++;}
          }
        } else {
          // single line select
          Select[zR] = true; SelCnt++;
        }
      }
    } SelLast = zR; drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

void toggleStepRate() {
  // user has clicked on step rate profile column
  boolean zC = false;    // something has changed flag
  int zM;                // rate mode flag
  int zS;                // current step value
  int zR = getRowMYD();
  zR = rowPnt + zR;
  if ((zR < 0) || (zR >= rowNum)) {return;}  // row is before/after list
  if (ArrayVal[zR * rowWidth] < 0) {return;}  // command line so exit
  
  // valid row pointer
  zR = (zR * rowWidth) + 12;
  zS = ArrayVal[zR];
  if (zS > 600) {zS -= 600; zM = 3;}
  else if (zS > 400) {zS -= 400; zM = 2;}
  else if (zS > 200) {zS -= 200; zM = 1;}
  else {zM = 0;}
    
  if (keySHIFT) {
    // special 'quick' copy/paste step rate function
    if (mB == LEFT) {
      // paste quick rate value, which has previously been set
      if (QuickRate >= 0) {zM = QuickRate; zC = true;}
    } else {
      // record 'quick' step rate value
      QuickRate = zM; zC = true;
      msgTx = "QuickRate = " + str(zM); drawFlag = 1; return;
    }
  } else {
    // toggle current rate value
    if (mB == LEFT) {zM++; if (zM > 3) {zM = 0;}}
    else {zM--; if (zM < 0) {zM = 3;}}
    zC = true;
  }
  if (zM > 2) {zS += 600;}
  else if (zM > 1) {zS += 400;}
  else if (zM > 0) {zS += 200;}
  ArrayVal[zR] = zS;
  if (zC) {drawFlag = 1;}
}

// ----------------------------------------------------------------------

void toggleUsbEn() {
  // toggles the usbEn flag to control messages sent to the Arduino
  //if (SelCnt > 0) {ClearSelect();}
  if (comFlag < 1) {return;}  // no COM port connection
  
  // if enabled and we are playiong a move, then stop it first
  if ((usbEn) && (playMode > 1)) {playStop();}
  playMode = 0;  // reset the flag in case of playing once
  // now toggle the flag
  usbEn = !usbEn; drawFlag = 1;
  
  // if now enabled then send the current line data
  if (usbEn) {
    sendRowData(rowPnt,rowPnt,1);
  }
}

// ----------------------------------------------------------------------

void transSel() {
  // transpose angles for selected lines
  msgTx = "No rows selected"; rateFlag = true; drawFlag = 1;  
  if (SelCnt < 1) {return;}
  usbEn = false;  // avoid any unintentional movement
  
  cursor(WAIT);
  int zP = 0; int zRP = 0; boolean zSR0 = false;
  for (int zR = 0; zR < rowNum; zR++) {
    // work through each selected row
    if (Select[zR]) {
      // row is selected
      zRP = zR * rowWidth;
      // check to see if this is a command row?
      if (ArrayVal[zRP] > 0) {
        // it is not Cmd so proceed
        // if current row then set a flag to update sliders at the end
        if (zR == rowPnt) {zSR0 = true;}
        // first grab row values in Temp[] store
        for (zP = 0; zP < (rowWidth - 1); zP++) {
          Temp[zP] = ArrayVal[zP + zRP];
        }
        // now perform transposition of angles
        // it is the tier 3 servos which need to be corrected
        ArrayVal[zRP]     = Temp[3];  // set FR1
        ArrayVal[zRP + 1] = Temp[4];  // set FR2
        ArrayVal[zRP + 2] = 180 - getAngVal(Temp[5]);  // set FR3
        if (AngRate) {ArrayVal[zRP + 2] += 400;}
        if (!AngPwr) {ArrayVal[zRP + 2] += 200;}
        
        ArrayVal[zRP + 3] = Temp[0];  // set FL1
        ArrayVal[zRP + 4] = Temp[1];  // set FL2
        ArrayVal[zRP + 5] = 180 - getAngVal(Temp[2]);  // set FL3
        if (AngRate) {ArrayVal[zRP + 5] += 400;}
        if (!AngPwr) {ArrayVal[zRP + 5] += 200;}
        
        ArrayVal[zRP + 6] = Temp[9];  // set RR1
        ArrayVal[zRP + 7] = Temp[10]; // set RR2
        ArrayVal[zRP + 8] = 180 - getAngVal(Temp[11]); // set RR3
        if (AngRate) {ArrayVal[zRP + 8] += 400;}
        if (!AngPwr) {ArrayVal[zRP + 8] += 200;}
        
        ArrayVal[zRP + 9] = Temp[6];  // set RL1
        ArrayVal[zRP + 10] = Temp[7]; // set RL1
        ArrayVal[zRP + 11] = 180 - getAngVal(Temp[8]); // set RL1
        if (AngRate) {ArrayVal[zRP + 11] += 400;}
        if (!AngPwr) {ArrayVal[zRP + 11] += 200;}
      }
    }
  }
  usbEn =false;
  if (zSR0) {setAngValFromRow();}  // set servo values and sliders
  msgTx = "Selected data transposed";
}

// ----------------------------------------------------------------------

void transFR1FL1() {
  // transpose the FR1 and FL1 angles
  int zP =  rowPnt * rowWidth;
  Temp[0] = ArrayVal[zP];  // copy FR1
  ArrayVal[zP] =  ArrayVal[zP + 3];  // get FR1 from FL1
  ArrayVal[zP + 3] = Temp[0];        // replace FL1
  usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  msgTx = "FR1 and FL1 transposed";
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void transFR2FL2() {
  // transpose the FR2 and FL2 angles
  int zP =  rowPnt * rowWidth;
  Temp[0] = ArrayVal[zP + 1];  // copy FR2
  ArrayVal[zP + 1] =  ArrayVal[zP + 4];  // get FR2 from FL2
  ArrayVal[zP + 4] = Temp[0];            // replace FL2
  usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  msgTx = "FR2 and FL2 transposed";
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void transFR3FL3() {
  // transpose the FR3 and FL3 angles
  int zP =  rowPnt * rowWidth;
  Temp[0] = ArrayVal[zP + 2];  // copy FR3
  ArrayVal[zP + 2] = 180 - getAngVal(ArrayVal[zP + 5]);  // get FR3 from FL3
  if (AngRate) {ArrayVal[zP + 2] += 400;}
  if (!AngPwr) {ArrayVal[zP + 2] += 200;}
  ArrayVal[zP + 5] = 180 - getAngVal(Temp[0]);  // replace FL3
  if (AngRate) {ArrayVal[zP + 5] += 400;}
  if (!AngPwr) {ArrayVal[zP + 5] += 200;}
  usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  msgTx = "FR3 and FL3 transposed";
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void transRR1RL1() {
  // transpose the RR1 and RL1 angles
  int zP =  rowPnt * rowWidth;
  Temp[0] = ArrayVal[zP + 6];  // copy RR1
  ArrayVal[zP + 6] = ArrayVal[zP + 9];  // get RR1 from RL1
  ArrayVal[zP + 9] = Temp[0];           // replace RL1
  usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  msgTx = "RR1 and RL1 transposed";
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void transRR2RL2() {
  // transpose the RR2 and RL2 angles
  int zP =  rowPnt * rowWidth;
  Temp[0] = ArrayVal[zP + 7];  // copy RR2
  ArrayVal[zP + 7] = ArrayVal[zP + 10];  // get RR2 from RL2
  ArrayVal[zP + 10] = Temp[0];           // replace RL2
  usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  msgTx = "RR2 and RL2 transposed";
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

void transRR3RL3() {
  // transpose the RR3 and RL3 angles
  int zP =  rowPnt * rowWidth;
  Temp[0] = ArrayVal[zP + 8];  // copy RR3
  ArrayVal[zP + 8] = 180 - getAngVal(ArrayVal[zP + 11]);  // get RR3 from RL3
  if (AngRate) {ArrayVal[zP + 8] += 400;}
  if (!AngPwr) {ArrayVal[zP + 8] += 200;}
  ArrayVal[zP + 11] = 180 - getAngVal(Temp[0]);            // replace RL3
  if (AngRate) {ArrayVal[zP + 11] += 400;}
  if (!AngPwr) {ArrayVal[zP + 11] += 200;}
  usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  msgTx = "RR3 and RL3 transposed";
  rateFlag = true; drawFlag = 1;
}


// ######################################################################


void usbPortOp() {
  // called from a timer to empty the usb buffer
  usbLen = usbBuff.length(); drawUsbLamp();
  if (usbLen < 1) {return;}  // nothing in buffer
  
  // we are going to send something from usbBuff
  if (usbLen <= WiFiPW) {
    // it should only take 2.8ms to send 32 bytes
    usbPort.write(usbBuff); Tx_t0 = millis();  // send small packet
    usbBuff = ""; usbLen = 0; // send small packet and clear buffer
  } else {
    usbPort.write(usbBuff.substring(0,WiFiPW)); // send WiFiPW length packet
    Tx_t0 = millis();
    usbBuff = usbBuff.substring(WiFiPW,usbBuff.length());  // shift buffer by 16 char
    usbLen = usbBuff.length();
  } Ping = PingMax;  // no need to ping if sending data, so reset timer
}

// ----------------------------------------------------------------------

void usbPortWrite(String zmsg) {
  // called to send text over the serial port
  // nothing is sent if comFlag or usbEn are false
  if (!zmsg.equals("#")) {
    usbMsg = zmsg;  // keep a record or what was last sent or intended, ignoring pings
    if ((comFlag > 0) && usbEn) {
      usbBuff = usbBuff + zmsg;  // add new message to buffer
      //println(zmsg);
    } else {usbBuff = "";}  // flush the buffer if not sending
  } else {
    // this is a 'ping' # which is always sent if comm port is open
    // the droids eyes should always be red when this app is active
    if (comFlag > 0) {usbBuff = usbBuff + zmsg;}  // add new message to buffer
  }
}


// ######################################################################


void Walk_Init() {
  // initiallises the 'Walk' simulator from the first row, which will be
  // drawn. A delay is set based on the row step value, but if the row
  // contains a Command then the timer is set to re-trigger immediately
  // first check for selected lines
  sim1st = 0;           // first row of data
  simEnd = rowNum - 1;  // last row of data
  if (SelCnt > 0) {
    // rows are selected so find first selected block
    // if multiple blocks or rows are selected we only play the first
    sim1st = -1;
    for (int zR = 0; zR < rowNum; zR++) {
      if (Select[zR]) {
        if (sim1st < 0) {sim1st = zR;}    // first selected row
        simEnd = zR;                      // track last selected row
      } else {if (sim1st >= 0) {break;}}  // exit at end of selected block
    }
  }
  
  rowPnt = sim1st;     // define the row to be drawn
  simPnt = sim1st;     // simulator start pointer
  CmdForNum = 0;  // For... Next loop count
  CmdForRow = 0;  // row pointer for For... Next loops
  WalkNext = millis();  // set the timer to trigger immediately
  msgTx = "Sim " + str(rowPnt);
  drawFlag = 1;
}

// ----------------------------------------------------------------------

void Walk_Sim() {
  // called from draw() function when enabled and timer reached
  // still within the current list
  msgTx = "Sim " + str(simPnt) + "   " + str(t1);
  int zCD = 1;       // command step delay, default
  int zDF = 0;       // temp drawFlag
  int zP = simPnt * rowWidth;
  int zCmdCode = 0;  // command code type
  int zCmdVal0 = 0;  // command value
  if (ArrayVal[zP] < 0) {
    // command line so check for Goto, Speed, Pause, etc
    zCmdCode = ArrayVal[zP + 1];
    zCmdVal0 = ArrayVal[zP + 2];
    // a blank line is simply ignored
    if (zCmdCode == 1) {
      // this is a Goto command so go to it if in continuous mode
      if (WalkSim == 2) {simPnt = ArrayVal[zP + 2] - 1;}
      // if not then stop the process
      else {WalkSim = 0; msgTx = "Sim Terminated";}
    }
    else if(zCmdCode == 2) {
      // this is a speed change command
      Speed = zCmdVal0;
    } 
    else if(zCmdCode == 3) {
      // this is a Pause command, so determine pause delay
      zCD = 1000/Speed; zCD = max(zCD,4);
      zCD = zCD * zCmdVal0;
    } 
    else if(zCmdCode == 4) {
      // this is a Flag command, so simply ignore it
    } 
    else if(zCmdCode == 5) {
      // this is a Balance command, so simply ignore it
    } 
    else if(zCmdCode == 6) {
      // this is a For... Next loop command, so set pointers
      CmdForNum = zCmdVal0; CmdForRow = simPnt;
    } 
    else if(zCmdCode == 7) {
      // this is a Next command in a For... Next loop
      // loop back to previous row if there is more to do
      if (CmdForNum > 1) {CmdForNum--; simPnt = CmdForRow;}
      else {CmdForNum = 0; CmdForRow = 0;}
    } 
    // Command lines do not consume step time so simply add 1ms default delay
    // unless a Pause command has been encountered
     WalkNext = millis() + zCD;
  } else {
    // not a command line so load the step delay
    rowPnt = simPnt;  // set the drawing pointer after the simulated delay
    getWalkDelay(); zDF = 1;
  }
  
  simPnt++;  // point at the next line
  if (simPnt > simEnd ){
    // at the last line in the list so either stop or go back to the beginning
    if (WalkSim == 2) {
      // in continuous mode so loop back to the beginning
      Walk_Init();
    } else {
      // switch OFF the Walk simulator
      WalkSim = 0;  // stop the simulator
      msgTx = "Sim End";
    }
  }
  drawFlag = zDF;
}

// ----------------------------------------------------------------------
