// ########################################################################
//
//  Move Trainer v0.02
//
//  Released:  12/10/2020    By: TechKnowTone
//
// ########################################################################
/*
  For RoboDog
  
  This code communicates with an ESP8266 micro via the USB serial interface
  to control a droids servo motors connected to a PCA9685 over I2C. 
  
  This version includes:
  * 200 row limit
  * 12 slider fields with increment/decrement zones
  * return to REST, READY and SQUAT buttons
  * copy/paste memory data to/from the clipboard
  * help system
  * select rows
  * 4 memory stores
  * calculates steps for selected lines
  * transposes angles in selected rows
  * inserts 'Goto' commands
  * inserts 'Speed' command
  * servo power ON/OFF buttons
  * quick SHIFT + mouse copy n paste features
  * rate mode, between start and finish angles
  * list screen
  * Multi-move controls + linked
  
*/

// declare libraries
import processing.serial.*;
Serial usbPort;

import java.awt.*;
//import java.awt.event.*;
//import java.awt.event.MouseEvent;
import java.awt.datatransfer.*;
import javax.swing.*;
import java.io.*;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.awt.Toolkit;


// declare variables
int AB;                   // text angle background Blue colour component
int AG;                   // text angle background Green colour component
int AngField = -1;        // if >= 0 then draw a box around the angle field
int AngLast;              // previous angle
int AngLast0;             // another previous angle
int AngDif;               // different in pervious to current angle
boolean AngPwr;           // used during drawing, true if the servo is powered, false if not
boolean AngRate;          // used during drawing, true if the angle is an auto-rate angle
int AnyInt;               // temp value used anywhere
int AR;                   // text angle background Red colour component
int BB;                   // text background Blue colour component
int BG;                   // text background Green colour component
boolean Blank;            // set == true is a blank line command is detected
int BR;                   // text background Red colour component
int CalcMaxHgtLft;        // calculated left-hand max height value
int CalcMaxHgtRht;        // calculated right-hand max height value
int CalcNomStepsLft;      // calculated left-hand steps value
int CalcNomStepsRht;      // calculated right-hand steps value
boolean Cmd;              // temporary flag used in display mode
int CmdCode;              // command code number
int CmdForNum;            // sim - number of loops in a For... Next command
int CmdForRow;            // sim - pointer for For... Next loop
boolean CmdRow;           // true if current row is a command line
int CmdType;              // type of command line
int CmdVal0;              // command 1st value
int comFlag;              // = 1 if COM serial available
int comListLength;        // number of items in COM connection list
String comName;           // serial port COM name
int comPnt;               // COM port list pointer
int CopyCnt;              // if > 0 then data is stored in CopyMem[]
String CopyData = "";     // data to be copied to the clipboard
String CR = "\r\n";       // carriage return for strings
String data = "";         // characters recevied from the usb port
int drawFlag;             // = 1 to force screen redraw
int DegStep;              // number of degrees per step in autostep calcs
int DX0,DX1,DY0,DY1;      // variables used in drawing walk co-ordinates
boolean Echo;             // true if Arduino responses are being received
int FieldTgt = 0;         // pointer to the field clicked
int FieldVal = 0;         // temp field value
int FL1;                  // array pointer
int FL2;                  // array pointer
int FL3;                  // array pointer
int Flag;                 // default 'Flag' value
int FldLL,FldUL;          // temp field limits
final int FldT1maxLft = 0;   // field reference pointer
final int FldT1maxRht = 1;   // field reference pointer
final int FldT2maxLft = 2;   // field reference pointer
final int FldT2minLft = 3;   // field reference pointer
final int FldT2maxRht = 4;   // field reference pointer
final int FldT2minRht = 5;   // field reference pointer
final int FldT1WlkLft = 6;   // field reference pointer
int FR1;                  // array pointer
int FR2;                  // array pointer
int FR3;                  // array pointer
int GndY;                 // simulator ground reference Y-axis
int GndYM;                // simulator ground Y-axis max
String Help$;             // message generated by help system
String HelpLast$;         // previous message generated by help system
int HelpOFF;              // if >0 then help system is disabled
int HelpRowPnt;           // >= 0 if mouse is over a valid row number
int HX0,HX1,HX2,HX3,HY0,HY1,HY2,HY3;  // X/Y co-ordinates for tier 1 leg highlight drawing
PImage img0;              // image loaded 0 as background
PImage img1;              // image loaded 1 as background
PImage img2;              // image loaded 1 as background
int Inc;                  // any value increment
int IncCnt;               // any increment counter
int index;
int interval;             // main loop interval in milliseconds
int JoyX0;                // x-coordinate of joystick slider left end
int JoyX1;                // x-coordinate of joystick slider right end
int keyDwn;               // tracks keyboard up/down activity, 0 = up, 1 = down
boolean keyESC;           // true if ESC key is pressed
boolean keySHIFT;         // true whenever the SHIFT key is pressed
int LastVal;              // any previous value
int LF;                   // ASCII line feed
int M0;                   // if >0 then data stored in temp memory
int M1;                   // if >0 then data stored in temp memory
int M2;                   // if >0 then data stored in temp memory
int M3;                   // if >0 then data stored in temp memory
int mDwn;                 // state of mouse button down timer
int mB, mON, mX, mXD, mY, mYD;    // mouse flag, x,y, down
String msgCOM;            // COM port name
String msgRx;             // message received from Arduino
String msgTx;             // message sent to Arduino
int Multi0,Multi1;        // servo pointers in multimove functions
int MultiPnt;             // multi-move pointer
int mXL,mYL;              // previous x,y values
int MXDL;                 // previous MXD value
int MYDL;                 // previous MXD value
int mXOff;                // mouse X offset to control clicking error
int nextWiFims;           // WiFi ms timer
int nextMillis;           // loop end time in millis
int Pause;                // Pause delay value, default = 1
int Ping = 50;            // ping timer to keep auto-OFF from kicking in
int PingMax = Ping;       // record the max delay between pings
int playMode;             // mode ove movement play, 0 = stop; 1 = once, 2 = continuous
int QuickAng;             // quick cut/paste angle value
int QuickRate;            // quick cut/paste step rate value, C,A,D,S
int QuickStep;            // quick cut/paste step value
boolean Rated;            // == true if there are rate values in the list
boolean rateFlag;         // == true if rates are to be calculated before a redraw
String Released$ = "12/10/2020";  // latest update
int RL1;                  // array pointer
int RL2;                  // array pointer
int RL3;                  // array pointer
int rowColPnt;            // Help pointer to row column
int rowPnt;               // pointer to current array row
int RR1;                  // array pointer
int RR2;                  // array pointer
int RR3;                  // array pointer
String Rx$;               // single character received from usb port
int RxB;                  // set true when a byte is received
int Scrn;                 // screen display pointer 0 or 1
boolean Sel;              // temporary selected flag
int SelCnt;               // number of selected lines
int SelLast;              // most recent selected line, -1 means no recent
boolean Send;             // general flag records change and need to send data
String Sep$;              // separator used in line listing
int servoOE;              // servo board Output Enable; default = 1  OFF
int servoTgt;             // target/pointer servo for manual setting
int sim1st;               // simulator start at row pointer
int simEnd;               // simulator end pointer
int simPnt;               // simulator row pointer. Note rowPnt is what is displayed
int SL0;                  // array pointer
int SL1;                  // array pointer
int SL1Av;                // average SL1 angle
int SL2;                  // array pointer
int SL2Av;                // average SL2 angle
int SL3;                  // array pointer
int SL3Av;                // average SL3 angle
int SL4;                  // array pointer
int sliderDwn;            // =1 when mouse is clicked on slider
String Slope$;            // temp string used in row list
int Speed;                // speed value to be used by Arduino
int SpeedDef;             // speed value to be used by Arduino
int SpeedLL;              // speed value to be used by Arduino
int SpeedUL;              // speed value to be used by Arduino
int SR0;                  // array pointer
int SR1;                  // array pointer
int SR1Av;                // average SR1 angle
int SR2;                  // array pointer
int SR2Av;                // average SR2 angle
int SR3;                  // array pointer
int SR3Av;                // average SR3 angle
int SR4;                  // array pointer
int Steps;                // Walk number of steps set by the user
int t0,t1;                // millis() timers
int T1maxAngLft;          // Walk Trainer T1 maxAng left side
int T1maxAngRht;          // Walk Trainer T1 maxAng right side
int T2maxAngLft;          // Walk Trainer T2 maxAng left side
int T2maxAngRht;          // Walk Trainer T2 maxAng right side
int T2midAngLft;          // Walk Trainer T2 midAng left side
int T2midAngRht;          // Walk Trainer T2 midAng right side
int T2minAngLft;          // Walk Trainer T2 minAng left side
int T2minAngRht;          // Walk Trainer T2 minAng right side
int TB;                   // text Blue colour component
int TG;                   // text Green colour component
int togPwr;               // >0 to set all servos ON/OFF
int TR;                   // text Red colour component
int Tx_t0;                // usb send timers
int Tx_t1;                // usb send timers
String usbBuff;           // buffer string used to prevent WiFi overload
boolean usbEn;            // true to enable USB messages
int usbLen;               // length of usbMsg
String usbMsg;            // latest message sent to the usb port
int usbOnCnt;             // Usn enabled warning timer
int WalkHeight;           // user set height in walk sequence
int WalkLift;             // lift angle during walking
int WalkNext;             // micros() timer for Walk simulator          
int WalkSim;              // a flag which controled the walk simulator
int WaklDelay;            // sets the walk delay timer
int WalkSteps;            // number of steps used in walk sequences
int WiFiInt;              // WiFi interval in ms
int WiFiPW;               // WiFi packet width, must be longer than the max command length

// declare arrays
int rowDepth = 200;        // the number of rows in the array
int rowMax = rowDepth - 1; // one less than the maximum number of rows supported by rowDepth
int rowNum = 1;            // number of rows in this list
int rowWidth = 13;         // the width of each row
int arrayMax = rowDepth * rowWidth;
//               FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,NA.
int [] AngVal = {  0,  0, 90,  0,  0, 90,  0,  0, 90,  0,  0, 90,  0};  // servo angles
int [] AngVLL = {  0,  0, 70,  0,  0, 50,  0,  0, 70,  0,  0, 50,  0};  // servo angle lower limit
int [] AngVUL = {160,140,130,160,140,110,160,140,130,160,140,110,  0};  // servo angle upper limit
int [] ArrayVal = new int [arrayMax]; // servo memory
int [] Ch     = {  1,  3,  5,  0,  2,  4, 14, 12, 10, 15, 13, 11};  // maps SRn/SLn array to servo channels
int [] CopyMem = new int [arrayMax]; // copy memory
int [] JoyXval = new int [13];  // joystick slider position values
//               FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,NA.
int [] JoyXLL = {  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0};  // form slider limits
int [] JoyXUL = {360,200,200,360,200,200,360,200,200,360,200,200,360};  // form slider limits
int [] Link   = {  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0};  // servo link flags
int [] Mem0 = new int [arrayMax];  // temp memory store
int [] Mem1 = new int [arrayMax];  // temp memory store
int [] Mem2 = new int [arrayMax];  // temp memory store
int [] Mem3 = new int [arrayMax];  // temp memory store
int [] MaxA = new int [rowWidth];  // temp angle max
int [] MinA = new int [rowWidth];  // temp angle min
boolean [] Select = new boolean [arrayMax];
int [] Temp = new int [rowWidth]; // temp row store

void setup() {
  // first code to run. Also run again by a soft reset
  setDefaults();
  ClearArray(0);
  ClearLinks();
  ClearSelect();
  setSliderX();
  size(824,760);  // screen size must match background bitmap
  surface.setTitle("RoboDog Move Trainer");
  img0 = loadImage("Move Trainer 0.png");
  img1 = loadImage("Move Trainer 1.png");
  img2 = loadImage("Move Trainer 2.png");
  getCOMPort();
  textSize(13);
  println("Ready!");
  nextMillis = millis() + interval;
  nextWiFims = millis() + WiFiInt;
}

// ----------------------------------------------------------------------

void draw() {
  // main loop. Comes here about 60 times a second
  if (drawFlag > 0) {
    if (rateFlag) {calcRates();}  // perform a rate calculation before a refresh
    if (Scrn == 0) {
      // Trainer screen
      background(img0);  // define background image
      drawAngles();      // put numbers in angle fields
      drawAngField();    // put red box round selected angle
      drawRates();       // fill in 'rate' buttons
      drawLinks();       // draw boxes around link buttons
      drawDegSteps();    // update Steps field
      drawSpeed();       // update Speed field
    } else if (Scrn == 1){
      // List screen
      background(img1); // define background image
      drawAngPos();      // draw angle position heighlights on RoboDog image
      drawAngField();    // put red box round respective servos
      drawDegSteps();    // update Steps field
      drawSpeed();       // update Speed field
    } else {
      // Simulator screen
      background(img2); // define background image
      drawWalkLegs();
      drawWalkAngles();
    }
    drawHidden();
    drawArrayRows();
    drawMemBars();
    drawCOMMsg(); drawRxMsg(); drawTxMsg(); drawUsbLamp(); drawTxRxTime();
    drawEchoLamp(); drawHelp();
    if (rowColPnt >= 0) {drawColHighlight();}  // draw these last/on top
    drawFlag = 0;
  }
  if (mON > 0) {drawMouseXY();}
  
  // perform usb buffer task
  if (millis() >= nextWiFims) {
    nextWiFims = nextWiFims + WiFiInt;
    // the following should never be true unless significant time is lost somewhere
    if (nextWiFims < millis()) {nextWiFims = millis() + WiFiInt;}
    // only send data to the usb port if it is active
    if (comFlag > 0) {usbPortOp();} else {usbBuff = "";}
  }
  
  // perform Walk simulator function
  if (WalkSim > 0) {
    // in active Walk mode, so check the timer
    // if WalkSim < 0 then in Pause state
    if (millis() >= WalkNext) {
      t1 = millis() - t0;  // capture slack space time
      Walk_Sim();
      t0 = millis();  // record start of free time
    }
  }
  
  // perform 20ms tasks
  if (millis() >= nextMillis) {
    nextMillis = nextMillis + interval;
    // the following should never be true unless significant time is lost somewhere
    if (nextMillis < millis()) {nextMillis = millis() + interval;}
    
    // perform auto-repeat on some buttons
    if (mDwn > 0) {
      // perform auto repeat functions
      mDwn--;
      if (mDwn < 1) {mDwn = 5; mouseDown();}
    }
    // send a 'ping' character every second to keep servo power ON
    Ping--; if (Ping < 1) {Ping = PingMax; usbPortWrite("#");}
    if (RxB > 0) {RxB--;}
    // Usb wardning timer
    usbOnCnt++;
    if (usbOnCnt == 10) {drawFlag = 1;;}
    if (usbOnCnt >= 20) {usbOnCnt = 0; drawFlag = 1;}
  }
}

// ----------------------------------------------------------------------

void keyPressed() {
  // used to receive keyboard keys
  if (key == ESC) {keyESC = true; key = 0; ESC(); return;}
  
  HelpRowPnt = -1;  // for Scrn1 the current row will drive the highlights
  keyDwn = 1;
       if (keyCode == DOWN) {incRowPnt();}
  else if (keyCode == SHIFT) {keySHIFT = true;}
  else if (keyCode == UP) {decRowPnt();}
  else if (keyCode == 88) {if ((int)key == 24) {cutRows();}}    // Ctrl+X key
  else if (keyCode == 86) {if ((int)key == 22) {pasteRows();}}  // Ctrl+V key
  else if (keyCode == 82) {keyToggleRate();} // R key
  else if (keyCode == 67) {if ((int)key ==  3) {copyRows();}}   // Ctrl+C key
  else {
   println("key: " + key);
   println("int key: " + (int) key);
   println("keyCode: " + keyCode);
   println();
  }
}

// ----------------------------------------------------------------------

void keyReleased() {
  // called whenever a key is released
  keyDwn = 0;keySHIFT = false; keyESC = false;
}

// ----------------------------------------------------------------------

void mouseDragged() {
  // mouse has been clicked and dragged on window
  mX = mouseX; mY = mouseY;
  
       if (Scrn == 0) {mouseDragged_0();}
  else if (Scrn == 1) {mouseDragged_1();}
  else {mouseDragged_2();}
}

// ----------------------------------------------------------------------

void mouseDragged_0() {
  // mouse has been clicked and dragged on window on Scrn0

  if (sliderDwn > 0) {
    // ignore mY if it goes outside of slider region
    readXSlider(servoTgt); return;
  }
  
  if (MultiPnt > 0) {
    MultiMoveXY(Multi0,Multi1); return;
  }
}

// ----------------------------------------------------------------------

void mouseDragged_1() {
  // mouse has been clicked and dragged on window on Scrn1

}

// ----------------------------------------------------------------------

void mouseDragged_2() {
  // mouse has been clicked and dragged on window on Scrn2

  if (sliderDwn > 0) {
    // ignore mY if it goes outside of slider region
    adjustField(FieldTgt); return;
  }
}

// ----------------------------------------------------------------------

void mouseDown() {
  // called from main loop to repeat certain events after a pause delay
  if (sliderDwn > 0) {mDwn = 0; return;}
  if (MultiPnt > 0) {mDwn = 0; return;}
  if (togPwr > 0) {togglePwrAll(); return;}
  
  if (Scrn == 0) {mouseDown_0();}
  else if (Scrn == 1) {mouseDown_1();}
  else {mouseDown_2();}

  //#####################################################################
  //
  // list lines region
  //
  //#####################################################################
  int zmY = 547; if (Scrn == 1) {zmY = 41;}
  if ((mY >= zmY) && (mY <= 686)) {
    // angles
    if ((mX >= 154) && (mX < 494)) {changeAngleVal(); return;}

    // steps field
    if ((mX >= 543) && (mX < 582)) {changeSteps(); return;} // inc/dec step value

    // rate field
    if ((mX >= 512) && (mX < 537)) {mDwn = 0; resetStepRates(); return;}
  }
  
  if (((mY >= 548) && (mY <= 684))) {
    // Memory sliders and buttons

    if ((mX >= 585) && (mX <= 611)) {
      // vertical slider region
      if (mY < 566) {usbEn = false; decRowPnt(); return;}
      if (mY >= 666) {usbEn = false; incRowPnt(); return;}
    }
    if (((mY >= 559) && (mY <= 595))) {
      // Memory buttons top row
      if ((mX >= 691) && (mX < 719)) {mDwn = 0; storeM0(); return;}  // M0 button held-down event
      if ((mX >= 719) && (mX < 749)) {mDwn = 0; storeM1(); return;}  // M1 button held-down event
      if ((mX >= 749) && (mX < 778)) {mDwn = 0; storeM2(); return;}  // M2 button held-down event
      if ((mX >= 778) && (mX <= 807)) {mDwn = 0; storeM3(); return;}  // M3 button held-down event
    }
    if ((mY >= 636) && (mY <= 671)) {
      // Memory buttons 3rd row
      if ((mX >= 691) && (mX < 749)) {ClearArray(1); ClearSelect(); mDwn = 0; cursor(ARROW); return;}
      if ((mX >= 778) && (mX < 807)) {
        // paste key
        getTextFromClipboard(); ClearSelect(); ClearLinks();
        mDwn = 0; cursor(ARROW); return;
      }
    }
  }
  
  if ((mY >= 703) && (mY < 741)) {
    // Processing logo - soft restart
    if ((mX >= 12) && (mX <= 49)) {
      // reset everything but keep the current screen mode
      AnyInt = Scrn; mDwn = 0; closeCOMPort(); setup(); Scrn = AnyInt;
    }
  }
}

// ----------------------------------------------------------------------

void mouseDown_0() {
  // called to repeat certain events after a pause delay on Scrn 0
  
  if ((mY >= 122) && (mY <= 148)) {
      // Level 3 text fields
      if ((mX >= 148) && (mX < 160)) {incXSlider(FR3); return;}
      if ((mX > 186) && (mX <= 198)) {decXSlider(FR3); return;}

      if ((mX >= 250) && (mX < 262)) {incXSlider(FL3); return;}
      if ((mX > 287) && (mX <= 299)) {decXSlider(FL3); return;}

      if ((mX >= 376) && (mX < 388)) {incXSlider(RR3); return;}
      if ((mX > 413) && (mX <= 425)) {decXSlider(RR3); return;}

      if ((mX >= 476) && (mX < 488)) {incXSlider(RL3); return;}
      if ((mX > 514) && (mX <= 526)) {decXSlider(RL3); return;}

  } else if ((mY >= 260) && (mY <= 286)) {
      // Level 2 text fields
      if ((mX >= 174) && (mX < 186)) {incXSlider(FR2); return;}
      if ((mX > 211) && (mX <= 223)) {decXSlider(FR2); return;}

      if ((mX >= 275) && (mX < 287)) {incXSlider(FL2); return;}
      if ((mX > 312) && (mX <= 324)) {decXSlider(FL2); return;}

      if ((mX >= 400) && (mX < 412)) {incXSlider(RR2); return;}
      if ((mX > 438) && (mX <= 450)) {decXSlider(RR2); return;}

      if ((mX >= 501) && (mX < 513)) {incXSlider(RL2); return;}
      if ((mX > 539) && (mX <= 551)) {decXSlider(RL2); return;}

  } else if ((mY >= 385) && (mY <= 411)) {
      // Level 1 text fields
      if ((mX >= 286) && (mX < 298)) {decXSlider(FR1); return;}
      if ((mX > 323) && (mX <= 335)) {incXSlider(FR1); return;}

      if ((mX >= 385) && (mX < 397)) {decXSlider(FL1); return;}
      if ((mX > 422) && (mX <= 434)) {incXSlider(FL1); return;}

      if ((mX >= 512) && (mX < 524)) {decXSlider(RR1); return;}
      if ((mX > 549) && (mX <= 561)) {incXSlider(RR1); return;}

      if ((mX >= 612) && (mX < 624)) {decXSlider(RL1); return;}
      if ((mX > 649) && (mX <= 661)) {incXSlider(RL1); return;}
  }
  
  //#####################################################################
  //
  // Multi-move components
  //
  //#####################################################################
  if ((mY >= 488) && (mY <= 545)) {
    if ((mX >= 222) && (mX <= 607)) {
      if (mY <= 502) {
        if ((mX >= 235) && (mX <= 265)) {MultiDecY(FR1,FR2); return;}
        if ((mX >= 337) && (mX <= 367)) {MultiDecY(FL1,FL2); return;}
        if ((mX >= 462) && (mX <= 491)) {MultiDecY(RR1,RR2); return;}
        if ((mX >= 564) && (mX <= 594)) {MultiDecY(RL1,RL2); return;}
      }
      else if (mY <= 531) {
        if (mX <= 235) {incXSlider(FR2);}
        if ((mX >= 265) && (mX <= 277)) {decXSlider(FR2); return;}

        if ((mX >= 324) && (mX <= 337)) {incXSlider(FL2); return;}
        if ((mX >= 367) && (mX <= 380)) {decXSlider(FL2); return;}

        if ((mX >= 449) && (mX <= 462)) {incXSlider(RR2); return;}
        if ((mX >= 491) && (mX <= 504)) {decXSlider(RR2); return;}

        if ((mX >= 551) && (mX <= 564)) {incXSlider(RL2); return;}
        if ((mX >= 594) && (mX <= 607)) {decXSlider(RL2); return;}
      }
      else {
        if ((mX >= 235) && (mX <= 265)) {MultiIncY(FR1,FR2); return;}
        if ((mX >= 337) && (mX <= 367)) {MultiIncY(FL1,FL2); return;}
        if ((mX >= 462) && (mX <= 491)) {MultiIncY(RR1,RR2); return;}
        if ((mX >= 564) && (mX <= 594)) {MultiIncY(RL1,RL2); return;}
      }
    }
  }
  
  if ((mY >= 515) && (mY <= 539)) {
    // Step buttons and field
         if ((mX >= 12) && (mX < 31)) {decDegSteps(); return;}
    else if ((mX >= 81) && (mX <= 99)) {incDegSteps(); return;}

    // Speed buttons and field
    else if ((mX >= 722) && (mX < 740)) {decSpeed(); return;}
    else if ((mX >= 791) && (mX <= 809)) {incSpeed(); return;}
  }
}

// ----------------------------------------------------------------------

void mouseDown_1() {
  // called to repeat certain events after a pause delay on Scrn 1
  if ((mY >= 515) && (mY <= 539)) {
    // Step buttons and field
         if ((mX >= 12) && (mX < 31)) {decDegSteps(); return;}
    else if ((mX >= 81) && (mX <= 99)) {incDegSteps(); return;}

    // Speed buttons and field
    else if ((mX >= 722) && (mX < 740)) {decSpeed(); return;}
    else if ((mX >= 791) && (mX <= 809)) {incSpeed(); return;}
  }
}

// ----------------------------------------------------------------------

void mouseDown_2() {
  // called to repeat certain events after a pause delay on Scrn 2
  
  if ((mY >= 404) && (mY <= 430)) {
    // T2 max field
         if ((mX >= 103) && (mX < 115)) {decField(FldT2maxLft); return;}
    else if ((mX >= 141) && (mX <= 153)) {incField(FldT2maxLft); return;}
 
    // T2 min field
         if ((mX >= 273) && (mX < 295)) {decField(FldT2minLft); return;}
    else if ((mX >= 310) && (mX <= 322)) {incField(FldT2minLft); return;}

    // T2 max right field
         if ((mX >= 567) && (mX < 579)) {decField(FldT2maxRht); return;}
    else if ((mX >= 605) && (mX <= 617)) {incField(FldT2maxRht); return;}

    // T2 min right field
         if ((mX >= 737) && (mX < 749)) {decField(FldT2minRht); return;}
    else if ((mX >= 774) && (mX <= 786)) {incField(FldT2minRht); return;}
 }

  if ((mY >= 454) && (mY <= 480)) {
    // Speed buttons and field
         if ((mX >= 368) && (mX < 386)) {decSpeed(); return;}
    else if ((mX >= 437) && (mX <= 455)) {incSpeed(); return;}
  }

  if ((mY >= 470) && (mY <= 496)) {
    // T1 max left field
         if ((mX >= 103) && (mX < 115)) {decField(FldT1maxLft); return;}
    else if ((mX >= 141) && (mX <= 153)) {incField(FldT1maxLft); return;}

    // T1 max right field
         if ((mX >= 567) && (mX < 579)) {decField(FldT1maxRht); return;}
    else if ((mX >= 605) && (mX <= 617)) {incField(FldT1maxRht); return;}
  }
}

// ----------------------------------------------------------------------

void mouseMoved() {
  // used primarily with the help system
  mX = mouseX; mY = mouseY; mXD = mX; mYD = mY;
  Help_Message();
  if (mON > 0) {drawFlag = 1;}
}

// ----------------------------------------------------------------------

void mousePressed() {
  // mouse has been clicked on window
  // there are common and screen specific functions
  // the common functions are here
  mX = mouseX; mY = mouseY; mB = mouseButton; mDwn = 25;
  cursor(HAND);

  if ((mX < 50) && (mY < 50)) {
    // toggle mouse X,Y display ON/OFF
    if (mON < 1) {mON = 1;} else {mON = 0; drawFlag = 1;}
  }
  
  mXD = mX; mYD = mY; // record mouse down co-ordinates
  grabRowInTemp(rowPnt);
  
  if ((mX >= 377) && (mX <= 446) && (mY < 36)) {
    // change the screen mode and reset flags
    if (mB == LEFT) {
      Scrn++; if (Scrn > 2) {Scrn = 0;}
    } else {
      Scrn--; if (Scrn < 0) {Scrn = 2;}
    }
    drawFlag = 1; mDwn = 0; WalkSim = 0; return;
  }

  if ((mY >= 13) && (mY <= 47)) {
    // top function 1st row buttons, REST, READY, STAND
    if ((mX >= 601) && (mX < 671)) {insertRest();}
    else if ((mX >= 671) && (mX < 741)) {insertReady();}
    else if ((mX >= 741) && (mX <= 810)) {insertStand();}
  }
  else if ((mY >= 48) && (mY <= 82)) {
    // top function 2nd row buttons: TBD, TBD, SQUAT
    if ((mX >= 601) && (mX < 671)) {return;}
    else if ((mX >= 671) && (mX < 741)) {return;}
    else if ((mX >= 741) && (mX <= 810)) {insertSquat();}
  }

  //#####################################################################
  //
  // Screen specific functions
  //
  //#####################################################################
  if (Scrn == 0) {mousePressed_0();}
  if (Scrn == 1) {mousePressed_1();}
  else {mousePressed_2();}
  

  //#####################################################################
  //
  // Common screen functions
  //
  //#####################################################################
  if ((mY >= 547) && (mY <= 686)) {
    // Memory lines, sliders and buttons
    if (mX <= 105) {
      // left-hand side buttons and arrows
      if (mY <= 580) {
        // left region top row
        if ((mX >= 72) && (mX <= 102)) {moveSelUp();}
      } else if (mY <= 616) {
        // left region 2nd row
             if (mX < 39) {cutRows();}
        else if ((mX >= 39) && (mX < 71)) {copyRows();}
      } else if (mY <= 650) {
        // left region 3rd row
             if ((mX >= 9) && (mX < 39)) {insertGoto();}    // insert a Goto selected row branch
        else if ((mX >= 39) && (mX < 71)) {calcSelSteps();}  // calc. steps for selected rows
        else if ((mX >= 71) && (mX <= 102)) {transSel();}   // transpose selected rows
      } else {
        // left region 4th row
             if ((mX >= 39) && (mX < 71)) {reverseSelRows();}
        else if ((mX >= 71) && (mX <= 102)) {moveSelDwn();}
      }
    }
    if (CopyCnt > 0) {  
      if ((mY > 575) && (mY <= 614)) {  
        if ((mX >= 71) && (mX <= 102)) {pasteRows();}
      }  
    }
    
    if ((mX >= 585) && (mX <= 611)) {
      // vertical slider region
           if (mY < 566) {decRowPnt();}
      else if ((mY >= 566) && (mY < 589)) {gotoFirst();}
      else if ((mY >= 642) && (mY < 665)) {gotoLast();}
      else if (mY >= 665) {incRowPnt();}
    }
    
    if ((mX >= 614) && (mX <= 643)) {
      //R-hand insert buttons Blank,Delete,Pause,Balance
           if ((mY >= 541) && (mY <= 576)) {insertBlank();}
      else if ((mY >= 578) && (mY <= 614)) {deleteRows();}
      else if ((mY >= 616) && (mY <= 651)) {insertPause();}
      else if ((mY >= 653) && (mY <= 689)) {insertBalance();}
    }
    if ((mX >= 646) && (mX <= 675)) {
      //R-hand insert buttons Speed,New,Flag,TBD
           if ((mY >= 541) && (mY <= 576)) {insertSpeed();}
      else if ((mY >= 587) && (mY <= 614)) {newRow();}
      else if ((mY >= 616) && (mY <= 651)) {InsertFlag();}
      else if ((mY >= 653) && (mY <= 689)) {InsertExCode();}
    }
    
    if ((mY >= 559) && (mY <= 595)) {
      // Memory buttons top row
           if ((mX >= 691) && (mX < 719)) {mDwn = 50;}  // M0 button. Set timer for held-down event
      else if ((mX >= 719) && (mX < 749)) {mDwn = 50;}  // M1 button. Set timer for held-down event
      else if ((mX >= 749) && (mX < 778)) {mDwn = 50;}  // M2 button. Set timer for held-down event
      else if ((mX >= 778) && (mX < 808)) {mDwn = 50;}  // M3 button. Set timer for held-down event
    }
    if ((mY >= 598) && (mY <= 633)) {
      // Play buttons 2nd row
      if (usbEn || (Scrn == 2)) {
             if ((mX >= 691) && (mX < 719)) {playOnce();} // play once
        else if ((mX >= 719) && (mX < 749)) {playConst();} // play continuously
        else if ((mX >= 749) && (mX < 778)) {togglePlayPause();} // pause
        else if ((mX >= 778) && (mX <= 807)) {playStop();} // stop playing
      }
    }
    if ((mY >= 636) && (mY <= 671)) {
      // 3rd row buttons - Clear,Copy,Paste
           if ((mX >= 691) && (mX < 749)) {mDwn = 50; cursor(WAIT);}  // clear memory if held down
      else if ((mX >= 749) && (mX < 778)) {copyToClipboard();}
      else if ((mX >= 778) && (mX < 807)) {mDwn = 50; cursor(WAIT);}  // paste from clipboard if held down
    }
  }

  //#####################################################################
  //
  // list lines region
  //
  //#####################################################################
  int zmY = 547; if (Scrn == 1) {zmY = 41;}
  if ((mY >= zmY) && (mY <= 686)) {
    if ((mX >= 105) && (mX < 582)) {
      // memory lines region
      if ((mX >= 105) && (mX < 138)) {toggleRowSelect(); return;}
      if ((mX >= 154) && (mX < 494)) {changeAngleVal(); return;}

      // rate profile field  
      if ((mX >= 512) && (mX < 537)) {mDwn = 50; toggleStepRate(); return;}

      // steps field
      if ((mX >= 543) && (mX < 582)) {changeSteps(); return;} // inc/dec step value
    }
  }
  
  if ((mY >= 703) && (mY < 741)) {
    // Processing logo - soft restart
    if ((mX >= 12) && (mX <= 49)) {
      mDwn = 50; cursor(WAIT);
    }
  }
  if ((mY >= 690) && (mY < 716)) {
    // Com, RX, Tx fields
    // clicked on COM port?
    if ((mX >= 105) && (mX <= 180)) {
      if (mouseButton == LEFT) {getCOMPort();}
      else if (mouseButton == RIGHT) {closeCOMPort();}
    }
    if ((mX >= 478) && (mX <= 681)) {toggleUsbEn();}
  }
  if ((mY >= 686) && (mY <= 721)) {
    // Echo button?
    if ((mX >= 749) && (mX <= 807)) {toggleEcho();}
  }
  // if none of the above check for cancelling selects[]
  if ((SelCnt > 0) && (Help$.equals("-"))) {ClearSelect(); drawFlag = 1;};

}

// ----------------------------------------------------------------------

void mousePressed_0() {
  // screen 0 specific mouse pressed functions
  
  if (ArrayVal[rowPnt * rowWidth] < 0) {
    // command so make top part of form inactive, except REST and READY
  } else {
    if (!Help$.equals("-")) {cursor(HAND);}
    
    //#####################################################################
    //
    // Tier 3 components
    //
    //#####################################################################
    if ((mY >= 122) && (mY < 148)) {
        // Level 3 text fields & links
        if ((mX >= 148) && (mX < 160)) {incXSlider(FR3);}
        else if ((mX >= 160) && (mX < 186)) {readXSlider(FR3);}
        else if ((mX >= 186) && (mX < 198)) {decXSlider(FR3);}
  
        else if ((mX >= 198) && (mX < 215)) {copyLftToRgt(FL3, FR3);}
        else if ((mX >= 215) && (mX < 233)) {toggleLink(FR3,FL3);}
        else if ((mX >= 233) && (mX < 250)) {copyRgtToLft(FR3, FL3);}
  
        else if ((mX >= 250) && (mX < 262)) {incXSlider(FL3);}
        else if ((mX >= 262) && (mX <= 287)) {readXSlider(FL3);}
        else if ((mX > 287) && (mX <= 299)) {decXSlider(FL3);}

        else if ((mX >= 327) && (mX <= 353)) {toggleMultiLink(FR3);}

        else if ((mX >= 376) && (mX < 388)) {incXSlider(RR3);}
        else if ((mX >= 388) && (mX <= 413)) {readXSlider(RR3);}
        else if ((mX > 413) && (mX <= 425)) {decXSlider(RR3);}
  
        else if ((mX >= 425) && (mX < 441)) {copyLftToRgt(RL3, RR3);}
        else if ((mX >= 441) && (mX < 459)) {toggleLink(RL3,RR3);}
        else if ((mX >= 459) && (mX < 476)) {copyRgtToLft(RR3, RL3);}
  
        else if ((mX >= 476) && (mX < 488)) {incXSlider(RL3);}
        else if ((mX >= 488) && (mX <= 514)) {readXSlider(RL3);}
        else if ((mX > 514) && (mX <= 526)) {decXSlider(RL3);}

    } else if ((mY >= 148) && (mY <= 168)) {
        // Level 3 transposes, Pwr & Rate buttons
        if ((mX >= 165) && (mX <= 182)) {togglePwr(FR3);}
        else if ((mX >= 183) && (mX <= 198)) {toggleRate(FR3);}
        else if ((mX >= 213) && (mX < 234)) {transFR3FL3();}
        else if ((mX >= 266) && (mX <= 284)) {togglePwr(FL3);}
        else if ((mX >= 285) && (mX <= 299)) {toggleRate(FL3);}
        else if ((mX >= 391) && (mX <= 409)) {togglePwr(RR3);}
        else if ((mX >= 410) && (mX <= 425)) {toggleRate(RR3);}
        else if ((mX >= 439) && (mX < 461)) {transRR3RL3();}
        else if ((mX >= 492) && (mX <= 510)) {togglePwr(RL3);}
        else if ((mX >= 510) && (mX <= 526)) {toggleRate(RL3);}

    //#####################################################################
    //
    // Tier 2 components
    //
    //#####################################################################
    } else if ((mY >= 260) && (mY < 286)) {
        // Level 2 text fields & links
        if ((mX >= 174) && (mX < 186)) {incXSlider(FR2);}
        else if ((mX >= 186) && (mX <= 211)) {readXSlider(FR2);}
        else if ((mX > 211) && (mX <= 223)) {decXSlider(FR2);}
  
        else if ((mX >= 223) && (mX < 240)) {copyLftToRgt(FL2, FR2);}
        else if ((mX >= 240) && (mX < 258)) {toggleLink(FR2,FL2);}
        else if ((mX >= 258) && (mX < 275)) {copyRgtToLft(FR2, FL2);}
  
        else if ((mX >= 275) && (mX < 287)) {incXSlider(FL2);}
        else if ((mX >= 287) && (mX <= 312)) {readXSlider(FL2);}
        else if ((mX > 312) && (mX <= 324)) {decXSlider(FL2);}

        else if ((mX >= 354) && (mX <= 379)) {toggleMultiLink(FR2);}
  
        else if ((mX >= 400) && (mX < 412)) {incXSlider(RR2);}
        else if ((mX >= 412) && (mX <= 438)) {readXSlider(RR2);}
        else if ((mX > 438) && (mX <= 450)) {decXSlider(RR2);}
  
        else if ((mX >= 450) && (mX < 467)) {copyLftToRgt(RL2, RR2);}
        else if ((mX >= 467) && (mX < 485)) {toggleLink(RR2,RL2);}
        else if ((mX >= 485) && (mX < 501)) {copyRgtToLft(RR2, RL2);}
  
        else if ((mX >= 501) && (mX < 513)) {incXSlider(RL2);}
        else if ((mX >= 513) && (mX <= 539)) {readXSlider(RL2);}
        else if ((mX > 539) && (mX <= 551)) {decXSlider(RL2);}

    } else if ((mY >= 286) && (mY <= 305)) {
        // Level 2 transposes, Pwr & Rate buttons
        if ((mX >= 174) && (mX <= 188)) {toggleRate(FR2);}
        else if ((mX >= 189) && (mX <= 206)) {togglePwr(FR2);}
        else if ((mX >= 238) && (mX < 258)) {transFR2FL2();}
        else if ((mX >= 275) && (mX <= 289)) {toggleRate(FL2);}
        else if ((mX >= 290) && (mX <= 308)) {togglePwr(FL2);}
        else if ((mX >= 400) && (mX <= 415)) {toggleRate(RR2);}
        else if ((mX >= 416) && (mX <= 434)) {togglePwr(RR2);}
        else if ((mX >= 465) && (mX < 485)) {transRR2RL2();}
        else if ((mX >= 500) && (mX <= 516)) {toggleRate(RL2);}
        else if ((mX >= 517) && (mX <= 535)) {togglePwr(RL2);}

    //#####################################################################
    //
    // Tier 1 components
    //
    //#####################################################################
    } else if ((mY >= 385) && (mY < 411)) {
        // Level 1 text fields & links
        if ((mX >= 286) && (mX < 298)) {decXSlider(FR1);}
        else if ((mX >= 298) && (mX <= 323)) {readXSlider(FR1);}
        else if ((mX > 323) && (mX <= 335)) {incXSlider(FR1);}
  
        else if ((mX >= 335) && (mX < 352)) {copyLftToRgt(FL1, FR1);}
        else if ((mX >= 352) && (mX < 370)) {toggleLink(FR1,FL1);}
        else if ((mX >= 370) && (mX < 385)) {copyRgtToLft(FR1, FL1);}
  
        else if ((mX >= 385) && (mX < 397)) {decXSlider(FL1);}
        else if ((mX >= 397) && (mX <= 422)) {readXSlider(FL1);}
        else if ((mX > 422) && (mX <= 434)) {incXSlider(FL1);}

        else if ((mX >= 463) && (mX <= 489)) {toggleMultiLink(FR1);}
  
        else if ((mX >= 512) && (mX < 524)) {decXSlider(RR1);}
        else if ((mX >= 524) && (mX <= 549)) {readXSlider(RR1);}
        else if ((mX > 549) && (mX <= 561)) {incXSlider(RR1);}
  
        else if ((mX >= 561) && (mX < 578)) {copyLftToRgt(RL1, RR1);}
        else if ((mX >= 578) && (mX < 596)) {toggleLink(RR1,RL1);}
        else if ((mX >= 596) && (mX < 612)) {copyRgtToLft(RR1, RL1);}
  
        else if ((mX >= 612) && (mX < 624)) {decXSlider(RL1);}
        else if ((mX >= 624) && (mX <= 649)) {readXSlider(RL1);}
        else if ((mX > 649) && (mX <= 661)) {incXSlider(RL1);}

    } else if ((mY >= 411) && (mY <= 430)) {
        // Level 1 transposes, Pwr & Ratebuttons
        if ((mX >= 286) && (mX <= 300)) {toggleRate(FR1);}
        else if ((mX >= 301) && (mX <= 320)) {togglePwr(FR1);}
        else if ((mX >= 350) && (mX < 370)) {transFR1FL1();}
        else if ((mX >= 385) && (mX <= 400)) {toggleRate(FL1);}
        else if ((mX >= 401) && (mX <= 419)) {togglePwr(FL1);}
        else if ((mX >= 511) && (mX <= 526)) {toggleRate(RR1);}
        else if ((mX >= 527) && (mX <= 546)) {togglePwr(RR1);}
        else if ((mX >= 576) && (mX < 596)) {transRR1RL1();}
        else if ((mX >= 612) && (mX <= 627)) {toggleRate(RL1);}
        else if ((mX >= 628) && (mX <= 646)) {togglePwr(RL1);}
    }
  }
  
  
  //#####################################################################
  //
  // Multi-move components
  //
  //#####################################################################
  if ((mY >= 488) && (mY <= 545)) {
    if ((mX >= 222) && (mX <= 607)) {
      if (mY <= 502) {
        if ((mX >= 235) && (mX <= 265)) {MultiDecY(FR1,FR2); return;}
        if ((mX >= 337) && (mX <= 367)) {MultiDecY(FL1,FL2); return;}
        if ((mX >= 462) && (mX <= 491)) {MultiDecY(RR1,RR2); return;}
        if ((mX >= 564) && (mX <= 594)) {MultiDecY(RL1,RL2); return;}
      }
      else if (mY <= 531) {
        if (mX <= 235) {incXSlider(FR2);}
        if ((mX > 235) && (mX < 265)) {MultiMoveXY(FR1,FR2); return;}
        if ((mX >= 265) && (mX <= 277)) {decXSlider(FR2); return;}

        if ((mX >= 324) && (mX <= 337)) {incXSlider(FL2); return;}
        if ((mX > 337) && (mX < 367)) {MultiMoveXY(FL1,FL2); return;}
        if ((mX >= 367) && (mX <= 380)) {decXSlider(FL2); return;}

        if ((mX >= 449) && (mX <= 462)) {incXSlider(RR2); return;}
        if ((mX > 462) && (mX < 491)) {MultiMoveXY(RR1,RR2); return;}
        if ((mX >= 491) && (mX <= 504)) {decXSlider(RR2); return;}

        if ((mX >= 551) && (mX <= 564)) {incXSlider(RL2); return;}
        if ((mX > 564) && (mX < 594)) {MultiMoveXY(RL1,RL2); return;}
        if ((mX >= 594) && (mX <= 607)) {decXSlider(RL2); return;}
      }
      else {
        if ((mX >= 235) && (mX <= 265)) {MultiIncY(FR1,FR2); return;}
        if ((mX >= 337) && (mX <= 367)) {MultiIncY(FL1,FL2); return;}
        if ((mX >= 462) && (mX <= 491)) {MultiIncY(RR1,RR2); return;}
        if ((mX >= 564) && (mX <= 594)) {MultiIncY(RL1,RL2); return;}
      }
    }
  }
  
  
  if ((mY >= 515) && (mY <= 539)) {
    // DegSteps & Speed buttons and field
         if ((mX >= 12) && (mX < 31)) {decDegSteps(); return;}
    else if ((mX >= 31) && (mX < 81)) {resetDegSteps(); return;}
    else if ((mX >= 81) && (mX <= 99)) {incDegSteps(); return;}
    
         if ((mX >= 722) && (mX < 740)) {decSpeed(); return;}
    else if ((mX >= 740) && (mX < 791)) {resetSpeed(); return;}
    else if ((mX >= 791) && (mX <= 809)) {incSpeed(); return;}
  }
}

// ----------------------------------------------------------------------

void mousePressed_1() {
  // screen 1 specific mouse pressed functions
  if ((mY >= 515) && (mY <= 539)) {
    // DegSteps & Speed buttons and field
         if ((mX >= 12) && (mX < 31)) {decDegSteps();}
    else if ((mX >= 31) && (mX < 81)) {resetDegSteps();}
    else if ((mX >= 81) && (mX <= 99)) {incDegSteps();}
    
         if ((mX >= 722) && (mX < 740)) {decSpeed();}
    else if ((mX >= 740) && (mX < 791)) {resetSpeed();}
    else if ((mX >= 791) && (mX <= 809)) {incSpeed();}
  }
}

// ----------------------------------------------------------------------

void mousePressed_2() {
  // screen 2 specific mouse pressed functions
  if ((mY >= 498) && (mY <= 533)) {
    if ((mX >= 376) && (mX <= 446)) {Calc_Walk(); return;}
  }

  if ((mY >= 256) && (mY <= 358)) {
    // lower leg ground region
    if ((mX >= 224) && (mX <= 566)) {setGndY(); drawFlag = 1;}
  }
  
  if ((mY >= 372) && (mY <= 406)) {
    // Titles, copy buttons and Link
         if ((mX >= 24) && (mX <= 78)) {resetLftFlds(); drawFlag = 1;}
    else if ((mX >= 489) && (mX <= 552)) {resetRhtFlds(); drawFlag = 1;}
  }
    
  if ((mY >= 404) && (mY <= 430)) {
    // T2 max left field
         if ((mX >= 103) && (mX < 115)) {decField(FldT2maxLft);}
    else if ((mX >= 115) && (mX < 141)) {adjustField(FldT2maxLft);}
    else if ((mX >= 141) && (mX <= 153)) {incField(FldT2maxLft);}

    // T2 min left field
         if ((mX >= 273) && (mX < 295)) {decField(FldT2minLft);}
    else if ((mX >= 295) && (mX < 310)) {adjustField(FldT2minLft);}
    else if ((mX >= 310) && (mX <= 322)) {incField(FldT2minLft);}

    // T2 max right field
         if ((mX >= 567) && (mX < 579)) {decField(FldT2maxRht);}
    else if ((mX >= 579) && (mX < 605)) {adjustField(FldT2maxRht);}
    else if ((mX >= 605) && (mX <= 617)) {incField(FldT2maxRht);}

    // T2 min right field
         if ((mX >= 737) && (mX < 749)) {decField(FldT2minRht);}
    else if ((mX >= 749) && (mX < 774)) {adjustField(FldT2minRht);}
    else if ((mX >= 774) && (mX <= 786)) {incField(FldT2minRht);}
  }

  if ((mY >= 454) && (mY <= 480)) {
    // Speed buttons and field
         if ((mX >= 368) && (mX < 386)) {decSpeed();}
    else if ((mX >= 386) && (mX < 437)) {resetSpeed();}
    else if ((mX >= 437) && (mX <= 455)) {incSpeed();}
  }

  if ((mY >= 470) && (mY <= 496)) {
    // T1 max left field
         if ((mX >= 103) && (mX < 115)) {decField(FldT1maxLft);}
    else if ((mX >= 115) && (mX < 141)) {adjustField(FldT1maxLft);}
    else if ((mX >= 141) && (mX <= 153)) {incField(FldT1maxLft);}

    // T1 walk lift field
         if ((mX >= 273) && (mX < 295)) {decField(FldT1WlkLft);}
    else if ((mX >= 295) && (mX < 310)) {adjustField(FldT1WlkLft);}
    else if ((mX >= 310) && (mX <= 322)) {incField(FldT1WlkLft);}

    // T1 max right field
         if ((mX >= 567) && (mX < 579)) {decField(FldT1maxRht);}
    else if ((mX >= 579) && (mX < 605)) {adjustField(FldT1maxRht);}
    else if ((mX >= 605) && (mX <= 617)) {incField(FldT1maxRht);}
  }
}

// ----------------------------------------------------------------------

void mouseReleased() {
  // mouse released from click and drag operation
  mX = mouseX; mY = mouseY;
  if (mDwn > 0) {
    // look for click events
    if (((mY >= 560) && (mY <= 596))) {
      // Memory buttons top row
           if ((mX >= 691) && (mX < 719)) {recallM0();}  // M0 store recall event
      else if ((mX >= 719) && (mX < 749)) {recallM1();}  // M1 store recall event
      else if ((mX >= 749) && (mX < 778)) {recallM2();}  // M2 store recall event
      else if ((mX >= 778) && (mX <= 807)) {recallM3();}  // M3 store recall event
    }
  } else {
    // look for held events which have cleared mDwn
  }
  Inc = 1;         // reset increment value
  IncCnt = 0;      // clear increment counter
  mDwn = 0;        // clear mouse down timer
  togPwr = 0;      // clear power ALL toggle flag
  sliderDwn = 0;   // clear slider flag
  MultiPnt = 0;    // cancel multi-move pointer
  Help_Message();  // update help messages
  cursor(ARROW);
}

// ----------------------------------------------------------------------

void mouseWheel(MouseEvent event) {
  // check for mouse wheel scrolling rows
  int zYLL = 548; if (Scrn == 1) {zYLL = 41;}
  HelpRowPnt = -1;  // for Scrn1 the current row will drive the highlights
  if ((mX >= 105) && (mX <= 582)) {
    if ((mY >= zYLL) && (mY <= 684)) {
      // we are within the row area
      int zNew = rowPnt + int(event.getCount());
      if ((zNew >= 0) && (zNew < rowNum)) {
        // changing row value
        rowPnt = zNew; sendRowData(rowPnt,rowPnt,1);
        setAngValFromRow();  // set servo values and sliders
        drawFlag = 1;
      }
    }
  }
}

// ----------------------------------------------------------------------

void serialEvent(Serial usbPort) {
  // respond to a serial data received event
  Echo = true;  // if we received something then set this flag
  if (Tx_t0 > 0) {
    Tx_t1 = millis() - Tx_t0; drawTxRxTime(); Tx_t0 = 0;
  drawFlag = 1;
  } RxB = 3; 
  //nextWiFims = millis() + 20;  // delay sending whilst receiving
  Ping = PingMax;  // no need to 'ping' if we are receiving data
  Rx$ = usbPort.readString(); print(Rx$);
  data += Rx$; int RxLF = data.indexOf(LF);
  if (RxLF >= 0) {
    // received data contains a LF/CR so extract the first message
    msgRx = data.substring(0, RxLF); drawFlag = 1;
    // decode messages
    if (msgRx.indexOf("MT ") >= 0) {
      // the is a MoveMemTgtLoader() row report so move row to track it
      // trap serial comms errors
      int zRP = int(trim(msgRx.substring(2)));
      if ((zRP >= 0) && (zRP < rowNum)) {
        rowPnt = zRP;
        setAngValFromRow();  // update text fields and sliders
      }
      else if (msgRx.indexOf("SW") >= 0) {
        // this is a SWn. report, user has pushed a button whilst walking
        WalkSim = 0; playMode = 0; usbEn = false;
      }
      else if (msgRx.indexOf("WalkEnd") >= 0) {
        // this is an end of walking message, send during a Play Once run
        WalkSim = 0; playMode = 0;
      }
    }
    // make sure message fits Rx field
    if (msgRx.length() > 24) {msgRx = msgRx.substring(0,20) + "...";}
    data = trim(data.substring(RxLF));
  }
}

// ----------------------------------------------------------------------

void setDefaults() {
  // load default values
  AB = 255;           // text angle background Blue colour component
  AG = 255;           // text angle background Green colour component
  AngDif = 0;         // different in pervious to current angle
  AngLast = 0;        // previous angle
  AngPwr = false;     // during drawing true if servo is powered, false if not
  AngRate = false;    // used during drawing, true if the angle is an auto-rate angle
  AngVal[ 0] = 0;     // FR1 default
  AngVal[ 1] = 0;     // FR2 default
  AngVal[ 2] = 90;    // FR3 default
  AngVal[ 3] = 0;     // FL1 default
  AngVal[ 4] = 0;     // FL2 default
  AngVal[ 5] = 90;    // FL3 default
  AngVal[ 6] = 0;     // RR1 default
  AngVal[ 7] = 0;     // RR2 default
  AngVal[ 8] = 90;    // RR3 default
  AngVal[ 9] = 0;     // RL1 default
  AngVal[10] = 0;     // RL2 default
  AngVal[11] = 90;    // RL3 default
  AR = 255;           // text angle background Red colour component
  BB = 255;           // text background Blue colour component
  BG = 255;           // text background Green colour component
  Blank = false;      // set == true is a blank line command is detected
  BR = 255;           // text background Red colour component
  CalcMaxHgtLft = 178;   // calculated left-hand max height value
  CalcMaxHgtRht = 178;   // calculated right-hand max height value
  CalcNomStepsLft = 41;  // calculated left-hand steps value
  CalcNomStepsRht = 41;  // calculated right-hand steps value
  Cmd = false;        // temporary flag used in display mode
  CmdCode = 0;        // command code number
  CmdRow = false;     // true if current row is a command line
  CmdVal0 = 0;        // command 1st value
  comFlag = 0;        // = 1 if COM serial available
  comListLength = 0;  // number of items in COM connection list
  comName = "";       // serial port COM name
  comPnt = 0;         // COM port list pointer
  CopyCnt = 0;        // if > 0 then data is stored in CopyMem[]
  CopyData = "";      // data to be copied to the clipboard
  CR = "\r\n";        // carriage return for strings
  data = "";
  DegStep = 1;        // number of degrees per step in autostep calcs
  drawFlag = 1;       // = 1 to force screen redraw
  Echo = true;        // true if Arduino responses are being received
  FL1 = 3;            // array pointer
  FL2 = 4;            // array pointer
  FL3 = 5;            // array pointer
  Flag = 1;                 // default 'Flag' value
  FR1 = 0;            // array pointer
  FR2 = 1;            // array pointer
  FR3 = 2;            // array pointer
  GndY = 358;         // simulator ground reference Y-axis
  Help$ = "";         // message generated by help system
  HelpLast$ = "";     // previous message generated by help system
  HelpOFF = 0;        // if >0 then help system is disabled, default is ON
  Inc = 1;            // any value increment
  IncCnt = 0;         // any increment counter
  index = 0;
  interval = 20;      // main loop interval in milliseconds
  JoyX0 = 40;         // x-coordinate of joystick slider left end
  JoyX1 = 151;        // x-coordinate of joystick slider right end
  keyDwn = 0;         // tracks keyboard up/down activity, 0 = up, 1 = down
  keyESC = false;     // true if ESC key is pressed
  keySHIFT = false;   // true whenever the SHIFT key is pressed
  LF = 10;            // ASCII line feed
  M0 = 0;             // if >0 then data stored in temp memory
  M1 = 0;             // if >0 then data stored in temp memory
  M2 = 0;             // if >0 then data stored in temp memory
  M3 = 0;             // if >0 then data stored in temp memory
  mDwn = 0;           // state of mouse button down timer
  msgCOM = "None";    // COM port name
  msgRx = "";         // message received from Arduino
  msgTx = "";         // message sent to Arduino
  MultiPnt = 0;       // multi-move pointer
  MXDL = 0;           // previous MXD value
  MYDL = 0;           // previous MXD value
  mXOff = 0;          // mouse X offset to control clicking error
  nextWiFims = 0;     // WiFi ms timer
  nextMillis = 0;     // loop end time in millis
  Pause = 1;          // Pause delay value, default = 1
  Ping = 50;          // ping timer to keep auto-OFF from kicking in
  PingMax = Ping;     // record the max delay between pings
  playMode = 0;       // mode ove movement play, 0 = stop; 1 = once, 2 = continuous
  QuickAng = -1;      // quick cut/paste angle value is invalid
  QuickRate = -1;     // quick cut/paste step rate value, C,A,D,S
  QuickStep = -1;     // quick cut/paste step value
  rateFlag = false;   // == true if rates are to be calculated before a redraw
  RL1 = 9;            // array pointer
  RL2 = 10;           // array pointer
  RL3 = 11;           // array pointer
  rowColPnt = -1;     // Help pointer to row column, -1 == off
  rowPnt = 0;         // pointer to current array row
  RR1 = 6;            // array pointer
  RR2 = 7;            // array pointer
  RR3 = 8;            // array pointer
  Rx$ = "";           // single character received from usb port
  RxB = 0;            // set true when a byte is received
  Scrn = 0;           // screen display pointer 0 or 1
  SelCnt = 0;         // number of selected lines
  SelLast = -1;       // most recent selected line, -1 means no recent
  Send = false;       // general flag records change and need to send data
  Sep$ = ",";         // separator used in line listing
  servoOE = 1;        // servo board Output Enable; default = 1  OFF
  servoTgt = 0;       // target/pointer servo for manual setting
  sim1st = 0;         // simulator start at row pointer
  simEnd = 0;         // simulator end pointer
  simPnt = 0;         // simulator row pointer. Note rowPnt is what is displayed
  SL0 = 5;            // array pointer
  SL1 = 6;            // array pointer
  SL1Av = 0;          // average SL1 angle
  SL2 = 7;            // array pointer
  SL2Av = 0;          // average SL2 angle
  SL3 = 8;            // array pointer
  SL3Av = 0;          // average SL3 angle
  SL4 = 9;            // array pointer
  sliderDwn = 0;      // =1 when mouse is clicked on slider
  Slope$ = "";        // temp string used in row list
  Speed = 50;         // speed value to be used by Arduino
  SpeedDef = 50;      // speed value to be used by Arduino
  SpeedLL = 1;        // speed value to be used by Arduino
  SpeedUL = 250;      // speed value to be used by Arduino
  SR0 = 0;            // array pointer
  SR1 = 1;            // array pointer
  SR1Av = 0;          // average SR1 angle
  SR2 = 2;            // array pointer
  SR2Av = 0;          // average SR2 angle
  SR3 = 3;            // array pointer
  SR3Av = 0;          // average SR3 angle
  SR4 = 4;            // array pointer
  Steps = 30;         // Walk number of steps set by the user
  TB = 0;             // text Blue colour component
  TG = 0;             // text Green colour component
  togPwr = 0;         // >0 to set all servos ON/OFF
  TR = 0;             // text Red colour component
  Tx_t0 = 0;          // usb send timers
  Tx_t1 = 0;          // usb send timers
  usbBuff = "";       // buffer string used to prevent WiFi overload
  usbEn = false;      // true to enable USB messages
  usbLen = 0;         // length of usbMsg
  usbMsg = "";        // latest message sent to the usb port
  WaklDelay = 25;     // sets the walk delay timer
  WalkHeight = 145;   // user set height in walk sequence
  WalkLift = 40;      // lift angle during walking
  WalkNext = 2147483640;  // millis() timer for Walk simulator          
  WalkSim = 0;        // a flag which controls the walk simulator, 0 == OFF
  WalkSteps = 30;     // number of steps used in walk sequences
  WiFiInt = 20;       // WiFi interval in ms
  WiFiPW = 64;        // WiFi packet width, must be longer than the max command length

  resetLftFlds(); resetRhtFlds();

}
