import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import processing.serial.*; 
import java.awt.*; 
import java.awt.datatransfer.*; 
import javax.swing.*; 
import java.io.*; 
import java.awt.datatransfer.Clipboard; 
import java.awt.datatransfer.Transferable; 
import java.awt.datatransfer.DataFlavor; 
import java.awt.datatransfer.UnsupportedFlavorException; 
import java.awt.Toolkit; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class Proc_Move_Trainer_02 extends PApplet {

// ########################################################################
//
//  Move Trainer v0.02
//
//  Released:  12/10/2020    By: TechKnowTone
//
// ########################################################################
/*
  For RoboDog
  
  This code communicates with an ESP8266 micro via the USB serial interface
  to control a droids servo motors connected to a PCA9685 over I2C. 
  
  This version includes:
  * 200 row limit
  * 12 slider fields with increment/decrement zones
  * return to REST, READY and SQUAT buttons
  * copy/paste memory data to/from the clipboard
  * help system
  * select rows
  * 4 memory stores
  * calculates steps for selected lines
  * transposes angles in selected rows
  * inserts 'Goto' commands
  * inserts 'Speed' command
  * servo power ON/OFF buttons
  * quick SHIFT + mouse copy n paste features
  * rate mode, between start and finish angles
  * list screen
  * Multi-move controls + linked
  
*/

// declare libraries

Serial usbPort;


//import java.awt.event.*;
//import java.awt.event.MouseEvent;










// declare variables
int AB;                   // text angle background Blue colour component
int AG;                   // text angle background Green colour component
int AngField = -1;        // if >= 0 then draw a box around the angle field
int AngLast;              // previous angle
int AngLast0;             // another previous angle
int AngDif;               // different in pervious to current angle
boolean AngPwr;           // used during drawing, true if the servo is powered, false if not
boolean AngRate;          // used during drawing, true if the angle is an auto-rate angle
int AnyInt;               // temp value used anywhere
int AR;                   // text angle background Red colour component
int BB;                   // text background Blue colour component
int BG;                   // text background Green colour component
boolean Blank;            // set == true is a blank line command is detected
int BR;                   // text background Red colour component
int CalcMaxHgtLft;        // calculated left-hand max height value
int CalcMaxHgtRht;        // calculated right-hand max height value
int CalcNomStepsLft;      // calculated left-hand steps value
int CalcNomStepsRht;      // calculated right-hand steps value
boolean Cmd;              // temporary flag used in display mode
int CmdCode;              // command code number
int CmdForNum;            // sim - number of loops in a For... Next command
int CmdForRow;            // sim - pointer for For... Next loop
boolean CmdRow;           // true if current row is a command line
int CmdType;              // type of command line
int CmdVal0;              // command 1st value
int comFlag;              // = 1 if COM serial available
int comListLength;        // number of items in COM connection list
String comName;           // serial port COM name
int comPnt;               // COM port list pointer
int CopyCnt;              // if > 0 then data is stored in CopyMem[]
String CopyData = "";     // data to be copied to the clipboard
String CR = "\r\n";       // carriage return for strings
String data = "";         // characters recevied from the usb port
int drawFlag;             // = 1 to force screen redraw
int DegStep;              // number of degrees per step in autostep calcs
int DX0,DX1,DY0,DY1;      // variables used in drawing walk co-ordinates
boolean Echo;             // true if Arduino responses are being received
int FieldTgt = 0;         // pointer to the field clicked
int FieldVal = 0;         // temp field value
int FL1;                  // array pointer
int FL2;                  // array pointer
int FL3;                  // array pointer
int Flag;                 // default 'Flag' value
int FldLL,FldUL;          // temp field limits
final int FldT1maxLft = 0;   // field reference pointer
final int FldT1maxRht = 1;   // field reference pointer
final int FldT2maxLft = 2;   // field reference pointer
final int FldT2minLft = 3;   // field reference pointer
final int FldT2maxRht = 4;   // field reference pointer
final int FldT2minRht = 5;   // field reference pointer
final int FldT1WlkLft = 6;   // field reference pointer
int FR1;                  // array pointer
int FR2;                  // array pointer
int FR3;                  // array pointer
int GndY;                 // simulator ground reference Y-axis
int GndYM;                // simulator ground Y-axis max
String Help$;             // message generated by help system
String HelpLast$;         // previous message generated by help system
int HelpOFF;              // if >0 then help system is disabled
int HelpRowPnt;           // >= 0 if mouse is over a valid row number
int HX0,HX1,HX2,HX3,HY0,HY1,HY2,HY3;  // X/Y co-ordinates for tier 1 leg highlight drawing
PImage img0;              // image loaded 0 as background
PImage img1;              // image loaded 1 as background
PImage img2;              // image loaded 1 as background
int Inc;                  // any value increment
int IncCnt;               // any increment counter
int index;
int interval;             // main loop interval in milliseconds
int JoyX0;                // x-coordinate of joystick slider left end
int JoyX1;                // x-coordinate of joystick slider right end
int keyDwn;               // tracks keyboard up/down activity, 0 = up, 1 = down
boolean keyESC;           // true if ESC key is pressed
boolean keySHIFT;         // true whenever the SHIFT key is pressed
int LastVal;              // any previous value
int LF;                   // ASCII line feed
int M0;                   // if >0 then data stored in temp memory
int M1;                   // if >0 then data stored in temp memory
int M2;                   // if >0 then data stored in temp memory
int M3;                   // if >0 then data stored in temp memory
int mDwn;                 // state of mouse button down timer
int mB, mON, mX, mXD, mY, mYD;    // mouse flag, x,y, down
String msgCOM;            // COM port name
String msgRx;             // message received from Arduino
String msgTx;             // message sent to Arduino
int Multi0,Multi1;        // servo pointers in multimove functions
int MultiPnt;             // multi-move pointer
int mXL,mYL;              // previous x,y values
int MXDL;                 // previous MXD value
int MYDL;                 // previous MXD value
int mXOff;                // mouse X offset to control clicking error
int nextWiFims;           // WiFi ms timer
int nextMillis;           // loop end time in millis
int Pause;                // Pause delay value, default = 1
int Ping = 50;            // ping timer to keep auto-OFF from kicking in
int PingMax = Ping;       // record the max delay between pings
int playMode;             // mode ove movement play, 0 = stop; 1 = once, 2 = continuous
int QuickAng;             // quick cut/paste angle value
int QuickRate;            // quick cut/paste step rate value, C,A,D,S
int QuickStep;            // quick cut/paste step value
boolean Rated;            // == true if there are rate values in the list
boolean rateFlag;         // == true if rates are to be calculated before a redraw
String Released$ = "12/10/2020";  // latest update
int RL1;                  // array pointer
int RL2;                  // array pointer
int RL3;                  // array pointer
int rowColPnt;            // Help pointer to row column
int rowPnt;               // pointer to current array row
int RR1;                  // array pointer
int RR2;                  // array pointer
int RR3;                  // array pointer
String Rx$;               // single character received from usb port
int RxB;                  // set true when a byte is received
int Scrn;                 // screen display pointer 0 or 1
boolean Sel;              // temporary selected flag
int SelCnt;               // number of selected lines
int SelLast;              // most recent selected line, -1 means no recent
boolean Send;             // general flag records change and need to send data
String Sep$;              // separator used in line listing
int servoOE;              // servo board Output Enable; default = 1  OFF
int servoTgt;             // target/pointer servo for manual setting
int sim1st;               // simulator start at row pointer
int simEnd;               // simulator end pointer
int simPnt;               // simulator row pointer. Note rowPnt is what is displayed
int SL0;                  // array pointer
int SL1;                  // array pointer
int SL1Av;                // average SL1 angle
int SL2;                  // array pointer
int SL2Av;                // average SL2 angle
int SL3;                  // array pointer
int SL3Av;                // average SL3 angle
int SL4;                  // array pointer
int sliderDwn;            // =1 when mouse is clicked on slider
String Slope$;            // temp string used in row list
int Speed;                // speed value to be used by Arduino
int SpeedDef;             // speed value to be used by Arduino
int SpeedLL;              // speed value to be used by Arduino
int SpeedUL;              // speed value to be used by Arduino
int SR0;                  // array pointer
int SR1;                  // array pointer
int SR1Av;                // average SR1 angle
int SR2;                  // array pointer
int SR2Av;                // average SR2 angle
int SR3;                  // array pointer
int SR3Av;                // average SR3 angle
int SR4;                  // array pointer
int Steps;                // Walk number of steps set by the user
int t0,t1;                // millis() timers
int T1maxAngLft;          // Walk Trainer T1 maxAng left side
int T1maxAngRht;          // Walk Trainer T1 maxAng right side
int T2maxAngLft;          // Walk Trainer T2 maxAng left side
int T2maxAngRht;          // Walk Trainer T2 maxAng right side
int T2midAngLft;          // Walk Trainer T2 midAng left side
int T2midAngRht;          // Walk Trainer T2 midAng right side
int T2minAngLft;          // Walk Trainer T2 minAng left side
int T2minAngRht;          // Walk Trainer T2 minAng right side
int TB;                   // text Blue colour component
int TG;                   // text Green colour component
int togPwr;               // >0 to set all servos ON/OFF
int TR;                   // text Red colour component
int Tx_t0;                // usb send timers
int Tx_t1;                // usb send timers
String usbBuff;           // buffer string used to prevent WiFi overload
boolean usbEn;            // true to enable USB messages
int usbLen;               // length of usbMsg
String usbMsg;            // latest message sent to the usb port
int usbOnCnt;             // Usn enabled warning timer
int WalkHeight;           // user set height in walk sequence
int WalkLift;             // lift angle during walking
int WalkNext;             // micros() timer for Walk simulator          
int WalkSim;              // a flag which controled the walk simulator
int WaklDelay;            // sets the walk delay timer
int WalkSteps;            // number of steps used in walk sequences
int WiFiInt;              // WiFi interval in ms
int WiFiPW;               // WiFi packet width, must be longer than the max command length

// declare arrays
int rowDepth = 200;        // the number of rows in the array
int rowMax = rowDepth - 1; // one less than the maximum number of rows supported by rowDepth
int rowNum = 1;            // number of rows in this list
int rowWidth = 13;         // the width of each row
int arrayMax = rowDepth * rowWidth;
//               FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,NA.
int [] AngVal = {  0,  0, 90,  0,  0, 90,  0,  0, 90,  0,  0, 90,  0};  // servo angles
int [] AngVLL = {  0,  0, 70,  0,  0, 50,  0,  0, 70,  0,  0, 50,  0};  // servo angle lower limit
int [] AngVUL = {160,140,130,160,140,110,160,140,130,160,140,110,  0};  // servo angle upper limit
int [] ArrayVal = new int [arrayMax]; // servo memory
int [] Ch     = {  1,  3,  5,  0,  2,  4, 14, 12, 10, 15, 13, 11};  // maps SRn/SLn array to servo channels
int [] CopyMem = new int [arrayMax]; // copy memory
int [] JoyXval = new int [13];  // joystick slider position values
//               FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,NA.
int [] JoyXLL = {  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0};  // form slider limits
int [] JoyXUL = {360,200,200,360,200,200,360,200,200,360,200,200,360};  // form slider limits
int [] Link   = {  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0};  // servo link flags
int [] Mem0 = new int [arrayMax];  // temp memory store
int [] Mem1 = new int [arrayMax];  // temp memory store
int [] Mem2 = new int [arrayMax];  // temp memory store
int [] Mem3 = new int [arrayMax];  // temp memory store
int [] MaxA = new int [rowWidth];  // temp angle max
int [] MinA = new int [rowWidth];  // temp angle min
boolean [] Select = new boolean [arrayMax];
int [] Temp = new int [rowWidth]; // temp row store

public void setup() {
  // first code to run. Also run again by a soft reset
  setDefaults();
  ClearArray(0);
  ClearLinks();
  ClearSelect();
  setSliderX();
    // screen size must match background bitmap
  surface.setTitle("RoboDog Move Trainer");
  img0 = loadImage("Move Trainer 0.png");
  img1 = loadImage("Move Trainer 1.png");
  img2 = loadImage("Move Trainer 2.png");
  getCOMPort();
  textSize(13);
  println("Ready!");
  nextMillis = millis() + interval;
  nextWiFims = millis() + WiFiInt;
}

// ----------------------------------------------------------------------

public void draw() {
  // main loop. Comes here about 60 times a second
  if (drawFlag > 0) {
    if (rateFlag) {calcRates();}  // perform a rate calculation before a refresh
    if (Scrn == 0) {
      // Trainer screen
      background(img0);  // define background image
      drawAngles();      // put numbers in angle fields
      drawAngField();    // put red box round selected angle
      drawRates();       // fill in 'rate' buttons
      drawLinks();       // draw boxes around link buttons
      drawDegSteps();    // update Steps field
      drawSpeed();       // update Speed field
    } else if (Scrn == 1){
      // List screen
      background(img1); // define background image
      drawAngPos();      // draw angle position heighlights on RoboDog image
      drawAngField();    // put red box round respective servos
      drawDegSteps();    // update Steps field
      drawSpeed();       // update Speed field
    } else {
      // Simulator screen
      background(img2); // define background image
      drawWalkLegs();
      drawWalkAngles();
    }
    drawHidden();
    drawArrayRows();
    drawMemBars();
    drawCOMMsg(); drawRxMsg(); drawTxMsg(); drawUsbLamp(); drawTxRxTime();
    drawEchoLamp(); drawHelp();
    if (rowColPnt >= 0) {drawColHighlight();}  // draw these last/on top
    drawFlag = 0;
  }
  if (mON > 0) {drawMouseXY();}
  
  // perform usb buffer task
  if (millis() >= nextWiFims) {
    nextWiFims = nextWiFims + WiFiInt;
    // the following should never be true unless significant time is lost somewhere
    if (nextWiFims < millis()) {nextWiFims = millis() + WiFiInt;}
    // only send data to the usb port if it is active
    if (comFlag > 0) {usbPortOp();} else {usbBuff = "";}
  }
  
  // perform Walk simulator function
  if (WalkSim > 0) {
    // in active Walk mode, so check the timer
    // if WalkSim < 0 then in Pause state
    if (millis() >= WalkNext) {
      t1 = millis() - t0;  // capture slack space time
      Walk_Sim();
      t0 = millis();  // record start of free time
    }
  }
  
  // perform 20ms tasks
  if (millis() >= nextMillis) {
    nextMillis = nextMillis + interval;
    // the following should never be true unless significant time is lost somewhere
    if (nextMillis < millis()) {nextMillis = millis() + interval;}
    
    // perform auto-repeat on some buttons
    if (mDwn > 0) {
      // perform auto repeat functions
      mDwn--;
      if (mDwn < 1) {mDwn = 5; mouseDown();}
    }
    // send a 'ping' character every second to keep servo power ON
    Ping--; if (Ping < 1) {Ping = PingMax; usbPortWrite("#");}
    if (RxB > 0) {RxB--;}
    // Usb wardning timer
    usbOnCnt++;
    if (usbOnCnt == 10) {drawFlag = 1;;}
    if (usbOnCnt >= 20) {usbOnCnt = 0; drawFlag = 1;}
  }
}

// ----------------------------------------------------------------------

public void keyPressed() {
  // used to receive keyboard keys
  if (key == ESC) {keyESC = true; key = 0; ESC(); return;}
  
  HelpRowPnt = -1;  // for Scrn1 the current row will drive the highlights
  keyDwn = 1;
       if (keyCode == DOWN) {incRowPnt();}
  else if (keyCode == SHIFT) {keySHIFT = true;}
  else if (keyCode == UP) {decRowPnt();}
  else if (keyCode == 88) {if ((int)key == 24) {cutRows();}}    // Ctrl+X key
  else if (keyCode == 86) {if ((int)key == 22) {pasteRows();}}  // Ctrl+V key
  else if (keyCode == 82) {keyToggleRate();} // R key
  else if (keyCode == 67) {if ((int)key ==  3) {copyRows();}}   // Ctrl+C key
  else {
   println("key: " + key);
   println("int key: " + (int) key);
   println("keyCode: " + keyCode);
   println();
  }
}

// ----------------------------------------------------------------------

public void keyReleased() {
  // called whenever a key is released
  keyDwn = 0;keySHIFT = false; keyESC = false;
}

// ----------------------------------------------------------------------

public void mouseDragged() {
  // mouse has been clicked and dragged on window
  mX = mouseX; mY = mouseY;
  
       if (Scrn == 0) {mouseDragged_0();}
  else if (Scrn == 1) {mouseDragged_1();}
  else {mouseDragged_2();}
}

// ----------------------------------------------------------------------

public void mouseDragged_0() {
  // mouse has been clicked and dragged on window on Scrn0

  if (sliderDwn > 0) {
    // ignore mY if it goes outside of slider region
    readXSlider(servoTgt); return;
  }
  
  if (MultiPnt > 0) {
    MultiMoveXY(Multi0,Multi1); return;
  }
}

// ----------------------------------------------------------------------

public void mouseDragged_1() {
  // mouse has been clicked and dragged on window on Scrn1

}

// ----------------------------------------------------------------------

public void mouseDragged_2() {
  // mouse has been clicked and dragged on window on Scrn2

  if (sliderDwn > 0) {
    // ignore mY if it goes outside of slider region
    adjustField(FieldTgt); return;
  }
}

// ----------------------------------------------------------------------

public void mouseDown() {
  // called from main loop to repeat certain events after a pause delay
  if (sliderDwn > 0) {mDwn = 0; return;}
  if (MultiPnt > 0) {mDwn = 0; return;}
  if (togPwr > 0) {togglePwrAll(); return;}
  
  if (Scrn == 0) {mouseDown_0();}
  else if (Scrn == 1) {mouseDown_1();}
  else {mouseDown_2();}

  //#####################################################################
  //
  // list lines region
  //
  //#####################################################################
  int zmY = 547; if (Scrn == 1) {zmY = 41;}
  if ((mY >= zmY) && (mY <= 686)) {
    // angles
    if ((mX >= 154) && (mX < 494)) {changeAngleVal(); return;}

    // steps field
    if ((mX >= 543) && (mX < 582)) {changeSteps(); return;} // inc/dec step value

    // rate field
    if ((mX >= 512) && (mX < 537)) {mDwn = 0; resetStepRates(); return;}
  }
  
  if (((mY >= 548) && (mY <= 684))) {
    // Memory sliders and buttons

    if ((mX >= 585) && (mX <= 611)) {
      // vertical slider region
      if (mY < 566) {usbEn = false; decRowPnt(); return;}
      if (mY >= 666) {usbEn = false; incRowPnt(); return;}
    }
    if (((mY >= 559) && (mY <= 595))) {
      // Memory buttons top row
      if ((mX >= 691) && (mX < 719)) {mDwn = 0; storeM0(); return;}  // M0 button held-down event
      if ((mX >= 719) && (mX < 749)) {mDwn = 0; storeM1(); return;}  // M1 button held-down event
      if ((mX >= 749) && (mX < 778)) {mDwn = 0; storeM2(); return;}  // M2 button held-down event
      if ((mX >= 778) && (mX <= 807)) {mDwn = 0; storeM3(); return;}  // M3 button held-down event
    }
    if ((mY >= 636) && (mY <= 671)) {
      // Memory buttons 3rd row
      if ((mX >= 691) && (mX < 749)) {ClearArray(1); ClearSelect(); mDwn = 0; cursor(ARROW); return;}
      if ((mX >= 778) && (mX < 807)) {
        // paste key
        getTextFromClipboard(); ClearSelect(); ClearLinks();
        mDwn = 0; cursor(ARROW); return;
      }
    }
  }
  
  if ((mY >= 703) && (mY < 741)) {
    // Processing logo - soft restart
    if ((mX >= 12) && (mX <= 49)) {
      // reset everything but keep the current screen mode
      AnyInt = Scrn; mDwn = 0; closeCOMPort(); setup(); Scrn = AnyInt;
    }
  }
}

// ----------------------------------------------------------------------

public void mouseDown_0() {
  // called to repeat certain events after a pause delay on Scrn 0
  
  if ((mY >= 122) && (mY <= 148)) {
      // Level 3 text fields
      if ((mX >= 148) && (mX < 160)) {incXSlider(FR3); return;}
      if ((mX > 186) && (mX <= 198)) {decXSlider(FR3); return;}

      if ((mX >= 250) && (mX < 262)) {incXSlider(FL3); return;}
      if ((mX > 287) && (mX <= 299)) {decXSlider(FL3); return;}

      if ((mX >= 376) && (mX < 388)) {incXSlider(RR3); return;}
      if ((mX > 413) && (mX <= 425)) {decXSlider(RR3); return;}

      if ((mX >= 476) && (mX < 488)) {incXSlider(RL3); return;}
      if ((mX > 514) && (mX <= 526)) {decXSlider(RL3); return;}

  } else if ((mY >= 260) && (mY <= 286)) {
      // Level 2 text fields
      if ((mX >= 174) && (mX < 186)) {incXSlider(FR2); return;}
      if ((mX > 211) && (mX <= 223)) {decXSlider(FR2); return;}

      if ((mX >= 275) && (mX < 287)) {incXSlider(FL2); return;}
      if ((mX > 312) && (mX <= 324)) {decXSlider(FL2); return;}

      if ((mX >= 400) && (mX < 412)) {incXSlider(RR2); return;}
      if ((mX > 438) && (mX <= 450)) {decXSlider(RR2); return;}

      if ((mX >= 501) && (mX < 513)) {incXSlider(RL2); return;}
      if ((mX > 539) && (mX <= 551)) {decXSlider(RL2); return;}

  } else if ((mY >= 385) && (mY <= 411)) {
      // Level 1 text fields
      if ((mX >= 286) && (mX < 298)) {decXSlider(FR1); return;}
      if ((mX > 323) && (mX <= 335)) {incXSlider(FR1); return;}

      if ((mX >= 385) && (mX < 397)) {decXSlider(FL1); return;}
      if ((mX > 422) && (mX <= 434)) {incXSlider(FL1); return;}

      if ((mX >= 512) && (mX < 524)) {decXSlider(RR1); return;}
      if ((mX > 549) && (mX <= 561)) {incXSlider(RR1); return;}

      if ((mX >= 612) && (mX < 624)) {decXSlider(RL1); return;}
      if ((mX > 649) && (mX <= 661)) {incXSlider(RL1); return;}
  }
  
  //#####################################################################
  //
  // Multi-move components
  //
  //#####################################################################
  if ((mY >= 488) && (mY <= 545)) {
    if ((mX >= 222) && (mX <= 607)) {
      if (mY <= 502) {
        if ((mX >= 235) && (mX <= 265)) {MultiDecY(FR1,FR2); return;}
        if ((mX >= 337) && (mX <= 367)) {MultiDecY(FL1,FL2); return;}
        if ((mX >= 462) && (mX <= 491)) {MultiDecY(RR1,RR2); return;}
        if ((mX >= 564) && (mX <= 594)) {MultiDecY(RL1,RL2); return;}
      }
      else if (mY <= 531) {
        if (mX <= 235) {incXSlider(FR2);}
        if ((mX >= 265) && (mX <= 277)) {decXSlider(FR2); return;}

        if ((mX >= 324) && (mX <= 337)) {incXSlider(FL2); return;}
        if ((mX >= 367) && (mX <= 380)) {decXSlider(FL2); return;}

        if ((mX >= 449) && (mX <= 462)) {incXSlider(RR2); return;}
        if ((mX >= 491) && (mX <= 504)) {decXSlider(RR2); return;}

        if ((mX >= 551) && (mX <= 564)) {incXSlider(RL2); return;}
        if ((mX >= 594) && (mX <= 607)) {decXSlider(RL2); return;}
      }
      else {
        if ((mX >= 235) && (mX <= 265)) {MultiIncY(FR1,FR2); return;}
        if ((mX >= 337) && (mX <= 367)) {MultiIncY(FL1,FL2); return;}
        if ((mX >= 462) && (mX <= 491)) {MultiIncY(RR1,RR2); return;}
        if ((mX >= 564) && (mX <= 594)) {MultiIncY(RL1,RL2); return;}
      }
    }
  }
  
  if ((mY >= 515) && (mY <= 539)) {
    // Step buttons and field
         if ((mX >= 12) && (mX < 31)) {decDegSteps(); return;}
    else if ((mX >= 81) && (mX <= 99)) {incDegSteps(); return;}

    // Speed buttons and field
    else if ((mX >= 722) && (mX < 740)) {decSpeed(); return;}
    else if ((mX >= 791) && (mX <= 809)) {incSpeed(); return;}
  }
}

// ----------------------------------------------------------------------

public void mouseDown_1() {
  // called to repeat certain events after a pause delay on Scrn 1
  if ((mY >= 515) && (mY <= 539)) {
    // Step buttons and field
         if ((mX >= 12) && (mX < 31)) {decDegSteps(); return;}
    else if ((mX >= 81) && (mX <= 99)) {incDegSteps(); return;}

    // Speed buttons and field
    else if ((mX >= 722) && (mX < 740)) {decSpeed(); return;}
    else if ((mX >= 791) && (mX <= 809)) {incSpeed(); return;}
  }
}

// ----------------------------------------------------------------------

public void mouseDown_2() {
  // called to repeat certain events after a pause delay on Scrn 2
  
  if ((mY >= 404) && (mY <= 430)) {
    // T2 max field
         if ((mX >= 103) && (mX < 115)) {decField(FldT2maxLft); return;}
    else if ((mX >= 141) && (mX <= 153)) {incField(FldT2maxLft); return;}
 
    // T2 min field
         if ((mX >= 273) && (mX < 295)) {decField(FldT2minLft); return;}
    else if ((mX >= 310) && (mX <= 322)) {incField(FldT2minLft); return;}

    // T2 max right field
         if ((mX >= 567) && (mX < 579)) {decField(FldT2maxRht); return;}
    else if ((mX >= 605) && (mX <= 617)) {incField(FldT2maxRht); return;}

    // T2 min right field
         if ((mX >= 737) && (mX < 749)) {decField(FldT2minRht); return;}
    else if ((mX >= 774) && (mX <= 786)) {incField(FldT2minRht); return;}
 }

  if ((mY >= 454) && (mY <= 480)) {
    // Speed buttons and field
         if ((mX >= 368) && (mX < 386)) {decSpeed(); return;}
    else if ((mX >= 437) && (mX <= 455)) {incSpeed(); return;}
  }

  if ((mY >= 470) && (mY <= 496)) {
    // T1 max left field
         if ((mX >= 103) && (mX < 115)) {decField(FldT1maxLft); return;}
    else if ((mX >= 141) && (mX <= 153)) {incField(FldT1maxLft); return;}

    // T1 max right field
         if ((mX >= 567) && (mX < 579)) {decField(FldT1maxRht); return;}
    else if ((mX >= 605) && (mX <= 617)) {incField(FldT1maxRht); return;}
  }
}

// ----------------------------------------------------------------------

public void mouseMoved() {
  // used primarily with the help system
  mX = mouseX; mY = mouseY; mXD = mX; mYD = mY;
  Help_Message();
  if (mON > 0) {drawFlag = 1;}
}

// ----------------------------------------------------------------------

public void mousePressed() {
  // mouse has been clicked on window
  // there are common and screen specific functions
  // the common functions are here
  mX = mouseX; mY = mouseY; mB = mouseButton; mDwn = 25;
  cursor(HAND);

  if ((mX < 50) && (mY < 50)) {
    // toggle mouse X,Y display ON/OFF
    if (mON < 1) {mON = 1;} else {mON = 0; drawFlag = 1;}
  }
  
  mXD = mX; mYD = mY; // record mouse down co-ordinates
  grabRowInTemp(rowPnt);
  
  if ((mX >= 377) && (mX <= 446) && (mY < 36)) {
    // change the screen mode and reset flags
    if (mB == LEFT) {
      Scrn++; if (Scrn > 2) {Scrn = 0;}
    } else {
      Scrn--; if (Scrn < 0) {Scrn = 2;}
    }
    drawFlag = 1; mDwn = 0; WalkSim = 0; return;
  }

  if ((mY >= 13) && (mY <= 47)) {
    // top function 1st row buttons, REST, READY, STAND
    if ((mX >= 601) && (mX < 671)) {insertRest();}
    else if ((mX >= 671) && (mX < 741)) {insertReady();}
    else if ((mX >= 741) && (mX <= 810)) {insertStand();}
  }
  else if ((mY >= 48) && (mY <= 82)) {
    // top function 2nd row buttons: TBD, TBD, SQUAT
    if ((mX >= 601) && (mX < 671)) {return;}
    else if ((mX >= 671) && (mX < 741)) {return;}
    else if ((mX >= 741) && (mX <= 810)) {insertSquat();}
  }

  //#####################################################################
  //
  // Screen specific functions
  //
  //#####################################################################
  if (Scrn == 0) {mousePressed_0();}
  if (Scrn == 1) {mousePressed_1();}
  else {mousePressed_2();}
  

  //#####################################################################
  //
  // Common screen functions
  //
  //#####################################################################
  if ((mY >= 547) && (mY <= 686)) {
    // Memory lines, sliders and buttons
    if (mX <= 105) {
      // left-hand side buttons and arrows
      if (mY <= 580) {
        // left region top row
        if ((mX >= 72) && (mX <= 102)) {moveSelUp();}
      } else if (mY <= 616) {
        // left region 2nd row
             if (mX < 39) {cutRows();}
        else if ((mX >= 39) && (mX < 71)) {copyRows();}
      } else if (mY <= 650) {
        // left region 3rd row
             if ((mX >= 9) && (mX < 39)) {insertGoto();}    // insert a Goto selected row branch
        else if ((mX >= 39) && (mX < 71)) {calcSelSteps();}  // calc. steps for selected rows
        else if ((mX >= 71) && (mX <= 102)) {transSel();}   // transpose selected rows
      } else {
        // left region 4th row
             if ((mX >= 39) && (mX < 71)) {reverseSelRows();}
        else if ((mX >= 71) && (mX <= 102)) {moveSelDwn();}
      }
    }
    if (CopyCnt > 0) {  
      if ((mY > 575) && (mY <= 614)) {  
        if ((mX >= 71) && (mX <= 102)) {pasteRows();}
      }  
    }
    
    if ((mX >= 585) && (mX <= 611)) {
      // vertical slider region
           if (mY < 566) {decRowPnt();}
      else if ((mY >= 566) && (mY < 589)) {gotoFirst();}
      else if ((mY >= 642) && (mY < 665)) {gotoLast();}
      else if (mY >= 665) {incRowPnt();}
    }
    
    if ((mX >= 614) && (mX <= 643)) {
      //R-hand insert buttons Blank,Delete,Pause,Balance
           if ((mY >= 541) && (mY <= 576)) {insertBlank();}
      else if ((mY >= 578) && (mY <= 614)) {deleteRows();}
      else if ((mY >= 616) && (mY <= 651)) {insertPause();}
      else if ((mY >= 653) && (mY <= 689)) {insertBalance();}
    }
    if ((mX >= 646) && (mX <= 675)) {
      //R-hand insert buttons Speed,New,Flag,TBD
           if ((mY >= 541) && (mY <= 576)) {insertSpeed();}
      else if ((mY >= 587) && (mY <= 614)) {newRow();}
      else if ((mY >= 616) && (mY <= 651)) {InsertFlag();}
      else if ((mY >= 653) && (mY <= 689)) {InsertExCode();}
    }
    
    if ((mY >= 559) && (mY <= 595)) {
      // Memory buttons top row
           if ((mX >= 691) && (mX < 719)) {mDwn = 50;}  // M0 button. Set timer for held-down event
      else if ((mX >= 719) && (mX < 749)) {mDwn = 50;}  // M1 button. Set timer for held-down event
      else if ((mX >= 749) && (mX < 778)) {mDwn = 50;}  // M2 button. Set timer for held-down event
      else if ((mX >= 778) && (mX < 808)) {mDwn = 50;}  // M3 button. Set timer for held-down event
    }
    if ((mY >= 598) && (mY <= 633)) {
      // Play buttons 2nd row
      if (usbEn || (Scrn == 2)) {
             if ((mX >= 691) && (mX < 719)) {playOnce();} // play once
        else if ((mX >= 719) && (mX < 749)) {playConst();} // play continuously
        else if ((mX >= 749) && (mX < 778)) {togglePlayPause();} // pause
        else if ((mX >= 778) && (mX <= 807)) {playStop();} // stop playing
      }
    }
    if ((mY >= 636) && (mY <= 671)) {
      // 3rd row buttons - Clear,Copy,Paste
           if ((mX >= 691) && (mX < 749)) {mDwn = 50; cursor(WAIT);}  // clear memory if held down
      else if ((mX >= 749) && (mX < 778)) {copyToClipboard();}
      else if ((mX >= 778) && (mX < 807)) {mDwn = 50; cursor(WAIT);}  // paste from clipboard if held down
    }
  }

  //#####################################################################
  //
  // list lines region
  //
  //#####################################################################
  int zmY = 547; if (Scrn == 1) {zmY = 41;}
  if ((mY >= zmY) && (mY <= 686)) {
    if ((mX >= 105) && (mX < 582)) {
      // memory lines region
      if ((mX >= 105) && (mX < 138)) {toggleRowSelect(); return;}
      if ((mX >= 154) && (mX < 494)) {changeAngleVal(); return;}

      // rate profile field  
      if ((mX >= 512) && (mX < 537)) {mDwn = 50; toggleStepRate(); return;}

      // steps field
      if ((mX >= 543) && (mX < 582)) {changeSteps(); return;} // inc/dec step value
    }
  }
  
  if ((mY >= 703) && (mY < 741)) {
    // Processing logo - soft restart
    if ((mX >= 12) && (mX <= 49)) {
      mDwn = 50; cursor(WAIT);
    }
  }
  if ((mY >= 690) && (mY < 716)) {
    // Com, RX, Tx fields
    // clicked on COM port?
    if ((mX >= 105) && (mX <= 180)) {
      if (mouseButton == LEFT) {getCOMPort();}
      else if (mouseButton == RIGHT) {closeCOMPort();}
    }
    if ((mX >= 478) && (mX <= 681)) {toggleUsbEn();}
  }
  if ((mY >= 686) && (mY <= 721)) {
    // Echo button?
    if ((mX >= 749) && (mX <= 807)) {toggleEcho();}
  }
  // if none of the above check for cancelling selects[]
  if ((SelCnt > 0) && (Help$.equals("-"))) {ClearSelect(); drawFlag = 1;};

}

// ----------------------------------------------------------------------

public void mousePressed_0() {
  // screen 0 specific mouse pressed functions
  
  if (ArrayVal[rowPnt * rowWidth] < 0) {
    // command so make top part of form inactive, except REST and READY
  } else {
    if (!Help$.equals("-")) {cursor(HAND);}
    
    //#####################################################################
    //
    // Tier 3 components
    //
    //#####################################################################
    if ((mY >= 122) && (mY < 148)) {
        // Level 3 text fields & links
        if ((mX >= 148) && (mX < 160)) {incXSlider(FR3);}
        else if ((mX >= 160) && (mX < 186)) {readXSlider(FR3);}
        else if ((mX >= 186) && (mX < 198)) {decXSlider(FR3);}
  
        else if ((mX >= 198) && (mX < 215)) {copyLftToRgt(FL3, FR3);}
        else if ((mX >= 215) && (mX < 233)) {toggleLink(FR3,FL3);}
        else if ((mX >= 233) && (mX < 250)) {copyRgtToLft(FR3, FL3);}
  
        else if ((mX >= 250) && (mX < 262)) {incXSlider(FL3);}
        else if ((mX >= 262) && (mX <= 287)) {readXSlider(FL3);}
        else if ((mX > 287) && (mX <= 299)) {decXSlider(FL3);}

        else if ((mX >= 327) && (mX <= 353)) {toggleMultiLink(FR3);}

        else if ((mX >= 376) && (mX < 388)) {incXSlider(RR3);}
        else if ((mX >= 388) && (mX <= 413)) {readXSlider(RR3);}
        else if ((mX > 413) && (mX <= 425)) {decXSlider(RR3);}
  
        else if ((mX >= 425) && (mX < 441)) {copyLftToRgt(RL3, RR3);}
        else if ((mX >= 441) && (mX < 459)) {toggleLink(RL3,RR3);}
        else if ((mX >= 459) && (mX < 476)) {copyRgtToLft(RR3, RL3);}
  
        else if ((mX >= 476) && (mX < 488)) {incXSlider(RL3);}
        else if ((mX >= 488) && (mX <= 514)) {readXSlider(RL3);}
        else if ((mX > 514) && (mX <= 526)) {decXSlider(RL3);}

    } else if ((mY >= 148) && (mY <= 168)) {
        // Level 3 transposes, Pwr & Rate buttons
        if ((mX >= 165) && (mX <= 182)) {togglePwr(FR3);}
        else if ((mX >= 183) && (mX <= 198)) {toggleRate(FR3);}
        else if ((mX >= 213) && (mX < 234)) {transFR3FL3();}
        else if ((mX >= 266) && (mX <= 284)) {togglePwr(FL3);}
        else if ((mX >= 285) && (mX <= 299)) {toggleRate(FL3);}
        else if ((mX >= 391) && (mX <= 409)) {togglePwr(RR3);}
        else if ((mX >= 410) && (mX <= 425)) {toggleRate(RR3);}
        else if ((mX >= 439) && (mX < 461)) {transRR3RL3();}
        else if ((mX >= 492) && (mX <= 510)) {togglePwr(RL3);}
        else if ((mX >= 510) && (mX <= 526)) {toggleRate(RL3);}

    //#####################################################################
    //
    // Tier 2 components
    //
    //#####################################################################
    } else if ((mY >= 260) && (mY < 286)) {
        // Level 2 text fields & links
        if ((mX >= 174) && (mX < 186)) {incXSlider(FR2);}
        else if ((mX >= 186) && (mX <= 211)) {readXSlider(FR2);}
        else if ((mX > 211) && (mX <= 223)) {decXSlider(FR2);}
  
        else if ((mX >= 223) && (mX < 240)) {copyLftToRgt(FL2, FR2);}
        else if ((mX >= 240) && (mX < 258)) {toggleLink(FR2,FL2);}
        else if ((mX >= 258) && (mX < 275)) {copyRgtToLft(FR2, FL2);}
  
        else if ((mX >= 275) && (mX < 287)) {incXSlider(FL2);}
        else if ((mX >= 287) && (mX <= 312)) {readXSlider(FL2);}
        else if ((mX > 312) && (mX <= 324)) {decXSlider(FL2);}

        else if ((mX >= 354) && (mX <= 379)) {toggleMultiLink(FR2);}
  
        else if ((mX >= 400) && (mX < 412)) {incXSlider(RR2);}
        else if ((mX >= 412) && (mX <= 438)) {readXSlider(RR2);}
        else if ((mX > 438) && (mX <= 450)) {decXSlider(RR2);}
  
        else if ((mX >= 450) && (mX < 467)) {copyLftToRgt(RL2, RR2);}
        else if ((mX >= 467) && (mX < 485)) {toggleLink(RR2,RL2);}
        else if ((mX >= 485) && (mX < 501)) {copyRgtToLft(RR2, RL2);}
  
        else if ((mX >= 501) && (mX < 513)) {incXSlider(RL2);}
        else if ((mX >= 513) && (mX <= 539)) {readXSlider(RL2);}
        else if ((mX > 539) && (mX <= 551)) {decXSlider(RL2);}

    } else if ((mY >= 286) && (mY <= 305)) {
        // Level 2 transposes, Pwr & Rate buttons
        if ((mX >= 174) && (mX <= 188)) {toggleRate(FR2);}
        else if ((mX >= 189) && (mX <= 206)) {togglePwr(FR2);}
        else if ((mX >= 238) && (mX < 258)) {transFR2FL2();}
        else if ((mX >= 275) && (mX <= 289)) {toggleRate(FL2);}
        else if ((mX >= 290) && (mX <= 308)) {togglePwr(FL2);}
        else if ((mX >= 400) && (mX <= 415)) {toggleRate(RR2);}
        else if ((mX >= 416) && (mX <= 434)) {togglePwr(RR2);}
        else if ((mX >= 465) && (mX < 485)) {transRR2RL2();}
        else if ((mX >= 500) && (mX <= 516)) {toggleRate(RL2);}
        else if ((mX >= 517) && (mX <= 535)) {togglePwr(RL2);}

    //#####################################################################
    //
    // Tier 1 components
    //
    //#####################################################################
    } else if ((mY >= 385) && (mY < 411)) {
        // Level 1 text fields & links
        if ((mX >= 286) && (mX < 298)) {decXSlider(FR1);}
        else if ((mX >= 298) && (mX <= 323)) {readXSlider(FR1);}
        else if ((mX > 323) && (mX <= 335)) {incXSlider(FR1);}
  
        else if ((mX >= 335) && (mX < 352)) {copyLftToRgt(FL1, FR1);}
        else if ((mX >= 352) && (mX < 370)) {toggleLink(FR1,FL1);}
        else if ((mX >= 370) && (mX < 385)) {copyRgtToLft(FR1, FL1);}
  
        else if ((mX >= 385) && (mX < 397)) {decXSlider(FL1);}
        else if ((mX >= 397) && (mX <= 422)) {readXSlider(FL1);}
        else if ((mX > 422) && (mX <= 434)) {incXSlider(FL1);}

        else if ((mX >= 463) && (mX <= 489)) {toggleMultiLink(FR1);}
  
        else if ((mX >= 512) && (mX < 524)) {decXSlider(RR1);}
        else if ((mX >= 524) && (mX <= 549)) {readXSlider(RR1);}
        else if ((mX > 549) && (mX <= 561)) {incXSlider(RR1);}
  
        else if ((mX >= 561) && (mX < 578)) {copyLftToRgt(RL1, RR1);}
        else if ((mX >= 578) && (mX < 596)) {toggleLink(RR1,RL1);}
        else if ((mX >= 596) && (mX < 612)) {copyRgtToLft(RR1, RL1);}
  
        else if ((mX >= 612) && (mX < 624)) {decXSlider(RL1);}
        else if ((mX >= 624) && (mX <= 649)) {readXSlider(RL1);}
        else if ((mX > 649) && (mX <= 661)) {incXSlider(RL1);}

    } else if ((mY >= 411) && (mY <= 430)) {
        // Level 1 transposes, Pwr & Ratebuttons
        if ((mX >= 286) && (mX <= 300)) {toggleRate(FR1);}
        else if ((mX >= 301) && (mX <= 320)) {togglePwr(FR1);}
        else if ((mX >= 350) && (mX < 370)) {transFR1FL1();}
        else if ((mX >= 385) && (mX <= 400)) {toggleRate(FL1);}
        else if ((mX >= 401) && (mX <= 419)) {togglePwr(FL1);}
        else if ((mX >= 511) && (mX <= 526)) {toggleRate(RR1);}
        else if ((mX >= 527) && (mX <= 546)) {togglePwr(RR1);}
        else if ((mX >= 576) && (mX < 596)) {transRR1RL1();}
        else if ((mX >= 612) && (mX <= 627)) {toggleRate(RL1);}
        else if ((mX >= 628) && (mX <= 646)) {togglePwr(RL1);}
    }
  }
  
  
  //#####################################################################
  //
  // Multi-move components
  //
  //#####################################################################
  if ((mY >= 488) && (mY <= 545)) {
    if ((mX >= 222) && (mX <= 607)) {
      if (mY <= 502) {
        if ((mX >= 235) && (mX <= 265)) {MultiDecY(FR1,FR2); return;}
        if ((mX >= 337) && (mX <= 367)) {MultiDecY(FL1,FL2); return;}
        if ((mX >= 462) && (mX <= 491)) {MultiDecY(RR1,RR2); return;}
        if ((mX >= 564) && (mX <= 594)) {MultiDecY(RL1,RL2); return;}
      }
      else if (mY <= 531) {
        if (mX <= 235) {incXSlider(FR2);}
        if ((mX > 235) && (mX < 265)) {MultiMoveXY(FR1,FR2); return;}
        if ((mX >= 265) && (mX <= 277)) {decXSlider(FR2); return;}

        if ((mX >= 324) && (mX <= 337)) {incXSlider(FL2); return;}
        if ((mX > 337) && (mX < 367)) {MultiMoveXY(FL1,FL2); return;}
        if ((mX >= 367) && (mX <= 380)) {decXSlider(FL2); return;}

        if ((mX >= 449) && (mX <= 462)) {incXSlider(RR2); return;}
        if ((mX > 462) && (mX < 491)) {MultiMoveXY(RR1,RR2); return;}
        if ((mX >= 491) && (mX <= 504)) {decXSlider(RR2); return;}

        if ((mX >= 551) && (mX <= 564)) {incXSlider(RL2); return;}
        if ((mX > 564) && (mX < 594)) {MultiMoveXY(RL1,RL2); return;}
        if ((mX >= 594) && (mX <= 607)) {decXSlider(RL2); return;}
      }
      else {
        if ((mX >= 235) && (mX <= 265)) {MultiIncY(FR1,FR2); return;}
        if ((mX >= 337) && (mX <= 367)) {MultiIncY(FL1,FL2); return;}
        if ((mX >= 462) && (mX <= 491)) {MultiIncY(RR1,RR2); return;}
        if ((mX >= 564) && (mX <= 594)) {MultiIncY(RL1,RL2); return;}
      }
    }
  }
  
  
  if ((mY >= 515) && (mY <= 539)) {
    // DegSteps & Speed buttons and field
         if ((mX >= 12) && (mX < 31)) {decDegSteps(); return;}
    else if ((mX >= 31) && (mX < 81)) {resetDegSteps(); return;}
    else if ((mX >= 81) && (mX <= 99)) {incDegSteps(); return;}
    
         if ((mX >= 722) && (mX < 740)) {decSpeed(); return;}
    else if ((mX >= 740) && (mX < 791)) {resetSpeed(); return;}
    else if ((mX >= 791) && (mX <= 809)) {incSpeed(); return;}
  }
}

// ----------------------------------------------------------------------

public void mousePressed_1() {
  // screen 1 specific mouse pressed functions
  if ((mY >= 515) && (mY <= 539)) {
    // DegSteps & Speed buttons and field
         if ((mX >= 12) && (mX < 31)) {decDegSteps();}
    else if ((mX >= 31) && (mX < 81)) {resetDegSteps();}
    else if ((mX >= 81) && (mX <= 99)) {incDegSteps();}
    
         if ((mX >= 722) && (mX < 740)) {decSpeed();}
    else if ((mX >= 740) && (mX < 791)) {resetSpeed();}
    else if ((mX >= 791) && (mX <= 809)) {incSpeed();}
  }
}

// ----------------------------------------------------------------------

public void mousePressed_2() {
  // screen 2 specific mouse pressed functions
  if ((mY >= 498) && (mY <= 533)) {
    if ((mX >= 376) && (mX <= 446)) {Calc_Walk(); return;}
  }

  if ((mY >= 256) && (mY <= 358)) {
    // lower leg ground region
    if ((mX >= 224) && (mX <= 566)) {setGndY(); drawFlag = 1;}
  }
  
  if ((mY >= 372) && (mY <= 406)) {
    // Titles, copy buttons and Link
         if ((mX >= 24) && (mX <= 78)) {resetLftFlds(); drawFlag = 1;}
    else if ((mX >= 489) && (mX <= 552)) {resetRhtFlds(); drawFlag = 1;}
  }
    
  if ((mY >= 404) && (mY <= 430)) {
    // T2 max left field
         if ((mX >= 103) && (mX < 115)) {decField(FldT2maxLft);}
    else if ((mX >= 115) && (mX < 141)) {adjustField(FldT2maxLft);}
    else if ((mX >= 141) && (mX <= 153)) {incField(FldT2maxLft);}

    // T2 min left field
         if ((mX >= 273) && (mX < 295)) {decField(FldT2minLft);}
    else if ((mX >= 295) && (mX < 310)) {adjustField(FldT2minLft);}
    else if ((mX >= 310) && (mX <= 322)) {incField(FldT2minLft);}

    // T2 max right field
         if ((mX >= 567) && (mX < 579)) {decField(FldT2maxRht);}
    else if ((mX >= 579) && (mX < 605)) {adjustField(FldT2maxRht);}
    else if ((mX >= 605) && (mX <= 617)) {incField(FldT2maxRht);}

    // T2 min right field
         if ((mX >= 737) && (mX < 749)) {decField(FldT2minRht);}
    else if ((mX >= 749) && (mX < 774)) {adjustField(FldT2minRht);}
    else if ((mX >= 774) && (mX <= 786)) {incField(FldT2minRht);}
  }

  if ((mY >= 454) && (mY <= 480)) {
    // Speed buttons and field
         if ((mX >= 368) && (mX < 386)) {decSpeed();}
    else if ((mX >= 386) && (mX < 437)) {resetSpeed();}
    else if ((mX >= 437) && (mX <= 455)) {incSpeed();}
  }

  if ((mY >= 470) && (mY <= 496)) {
    // T1 max left field
         if ((mX >= 103) && (mX < 115)) {decField(FldT1maxLft);}
    else if ((mX >= 115) && (mX < 141)) {adjustField(FldT1maxLft);}
    else if ((mX >= 141) && (mX <= 153)) {incField(FldT1maxLft);}

    // T1 walk lift field
         if ((mX >= 273) && (mX < 295)) {decField(FldT1WlkLft);}
    else if ((mX >= 295) && (mX < 310)) {adjustField(FldT1WlkLft);}
    else if ((mX >= 310) && (mX <= 322)) {incField(FldT1WlkLft);}

    // T1 max right field
         if ((mX >= 567) && (mX < 579)) {decField(FldT1maxRht);}
    else if ((mX >= 579) && (mX < 605)) {adjustField(FldT1maxRht);}
    else if ((mX >= 605) && (mX <= 617)) {incField(FldT1maxRht);}
  }
}

// ----------------------------------------------------------------------

public void mouseReleased() {
  // mouse released from click and drag operation
  mX = mouseX; mY = mouseY;
  if (mDwn > 0) {
    // look for click events
    if (((mY >= 560) && (mY <= 596))) {
      // Memory buttons top row
           if ((mX >= 691) && (mX < 719)) {recallM0();}  // M0 store recall event
      else if ((mX >= 719) && (mX < 749)) {recallM1();}  // M1 store recall event
      else if ((mX >= 749) && (mX < 778)) {recallM2();}  // M2 store recall event
      else if ((mX >= 778) && (mX <= 807)) {recallM3();}  // M3 store recall event
    }
  } else {
    // look for held events which have cleared mDwn
  }
  Inc = 1;         // reset increment value
  IncCnt = 0;      // clear increment counter
  mDwn = 0;        // clear mouse down timer
  togPwr = 0;      // clear power ALL toggle flag
  sliderDwn = 0;   // clear slider flag
  MultiPnt = 0;    // cancel multi-move pointer
  Help_Message();  // update help messages
  cursor(ARROW);
}

// ----------------------------------------------------------------------

public void mouseWheel(MouseEvent event) {
  // check for mouse wheel scrolling rows
  int zYLL = 548; if (Scrn == 1) {zYLL = 41;}
  HelpRowPnt = -1;  // for Scrn1 the current row will drive the highlights
  if ((mX >= 105) && (mX <= 582)) {
    if ((mY >= zYLL) && (mY <= 684)) {
      // we are within the row area
      int zNew = rowPnt + PApplet.parseInt(event.getCount());
      if ((zNew >= 0) && (zNew < rowNum)) {
        // changing row value
        rowPnt = zNew; sendRowData(rowPnt,rowPnt,1);
        setAngValFromRow();  // set servo values and sliders
        drawFlag = 1;
      }
    }
  }
}

// ----------------------------------------------------------------------

public void serialEvent(Serial usbPort) {
  // respond to a serial data received event
  Echo = true;  // if we received something then set this flag
  if (Tx_t0 > 0) {
    Tx_t1 = millis() - Tx_t0; drawTxRxTime(); Tx_t0 = 0;
  drawFlag = 1;
  } RxB = 3; 
  //nextWiFims = millis() + 20;  // delay sending whilst receiving
  Ping = PingMax;  // no need to 'ping' if we are receiving data
  Rx$ = usbPort.readString(); print(Rx$);
  data += Rx$; int RxLF = data.indexOf(LF);
  if (RxLF >= 0) {
    // received data contains a LF/CR so extract the first message
    msgRx = data.substring(0, RxLF); drawFlag = 1;
    // decode messages
    if (msgRx.indexOf("MT ") >= 0) {
      // the is a MoveMemTgtLoader() row report so move row to track it
      // trap serial comms errors
      int zRP = PApplet.parseInt(trim(msgRx.substring(2)));
      if ((zRP >= 0) && (zRP < rowNum)) {
        rowPnt = zRP;
        setAngValFromRow();  // update text fields and sliders
      }
      else if (msgRx.indexOf("SW") >= 0) {
        // this is a SWn. report, user has pushed a button whilst walking
        WalkSim = 0; playMode = 0; usbEn = false;
      }
      else if (msgRx.indexOf("WalkEnd") >= 0) {
        // this is an end of walking message, send during a Play Once run
        WalkSim = 0; playMode = 0;
      }
    }
    // make sure message fits Rx field
    if (msgRx.length() > 24) {msgRx = msgRx.substring(0,20) + "...";}
    data = trim(data.substring(RxLF));
  }
}

// ----------------------------------------------------------------------

public void setDefaults() {
  // load default values
  AB = 255;           // text angle background Blue colour component
  AG = 255;           // text angle background Green colour component
  AngDif = 0;         // different in pervious to current angle
  AngLast = 0;        // previous angle
  AngPwr = false;     // during drawing true if servo is powered, false if not
  AngRate = false;    // used during drawing, true if the angle is an auto-rate angle
  AngVal[ 0] = 0;     // FR1 default
  AngVal[ 1] = 0;     // FR2 default
  AngVal[ 2] = 90;    // FR3 default
  AngVal[ 3] = 0;     // FL1 default
  AngVal[ 4] = 0;     // FL2 default
  AngVal[ 5] = 90;    // FL3 default
  AngVal[ 6] = 0;     // RR1 default
  AngVal[ 7] = 0;     // RR2 default
  AngVal[ 8] = 90;    // RR3 default
  AngVal[ 9] = 0;     // RL1 default
  AngVal[10] = 0;     // RL2 default
  AngVal[11] = 90;    // RL3 default
  AR = 255;           // text angle background Red colour component
  BB = 255;           // text background Blue colour component
  BG = 255;           // text background Green colour component
  Blank = false;      // set == true is a blank line command is detected
  BR = 255;           // text background Red colour component
  CalcMaxHgtLft = 178;   // calculated left-hand max height value
  CalcMaxHgtRht = 178;   // calculated right-hand max height value
  CalcNomStepsLft = 41;  // calculated left-hand steps value
  CalcNomStepsRht = 41;  // calculated right-hand steps value
  Cmd = false;        // temporary flag used in display mode
  CmdCode = 0;        // command code number
  CmdRow = false;     // true if current row is a command line
  CmdVal0 = 0;        // command 1st value
  comFlag = 0;        // = 1 if COM serial available
  comListLength = 0;  // number of items in COM connection list
  comName = "";       // serial port COM name
  comPnt = 0;         // COM port list pointer
  CopyCnt = 0;        // if > 0 then data is stored in CopyMem[]
  CopyData = "";      // data to be copied to the clipboard
  CR = "\r\n";        // carriage return for strings
  data = "";
  DegStep = 1;        // number of degrees per step in autostep calcs
  drawFlag = 1;       // = 1 to force screen redraw
  Echo = true;        // true if Arduino responses are being received
  FL1 = 3;            // array pointer
  FL2 = 4;            // array pointer
  FL3 = 5;            // array pointer
  Flag = 1;                 // default 'Flag' value
  FR1 = 0;            // array pointer
  FR2 = 1;            // array pointer
  FR3 = 2;            // array pointer
  GndY = 358;         // simulator ground reference Y-axis
  Help$ = "";         // message generated by help system
  HelpLast$ = "";     // previous message generated by help system
  HelpOFF = 0;        // if >0 then help system is disabled, default is ON
  Inc = 1;            // any value increment
  IncCnt = 0;         // any increment counter
  index = 0;
  interval = 20;      // main loop interval in milliseconds
  JoyX0 = 40;         // x-coordinate of joystick slider left end
  JoyX1 = 151;        // x-coordinate of joystick slider right end
  keyDwn = 0;         // tracks keyboard up/down activity, 0 = up, 1 = down
  keyESC = false;     // true if ESC key is pressed
  keySHIFT = false;   // true whenever the SHIFT key is pressed
  LF = 10;            // ASCII line feed
  M0 = 0;             // if >0 then data stored in temp memory
  M1 = 0;             // if >0 then data stored in temp memory
  M2 = 0;             // if >0 then data stored in temp memory
  M3 = 0;             // if >0 then data stored in temp memory
  mDwn = 0;           // state of mouse button down timer
  msgCOM = "None";    // COM port name
  msgRx = "";         // message received from Arduino
  msgTx = "";         // message sent to Arduino
  MultiPnt = 0;       // multi-move pointer
  MXDL = 0;           // previous MXD value
  MYDL = 0;           // previous MXD value
  mXOff = 0;          // mouse X offset to control clicking error
  nextWiFims = 0;     // WiFi ms timer
  nextMillis = 0;     // loop end time in millis
  Pause = 1;          // Pause delay value, default = 1
  Ping = 50;          // ping timer to keep auto-OFF from kicking in
  PingMax = Ping;     // record the max delay between pings
  playMode = 0;       // mode ove movement play, 0 = stop; 1 = once, 2 = continuous
  QuickAng = -1;      // quick cut/paste angle value is invalid
  QuickRate = -1;     // quick cut/paste step rate value, C,A,D,S
  QuickStep = -1;     // quick cut/paste step value
  rateFlag = false;   // == true if rates are to be calculated before a redraw
  RL1 = 9;            // array pointer
  RL2 = 10;           // array pointer
  RL3 = 11;           // array pointer
  rowColPnt = -1;     // Help pointer to row column, -1 == off
  rowPnt = 0;         // pointer to current array row
  RR1 = 6;            // array pointer
  RR2 = 7;            // array pointer
  RR3 = 8;            // array pointer
  Rx$ = "";           // single character received from usb port
  RxB = 0;            // set true when a byte is received
  Scrn = 0;           // screen display pointer 0 or 1
  SelCnt = 0;         // number of selected lines
  SelLast = -1;       // most recent selected line, -1 means no recent
  Send = false;       // general flag records change and need to send data
  Sep$ = ",";         // separator used in line listing
  servoOE = 1;        // servo board Output Enable; default = 1  OFF
  servoTgt = 0;       // target/pointer servo for manual setting
  sim1st = 0;         // simulator start at row pointer
  simEnd = 0;         // simulator end pointer
  simPnt = 0;         // simulator row pointer. Note rowPnt is what is displayed
  SL0 = 5;            // array pointer
  SL1 = 6;            // array pointer
  SL1Av = 0;          // average SL1 angle
  SL2 = 7;            // array pointer
  SL2Av = 0;          // average SL2 angle
  SL3 = 8;            // array pointer
  SL3Av = 0;          // average SL3 angle
  SL4 = 9;            // array pointer
  sliderDwn = 0;      // =1 when mouse is clicked on slider
  Slope$ = "";        // temp string used in row list
  Speed = 50;         // speed value to be used by Arduino
  SpeedDef = 50;      // speed value to be used by Arduino
  SpeedLL = 1;        // speed value to be used by Arduino
  SpeedUL = 250;      // speed value to be used by Arduino
  SR0 = 0;            // array pointer
  SR1 = 1;            // array pointer
  SR1Av = 0;          // average SR1 angle
  SR2 = 2;            // array pointer
  SR2Av = 0;          // average SR2 angle
  SR3 = 3;            // array pointer
  SR3Av = 0;          // average SR3 angle
  SR4 = 4;            // array pointer
  Steps = 30;         // Walk number of steps set by the user
  TB = 0;             // text Blue colour component
  TG = 0;             // text Green colour component
  togPwr = 0;         // >0 to set all servos ON/OFF
  TR = 0;             // text Red colour component
  Tx_t0 = 0;          // usb send timers
  Tx_t1 = 0;          // usb send timers
  usbBuff = "";       // buffer string used to prevent WiFi overload
  usbEn = false;      // true to enable USB messages
  usbLen = 0;         // length of usbMsg
  usbMsg = "";        // latest message sent to the usb port
  WaklDelay = 25;     // sets the walk delay timer
  WalkHeight = 145;   // user set height in walk sequence
  WalkLift = 40;      // lift angle during walking
  WalkNext = 2147483640;  // millis() timer for Walk simulator          
  WalkSim = 0;        // a flag which controls the walk simulator, 0 == OFF
  WalkSteps = 30;     // number of steps used in walk sequences
  WiFiInt = 20;       // WiFi interval in ms
  WiFiPW = 64;        // WiFi packet width, must be longer than the max command length

  resetLftFlds(); resetRhtFlds();

}

// ----------------------------------------------------------------------

public void drawAngField() {
  // called to draw angle highlight boxes around angle fields and servos
  // Not used by Scrn 2 simulator screen
  if (AngField < 0) {return;}

  if (Scrn == 0) {
    stroke(255,80,80); noFill(); strokeWeight(4);
    if (AngField ==  0) {rect(286,385,48,24); return;}
    if (AngField ==  1) {rect(173,261,48,24); return;}
    if (AngField ==  2) {rect(148,122,48,24); return;}
    if (AngField ==  3) {rect(385,385,48,24); return;}
    if (AngField ==  4) {rect(275,261,48,24); return;}
    if (AngField ==  5) {rect(250,122,48,24); return;}
    if (AngField ==  6) {rect(512,385,48,24); return;}
    if (AngField ==  7) {rect(400,261,48,24); return;}
    if (AngField ==  8) {rect(376,122,48,24); return;}
    if (AngField ==  9) {rect(612,385,48,24); return;}
    if (AngField == 10) {rect(501,261,48,24); return;}
    if (AngField == 11) {rect(476,122,48,24); return;}
  } else {
    stroke(255,80,80); noFill(); strokeWeight(4);
    if (AngField ==  0) {rect(616,308,22,20); return;}  // FR1
    if (AngField ==  1) {rect(602,244,22,20); return;}  // FR2
    if (AngField ==  2) {rect(597,190,24,20); return;}  // FR3
    
    if (AngField ==  3) {rect(721,308,24,20); return;}  // FL1
    if (AngField ==  4) {rect(706,244,24,20); return;}  // FL2
    if (AngField ==  5) {rect(702,190,24,20); return;}  // FL3
    
    if (AngField ==  6) {rect(657,308,22,20); return;}  // RR1
    if (AngField ==  7) {rect(646,244,24,20); return;}  // RR2
    if (AngField ==  8) {rect(637,190,24,20); return;}  // RR3
    
    if (AngField ==  9) {rect(762,308,24,20); return;}  // RL1
    if (AngField == 10) {rect(748,244,24,20); return;}  // RL2
    if (AngField == 11) {rect(743,190,24,20); return;}  // RL3
  }
}

// ----------------------------------------------------------------------

public void drawAngFill(int zX,int zY) {
  // draws a filled rectangle based on pre-determined fill colours
  stroke(AR,AG,AB); fill(AR,AG,AB); rect(zX+2,zY-1,26,18);
}

// ----------------------------------------------------------------------

public void drawAngles() {
  // update servo angle value fields
  // servo array data is defined as:
  // SR0,SR1,SR2,SR3,SR4,SL0,SL1,SL2,SL3,SL4,Stp
  BR = 220; BG = 230; BB = 242;
  int zAng;
  int zAPnt = rowPnt * rowWidth;
  int zY1 = 391; int zY2 = 265; int zY3 = 128;
  if (ArrayVal[zAPnt] < 0) {return;}  // command so don't display angles

  // draw FR1
  TB = 255; TR = 0; zAng = getAngVal(ArrayVal[zAPnt]); if (AngRate) {TB = 0; TR = 255;}
  //drawTxtCtrField("       ", 311, zY1); 
  drawTxtCtrField(str(zAng) + "°", 311, zY1);
  if (!AngPwr) {stroke(0,0,0); fill(200,0,0); strokeWeight(1); circle(311,421,18);}

  // draw FR2
  zAPnt++;
  TB = 255; TR = 0; zAng = getAngVal(ArrayVal[zAPnt]); if (AngRate) {TB = 0; TR = 255;}
  //drawTxtCtrField("       ", 199, zY2);
  drawTxtCtrField(str(zAng) + "°", 199, zY2);
  if (!AngPwr) {stroke(0,0,0); fill(200,0,0); strokeWeight(1); circle(198,296,18);}

  // draw FR3
  zAPnt++;
  TB = 255; TR = 0; zAng = getAngVal(ArrayVal[zAPnt]); if (AngRate) {TB = 0; TR = 255;}
  //drawTxtCtrField("       ", 174, zY3);
  drawTxtCtrField(str(zAng) + "°", 174, zY3);
  if (!AngPwr) {stroke(0,0,0); fill(200,0,0); strokeWeight(1); circle(173,158,18);}

  // draw FL1
  zAPnt++;
  TB = 255; TR = 0; zAng = getAngVal(ArrayVal[zAPnt]); if (AngRate) {TB = 0; TR = 255;}
  //drawTxtCtrField("       ", 411, zY1);
  drawTxtCtrField(str(zAng) + "°", 411, zY1);
  if (!AngPwr) {stroke(0,0,0); fill(200,0,0); strokeWeight(1); circle(410,421,18);}

  // draw FL2
  zAPnt++;
  TB = 255; TR = 0; zAng = getAngVal(ArrayVal[zAPnt]); if (AngRate) {TB = 0; TR = 255;}
  //drawTxtCtrField("       ", 301, zY2);
  drawTxtCtrField(str(zAng) + "°", 301, zY2);
  if (!AngPwr) {stroke(0,0,0); fill(200,0,0); strokeWeight(1); circle(299,296,18);}

  // draw FL3
  zAPnt++;
  TB = 255; TR = 0; zAng = getAngVal(ArrayVal[zAPnt]); if (AngRate) {TB = 0; TR = 255;}
  //drawTxtCtrField("       ", 274, zY3);
  drawTxtCtrField(str(zAng) + "°", 274, zY3);
  if (!AngPwr) {stroke(0,0,0); fill(200,0,0); strokeWeight(1); circle(275,158,18);}

  // draw RR1
  zAPnt++;
  TB = 255; TR = 0; zAng = getAngVal(ArrayVal[zAPnt]); if (AngRate) {TB = 0; TR = 255;}
  //drawTxtCtrField("       ", 535, zY1);
  drawTxtCtrField(str(zAng) + "°", 535, zY1);
  if (!AngPwr) {stroke(0,0,0); fill(200,0,0); strokeWeight(1); circle(537,421,18);}

  // draw RR2
  zAPnt++;
  TB = 255; TR = 0; zAng = getAngVal(ArrayVal[zAPnt]); if (AngRate) {TB = 0; TR = 255;}
  //drawTxtCtrField("       ", 425, zY2);
  drawTxtCtrField(str(zAng) + "°", 425, zY2);
  if (!AngPwr) {stroke(0,0,0); fill(200,0,0); strokeWeight(1); circle(425,296,18);}

  // draw RR3
  zAPnt++;
  TB = 255; TR = 0; zAng = getAngVal(ArrayVal[zAPnt]); if (AngRate) {TB = 0; TR = 255;}
  //drawTxtCtrField("       ", 401, zY3);
  drawTxtCtrField(str(zAng) + "°", 401, zY3);
  if (!AngPwr) {stroke(0,0,0); fill(200,0,0); strokeWeight(1); circle(401,158,18);}

  // draw RL1
  zAPnt++;
  TB = 255; TR = 0; zAng = getAngVal(ArrayVal[zAPnt]); if (AngRate) {TB = 0; TR = 255;}
  //drawTxtCtrField("       ", 635, zY1);
  drawTxtCtrField(str(zAng) + "°", 635, zY1);
  if (!AngPwr) {stroke(0,0,0); fill(200,0,0); strokeWeight(1); circle(637,421,18);}

  // draw RL2
  zAPnt++;
  TB = 255; TR = 0; zAng = getAngVal(ArrayVal[zAPnt]); if (AngRate) {TB = 0; TR = 255;}
  //drawTxtCtrField("       ", 526, zY2);
  drawTxtCtrField(str(zAng) + "°", 526, zY2);
  if (!AngPwr) {stroke(0,0,0); fill(200,0,0); strokeWeight(1); circle(526,296,18);}

  // draw RL3
  zAPnt++;
  TB = 255; TR = 0; zAng = getAngVal(ArrayVal[zAPnt]); if (AngRate) {TB = 0; TR = 255;}
  //drawTxtCtrField("       ", 500, zY3);
  drawTxtCtrField(str(zAng) + "°", 500, zY3);
  if (!AngPwr) {stroke(0,0,0); fill(200,0,0); strokeWeight(1); circle(502,158,18);}
}

// ----------------------------------------------------------------------

public void drawAngPos() {
  // overlays coloured semi-transparent boxes on the RoboDog image for the
  // current row values in Scrn1 only
  int zAPnt;
  if (HelpRowPnt < 0) {zAPnt = rowPnt * rowWidth;}
  else {zAPnt = HelpRowPnt * rowWidth;}
  if (ArrayVal[zAPnt] < 0) {return;}  // command line
  
  strokeWeight(1);  // applies to all boxes
  int zX;  // co-ordinate for rectangle
  int zY;  // co-ordinate for rectangle

  // FR1,RR1,FL1,RL1 at 70% transparency == 77 (0 = transparent, 255 = opaque)
  fill(255,67,67,120); stroke(255,67,67,120);
  zY = mapInt(getAngVal(ArrayVal[zAPnt]),0,160,337,424);    // FR1
  rect(614,zY,32,10);
  zY = mapInt(getAngVal(ArrayVal[zAPnt+3]),0,160,337,424);  // FL1
  rect(704,zY,32,10);
  zY = mapInt(getAngVal(ArrayVal[zAPnt+6]),0,160,337,424);  // RR1
  rect(654,zY,32,10);
  zY = mapInt(getAngVal(ArrayVal[zAPnt+9]),0,160,337,424);  // RL1
  rect(746,zY,32,10);

  // FR2,RR2,FL2,RL2 at 70% transparency == 77 (0 = transparent, 255 = opaque)
  fill(29,171,76,120); stroke(29,171,76,120);
  zY = mapInt(getAngVal(ArrayVal[zAPnt+1]),0,160,237,336);  // FR2
  rect(614,zY,32,10);
  zY = mapInt(getAngVal(ArrayVal[zAPnt+4]),0,160,237,336);  // FL2
  rect(704,zY,32,10);
  zY = mapInt(getAngVal(ArrayVal[zAPnt+7]),0,160,237,336);  // RR2
  rect(654,zY,32,10);
  zY = mapInt(getAngVal(ArrayVal[zAPnt+10]),0,160,237,336); // RL2
  rect(746,zY,32,10);

  // FR3,RR3,FL3,RL3 at 70% transparency == 77 (0 = transparent, 255 = opaque)
  fill(0,112,192,120); stroke(0,112,192,120);
  zX = mapInt(getAngVal(ArrayVal[zAPnt+2]),70,130,614,583);  // FR3
  rect(zX,184,10,32);
  zX = mapInt(getAngVal(ArrayVal[zAPnt+5]),50,110,729,698);  // FL3
  rect(zX,184,10,32);
  zX = mapInt(getAngVal(ArrayVal[zAPnt+8]),70,130,654,623);  // RR3
  rect(zX,184,10,32);
  zX = mapInt(getAngVal(ArrayVal[zAPnt+11]),50,110,769,738); // RL3
  rect(zX,184,10,32);
}

// ----------------------------------------------------------------------

public void drawArrayRows() {
  // draw array data as a list of rows
  int zCX = 323;   // row left X co-ordinate
  int zDX = 150;   // row centre X co-ordinate
  int zPnt;        // row pointer
  //int zNL = 5;     // number of lines
  //int zLY = 607 - ((zNL/2) * 24);   // first row Y co-ordinate
  int zNL = 6;     // number of lines
  int zLY = 622 - ((zNL * 24)/2);   // first row Y co-ordinate
  String zR = "";  // temp row string
  int zRnX = 120;  // row number field X co-ordinate
  int zSlX = 518;  // steps field X co-ordinate 
  int zSP = 12;    // offset to array steps value
  if (Scrn == 1) {
    // list screen
    zNL = 27;      // number of lines
    zLY = 356 - ((zNL/2) * 24);  // first row Y co-ordinate
  }
  int zH0;
  if (Scrn != 1) {zH0 = (zNL/2) - rowPnt - 2;}
  else {zH0 = (zNL/2) - rowPnt - 1;}
  int zH1 = zH0 + rowNum + 1;

  getMinMaxAng();
  TR = 0; TG = 0; TB = 0;
  for (int zLP = 0; zLP < zNL; zLP++) {
    if ((zLP > zH0) && (zLP < zH1)) {
      int zRow = zLP - zH0 - 1;
      if (Select[zRow]) {BB = 60; BG = 255; BR = 255; Sel = true;} else {BB = 220; BG = 220; BR = 220; Sel = false;}
      zPnt =  zRow * rowWidth; TB = 0; TR = 0;
      drawTxtCtrField(str(zRow), zRnX, zLY);  // row number field
      drawRowValues(zPnt, zDX, zLY); TR = 0; TG = 0; TB = 0;
      drawTxtCtrField(getStrArrayVal(zPnt + zSP), 554, zLY);
      drawTxtField(Slope$,zSlX,zLY);  // Slope$ fields
    }
    else if ((zLP == zH0) || (zLP == zH1)) {
      // draw a header line
      BB = 220; BG = 220; BR = 220; TB = 0;
      drawTxtCtrField("--", zRnX, zLY);  // row number field
      drawTxtCtrField("FR1 FR2 FR3 FL1  FL2  FL3 RR1  RR2  RR3 RL1 RL2 RL3", zCX, zLY);
      drawTxtField("R", zSlX, zLY);  // Slope$ fields
      drawTxtCtrField("Stp", 554, zLY);
    }
    zLY += 24; TR = 0; TG = 0; TB = 0;
  }
  // reset text field backgrounds
  BB = 255; BG = 255; BR = 255;
}

// ----------------------------------------------------------------------

public void drawColHighlight() {
  //draws a rectangle around the associated active row column
  int zX = 0; int zY = 592; if (Scrn == 1) {zY = 351;}
  switch (rowColPnt) {
    case  0: zX = 151; break;  // FR1
    case  1: zX = 180; break;  // FR2
    case  2: zX = 208; break;  // FR3
    case  3: zX = 237; break;  // FL1
    case  4: zX = 267; break;  // FL2
    case  5: zX = 294; break;  // FL3
    case  6: zX = 323; break;  // RR1
    case  7: zX = 353; break;  // RR2
    case  8: zX = 381; break;  // RR3
    case  9: zX = 409; break;  // RL1
    case 10: zX = 440; break;  // RL2
    case 11: zX = 467; break;  // RL3
    default: return;
  }
  stroke(255,0,255); noFill(); strokeWeight(4); rect(zX,zY,29,27);
}

// ----------------------------------------------------------------------

public void drawCOMMsg() {
  // put text in COM value field
  if (comFlag < 1) {TR = 255;}
  drawTxtField(msgCOM, 120, 697); TR = 0;
}

// ----------------------------------------------------------------------

public void drawDegSteps() {
  // put Degrees/Step value in text field
  TB = 255; BR = 255; BG = 255; BB = 255;
  //drawTxtCtrField("    ", 55, 520);
  drawTxtCtrField(str(DegStep), 55, 520);
}

// ----------------------------------------------------------------------

public void drawEchoLamp() {
  // draws a lamp within the Echo button
  strokeWeight(1);
  if (Echo) {
    // echo is ON
    stroke(80,120,255); fill(80,120,255);
  } else {
    // echo is OFF
    stroke(255,255,255); fill(255,255,255);
  } rect(791,695,8,17);
}

// ----------------------------------------------------------------------

public void drawHelp() {
  // display help messages
  TB = 255; drawTxtCtrField(Help$, 408, 729);
}

// ----------------------------------------------------------------------

public void drawHidden() {
  // effectively hides features which are not always accessible
  stroke(191,191,191); fill(191,191,191); strokeWeight(1);
  //stroke(191,0,0); fill(191,0,0); strokeWeight(1);  // for test
  
  // X,Y co-ordinates are defined as:
  // Col 0 =   8, Col 1 =  40, Col 2 =  71
  // Row 0 = 542, Row 1 = 579, Row 2 = 616, Row 3 = 653
  
  // always hide these unused buttons
  rect( 8,542,30,36);  // col 0, row 0, Nop
  rect(40,542,30,36);  // col 1, row 0, Nop
  rect( 8,653,30,36);  // col 0, row 3, Nop
  
  if (SelCnt < 1) {
    // no rows are selected
    // hide shift, cut, paste, step and transform keys
    rect(71,542,31,36);  // col 2, row 0, Up arrow
    rect( 8,579,30,36);  // col 0, row 1, Cut
    rect(40,579,30,36);  // col 1, row 1, Copy
    rect( 8,616,30,36);  // col 0, row 2, Goto
    rect(40,616,30,36);  // col 1, row 2, Steps
    rect(71,616,31,36);  // col 2, row 2, Transpose
    rect(40,653,30,36);  // col 1, row 3, Reverse order
    rect(71,653,31,36);  // col 2, row 3, Down arrow
  }
  if (CopyCnt < 1) {
    // nothing has been copied
    // this hides the Paste button if nothing has been copied
    if (SelCnt < 1) {
      // no rows are selected
      rect(71,579,31,36);  // col 2, row 1, Paste
    }
  }
  if (!usbEn && (Scrn != 2)) {
    // hide play, repeat, pause, stop buttons when in Scrn0 and the
    // USB port messaging is not enabled
    rect(690,597,117,36);
  }
}

// ----------------------------------------------------------------------

public void drawLinks() {
  // if link flags are set draw boxes around buttons
  if (ArrayVal[rowPnt * rowWidth] < 0) {return;}  // command so don't draw angles

  stroke(255,80,80); noFill();
  strokeWeight(3);
  if (Link[FR3] > 0) {rect(213,121,22,29);}
  if (Link[FR3] > 1) {rect(326,124,29,22);}  // multi-link
  if (Link[RR3] > 0) {rect(439,121,22,29);}
  if (Link[FR2] > 0) {rect(238,259,22,29);}
  if (Link[FR2] > 1) {rect(352,262,29,22);}  // multi-link
  if (Link[RR2] > 0) {rect(465,259,22,30);}
  if (Link[FR1] > 0) {rect(350,383,22,30);}
  if (Link[FR1] > 1) {rect(462,386,29,22);}  // multi-link
  if (Link[RR1] > 0) {rect(576,383,22,30);}
}

// ----------------------------------------------------------------------

public void drawMemBars() {
  // colour in memory bars if data is stored
  stroke(80,200,220); fill(80,200,220); strokeWeight(1);
  if (M0 > 0) {rect(697,565,16,4);}
  if (M1 > 0) {rect(726,565,16,4);}
  if (M2 > 0) {rect(755,565,16,4);}
  if (M3 > 0) {rect(784,565,16,4);}
}

// ----------------------------------------------------------------------

public void drawMouseXY() {
  // draws the value of mouse x,y top left
  // draws a full X/Y axis on the form
  if ((mX != mXL) || (mY != mYL)) {
    // draw full X/Y axis
    stroke(255,0,0,128); strokeWeight(1);
    line(0,mY,823,mY); line(mX,0,mX,759);

    int zX,zY;
    if (mX > 412) {zX = mX - 58;} else {zX = mX + 8;}
    if (mY > 380) {zY = mY - 38;} else {zY = mY + 8;}
    stroke(0,0,0); fill(255,255,255); rect(zX,zY,50,30);
    fill(0,0,0); text("X=" + mX, zX + 4, zY + 13); text("Y=" + mY, zX + 4, zY + 27);
    mXL = mX; mYL = mY;
  }
}

// ----------------------------------------------------------------------

public void drawRates() {
  // fill in the 'Rate' buttons, faint or in RED
  int zAPnt = rowPnt * rowWidth;
  if (ArrayVal[zAPnt] < 0) {return;}  // command so don't display angle rates

  // FR1
  fill (120,120,120);if (ArrayVal[zAPnt] >= 400) {fill (255,60,60);}
  text("R",289,424); zAPnt++;

  // FR2
  fill (120,120,120);if (ArrayVal[zAPnt] >= 400) {fill (255,60,60);}
  text("R",177,299); zAPnt++;

  // FR3
  fill (120,120,120);if (ArrayVal[zAPnt] >= 400) {fill (255,60,60);}
  text("R",186,162); zAPnt++;

  // FL1
  fill (120,120,120);if (ArrayVal[zAPnt] >= 400) {fill (255,60,60);}
  text("R",388,424); zAPnt++;

  // FL2
  fill (120,120,120);if (ArrayVal[zAPnt] >= 400) {fill (255,60,60);}
  text("R",278,299); zAPnt++;

  // FL3
  fill (120,120,120);if (ArrayVal[zAPnt] >= 400) {fill (255,60,60);}
  text("R",288,162); zAPnt++;

  // RR1
  fill (120,120,120);if (ArrayVal[zAPnt] >= 400) {fill (255,60,60);}
  text("R",515,424); zAPnt++;

  // RR2
  fill (120,120,120);if (ArrayVal[zAPnt] >= 400) {fill (255,60,60);}
  text("R",403,299); zAPnt++;

  // RR3
  fill (120,120,120);if (ArrayVal[zAPnt] >= 400) {fill (255,60,60);}
  text("R",413,162); zAPnt++;

  // RL1
  fill (120,120,120);if (ArrayVal[zAPnt] >= 400) {fill (255,60,60);}
  text("R",615,424); zAPnt++;

  // RL2
  fill (120,120,120);if (ArrayVal[zAPnt] >= 400) {fill (255,60,60);}
  text("R",504,299); zAPnt++;

  // RL3
  fill (120,120,120);if (ArrayVal[zAPnt] >= 400) {fill (255,60,60);}
  text("R",514,162); zAPnt++;
}

// ----------------------------------------------------------------------

public void drawRowValues(int zPnt,int zDX,int zLY) {
  // draw a line of row angles in the list, with formatting
  int zAng;
  int zCP = 0;  // array column pointer
  String zV$ = "";
  Cmd = false; Blank = false; Slope$ = "C";
  
  // FR1 or Cmd
  if (Cmd) {zAng = ArrayVal[zPnt];} else {zAng = getAngVal(ArrayVal[zPnt]);}
  zPnt++;
  getAngFillCol(zAng,zCP); zCP++;  // determine the background fill colour
  if (zAng < 0) {
    Cmd = true;
    zAng = ArrayVal[zPnt];    // already pointing at next value
    Blank = true; Slope$ = "   ";
         if (zAng == 0) {return;}  // blank line command
    else if (zAng < 6) {zV$ = "cmd";}
    else {zV$ = "exe";}
  } else {
    zV$ = str(zAng);
    if (zAng < 100) {zV$ = "  " + zV$;}
    if (zAng < 10) {zV$ = "  " + zV$;}
  }
  if (AngRate || Cmd) {TR = 190;} else {TR = 0;}
  if (Sel) {drawTxtField(zV$,zDX,zLY);}
  else {
    if (!Cmd) {drawAngFill(zDX,zLY);}
    drawTxtNF(zV$,zDX,zLY);
  }
  if (Cmd) {Sel = false;}

  zDX += 28;  // FR2 or CmdCode
  if (Cmd) {zAng = ArrayVal[zPnt];} else {zAng = getAngVal(ArrayVal[zPnt]);}
  zPnt++;
  getAngFillCol(zAng,zCP); zCP++;  // determine the background fill colour
  zV$ = str(zAng); CmdCode = zAng;
  if (zAng < 100) {zV$ = "  " + zV$;}
  if (zAng < 10) {zV$ = "  " + zV$;}
  if (AngRate || Cmd) {TR = 190;} else {TR = 0;}
  if (Sel) {drawTxtField(zV$,zDX,zLY);}
  else {
    if (!Cmd) {drawAngFill(zDX,zLY);}
    drawTxtNF(zV$,zDX,zLY);
  }

  zDX += 29;  // FR3 or CmdVal0
  if (Cmd) {zAng = ArrayVal[zPnt];} else {zAng = getAngVal(ArrayVal[zPnt]);}
  zPnt++;
  getAngFillCol(zAng,zCP); zCP++;  // determine the background fill colour
  zV$ = str(zAng); CmdVal0 = zAng;
  if (zAng < 100) {zV$ = "  " + zV$;}
  if (zAng < 10) {zV$ = "  " + zV$;}
  if (AngRate || Cmd) {TR = 190;} else {TR = 0;}
  if (Sel) {drawTxtField(zV$,zDX,zLY);}
  else {
    if (!Cmd) {drawAngFill(zDX,zLY);}
    drawTxtNF(zV$,zDX,zLY);
  }

  zDX += 29;  // FL1
  zAng = getAngVal(ArrayVal[zPnt]); zPnt++;
  getAngFillCol(zAng,zCP); zCP++;  // determine the background fill colour
  if (Cmd && (zAng == 0)) {
    switch(CmdCode) {
      case 1: zV$ = " Goto row " + str(CmdVal0); break;
      case 2: zV$ = " Set Speed to " + str(CmdVal0); break;
      case 3: zV$ = "  Pause for " + str(CmdVal0) + " step"; if (CmdVal0 > 1) {zV$ += "s";} break;
      case 4: zV$ = " Set Flag to " + str(CmdVal0); break;
      case 5: zV$ = " Set Balance to ";
             if (CmdVal0 == 0) {zV$+= "OFF";}
        else if (CmdVal0 == 1) {zV$+= "topple only";}
        else if (CmdVal0 == 2) {zV$+= "balance only";}
        else if (CmdVal0 == 3) {zV$+= "topple and balance";}
        break;
      case 6: zV$ = " For...1 to  ..." + str(CmdVal0) + " loop"; break;
      case 7: zV$ = " Next after a For... loop"; break;
    } drawTxtNF(zV$,zDX,zLY);
    return;
  } else {
    zV$ = str(zAng);
    if (zAng < 100) {zV$ = "  " + zV$;}
    if (zAng < 10) {zV$ = "  " + zV$;}
  } 
  if (AngRate || Cmd) {TR = 190;} else {TR = 0;}
  if (Sel) {drawTxtField(zV$,zDX,zLY);}
  else {
    if (!Cmd) {drawAngFill(zDX,zLY);}
    drawTxtNF(zV$,zDX,zLY);
  }

  zDX += 29;  // FL2
  zAng = getAngVal(ArrayVal[zPnt]); zPnt++;
  getAngFillCol(zAng,zCP); zCP++;  // determine the background fill colour
  zV$ = str(zAng);
  if (zAng < 100) {zV$ = "  " + zV$;}
  if (zAng < 10) {zV$ = "  " + zV$;}
  if (AngRate || Cmd) {TR = 190;} else {TR = 0;}
  if (Sel) {drawTxtField(zV$,zDX,zLY);}
  else {
    if (!Cmd) {drawAngFill(zDX,zLY);}
    drawTxtNF(zV$,zDX,zLY);
  }

  zDX += 28;  // FL3
  zAng = getAngVal(ArrayVal[zPnt]); zPnt++;
  getAngFillCol(zAng,zCP); zCP++;  // determine the background fill colour
  zV$ = str(zAng);
  if (zAng < 100) {zV$ = "  " + zV$;}
  if (zAng < 10) {zV$ = "  " + zV$;}
  if (AngRate || Cmd) {TR = 190;} else {TR = 0;}
  if (Sel) {drawTxtField(zV$,zDX,zLY);}
  else {
    if (!Cmd) {drawAngFill(zDX,zLY);}
    drawTxtNF(zV$,zDX,zLY);
  }

  zDX += 29;  // RR1
  zAng = getAngVal(ArrayVal[zPnt]); zPnt++;
  getAngFillCol(zAng,zCP); zCP++;  // determine the background fill colour
  zV$ = str(zAng);
  if (zAng < 100) {zV$ = "  " + zV$;}
  if (zAng < 10) {zV$ = "  " + zV$;}
  if (AngRate || Cmd) {TR = 190;} else {TR = 0;}
  if (Sel) {drawTxtField(zV$,zDX,zLY);}
  else {
    if (!Cmd) {drawAngFill(zDX,zLY);}
    drawTxtNF(zV$,zDX,zLY);
  }

  zDX += 29;  // RR2
  zAng = getAngVal(ArrayVal[zPnt]); zPnt++;
  getAngFillCol(zAng,zCP); zCP++;  // determine the background fill colour
  zV$ = str(zAng);
  if (zAng < 100) {zV$ = "  " + zV$;}
  if (zAng < 10) {zV$ = "  " + zV$;}
  if (AngRate || Cmd) {TR = 190;} else {TR = 0;}
  if (Sel) {drawTxtField(zV$,zDX,zLY);}
  else {
    if (!Cmd) {drawAngFill(zDX,zLY);}
    drawTxtNF(zV$,zDX,zLY);
  }

  zDX += 29;  // RR3
  zAng = getAngVal(ArrayVal[zPnt]); zPnt++;
  getAngFillCol(zAng,zCP); zCP++;  // determine the background fill colour
  zV$ = str(zAng);
  if (zAng < 100) {zV$ = "  " + zV$;}
  if (zAng < 10) {zV$ = "  " + zV$;}
  if (AngRate || Cmd) {TR = 190;} else {TR = 0;}
  if (Sel) {drawTxtField(zV$,zDX,zLY);}
  else {
    if (!Cmd) {drawAngFill(zDX,zLY);}
    drawTxtNF(zV$,zDX,zLY);
  }

  zDX += 29;  // RL1
  zAng = getAngVal(ArrayVal[zPnt]); zPnt++;
  getAngFillCol(zAng,zCP); zCP++;  // determine the background fill colour
  zV$ = str(zAng);
  if (zAng < 100) {zV$ = "  " + zV$;}
  if (zAng < 10) {zV$ = "  " + zV$;}
  if (AngRate || Cmd) {TR = 190;} else {TR = 0;}
  if (Sel) {drawTxtField(zV$,zDX,zLY);}
  else {
    if (!Cmd) {drawAngFill(zDX,zLY);}
    drawTxtNF(zV$,zDX,zLY);
  }

  zDX += 29;  // RL2
  zAng = getAngVal(ArrayVal[zPnt]); zPnt++;
  getAngFillCol(zAng,zCP); zCP++;  // determine the background fill colour
  zV$ = str(zAng);
  if (zAng < 100) {zV$ = "  " + zV$;}
  if (zAng < 10) {zV$ = "  " + zV$;}
  if (AngRate || Cmd) {TR = 190;} else {TR = 0;}
  if (Sel) {drawTxtField(zV$,zDX,zLY);}
  else {
    if (!Cmd) {drawAngFill(zDX,zLY);}
    drawTxtNF(zV$,zDX,zLY);
  }

  zDX += 28;  // RL3
  zAng = getAngVal(ArrayVal[zPnt]); zPnt++;
  getAngFillCol(zAng,zCP); zCP++;  // determine the background fill colour
  zV$ = str(zAng);
  if (zAng < 100) {zV$ = "  " + zV$;}
  if (zAng < 10) {zV$ = "  " + zV$;}
  if (AngRate || Cmd) {TR = 190;} else {TR = 0;}
  if (Sel) {drawTxtField(zV$,zDX,zLY);}
  else {
    if (!Cmd) {drawAngFill(zDX,zLY);}
    drawTxtNF(zV$,zDX,zLY);
  }
}

// ----------------------------------------------------------------------

public void drawRxMsg() {
  // put text in the Rx field
  TB = 255;
  //drawTxtField("                                                    ", 223, 697);
  drawTxtField(msgRx, 223, 697);
}

// ----------------------------------------------------------------------

public void drawSpeed() {
  // put Speed value in text field
  TB = 255;
  //drawTxtCtrField("    ", 766, 520);
  drawTxtCtrField(str(Speed), 766, 520);
}

// ----------------------------------------------------------------------

public void drawTxtCtrField(String zM,int zX, int zY) {
  // called by other draw text functions. This centres the text around zX
  zX = zX - (int)(textWidth(zM)/2.0f);
  stroke(BR,BG,BB); fill(BR,BG,BB); 
  rect(zX, zY, textWidth(zM), 15);
  fill(TR,TG,TB); text(zM, zX, zY+12);
}

// ----------------------------------------------------------------------

public void drawTxtField(String zM,int zX, int zY) {
  // called by other draw text functions
  stroke(242,242,242); fill(BR,BG,BB); 
  rect(zX, zY, textWidth(zM) + 2, 14);
  fill(TR,TG,TB); text(zM, zX+4, zY+12);
}

// ----------------------------------------------------------------------

public void drawTxMsg() {
  // put text in Tx field
  if (msgTx.length() > 27) {msgTx = msgTx.substring(0,23) + "...";}
  if (usbEn) {TR = 255; TG = 0; TB = 0;}
  else {TR = 30; TG = 30; TB = 30;}
  //drawTxtField("                                              ", 486, 697);
  drawTxtField(msgTx, 486, 697);

  noFill(); 
  if (usbEn) {
    if (usbOnCnt >= 10) {
      stroke(255,80,80); strokeWeight(4); rect(477,689,206,30);
    }
  }
  else {
    stroke(80,255,80); strokeWeight(3); rect(477,689,206,30);  // green box
  } TR = 0; TG = 0; TB = 0;
}

// ----------------------------------------------------------------------

public void drawTxtNF(String zM,int zX, int zY) {
  // draws text without a filled background
  fill(TR,TG,TB); text(zM, zX+4, zY+12);
}

// ----------------------------------------------------------------------

public void drawTxRxTime() {
  // display the Tx_t1 - Tx_t0 time
  if (Tx_t1 > 999) {Tx_t1 = 999;}  // limit output to 1 second
  TB = 0; drawTxtField(str(Tx_t1), 718, 696);
}

// ----------------------------------------------------------------------

public void drawUsbLamp() {
  // draws a lamp next to the Tx field to indicate usb transmissions
  strokeWeight(1);
  if (usbLen > 0) {
    stroke(255,80,80); fill(255,80,80);
  } else if (RxB > 0) {
    stroke(80,255,80); fill(80,255,80);
  } else {
    stroke(255,255,255); fill(255,255,255);
  } rect(698,695,10,18);
}

// ----------------------------------------------------------------------

public void drawWalkAngles() {
  // update left and right Walk Trainer angle value fields
  // draw first blue text fields on light blue
  BR = 220; BG = 230; BB = 242;
  TR = 0; TG = 0; TB = 255;
  int zY0 = 410;
  // T2maxAngLft
  //drawTxtCtrField("       ", 128, zY0);
  drawTxtCtrField(str(T2maxAngLft) + "°", 128, zY0);
  // T2minAngLft
  //drawTxtCtrField("       ", 298, zY0);
  drawTxtCtrField(str(T2minAngLft) + "°", 298, zY0);
  // T2maxAngRht
  //drawTxtCtrField("       ", 591, zY0);
  drawTxtCtrField(str(T2maxAngRht) + "°", 591, zY0);
  // T2minAngRht
  //drawTxtCtrField("       ", 762, zY0);
  drawTxtCtrField(str(T2minAngRht) + "°", 762, zY0);

  // T1maxAngLft
  zY0 = 477;
  //drawTxtCtrField("       ", 128, zY0);
  drawTxtCtrField(str(T1maxAngLft) + "°", 128, zY0);
  // T1 Lift
  //drawTxtCtrField("       ", 298, zY0);
  drawTxtCtrField(str(WalkLift), 298, zY0);
  // T1maxAngRht
  //drawTxtCtrField("       ", 591, zY0);
  drawTxtCtrField(str(T1maxAngRht) + "°", 591, zY0);


  // draw calc. black text fields light grey
  BR = 191; BG = 191; BB = 191;
  TR = 0; TG = 0; TB = 0;
  // CalcNomStepsLft
  zY0 = 443;
  //drawTxtCtrField("       ", 128, zY0);
  drawTxtCtrField(str(CalcNomStepsLft), 128, zY0);
  // CalcMidT2AngLft
  //drawTxtCtrField("       ", 298, zY0);
  drawTxtCtrField(str(T2midAngLft), 298, zY0);
  // CalcNomStepsRht
  //drawTxtCtrField("       ", 591, zY0);
  drawTxtCtrField(str(CalcNomStepsRht), 591, zY0);
  // CalcMidT2AngRht
  //drawTxtCtrField("       ", 762, zY0);
  drawTxtCtrField(str(T2midAngRht), 762, zY0);

  // Walk lift
  zY0 = 477;
  //drawTxtCtrField("       ", 762, zY0);
  drawTxtCtrField(str(WalkLift), 762, zY0);

  // CalcMaxHgtLft
  zY0 = 509;
  //drawTxtCtrField("       ", 128, zY0);
  drawTxtCtrField(str(CalcMaxHgtLft), 128, zY0);
  // CalcMaxHgtRht
  //drawTxtCtrField("       ", 591, zY0);
  drawTxtCtrField(str(CalcMaxHgtRht), 591, zY0);

  // draw black text fields white background
  // Speed adjust
  BR = 255; BG = 255; BB = 255;
  zY0 = 460;
  //drawTxtCtrField("       ", 412, zY0);
  drawTxtCtrField(str(Speed), 412, zY0);
}

// ----------------------------------------------------------------------

public void drawWalkLegs() {
  // called in Scrn 2 mode to draw legs onto the robots image
  int zAPnt;
  if (HelpRowPnt < 0) {zAPnt = rowPnt * rowWidth;}  // get array pointer to FR1
  else {zAPnt = HelpRowPnt * rowWidth; HelpRowPnt = -1;}
  float zAng = 0.0f;  // any angle in degrees
  float zRad = 0.0f;  // any angle in radians
  float zSL = 30.0f;  // length of shoulder spigots
  float zUL = 67.0f;  // length of upper leg in pixels
  float zLL = 93.0f;  // length of lower leg in pixels
  // check for command line
  if (ArrayVal[zAPnt] < 0) {return;}
  
  // draw ground line
  stroke(190,190,190); line(230,GndY,560,GndY);
  GndYM = 0;  // reset the max dynamic Y reference
  
  // draw FR3
  DX0 = 285; DY0 = 112;
  if (AngField != 2) {stroke(50,50,140); fill(50,50,140);} else {stroke(140,160,255); fill(140,160,255);}
  strokeWeight(1); circle(DX0,DY0,26);
  zAng = (float)getAngVal(ArrayVal[zAPnt + 2]); zRad = zAng /57.2958f;
  DX1 = DX0 + PApplet.parseInt(zSL * cos(zRad)); DY1 = DY0 + PApplet.parseInt(zSL * sin(zRad));
  strokeWeight(10); line(DX0,DY0,DX1,DY1);
  // draw FR2
  DX0 = 285; DY0 = 184;
  if (AngField != 1) {stroke(0,90,0); fill(0,90,0);} else {stroke(0,255,0); fill(0,255,0);} 
  strokeWeight(1); circle(DX0,DY0,26);
  zAng = (float)getAngVal(ArrayVal[zAPnt + 1]); zRad = zAng /57.2958f;
  DX1 = DX0 + PApplet.parseInt(zUL * cos(zRad)); DY1 = DY0 + PApplet.parseInt(zUL * sin(zRad));
  strokeWeight(8);
  line(DX0,DY0,DX1,DY1); line(DX0 - 8,DY0,DX1 - 4,DY1); line(DX0 + 8,DY0,DX1 + 4,DY1); DY1 +=8;
  // draw FR1
  if (AngField != 0) {stroke(90,0,0); fill(90,0,0);} else {stroke(255,0,0); fill(255,0,0);} 
  strokeWeight(1); circle(DX1,DY1,24);
  HX0 = DX1; HY0 = DY1;
  zAng = zAng + 180.0f - (float)getAngVal(ArrayVal[zAPnt]); zRad = zAng /57.2958f;
  DX0 = DX1 + PApplet.parseInt(zLL * cos(zRad)); DY0 = DY1 + PApplet.parseInt(zLL * sin(zRad));
  strokeWeight(10); line(DX1,DY1,DX0,DY0);
  // draw foot
  fill(30,30,30); strokeWeight(1); circle(DX0,DY0,24);
  if (DY0 > GndYM) {GndYM = DY0;}
  
  // draw FL3
  DX0 = 338; DY0 = 112;
  if (AngField != 5) {stroke(120,120,210); fill(120,120,210);} else {stroke(190,190,255); fill(190,190,255);}
  strokeWeight(1); circle(DX0,DY0,26);
  zAng = (float)getAngVal(ArrayVal[zAPnt + 5]); zRad = zAng /57.2958f;
  DX1 = DX0 + PApplet.parseInt(zSL * cos(zRad)); DY1 = DY0 + PApplet.parseInt(zSL * sin(zRad));
  strokeWeight(10); line(DX0,DY0,DX1,DY1);
  // draw FL2
  DX0 = 338; DY0 = 184;
  if (AngField != 4) {stroke(0,160,0); fill(0,160,0);} else {stroke(0,255,0); fill(0,255,0);} 
  strokeWeight(1); circle(DX0,DY0,26);
  zAng = (float)getAngVal(ArrayVal[zAPnt + 4]); zRad = zAng /57.2958f;
  DX1 = DX0 + PApplet.parseInt(zUL * cos(zRad)); DY1 = DY0 + PApplet.parseInt(zUL * sin(zRad));
  strokeWeight(8);
  line(DX0,DY0,DX1,DY1); line(DX0 - 8,DY0,DX1 - 4,DY1); line(DX0 + 8,DY0,DX1 + 4,DY1); DY1 +=8;
  // draw FL1
  if (AngField != 3) {stroke(160,0,0); fill(160,0,0);} else {stroke(255,0,0); fill(255,0,0);} 
  strokeWeight(1); circle(DX1,DY1,24);
  HX1 = DX1; HY1 = DY1;
  zAng = zAng + 180.0f - (float)getAngVal(ArrayVal[zAPnt + 3]); zRad = zAng /57.2958f;
  DX0 = DX1 + PApplet.parseInt(zLL * cos(zRad)); DY0 = DY1 + PApplet.parseInt(zLL * sin(zRad));
  strokeWeight(10); line(DX1,DY1,DX0,DY0);
  // draw foot
  fill(110,110,110); strokeWeight(1); circle(DX0,DY0,24);
  if (DY0 > GndYM) {GndYM = DY0;}
  
  // draw RR3
  DX0 = 476; DY0 = 112;
  if (AngField != 8) {stroke(50,50,140); fill(50,50,140);} else {stroke(140,160,255); fill(140,160,255);}
  strokeWeight(1); circle(DX0,DY0,26);
  zAng = (float)getAngVal(ArrayVal[zAPnt + 8]); zRad = zAng /57.2958f;
  DX1 = DX0 + PApplet.parseInt(zSL * cos(zRad)); DY1 = DY0 + PApplet.parseInt(zSL * sin(zRad));
  strokeWeight(10); line(DX0,DY0,DX1,DY1);
  // draw RR2
  DX0 = 476; DY0 = 184;
  if (AngField != 7) {stroke(0,90,0); fill(0,90,0);} else {stroke(0,255,0); fill(0,255,0);} 
  strokeWeight(1); circle(DX0,DY0,26);
  zAng = (float)getAngVal(ArrayVal[zAPnt + 7]); zRad = zAng /57.2958f;
  DX1 = DX0 + PApplet.parseInt(zUL * cos(zRad)); DY1 = DY0 + PApplet.parseInt(zUL * sin(zRad));
  strokeWeight(8);
  line(DX0,DY0,DX1,DY1); line(DX0 - 8,DY0,DX1 - 4,DY1); line(DX0 + 8,DY0,DX1 + 4,DY1); DY1 +=8;
  // draw RR1
  if (AngField != 6) {stroke(90,0,0); fill(90,0,0);} else {stroke(255,0,0); fill(255,0,0);} 
  strokeWeight(1); circle(DX1,DY1,24);
  HX2 = DX1; HY2 = DY1;
  zAng = zAng + 180.0f - (float)getAngVal(ArrayVal[zAPnt + 6]); zRad = zAng /57.2958f;
  DX0 = DX1 + PApplet.parseInt(zLL * cos(zRad)); DY0 = DY1 + PApplet.parseInt(zLL * sin(zRad));
  strokeWeight(10); line(DX1,DY1,DX0,DY0);
  // draw foot
  fill(30,30,30); strokeWeight(1); circle(DX0,DY0,24);
  if (DY0 > GndYM) {GndYM = DY0;}
  
  // draw RL3
  DX0 = 529; DY0 = 112;
  if (AngField != 11) {stroke(120,120,210); fill(120,120,210);} else {stroke(190,190,255); fill(190,190,255);}
  strokeWeight(1); circle(DX0,DY0,26);
  zAng = (float)getAngVal(ArrayVal[zAPnt + 11]); zRad = zAng /57.2958f;
  DX1 = DX0 + PApplet.parseInt(zSL * cos(zRad)); DY1 = DY0 + PApplet.parseInt(zSL * sin(zRad));
  strokeWeight(10); line(DX0,DY0,DX1,DY1);
  // draw RL2
  DX0 = 529; DY0 = 184;
  if (AngField != 10) {stroke(0,160,0); fill(0,160,0);} else {stroke(0,255,0);fill(0,255,0);} 
  strokeWeight(1); circle(DX0,DY0,26);
  zAng = (float)getAngVal(ArrayVal[zAPnt + 10]); zRad = zAng /57.2958f;
  DX1 = DX0 + PApplet.parseInt(zUL * cos(zRad)); DY1 = DY0 + PApplet.parseInt(zUL * sin(zRad));
  strokeWeight(8);
  line(DX0,DY0,DX1,DY1); line(DX0 - 8,DY0,DX1 - 4,DY1); line(DX0 + 8,DY0,DX1 + 4,DY1); DY1 +=8;
  // draw RL1
  if (AngField != 9) {stroke(160,0,0); fill(160,0,0);} else {stroke(255,0,0); fill(255,0,0);} 
  strokeWeight(1); circle(DX1,DY1,24);
  HX3 = DX1; HY3 = DY1;
  zAng = zAng + 180.0f - (float)getAngVal(ArrayVal[zAPnt + 9]); zRad = zAng /57.2958f;
  DX0 = DX1 + PApplet.parseInt(zLL * cos(zRad)); DY0 = DY1 + PApplet.parseInt(zLL * sin(zRad));
  strokeWeight(10); line(DX1,DY1,DX0,DY0);
  // draw foot
  fill(110,110,110); strokeWeight(1); circle(DX0,DY0,24);
  if (DY0 > GndYM) {GndYM = DY0;}
}

// ----------------------------------------------------------------------

// ----------------------------------------------------------------------

public void adjustField(int zFieldTgt) {
  // adjusts the slider field value for zFld
  FieldTgt = zFieldTgt;
  int zVal;  // limits
  if (sliderDwn < 1) {
    // just been clicked so get the current value
    switch(FieldTgt) {
      case FldT1maxLft: FieldVal = T1maxAngLft; FldLL = 0; FldUL = 160; break;
      case FldT1maxRht: FieldVal = T1maxAngRht; FldLL = 0; FldUL = 160; break;
      case FldT2maxLft: FieldVal = T2maxAngLft; FldLL = 10; FldUL = 140; break;
      case FldT2minLft: FieldVal = T2minAngLft; FldLL = 0; FldUL = 130; break;
      case FldT2maxRht: FieldVal = T2maxAngRht; FldLL = 10; FldUL = 140; break;
      case FldT2minRht: FieldVal = T2minAngRht; FldLL = 0; FldUL = 130; break;
      case FldT1WlkLft: FieldVal = WalkLift; FldLL = 10; FldUL = 60; break;
    }
  }
  // calculate new value and limit
  zVal = FieldVal + (mX - mXD)/3;
  if (zVal > FldUL) {zVal = FldUL;}
  if (zVal < FldLL) {zVal = FldLL;}
  switch(FieldTgt) {
    case FldT1maxLft: T1maxAngLft = zVal; calcRideHeights(); break;
    case FldT1maxRht: T1maxAngRht = zVal; calcRideHeights(); break;
    case FldT2maxLft: T2maxAngLft = zVal; calcMidSteps(); break;
    case FldT2minLft: T2minAngLft = zVal; calcMidSteps(); break;
    case FldT2maxRht: T2maxAngRht = zVal; calcMidSteps(); break;
    case FldT2minRht: T2minAngRht = zVal; calcMidSteps(); break;
    case FldT1WlkLft: WalkLift = zVal; break;
  } 
  checkDepFields();
  rateFlag = true; drawFlag = 1;
  sliderDwn = 1; // set slider down flag to prevent inc/dec from false triggering
}

// ----------------------------------------------------------------------

public void adjustLinkVal(int zT) {
  // called from the linking process, adjust an angle by AngDif
  AngVal[zT] = getAngVal(AngVal[zT]) + AngDif;
  if (AngVal[zT] < AngVLL[zT]) {
    AngVal[zT] = AngVLL[zT];
  }
  else if (AngVal[zT] > AngVUL[zT]) {
    AngVal[zT] = AngVUL[zT];
  }
  ArrayVal[(rowPnt * rowWidth) + zT] = AngVal[zT];
  setSliderX();

  // send value to usb port
  msgTx += "ST" + str(Ch[zT]) + "." +"SA" + str(AngVal[zT]) + ".";
}


// ######################################################################


public int Calc_AngLA(int zAng2,float zR_RH) {
  // calculates intermediate tier 1 angles from the parameters passed
  float zUB = (float)zAng2;
  float zUH = 75.22f * sin(zUB/57.2958f);
  float zLHT = zR_RH - zUH;
  //print("zUH "); print(zUH); print("\tzLHT "); println(zLHT);
  float zLB = 90.0f - degrees(asin(zLHT/103.5f));
  //print("zUB "); print(zUB); print("\tzLB "); println(zLB);
  return (int)(90.0f + zUB - zLB);
}

// ----------------------------------------------------------------------

public void calcMidSteps() {
  // simulator angles have been adjusted so recalculate mid points
  T2midAngLft = T2minAngLft + ((T2maxAngLft - T2minAngLft)/2);
  T2midAngRht = T2minAngRht + ((T2maxAngRht - T2minAngRht)/2);
  CalcNomStepsLft = T2maxAngLft - T2minAngLft + 1;
  CalcNomStepsRht = T2maxAngRht - T2minAngRht + 1;
}

// ----------------------------------------------------------------------

public void calcRates() {
  // work through the list and calculate rate angle values
  // we determine the start angle, end angle and number of intermediate steps to do this
  HelpOFF++;  // turn off help system if not already off
  //println("calcRates()...");
  //t0 = millis();  // used to time function duration
  rateFlag = false;
  int zFA = 0;  // finish angle
  int zSA = 0;  // start angle
  int zSR = 0;  // start angle row number
  for (int zC = 0; zC < (rowWidth - 1); zC++) {
    // we work down each column, one angle at a time
    // only the first column can contain commnd lines
    int zI = 0;  // counts the number of steps associated with the rate angles and the end fixed angle
    int zIT = 0; // temporary step value
    int zS = 0;  // state machine pointer
    for (int zP = 0; zP < rowNum; zP++) {
      int zR = zP * rowWidth;  // initialise the row pointer for each column
      //if ((zC > 0) || (ArrayVal[zR + zC] >= 0)) {
      if (ArrayVal[zR] >= 0) {
        // the row is not a command line so use the data
        //println("col: " + str(zC) + "  row: " + str(zP) + "  zS=" + str(zS) + "  Ang: " + str(ArrayVal[zR + zC]));
        switch (zS) {
          case 0:  // looking for first angle(s)
            // in this state if we find a rate angle before a valid angle we clear the rate flag
            if (ArrayVal[zR + zC] >= 400) {
              // this is the error condition, this rate angle does not have a fixed angle before it
              // so we remove the rate flag and set it as a fixed angle
              ArrayVal[zR + zC] -= 400;  // remove the rate flag from this angle
            } 
            zSA = getAngVal(ArrayVal[zR + zC]);  // store the fixed angle as the start angle
            zS++; break;  // move onto the next task for the next angle
          case 1:  // tracking fixed angles until we find a rate angle
            if (ArrayVal[zR + zC] >= 400) {
              // this is the first rate angle, so we now change the task to look for the next
              // fixed angle before calculating the rate(s)
              zSR = zP;  // store the row reference for this angle
              zIT = ArrayVal[zR + rowWidth -1];  // collect the number of steps
              while (zIT > 200) {zIT -= 200;}  // strip off any rate profile
              zI = zIT; zS++;
            } else {
              // this is a fixed angle so record it as a valid start angle
              zSA = getAngVal(ArrayVal[zR + zC]);  // store the fixed angle as the start angle
            } break;
          case 2:  // look for a fixed angle to set a target range for the rate angles
            if (ArrayVal[zR + zC] >= 400) {
              // we've encountered another rate angle so collect the number of steps
              zIT = ArrayVal[zR + rowWidth -1]; // add on the number of steps
              while (zIT > 200) {zIT -= 200;}    // strip off any rate profile
              zI += zIT;
            } else {
              // this is the terminating angle we are looking for so we can no calculate the rate angles
              zIT = ArrayVal[zR + rowWidth -1];  // add on the number of steps to get the total
              while (zIT > 200) {zIT -= 200;}    // strip off any rate profile
              zI += zIT;
              // now recalculate each rate angle based on its associate number of steps
              int zIC = 0;  // initiatlise the cumulative step value
              zFA = getAngVal(ArrayVal[zR + zC]);  // note finish angle
              
              // now set the rates, within the range, ignoring Cmd lines
              for (int zN = zSR; zN < zP; zN++) {
                //print("Adjusting rate in col: " + str(zC) + "  row: " + str(zN));
                int zRN = zN * rowWidth;
                if (ArrayVal[zRN] >= 0) {
                  // this is not a command line
                  zIT =  ArrayVal[zRN + rowWidth -1]; // get the cumulative steps
                  while (zIT > 200) {zIT -= 200;}    // strip off any rate profile
                  zIC += zIT;
                  int zAng = zSA + ((zIC * (zFA - zSA))/zI);
                  //println("  to: " + zAng);
                  //println("zSA: " + str(zSA) + 
                  //println("zIC: " + str(zIC) + "  zI: " + str(zI));
                  ArrayVal[zRN + zC] = zAng + 400;
                }
              }
              // now restart the tasks from this fixed angle
              zI = 0;    // reset the number of steps associated with the rate angles and the end fixed angle
              zSA = zFA; // store the fixed finish angle as the new start angle
              zSR = zP;  // store the row reference for this angle
              zS = 1;    // reset the task pointer to look for more rate angles
            }
            break;
        }
      }
    }
  }
  //println("calcRates() completed!");
  //println("t1-t0 = " + str(millis() - t0));
  HelpOFF--;  // turn on help system if held off by this function
}

// ----------------------------------------------------------------------

public void calcRideHeights() {
  // calculate the ride heights due to a T1 max angle change
  // initial calcs
  float zR_UC = T2maxAngRht - 90.0f;           // forward angles
  float zL_UC = T2maxAngLft - 90.0f;
  float zR_LB = 180.0f - T1maxAngRht + zR_UC;  // T1 vertical angle
  float zL_LB = 180.0f - T1maxAngLft + zL_UC;
  // conversion factors:
  //   90° == 1.57079557 rads
  // 1 rad == 57.2958°
  //float zR_UH = 75.22 * sin(1.57079557 - (zR_UC/57.2958));  // height contribution of upper leg
  float zR_UH = 75.22f * sin(radians(90.0f - zR_UC));  // height contribution of upper leg
  float zL_UH = 75.22f * sin(radians(90.0f - zL_UC));
  float zR_LH = 103.5f * sin(radians(90.0f - zR_LB));  // height contribution of lower leg
  float zL_LH = 103.5f * sin(radians(90.0f - zL_LB));
  float zR_RH = zR_UH + zR_LH;  // combined 'ride' height
  float zL_RH = zL_UH + zL_LH;
  CalcMaxHgtRht = (int)zR_RH;  // display the ride height
  CalcMaxHgtLft = (int)zL_RH;
}

// ----------------------------------------------------------------------

public void calcSelSteps() {
  // calculate steps for selected rows using difference with previous
  // row, providing it is not a command line
  if (SelCnt < 1) {return;}
   drawFlag = 1;
  if (rowNum < 2) {msgTx = "Error - < 2 lines"; return;}
  
  int zA = 0;       // any Temp[] angle
  int zB = 0;       // any row angle
  int zD = 0;       // differential angle range
  int zM;           // flag recording rate profile for this line
  int zN = 0;       // number of lines set/modified
  boolean zNS = false;  // == true if a new step value is determined
  int zP = 0;       // servo angle pointer
  int zRP;          // array memory pointer
  int zSteps =  1;  // min num steps
  int zT = 0;       // flag > 0 if Temp[] holds a reference line

  for (int zR = 0; zR < rowNum; zR++) {
    zRP = zR * rowWidth;        // current row data pointer
    if (!Select[zR]) {
      // this line is not selected so store it as a previous reference if not a command line
      if (ArrayVal[zRP] >= 0) {grabRowInTemp(zR); zT = 1;}
    } else {
      if (ArrayVal[zRP] >= 0) {
        // not a command row so perform the analysis
        if (zT < 1) {
          // either the first row or the first row with angle data, so use it as reference
          grabRowInTemp(zR); zT = 1;
        } else {
          // this row is selected so, and we have a reference line, so do the calculation
          // first grab the rate profile, to re-apply later
          zSteps = ArrayVal[zRP + rowWidth - 1];
          zM = 0;
          if (zSteps > 600) {zM = 3;}
          else if (zSteps > 400) {zM = 2;}
          else if (zSteps > 200) {zM = 1;}
          zSteps =  1;  // reset the temp step value for each row
          zNS = false;  // clear the new step flag
          for (zP = 0; zP < (rowWidth - 1); zP++) {
            // compare each angle with the Temp[] reference row, ignoring rate angles
            zA = getAngVal(Temp[zP]);  // strip off rate and Pwr flags
            if (!AngRate) {
              // the Temp[] reference is not a rate angle
              zB = getAngVal(ArrayVal[zP + zRP]);
              if (!AngRate) {
                // the current angle is not a rate angle too, so do the comparison
                zD = abs(zB - zA);
                zD /= DegStep;  // divide the angle range by DegStep
                if (zD > zSteps) {zSteps = zD; zNS = true;}
              }
            }
          }
          if (keySHIFT) {zSteps = 30;}  // using default value over-ride
          if (keySHIFT || zNS) {
            if (zM == 3) {zSteps += 600;}
            else if (zM == 2) {zSteps += 400;}
            else if (zM == 1) {zSteps += 200;}
            ArrayVal[zRP + rowWidth - 1] = zSteps; zN++;
          }
          // now grab this modified row as the reference for the next selected row
          grabRowInTemp(zR);
        }
      }
    }
  } 
  if (zT < 1) {msgTx = "Error - no ref line"; return;}
  if (zN < 1) {msgTx = "Error - no lines set"; return;}

  msgTx = str(zN) + " line steps set";
  rateFlag = true;
}

// ----------------------------------------------------------------------

public void Calc_Walk() {
  // calculates a walk sequence based on the Scrn1 settings
  WalkSim = 0;  // ensure simulator is stopped
  cursor(WAIT); ClearArray(0); ClearSelect(); mDwn = 0;
  int zFR1,zFR2,zFL1,zFL2,zRR1,zRR2,zRL1,zRL2;
  int zRP = 0;  // rowPnt
  
  // calculate nominal number of steps, but Steps is actually used
  CalcNomStepsLft = T2maxAngLft - T2minAngLft + 1;
  CalcNomStepsRht = T2maxAngRht - T2minAngRht + 1;
  
  // initial calcs
  float zR_UC = T2maxAngRht - 90.0f;           // forward angles
  float zL_UC = T2maxAngLft - 90.0f;
  float zR_LB = 180.0f - T1maxAngRht + zR_UC;  // T1 vertical angle
  float zL_LB = 180.0f - T1maxAngLft + zL_UC;
  // conversion factors:
  //   90° == 1.57079557 rads
  // 1 rad == 57.2958°
  //float zR_UH = 75.22 * sin(1.57079557 - (zR_UC/57.2958));  // height contribution of upper leg
  float zR_UH = 75.22f * sin(radians(90.0f - zR_UC));  // height contribution of upper leg
  float zL_UH = 75.22f * sin(radians(90.0f - zL_UC));
  float zR_LH = 103.5f * sin(radians(90.0f - zR_LB));  // height contribution of lower leg
  float zL_LH = 103.5f * sin(radians(90.0f - zL_LB));
  float zR_RH = zR_UH + zR_LH;  // combined 'ride' height
  float zL_RH = zL_UH + zL_LH;
  CalcMaxHgtRht = (int)zR_RH;  // display the ride height
  CalcMaxHgtLft = (int)zL_RH;
  if (WalkHeight > CalcMaxHgtRht) {WalkHeight = CalcMaxHgtRht;}
  if (WalkHeight > CalcMaxHgtLft) {WalkHeight = CalcMaxHgtLft;}
  
  // create the move to ready line
  int zMR = T2minAngRht + ((T2maxAngRht - T2minAngRht)/2);  // calc tier 2 mid-angle for ready position
  zFR2 = zMR; zRR2 = zMR;
  int zML = T2minAngLft + ((T2maxAngLft - T2minAngLft)/2);  // calc tier 2 mid-angle for ready position
  zFL2 = zML; zRL2 = zML;
  zFR1 = Calc_AngLA(zFR2,zR_RH); zRR1 = zFR1;
  zFL1 = Calc_AngLA(zFL2,zL_RH); zRL1 = zFL1;
  //         Row,zFR1,zFR2,zFR3,zFL1,zFL2,zFL3,zRR1,zRR2,zRR3,zRL1,zRL2,zRL3,zStp
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,  30);  // overwrite first row
  
  // Phase A - create the left up for fly step
  // zFR2, zFL2, zRR2 & zRL2 do not move in this first step
  zFL1 -= WalkLift;    // zFL2 lifts ready to fly forward
  zRR1 -= WalkLift;    // zRR2 lifts ready to fly forward
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  
  // Phase B - create the left up fly step
  zFR2 --; zFR1 = Calc_AngLA(zFR2,zR_RH);                  // FR leg pulling back
  zFL2 = T2maxAngLft; zFL1 = Calc_AngLA(zFL2,zL_RH) - WalkLift;  // FL fly to angles
  zRR2 = T2maxAngRht; zRR1 = Calc_AngLA(zRR2,zR_RH) - WalkLift;  // RR fly to angles
  zRL2 --; zRL1 = Calc_AngLA(zRL2,zL_RH);                  // RL leg pulling back
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row

  // Phase C - create the left landing step
  zFR2 --; zFR1 = Calc_AngLA(zFR2,zR_RH);            // FR leg pulling back
  zFL2 = T2maxAngLft; zFL1 = Calc_AngLA(zFL2,zL_RH); // FL landing at max fwd angle
  zRR2 = T2maxAngRht; zRR1 = Calc_AngLA(zRR2,zR_RH); // FL landing at max fwd angle
  zRL2 --; zRR1 = Calc_AngLA(zRR2,zL_RH);            // RL leg pulling back
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  
  // Phase D - create the 1st phase of pull back
  while (zFR2 >= T2minAngRht) {
    zFR2 --; zFR1 = Calc_AngLA(zFR2,zR_RH);   // FR leg pulling back
    zFL2 --; zFL1 = Calc_AngLA(zFL2,zL_RH);   // FL leg pulling back
    zRR2 --; zRR1 = Calc_AngLA(zRR2,zR_RH);   // RR leg pulling back
    zRL2 --; zRL1 = Calc_AngLA(zRL2,zL_RH);   // RL leg pulling back
    zRP++;
    setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  }
  
  // Phase E - create the right up fly step
  zFR1 -= WalkLift;                                // FR fly to angles
  zFL2 --; zFL1 = Calc_AngLA(zFL2,zL_RH);    // FL leg pulling back
  zRR2 --; zRR1 = Calc_AngLA(zRR2,zR_RH);    // RR leg pulling back
  zRL1 -= WalkLift;                                // RL fly to angles
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  
  // Phase F1 - create intermediate point
  zFR2 = zMR; zFR1 = Calc_AngLA(zFR2,zR_RH) - WalkLift;  // FR mid fly to angles
  zFL2 --; zFL1 = Calc_AngLA(zFL2,zL_RH);          // FL leg pulling back
  zRR2 --; zRR1 = Calc_AngLA(zRR2,zR_RH);          // RR leg pulling back
  zRL2 = zMR; zRL1 = Calc_AngLA(zRL2,zL_RH) - WalkLift;  // RL mid fly to angles
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row

  // Phase F2 - create max reach step
  zFR2 = T2maxAngRht; zFR1 = Calc_AngLA(zFR2,zR_RH) - WalkLift;  // FR max fwd angle
  zFL2 --; zFL1 = Calc_AngLA(zFL2,zL_RH);                  // FL leg pulling back
  zRR2 --; zRR1 = Calc_AngLA(zRR2,zR_RH);                  // RR leg pulling back
  zRL2 = T2maxAngLft; zRL1 = Calc_AngLA(zRL2,zL_RH) - WalkLift;  // RL max fwd angle
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row

  // Phase G - create the right landing step
  zFR2 = T2maxAngRht; zFR1 = Calc_AngLA(zFR2,zR_RH); // FR landing at max fwd angle
  zFL2 --; zFL1 = Calc_AngLA(zFL2,zL_RH);            // FL leg pulling back
  zRR2 --; zRR1 = Calc_AngLA(zRR2,zR_RH);            // RR leg pulling back
  zRL2 = T2maxAngLft; zRL1 = Calc_AngLA(zRL2,zL_RH); // RL landing at max fwd angle
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  
  // Phase H - create the 2nd phase of pull back
  while (zFR2 >= (zMR - 2)) {
    zFR2 --; zFR1 = Calc_AngLA(zFR2,zR_RH);   // FR leg pulling back
    zFL2 --; zFL1 = Calc_AngLA(zFL2,zL_RH);   // FL leg pulling back
    zRR2 --; zRR1 = Calc_AngLA(zRR2,zR_RH);   // RR leg pulling back
    zRL2 --; zRL1 = Calc_AngLA(zRL2,zL_RH);   // RL leg pulling back
    zRP++;
    setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  }
  
  // Phase J - create the left up fly step
  zFR2 --; zFR1 = Calc_AngLA(zFR2,zR_RH);    // FR leg pulling back
  zFL1 -= WalkLift;                                // FL fly to angles
  zRR1 -= WalkLift;                                // RR fly to angles
  zRL2 --; zRL1 = Calc_AngLA(zRL2,zL_RH);    // RL leg pulling back
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  
  // Phase K - create intermediate point
  zFR2 --; zFR1 = Calc_AngLA(zFR2,zR_RH);          // FR leg pulling back
  zFL2 = zML; zFL1 = Calc_AngLA(zFL2,zL_RH) - WalkLift;  // FL mid fly to angles
  zRR2 = zMR; zRR1 = Calc_AngLA(zRR2,zR_RH) - WalkLift;  // RL mid fly to angles
  zRL2 --; zRL1 = Calc_AngLA(zRL2,zL_RH);          // RR leg pulling back
  zRP++;
  setRowData(zRP,zFR1,zFR2,  90,zFL1,zFL2,  90,zRR1,zRR2,  90,zRL1,zRL2,  90,   1); rowNum++;  // add new row
  
  // now branch to the second row
  zRP++;
  setRowData(zRP,  -1,   1,   1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0); rowNum++;  // add new row
  
  setAngValFromRow();  // set servo values and sliders
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void changeAllSelVal(int zCol,int zDif) {
  // changes all associated selected row column values by zDif
  int zA = 0;   // any angle
  int zC0 = 0;  // any colum ref
  int zC1 = 0;  // any colum ref
  int zC2 = 0;  // any colum ref
  int zC3 = 0;  // any colum ref
  for (int zRp = 0; zRp < rowNum; zRp++) {
    if (Select[zRp]) {
      // this is a selected row so check for command line
      int zMp = zRp * rowWidth;
      if (ArrayVal[zMp] >= 0) {
        // not a command line so modify column values
        switch (zCol) {
          case  0: zC0 = 0; zC1 = 3; zC2 = 6; zC3 =  9; break;  // FR1
          case  1: zC0 = 1; zC1 = 4; zC2 = 7; zC3 = 10; break;  // FR2
          case  2: zC0 = 2; zC1 = 5; zC2 = 8; zC3 = 11; break;  // FR3
          case  3: zC0 = 0; zC1 = 3; zC2 = 6; zC3 =  9; break;  // FL1
          case  4: zC0 = 1; zC1 = 4; zC2 = 7; zC3 = 10; break;  // FL2
          case  5: zC0 = 2; zC1 = 5; zC2 = 8; zC3 = 11; break;  // FL3
          case  6: zC0 = 0; zC1 = 3; zC2 = 6; zC3 =  9; break;  // RR1
          case  7: zC0 = 1; zC1 = 4; zC2 = 7; zC3 = 10; break;  // RR2
          case  8: zC0 = 2; zC1 = 5; zC2 = 8; zC3 = 11; break;  // RR3
          case  9: zC0 = 0; zC1 = 3; zC2 = 6; zC3 =  9; break;  // RL1
          case 10: zC0 = 1; zC1 = 4; zC2 = 7; zC3 = 10; break;  // RL2
          case 11: zC0 = 2; zC1 = 5; zC2 = 8; zC3 = 11; break;  // RL3
        }
        zA = getAngVal(ArrayVal[zMp + zC0]);
        if (zA < AngVUL[zC0]) {ArrayVal[zMp + zC0] += zDif;}
        zA = getAngVal(ArrayVal[zMp + zC1]);
        if (zA < AngVUL[zC1]) {ArrayVal[zMp + zC1] += zDif;}
        zA = getAngVal(ArrayVal[zMp + zC2]);
        if (zA < AngVUL[zC2]) {ArrayVal[zMp + zC2] += zDif;}
        zA = getAngVal(ArrayVal[zMp + zC3]);
        if (zA < AngVUL[zC3]) {ArrayVal[zMp + zC3] += zDif;}
      }
    }
  }
}

// ----------------------------------------------------------------------

public void changeAngleVal() {
  // user has clicked on an angle value in the list window so change it
  
  int zR = getRowMYD();  // row pointer +/-, == 0 if current row
  int zRP = rowPnt + zR;
  if ((zRP < 0) || (zRP >= rowNum)) {return;}  // row is before/after list
  if (ArrayVal[zRP * rowWidth] < 0) {changeCmdVal(); return;}  // command line so exit
  
  int zA = getAngleMXD();  // array column pointer based on mouse position in list
  int zAV = zA + (zRP * rowWidth);  // pointer to the angle in ArrayVal[]
  boolean zC = false;  // change flag, == true if a change has occured
  if (keySHIFT) {
    // special 'quick' copy/paste function
    if (mB == LEFT) {
      // paste quick angle value, which has previously been set
      // this angle will not have any Pwr or Rate attributes
      if (QuickAng >= 0) {
        if ((QuickAng >= AngVLL[zA]) && (QuickAng <= AngVUL[zA])) {
          ArrayVal[zAV] = QuickAng; zC = true; rateFlag = true;
        }
      }
    } else {
      // record 'quick' angle value
      // only record angle, not Pwr or Rate attributes
      QuickAng = getAngVal(ArrayVal[zAV]);
      msgTx = "QuickAng = " + str(QuickAng);
    }
  } else {
    // normal increment/decrement process
    // the angle may be 'rated' if so this will remove the rate flag
    // the ON/OFF power factor will not be affected
    if (mB == LEFT) {
      // increment angle value
      AngLast = getAngVal(ArrayVal[zAV]);
      if (AngLast < AngVUL[zA]) {
        if (Select[zRP]) {
          // this row is selected so change all column values by the same
          // amount as if they were all linked together
          changeAllSelVal(zA,1);
          if (Select[rowPnt]) {zR = 0;}  // current row is affected, so send an update
        } else {
          // normal row
          ArrayVal[zAV] = 1 + AngLast;
        }
        zC = true; AngDif = 1;
      }
    } else {
      // decrement angle value
      AngLast = getAngVal(ArrayVal[zAV]);
      if (AngLast > AngVLL[zA]) {
        if (Select[zRP]) {
          // this row is selected so change all column values by the same
          // amount as if they were all linked together
          changeAllSelVal(zA,-1);
          if (Select[rowPnt]) {zR = 0;}  // current row is affected, so send an update
        } else {
          // normal row
          ArrayVal[zAV] = AngLast - 1;
        }
        zC = true; AngDif = 1;
      }
    }
    if (!AngPwr) {ArrayVal[zAV] += 200;}  // assign the power setting
  }
  
  if (zC) {
    if (zR == 0) {
      // current row changed
      setAngValFromRow();     // change sliders
      msgTx = "ST" + str(Ch[zA]) + "." +"SA" + str(AngVal[zA]) + ".";
      if (Select[zRP]) {setSliderX(); sendRowData(rowPnt,rowPnt,1);}
      else if (Link[zA] > 0) {moveLinkedVal(zA);}
      else {setSliderX(); usbPortWrite(msgTx);}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
    }
    rateFlag = true; drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

public void changeCmdType() {
  // it is possible to change the command 'Type' for executeable commands
  int zR = getRowMYD();                // row pointer +/-, == 0 if current row
  int zCP = (rowPnt + zR) * rowWidth;  // memory pointer
  int zXM = 7;                         // max Exe code value
  CmdCode = ArrayVal[zCP + 1];         // get command type code
  if (CmdCode < 6) {return;}           // not an executaable code
  
  if (mB == LEFT) {
    CmdCode++; if (CmdCode > zXM) {CmdCode = 6;}
  } else {
    CmdCode--; if (CmdCode < 6) {CmdCode = zXM;}
  }
  ArrayVal[zCP + 1] = CmdCode;
  // now set defaults for new Cmd line
  ArrayVal[zCP + 2] = 0; ArrayVal[zCP + 3] = 0; ArrayVal[zCP + 4] = 0; ArrayVal[zCP + 5] = 0; ArrayVal[zCP + 6] = 0;  // clear values
  switch (CmdCode) {
    case 6:  // For... next loop
      ArrayVal[zCP + 2] = 1; break;
    case 7:  // ...Next in For... loop
      break;
  }
  drawFlag = 1;
}

// ----------------------------------------------------------------------

public void changeCmdVal() {
  // user has clicked in list line region on a command line
  int zR = getRowMYD();  // row pointer +/-, == 0 if current row
  int zCP = (rowPnt + zR) * rowWidth;
  int zC2 = 0;  // changed 1st value flag
  AngField = getAngleMXD();
  CmdCode = ArrayVal[zCP + 1] ;  // get command type code
  CmdVal0 = ArrayVal[zCP + 2] ;  // get command 1st value
  if (CmdCode == 0) {return;}    // blank line
  else if (CmdCode == 1) {
     // Goto command
     if (mB == LEFT) {if (CmdVal0 < (rowNum -1)) {CmdVal0++; zC2++;}}
     else {if (CmdVal0 > 0) {CmdVal0--; zC2++;}}
  }
  else if (CmdCode == 2) {
     // Speed command
     if (mB == LEFT) {if (CmdVal0 < 250) {CmdVal0++; zC2++;}}
     else {if (CmdVal0 > 1) {CmdVal0--; zC2++;}}
  }
  else if (CmdCode == 3) {
     // Pause command
     if (mB == LEFT) {
       if (CmdVal0 < 1000) {
         CmdVal0+= Inc; zC2++; IncCnt++;
         if (IncCnt > 10) {IncCnt = 0; Inc += 5;}
         if (CmdVal0 > 1000) {CmdVal0 = 1000;}  // don't overshoot
       }
     } else {
       if (CmdVal0 > 1) {
         CmdVal0-= Inc; zC2++; IncCnt++;
         if (IncCnt > 10) {IncCnt = 0; Inc += 5;}
         if (CmdVal0 < 1) {CmdVal0 = 1;}  // don't undershoot
       }
     }
     Pause = CmdVal0;  // set the default value to match this
  }
  else if (CmdCode == 4) {
     // Flag command
     if (mB == LEFT) {if (CmdVal0 < 180) {CmdVal0++; zC2++;}}
     else {if (CmdVal0 > 1) {CmdVal0--; zC2++;}}
     Flag = CmdVal0;  // set the default value to match this
  }
  else if (CmdCode == 5) {
     // Balance command
          if ((mB == LEFT) && (CmdVal0 < 3)) {CmdVal0++; zC2++;}
     else if ((mB == RIGHT) && (CmdVal0 > 0)) {CmdVal0--; zC2++;}
  }
  else if (CmdCode == 6) {
     // For... before Next command
          if (AngField == 1) {changeCmdType();}
     else if (AngField >= 2) {
       if (mB == LEFT) {if (CmdVal0 < 1000) {CmdVal0++; zC2++;}}
       else {if (CmdVal0 > 1) {CmdVal0--; zC2++;}}
     } AngField = -1;
  }
  else if (CmdCode == 7) {
     // Next in ...For loop command
     if (AngField == 1) {changeCmdType();} 
  }

  if (zC2 > 0) {ArrayVal[zCP + 2] = CmdVal0; drawFlag = 1;}
}

// ----------------------------------------------------------------------

public void changeSteps() {
  // user clicked on steps field so inc/dec value
  // check for command row
  boolean zC = false;  // something changed flag
  int zSteps =  1;
  int zD = 0;
  int zP = 0;
  int zM = 0;          // rate mode flag
  int zR = getRowMYD();
  zR = rowPnt + zR;
  if ((zR < 0) || (zR >= rowNum)) {return;}  // row is before list
  if (ArrayVal[zR * rowWidth] < 0) {return;}  // command line so exit
  
  // strip off step profile zM, to add it later
  zSteps =  ArrayVal[(zR * rowWidth) + 12];
       if (zSteps > 600) {zSteps -= 600; zM = 3;}
  else if (zSteps > 400) {zSteps -= 400; zM = 2;}
  else if (zSteps > 200) {zSteps -= 200; zM = 1;}
  if (keySHIFT) {
    // special 'quick' copy/paste function
    if (mB == LEFT) {
      // paste quick step value, which has previously been set
      // this angle will not have any Pwr or Rate attributes
      if (QuickStep >= 0) {
        zSteps = QuickStep; zC = true;
             if (zM > 2) {zSteps += 600;}
        else if (zM > 1) {zSteps += 400;}
        else if (zM > 0) {zSteps += 200;}
        ArrayVal[(zR * rowWidth) + 12] = zSteps;
      }
    } else {
      // record 'quick' angle value
      // only record angle, not Pwr or Rate attributes
      QuickStep = zSteps;
      msgTx = "QuickStep = " + str(QuickStep); drawFlag = 1;
    }
  } else {
    // normal inc/dec process
    if (mB == LEFT) {
      // left mouse button increments the value
      if (zSteps < 180) { // max steps value is 180
        zSteps+= Inc; zC = true; IncCnt++;
        if (IncCnt > 10) {IncCnt = 0; Inc++;}
        if (zSteps > 180) {zSteps = 180;}  // don't overshoot
      }
    } else {
      // right mouse button decrements the value
      if (zSteps > 1) { // min steps value is 1
        zSteps-= Inc; zC = true; IncCnt++;
        if (IncCnt > 10) {IncCnt = 0; Inc++;}
        if (zSteps < 1) {zSteps = 1;}  // don't overshoot
      }
    }
    if (zM > 2) {zSteps += 600;}
    else if (zM > 1) {zSteps += 400;}
    else if (zM > 0) {zSteps += 200;}
    ArrayVal[(zR * rowWidth) + 12] = zSteps;
  }
  if (zC) {rateFlag = true; drawFlag = 1;}
}

// ----------------------------------------------------------------------

public void checkDepFields() {
  // simulator field values have changed, ensure that they don't overlap
  switch(FieldTgt) {
    case FldT2maxLft:
      if (T2maxAngLft < (T2minAngLft + 10)) {T2minAngLft = T2maxAngLft - 10;}
      break;
    case FldT2minLft:
      if (T2minAngLft > (T2maxAngLft - 10)) {T2maxAngLft = T2minAngLft + 10;}
      break;
    case FldT2maxRht:
      if (T2maxAngRht < (T2minAngRht + 10)) {T2minAngRht = T2maxAngRht - 10;}
      break;
    case FldT2minRht:
      if (T2minAngRht > (T2maxAngRht - 10)) {T2maxAngRht = T2minAngRht + 10;}
      break;
  }
}

// ----------------------------------------------------------------------

public void checkGotoCmdDel(int zD,int zCnt) {
  // called when row zD has zCnt deleted to correct any goto commands
  int zP;
  for (int zR = 0; zR < rowNum; zR++) {
    // look for goto commands
    zP = zR * rowWidth;
    if (ArrayVal[zP] < 0) {
      // this is a commands line, but is it a goto?
      if (ArrayVal[zP + 1] == 1) {
        // this is a Goto command so check that zD is not less than branched row
        if (ArrayVal[zP + 2] > zD) {
          ArrayVal[zP + 2]-= zCnt;
          println("Goto decremented");
        }
      }
    }
  }
}

// ----------------------------------------------------------------------

public void checkGotoCmdIns(int zI,int zCnt) {
  // called when row zI has zCnt inserted so correct any goto commands
  int zP;
  for (int zR = 0; zR < rowNum; zR++) {
    // look for goto commands
    zP = zR * rowWidth;
    if (ArrayVal[zP] < 0) {
      // this is a command line, but is it a goto?
      if (ArrayVal[zP + 1] == 1) {
        // this is a Goto command so check that zI is not less than branched row
        if (ArrayVal[zP + 2] >= zI) {
          ArrayVal[zP + 2]+= zCnt;
          println("Goto incremented");
        }
      }
    }
  }
}

// ----------------------------------------------------------------------

public void ClearArray(int zM) {
  // clear the values in an array to 90 degrees
  // zM = 0  just reset the array data only
  // zM = 1  update angles and send instruction over usb port
  cursor(WAIT);
  int zS = 0; int zW = rowWidth;
  for (int zC = 0; zC < arrayMax; zC++) {
    zS++; 
    if (zS < zW) {
           if (zS ==  1) {ArrayVal[zC] =  0;}
      else if (zS ==  2) {ArrayVal[zC] =  0;}
      else if (zS ==  3) {ArrayVal[zC] = 95;}
      else if (zS ==  4) {ArrayVal[zC] =  0;}
      else if (zS ==  5) {ArrayVal[zC] =  0;}
      else if (zS ==  6) {ArrayVal[zC] = 85;}
      else if (zS ==  7) {ArrayVal[zC] =  0;}
      else if (zS ==  8) {ArrayVal[zC] =  0;}
      else if (zS ==  9) {ArrayVal[zC] = 95;}
      else if (zS == 10) {ArrayVal[zC] =  0;}
      else if (zS == 11) {ArrayVal[zC] =  0;}
      else if (zS == 12) {ArrayVal[zC] = 85;}
    } else {ArrayVal[zC] = Steps; zS = 0;}
  }
  // reset pointers
  rowPnt = 0; rowNum = 1; drawFlag = 1;
  if (zM > 0) {return;}
  
  setAngValFromRow();  // set servo values and sliders
  // update Arduino
  usbEn = true; msgTx = "MC."; usbPortWrite(msgTx);
  usbEn = false; cursor(ARROW);
}

// ----------------------------------------------------------------------

public void ClearLinks() {
  // resets the link array contents
  for (int zL = 0; zL < rowWidth;zL++) {Link[zL] = 0;}
}

// ----------------------------------------------------------------------

public void ClearRow(int zR) {
  // sets row zR to default angle values and steps
  zR = zR * rowWidth;
  for (int zP = zR; zP < (zR + rowWidth - 1); zP++) {
         if (zP ==  0) {ArrayVal[zP] =  0;}
    else if (zP ==  1) {ArrayVal[zP] =  0;}
    else if (zP ==  2) {ArrayVal[zP] = 95;}
    else if (zP ==  3) {ArrayVal[zP] =  0;}
    else if (zP ==  4) {ArrayVal[zP] =  0;}
    else if (zP ==  5) {ArrayVal[zP] = 85;}
    else if (zP ==  6) {ArrayVal[zP] =  0;}
    else if (zP ==  7) {ArrayVal[zP] =  0;}
    else if (zP ==  8) {ArrayVal[zP] = 95;}
    else if (zP ==  9) {ArrayVal[zP] =  0;}
    else if (zP == 10) {ArrayVal[zP] =  0;}
    else if (zP == 11) {ArrayVal[zP] = 85;}
} ArrayVal[zR + rowWidth -1] = Steps;
}

// ----------------------------------------------------------------------

public void ClearRowBlockFT(int zF, int zT) {
  // clears a region from row zF to row zT inclusive
  for (int zR = zF; zR <= zT; zR++) {ClearRow(zR);}
}

// ----------------------------------------------------------------------

public void ClearSelect() {
  // clear the contents of the Select[] array
  cursor(WAIT);
  for (int zS = 0; zS < arrayMax; zS++) {Select[zS] = false;}
  SelCnt = 0; SelLast = -1;
  cursor(ARROW);
}

// ----------------------------------------------------------------------

public void closeCOMPort() {
  // Close the COM port and empty the serial buffer
  if (comFlag > 0) {
    usbPort.clear(); usbPort.stop();
    comFlag = 0; msgCOM = "----"; msgRx = ""; msgTx = ""; drawFlag = 1;
    println("COM Port Closed");
    servoOE = 1;  // turn OFF OE enabled lamp
  }
  usbEn = false; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void copyCopyTo(int zR) {
  // copies the contents of the copy memory into row memory starting row zR
  // ensure that we don't exceed the array
  int zT = CopyCnt; if ((zT + zR) >= rowDepth) {zT = rowDepth - zR;}
  zR = zR * rowWidth;
  for (int zP = 0; zP < (zT * rowWidth); zP++) {
    ArrayVal[zR + zP] = CopyMem[zP];
  }
}

// ----------------------------------------------------------------------

public void copyLftToRgt(int zF,int zT) {
  // copy the left servo value into the right-hand side servo value
  ArrayVal[(rowPnt * rowWidth) + zT] = ArrayVal[(rowPnt * rowWidth) + zF];
  // update angle array
  AngVal[zT] = AngVal[zF];
  // update sliders
  setSliderX();
  // send the angle
  msgTx = "ST" + str(Ch[zT]) + "." +"SA" + str(AngVal[zT]) + ".";
  usbPortWrite(msgTx); rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void copyLinkedVal(int zS) {
  // servos are linked so copy target to linked pair
  switch(zS) {
    case 0: copyRgtToLft(FR1,FL1); break;  // FR1
    case 1: copyRgtToLft(FR2,FL2); break;  // FR2
    case 2: copyRgtToLft(FR3,FL3); break;  // FR3
    
    case 3: copyLftToRgt(FL1,FR1); break;  // FL1
    case 4: copyLftToRgt(FL2,FR2); break;  // FL2
    case 5: copyLftToRgt(FL3,FR3); break;  // FL3

    case 6: copyRgtToLft(RR1,RL1); break;  // RR1
    case 7: copyRgtToLft(RR2,RL2); break;  // RR2
    case 8: copyRgtToLft(RR3,RL3); break;  // RR3

    case 9: copyLftToRgt(RL1,RR1); break;  // RL1
    case 10: copyLftToRgt(RL2,RR2); break;  // RL2
    case 11: copyLftToRgt(RL3,RR3); break;  // RL3
  }
}

// ----------------------------------------------------------------------

public void copyRgtToLft(int zF,int zT) {
  // copy the left servo value into the right-hand side servo value
  ArrayVal[(rowPnt * rowWidth) + zT] = ArrayVal[(rowPnt * rowWidth) + zF];
  // update angle array
  AngVal[zT] = AngVal[zF];
  // update sliders
  setSliderX();
  // send the angle
  msgTx = "ST" + str(Ch[zT]) + "." +"SA" + str(AngVal[zT]) + ".";
  usbPortWrite(msgTx); rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void copyRows() {
  // copy one or more rows into the CopyMem[] store
    int zP = 0; int zRP = rowPnt * rowWidth; int zCP; CopyCnt = 0;
  if (SelCnt < 1) {
    // copy current row
    for (zP = 0; zP < rowWidth; zP++) {
      CopyMem[zP] = ArrayVal[zP + zRP];
    }
    CopyCnt = 1;  // copied count
  } else {
    // copy selected row(s)
    cursor(WAIT);
    for (int zR = 0; zR < rowNum; zR++) {
      // work through each selected row
      if (Select[zR]) {
        // row is selected
        zRP = zR * rowWidth; zCP = CopyCnt * rowWidth;
        // grab the selected row values
        for (zP = 0; zP < rowWidth; zP++) {
          CopyMem[zP + zCP] = ArrayVal[zP + zRP];
        } CopyCnt++;  // count the copied rows
     }
    }
  }
  msgTx = str(CopyCnt) + " rows copied"; drawFlag = 1;
  cursor(ARROW);
}

// ----------------------------------------------------------------------

public void copyRowBlockFT(int zF, int zT, int zN) {
  // copy a block of rows zF to row zT, zN depth
  if (zT < zF) {
    // copying up so it's simple
    if ((zF + zN) >= rowDepth) {zN = rowDepth - zF -1;}
    for (int zP = 0; zP < zN; zP++) {
      copyRowFT(zF + zP, zT + zP);
    }
  } else {
    // copying down so start at the bottom
    if ((zT + zN) >= rowDepth) {zN = rowDepth - zT -1;}
    for (int zP = zN -1; zP >= 0; zP--) {
      copyRowFT(zF + zP, zT + zP);
    }
  }
}

// ----------------------------------------------------------------------

public void copyRowFT(int zF, int zT) {
  // copy a single row zF to row zT
  zF = zF * rowWidth; zT = zT * rowWidth;
  for (int zP = 0; zP < rowWidth; zP++) {
    ArrayVal[zP + zT] = ArrayVal[zP + zF];
  }
}

// ----------------------------------------------------------------------

public void copyToClipboard() {
  // called when user clicks on 'Copy' button
  // if lines are selected it will only copy them
  cursor(WAIT);
  int zC = 0;  // column reference
  CopyData = "  //     FR1,FR2,FR3,FL1,FL2,FL3,RR1,RR2,RR3,RL1,RL2,RL3,Stp " + CR;
  for (int zRow = 0; zRow < rowNum; zRow++) {
    if ((SelCnt == 0) || Select[zRow]) {
      int zP = zRow * rowWidth; zC = zP;
      CopyData += "  MemSet(" + Num3Dig(ArrayVal[zP] ) + ","; zP++;  // FR1
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // FR2
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // FR3
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // FL1
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // FL2
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // FL3
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // RR1
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // RR2
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // RR3
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // RL1
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // RL2
      CopyData += Num3Dig(ArrayVal[zP]) + ","; zP++;  // RL3
      CopyData += Num3Dig(ArrayVal[zP]) + ");";  // Stp
      
      // append CMD text?
      if (ArrayVal[zC] < 0) {
        // yes this is a command line
        CopyData += "  // ";  // C++ comment
        switch (ArrayVal[zC + 1]) {
          case 0: CopyData += "---"; break;
          case 1: CopyData += "Goto row " + str(ArrayVal[zC + 2]); break;
          case 2: CopyData += "Speed =  " + str(ArrayVal[zC + 2]); break;
          case 3: CopyData += "Pause for " + str(ArrayVal[zC + 2]); break;
          case 4: CopyData += "Flag = " + str(ArrayVal[zC + 2]); break;
          case 5: CopyData += "Balance = ";
                 if (ArrayVal[zC + 2] == 0) {CopyData += "OFF";}
            else if (ArrayVal[zC + 2] == 1) {CopyData += "topple only";} 
            else if (ArrayVal[zC + 2] == 2) {CopyData += "balance only";} 
            else if (ArrayVal[zC + 2] == 3) {CopyData += "topple and balance";} 
            break;
          case 6: CopyData += "For...1 to ..." + str(ArrayVal[zC + 2]) + " loop"; break;
          case 7: CopyData += "Next after a For... loop"; break;
        }
      }
      CopyData += CR;  // end of line CRLF
    }
  }
  msgTx = "Memory copied"; drawTxMsg();

  // now copy it
  StringSelection data = new StringSelection(CopyData);
  Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
  clipboard.setContents(data, data);
  drawFlag = 1; cursor(ARROW);
}

// ----------------------------------------------------------------------

public void cutRows() {
  // copy and cut current or selected rows
  if (SelCnt < 1) {
    // no rows selected so use current row
    Select[rowPnt] = true;  // fake current row selected
  }
  
  cursor(WAIT);
  // first copy the selected rows
  copyRows();
  
  if (CopyCnt == rowNum) {
    // all of the rows were selected, so reset them as default rows
    // then reset the flags and row count
    for (int zR = 0; zR < rowNum; zR++) {ClearRow(zR);}
    rowPnt = 0; rowNum = 1;
  } else {
    // copy unselected rows up then reset the rest
    int zN = 0;   // counts the unselected rows
    int zT = -1;  // row target pointer for copying
    for (int zR = 0; zR < rowNum; zR++) {
      // work through each selected row
      if (Select[zR]) {
        // row is selected so reset it
        ClearRow(zR);
        // move the target pointer to the first selected row
        if (zT < 0) {zT = zR;}
     } else {
       zN++;  // count the number of unselected rows
       if ((zR > zT) && (zT >= 0)) {
         // move the unselected row to the previously selected one
         copyRowFT(zR,zT); zT++;
         ClearRow(zR);
       }
     }
    }
    // reset flags and numbers  
    rowNum = zN; if (rowPnt >= rowNum) {rowPnt = rowNum - 1;}
  }
  ClearSelect();
  sendRowData(rowPnt,rowPnt,1);  // send data
  setAngValFromRow();     // set servo values and sliders
  msgTx = str(CopyCnt) + " rows copied and cut"; drawFlag = 1;
  cursor(ARROW);
}


// ######################################################################


public void decDegSteps() {
  // degrements the DegStep value
  if (DegStep > 1) {DegStep--; rateFlag = true; rateFlag = true; drawFlag = 1;}
}

// ----------------------------------------------------------------------

public void decField(int zFld) {
  // decrements the value of field zFld
  switch(zFld) {
    case FldT1maxLft:
      if (T1maxAngLft > 0) {T1maxAngLft--; calcRideHeights();} break;
    case FldT1maxRht:
      if (T1maxAngRht > 0) {T1maxAngRht--; calcRideHeights();} break;
    case FldT2maxLft:
      if (T2maxAngLft > 10) {T2maxAngLft--;
        if (T2maxAngLft < (T2minAngLft + 10)) {T2minAngLft = T2maxAngLft - 10;}
      calcMidSteps();} break;
    case FldT2minLft:
      if (T2minAngLft > 0) {T2minAngLft--; calcMidSteps();} break;
    case FldT2maxRht:
      if (T2maxAngRht > 10) {T2maxAngRht--;;
        if (T2maxAngRht < (T2minAngRht + 10)) {T2minAngRht = T2maxAngRht - 10;}
      calcMidSteps();} break;
    case FldT2minRht:
      if (T2minAngRht > 0) {T2minAngRht--; calcMidSteps();} break;
    case FldT1WlkLft:
      if (WalkLift > 10) {WalkLift--; drawFlag = 1;} break;
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void decRowPnt() {
  // decrement the memory row point if it is greater than zero
  if (rowPnt > 0) {
    rowPnt--; sendRowData(rowPnt,rowPnt,1);
    setAngValFromRow();  // set servo values and sliders
    drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

public void decSpeed() {
  // decrement Speed value towards minimum
  if (Speed > SpeedLL) {
    Speed-= Inc; IncCnt++;  if (keySHIFT) {Speed -= 9;}
    if (IncCnt > 10) {IncCnt = 0; Inc++;}
    if (Speed < SpeedLL) {Speed = SpeedLL;}
    sendSpeed(); drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

public void decXSlider(int zS) {
  // decrement target slider value
  servoTgt = zS;
  if (getAngVal(AngVal[zS]) > AngVLL[zS]) {
    AngLast = getAngVal(AngVal[zS]);  // get angle and strip off rate and pwr
    AngVal[zS] = AngLast - 1;
    AngDif = AngVal[zS] - AngLast;
    // update value in memory array
    ArrayVal[(rowPnt * rowWidth) + zS] = AngVal[zS];
    // send value to usb port
    msgTx = "ST" + str(Ch[zS]) + "." +"SA" + str(AngVal[zS]) + ".";
    if (Link[zS] > 0) {moveLinkedVal(zS);}
    else {setSliderX(); usbPortWrite(msgTx); rateFlag = true; drawFlag = 1;}
    if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
  }
}

// ----------------------------------------------------------------------

public void deleteRows() {
  // delete the current row until last, then reset row to default
  // if some selected then delete all of them
  // if the USB port is active send the revised row to the robot
  cursor(WAIT);
  if (rowNum == 1) {
    // only one row, so set it to default
    ClearRow(rowPnt);
  } else if (SelCnt > 0) {
    // one or more rows are selected
    if (SelCnt == rowNum) {
      // all of the rows are selected so clear memory and insert new row
      ClearSelect();
      ClearArray(1);
    } else {
      // only some rows are selected so just delete them
      int zNum = rowNum; int zS = 0;
      while (zS < rowNum) {
        // work through rows and delete selected ones
        if (Select[zS]) {
          // this is a selected row, copy data up to overwrite it, then
          // clear the last line
          moveRowBlockFT(zS + 1, zS, zNum - zS - 1);
          moveSelArrayUp();       // shift selects to match row deletion
          checkGotoCmdDel(zS,1);  // check for any goto commands and correct them if needed
          zNum--;
        } else {
          // not a selected row so simply increment the pointer
          zS++;
        }
      } rowNum = rowNum - SelCnt; if (rowPnt >= rowNum) {rowPnt = rowNum - 1;}
    }
  } else {
    // no rows are selected
    if (rowPnt == (rowNum - 1)) {
      // looking at last row so clear it and reduce the pointer and number of rows
      ClearRow(rowPnt);
      rowPnt--; rowNum--;
    } else {
      // delete a row somewhere within the list of rows by copying values up
      // and clearing the last line before reducing pointers
      moveRowBlockFT(rowPnt + 1, rowPnt, rowNum - rowPnt - 1);
      checkGotoCmdDel(rowPnt,1);  // check for any goto commands and correct them if needed
      // now reset the last line before adjusting pointers
      rowNum--;
    }
  } 
  ClearSelect(); rateFlag = true; drawFlag = 1;
  if (usbEn) {sendRowData(rowPnt,rowPnt,1);}
}


// ######################################################################


public void ESC() {
  // user has pressed the ESC key. If in the middle of a click-n-drag
  // this will be terminated and the droid returned to its original position.
  if ((sliderDwn < 1) && (MultiPnt < 1)) {return;}  // not a click-n-drag
  
  // original position is stored in Temp[] on mouswe click, so return to 
  // that position after killing all comm events
  usbBuff = "***";  // send ann abort '*' to arduinio, several to make sure
  RetRowFromTemp(); // recall the original values
  if (MultiPnt > 0) {
    // multipoint move so send full row of data
    sendRowData(rowPnt,rowPnt,1);  // send current row data to Arduino
  } else {
    // slider move so send original servo value
    // send value to usb port
    if (Link[servoTgt] > 0) {sendRowData(rowPnt,rowPnt,1);}  // more than one change
    else {
      // only one slider active so reset 6that value
      msgTx = "ST" + str(Ch[servoTgt]) + "." +"SA" + str(ArrayVal[(rowPnt * rowWidth) + servoTgt]) + ".";
      usbPortWrite(msgTx);
    }
  }
  setAngValFromRow();  // set servo values and sliders
  mDwn = 0; sliderDwn = 0; MultiPnt = 0; Help_Message();
  rateFlag = true; drawFlag = 1;
  cursor(ARROW);
}

// ----------------------------------------------------------------------

public void exit() {
  // program is closing so reset Arduino...
  usbPortWrite("XX.");
  delay(20); // allow 20ms for data to be sent
  super.exit();
}


// ######################################################################


public void getAngFillCol(int zAng,int zCP) {
  // sets background colours based on the column, angle and min/max values
  switch(zCP) {
    case 0:  // FR1
      AR = 255;
      AG = mapInt(zAng,MinA[0],MaxA[0],210,67);
      AB = mapInt(zAng,MinA[0],MaxA[0],210,67);
      break;
    case 1:  //FR2
      AR = mapInt(zAng,MinA[1],MaxA[1],210,27);
      AG = 255;
      AB = mapInt(zAng,MinA[1],MaxA[1],210,76);
      break;
    case 2:  //FR3
      AR = mapInt(zAng,MinA[2],MaxA[2],200,20);
      AG = mapInt(zAng,MinA[2],MaxA[2],220,120);
      AB = 255;
      break;
      
    case 3:  // FL1
      AR = 255;
      AG = mapInt(zAng,MinA[3],MaxA[3],210,67);
      AB = mapInt(zAng,MinA[3],MaxA[3],210,67);
      break;
    case 4:  // FL2
      AR = mapInt(zAng,MinA[4],MaxA[4],210,27);
      AG = 255;
      AB = mapInt(zAng,MinA[4],MaxA[4],210,76);
      break;
    case 5:  //FL3
      AR = mapInt(zAng,MinA[5],MaxA[5],200,20);
      AG = mapInt(zAng,MinA[5],MaxA[5],220,120);
      AB = 255;
      break;
      
    case 6:  // RR1
      AR = 255;
      AG = mapInt(zAng,MinA[6],MaxA[6],210,67);
      AB = mapInt(zAng,MinA[6],MaxA[6],210,67);
      break;
    case 7:  // RR2
      AR = mapInt(zAng,MinA[7],MaxA[7],210,27);
      AG = 255;
      AB = mapInt(zAng,MinA[7],MaxA[7],210,76);
      break;
    case 8:  //RR3
      AR = mapInt(zAng,MinA[8],MaxA[8],200,20);
      AG = mapInt(zAng,MinA[8],MaxA[8],220,120);
      AB = 255;
      break;
      
    case 9:  // RL1
      AR = 255;
      AG = mapInt(zAng,MinA[9],MaxA[9],210,67);
      AB = mapInt(zAng,MinA[9],MaxA[9],210,67);
      break;
    case 10:  //RL2
      AR = mapInt(zAng,MinA[10],MaxA[10],210,27);
      AG = 255;
      AB = mapInt(zAng,MinA[10],MaxA[10],210,76);
      break;
    case 11:  //RL3
      AR = mapInt(zAng,MinA[11],MaxA[11],200,20);
      AG = mapInt(zAng,MinA[11],MaxA[11],220,120);
      AB = 255;
      break;
  }
}

// ----------------------------------------------------------------------

public int getAngleMXD() {
  // returns an array pointer based on the mXD value
  int zA;
  zA = (mXD - 154) * 12/(495 - 154);
  zA = min(zA,11); zA = max(zA,0);
  //print("zA = "); println(zA);
  return zA;
}

// ----------------------------------------------------------------------

public int getAngVal(int zAng) {
  // strips off any auto-rate offset and sets the temp AngRate flag
  // strips off any Pwr offset and sets the temp AngPwr flag
  if (zAng >= 400) {zAng -= 400; AngRate = true;} else {AngRate = false;}
  if (zAng >= 200) {zAng -= 200; AngPwr = false;} else {AngPwr = true;}
  return zAng;
}

// ----------------------------------------------------------------------

public void getCOMPort() {
  // initialise serial comms if a port exists
  if (comFlag > 0) {
    // stops an existing connection if already made
    usbPort.clear(); usbPort.stop();
    comFlag = 0; msgCOM = "----"; delay(100);
  }
  comListLength = Serial.list().length;
  if (comListLength > 0) {
    comPnt++; if (comPnt >= comListLength) {comPnt = 0;}
    // test available serial port
    comName = Serial.list()[comPnt];
    println(comName);
    try {
      usbPort = new Serial(this, comName, 115200);
      //if (pingON > 0) {tick = 100;} // don't send ticks too early or board wont reset!
      //usbPort.bufferUntil(LF);
      comFlag = 1; msgCOM = comName;  msgTx = "USB port connected";  // USB port connected
    } catch(Exception e) {
      comFlag = 0; msgCOM = "-Error-"; msgTx = "Failed to connect";
    }
  } else {
    // no serial port available
    comFlag = 0; msgCOM = "-NA-"; msgRx = ""; msgTx = "No port available";
  }
  // set conditions
  usbEn = false;  // always disable usb port messaging
  drawFlag = 1;
}

// ----------------------------------------------------------------------

public Object getFromClipboard (DataFlavor flavor) {

  Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard(); 
  Transferable contents = clipboard.getContents(null);
  Object object = null;

  if (contents != null && contents.isDataFlavorSupported(flavor))
  {
    try
    {
      object = contents.getTransferData(flavor);
      println("Clipboard.getFromClipboard() >> Object transferred from clipboard.");
    }

    catch (UnsupportedFlavorException e1) // Unlikely but we must catch it
    {
      println("Clipboard.getFromClipboard() >> Unsupported flavor: " + e1);
      //~  e1.printStackTrace();
    }

    catch (java.io.IOException e2)
    {
      println("Clipboard.getFromClipboard() >> Unavailable data: " + e2);
      //~  e2.printStackTrace();
    }
  }

  return object;
} 

// ----------------------------------------------------------------------

public void getMinMaxAng() {
  // load the min max array from the current list
  // set all min values to max angle
  MinA[ 0] = 180; MinA[ 1] = 180; MinA[ 2] = 180; MinA[ 3] = 180;
  MinA[ 4] = 180; MinA[ 5] = 180; MinA[ 6] = 180; MinA[ 7] = 180;
  MinA[ 8] = 180; MinA[ 9] = 180; MinA[10] = 180; MinA[11] = 180;
  // set all max values to 1
  MaxA[ 0] = 0; MaxA[ 1] = 0; MaxA[ 2] = 0; MaxA[ 3] = 0;
  MaxA[ 4] = 0; MaxA[ 5] = 0; MaxA[ 6] = 0; MaxA[ 7] = 0;
  MaxA[ 8] = 0; MaxA[ 9] = 0; MaxA[10] = 0; MaxA[11] = 0;
  
  // now get the min/max values
  int zA;  // temp angle value
  for (int zR = 0; zR < rowNum; zR++) {
    int zP = zR * rowWidth;  // set the row memory pointer
    if (ArrayVal[zP] >= 0) {
      // not a Cmd line
      for (int zI = 0; zI < (rowWidth - 1); zI++) {
        zA = getAngVal(ArrayVal[zP+zI]);  
        if (zA > MaxA[zI]) {MaxA[zI] = zA;}
        else if (zA < MinA[zI]) {MinA[zI] = zA;}
      }
    }
  }
}

// ----------------------------------------------------------------------

public boolean getRated() {
  // scans the array to see if any angles are set to rate if so return true
  for (int zL = 0; zL < rowNum; zL++) {
    int zR = zL * rowWidth;
    if (ArrayVal[zR] >= 0) {
      // not a command line so check for rate value (400) added to angle?
      // as soon as we find one we return with true
      for (int zP = 0; zP < (rowWidth - 1); zP++) {
        if (ArrayVal[zP + zR] >= 400) {return true;}
      }
    }
  }
  return false;
}

// ----------------------------------------------------------------------

public int getProfile(int zR) {
  // returns the rate profile as a number for the row zR
  zR*= rowWidth; zR = ArrayVal[zR + 12];
       if (zR > 600) {zR = 3;}  // switch acceleration/deceleration
  else if (zR > 400) {zR = 2;}  // deceleration
  else if (zR > 200) {zR = 1;}  // acceleration
  else {zR = 0;}                // constant rate
  return zR;
}

// ----------------------------------------------------------------------

public int getRowMYD() {
  // returns the row offset based on mYD
  // add this to rowPnt to get the actual row pointer under the mouse
  int zR;
  if (Scrn != 1) {
    // SCrn0 or Scrn2
    zR = -2 + (6 * (mYD - 548)/(688 - 548));
    zR = max(zR,-2); zR = min(zR,3);
  } else {
    // Scrn1
    zR = -13 + (27 * (mYD - 42)/(686 - 42));
    zR = max(zR,-13); zR = min(zR,26);
  }
  //print("Row mYD = "); println(zR);
  return zR;
}

// ----------------------------------------------------------------------

public String getRowString(int zPnt) {
  // returns a string representing the row of data starting at zPnt
  String zR; Blank = false; Sep$ = ","; Slope$ = "C";
  zR = getStrArrayVal(zPnt) + Sep$; zPnt++;       // FR1
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // FR2
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // FR3
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // FL1
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // FL2
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // FL3
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // RR1
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // RR2
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // RR3
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // RL1
  zR = zR + getStrArrayVal(zPnt) + Sep$; zPnt++;  // RL2
  zR = zR + getStrArrayVal(zPnt); zPnt++;         // RL3;
  return zR;
}

// ----------------------------------------------------------------------

public String getStrArrayVal(int zPnt) {
  // returns the array value as a 3 character string, with leading spaces
  // if not a command line then also set value of Slope$
  // an angle value will be 0 - 180, plus an offset of 200 if Pwr is OFF
  // a step value will be 0 - 180, plus move profiles of 200,400 and 600
  String zS = "";
  if (Blank) {zS = "      "; return zS;}
  
  int zV = ArrayVal[zPnt];
  if (zV < 0) {
    zS = "Cmd"; TR = 255; TB = 0;
    if (ArrayVal[zPnt + 1] == 0) {
      Blank = true; zS = "   "; Sep$ = "  "; Slope$ = "   ";
    } return zS;
  }

  Slope$ = "C";
  if (zV >= 600) {zV -= 600; Slope$ = "S";}
  else if (zV >= 400) {zV -= 400; Slope$ = "D";}
  else if (zV >= 200) {zV -= 200; Slope$ = "A";}
  
  zS = str(zV);
  if (zV < 100) {zS = "  " + zS;}
  if (zV < 10) {zS = "  " + zS;}
  return zS;
}

// ----------------------------------------------------------------------

public void getTextFromClipboard() {
  // gets text from clipboard and analyses it as an import
  String text = (String) getFromClipboard(DataFlavor.stringFlavor);
  if (text == null) {
    msgTx = "Clipboard is empty!"; drawTxMsg();
    return;
  }
  msgTx = "Reading Clipboard data"; drawTxMsg();
  text = trim(text);
  if (text.indexOf("MemSet") >= 0) {
    // we are going to load one or more lines of data so clear the array
    // and prepare pointers
    cursor(WAIT);
    int zS = 0; int zW = rowWidth;
    for (int zC = 0; zC < arrayMax; zC++) {
      zS++; if (zS < zW) {ArrayVal[zC] = 90;}
      else {ArrayVal[zC] = Steps; zS = 0;}
    }
    // reset pointers
    String zLft = ""; int zEOL = 0; int zRow = 0; int zNum = 0;
    // now parse the data
    while (text.indexOf("MemSet(") >= 0) {
      // remove any preceeding text
      text = text.substring(text.indexOf("MemSet("),text.length());
      // find the end of the row data
      zEOL = text.indexOf(";");
      if (zEOL < 0) {break;}  // end of data has been reached
      // grab the row of data
      zLft = text.substring(7,zEOL);
      println(zLft);
      // now extract the numbers
      int zP = zRow * rowWidth; zNum++;
      while(true) {
        zEOL = zLft.indexOf(",");
        if (zEOL >= 0) {
          ArrayVal[zP] = PApplet.parseInt(trim(zLft.substring(0,zEOL))); zP++;
          //print(zLft.substring(0,zEOL)); print("\t"); println(int(trim(zLft.substring(0,zEOL))));
          zLft = zLft.substring(zEOL+1,zLft.length());
        } else { 
          zEOL = zLft.indexOf(")");
          if (zEOL >= 0) {
            ArrayVal[zP] = PApplet.parseInt(trim(zLft.substring(0,zEOL))); zP++;  // SR1
            //print(zLft.substring(0,zEOL)); print("\t"); println(int(trim(zLft.substring(0,zEOL))));
            break;
          } else {break;}
        }
      }
      
      zRow++;
  
      if ((zEOL + 1) >= text.length()) {break;}
      // remove the left portion
      text = text.substring(zEOL + 1,text.length());
    }
    rowPnt = 0; rowNum = zNum; usbEn =false;
    setAngValFromRow();  // set servo values and sliders
    cursor(ARROW);
  } else {
    msgTx = "Clipboard data is invalid"; drawTxMsg();
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void getWalkDelay() {
  // sets the Walk timer based on the current rows step value
  // the calling code must check for Command lines
  int zP = simPnt * rowWidth;
  WaklDelay = 1000/Speed;
  if (WaklDelay < 4) {WaklDelay = 4;}  // limit to 4ms minimum
  
  // check for and remove any step profile shaping
  int zStep = ArrayVal[zP + 12];
  if (zStep > 600) {zStep -= 600;}
  else if (zStep > 400) {zStep -= 400;}
  else if (zStep > 200) {zStep -= 200;}
  WalkNext = millis() + (WaklDelay * zStep);
}

// ----------------------------------------------------------------------

public void getXSliderOffset(int zS) {
  // called when mouse is clicked on a slider to determine the offset
  mXOff = JoyXval[zS] - mX;
}
// ----------------------------------------------------------------------

public void gotoFirst() {
  // goto the first row
  if (rowPnt > 1) {usbEn = false;}
  rowPnt = 0; drawFlag = 1;
  sendRowData(rowPnt,rowPnt,1);  // send row data
  setAngValFromRow();     // set servo values and sliders
}

// ----------------------------------------------------------------------

public void gotoLast() {
  // goto the last row
  if ((rowNum - rowPnt) > 2) {usbEn = false;}
  rowPnt = rowNum - 1; drawFlag = 1;
  sendRowData(rowPnt,rowPnt,1);  // send row data
  setAngValFromRow();     // set servo values and sliders
}

// ----------------------------------------------------------------------

public void grabRowInTemp(int zRP) {
  // copy the row zRP values to Temp[] store
  zRP *= rowWidth;
  for (int zP = 0; zP < (rowWidth - 1); zP++) {
    Temp[zP] = ArrayVal[zP + zRP];
  }
}


// ######################################################################


public void incDegSteps() {
  // increments the DegStep value
  if (DegStep < 30) {DegStep++; rateFlag = true; drawFlag = 1;}
}

// ----------------------------------------------------------------------

public void incField(int zFld) {
  // increments the value of field zFld
  switch(zFld) {
    case FldT1maxLft:
      if (T1maxAngLft < 160) {T1maxAngLft++; calcRideHeights();} break;
    case FldT1maxRht:
      if (T1maxAngRht < 160) {T1maxAngRht++; calcRideHeights();} break;
    case FldT2maxLft:
      if (T2maxAngLft < 140) {T2maxAngLft++; calcMidSteps();} break;
    case FldT2minLft:
      if (T2minAngLft < 130) {T2minAngLft++;
        if (T2minAngLft > (T2maxAngLft - 10)) {T2maxAngLft = T2minAngLft + 10;}
      calcMidSteps();} break;
    case FldT2maxRht:
      if (T2maxAngRht < 140) {T2maxAngRht++; calcMidSteps();} break;
    case FldT2minRht:
      if (T2minAngRht < 130) {T2minAngRht++;
        if (T2minAngRht > (T2maxAngRht - 10)) {T2maxAngRht = T2minAngRht + 10;}
      calcMidSteps();} break;
    case FldT1WlkLft:
      if (WalkLift <60) {WalkLift++;} break;
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void incRowPnt() {
  // increment the memory row point if it is less than max rows
  if (rowPnt < (rowNum - 1)) {
    rowPnt++; sendRowData(rowPnt,rowPnt,1);
    setAngValFromRow();  // set servo values and sliders
    drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

public void incSpeed() {
  // increment Speed value towards maximum
  if (Speed < SpeedUL) {
    Speed+= Inc; IncCnt++; if (keySHIFT) {Speed += 9;}
    if (IncCnt > 10) {IncCnt = 0; Inc++;}
    if (Speed > SpeedUL) {Speed = SpeedUL;}
    sendSpeed(); drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

public void incXSlider(int zS) {
  // increment slider value
  servoTgt = zS;
  if (getAngVal(AngVal[zS]) < AngVUL[zS]) {
    AngLast = getAngVal(AngVal[zS]);
    AngVal[zS] = AngLast + 1;
    AngDif = AngVal[zS] - AngLast;
    // update value in memory array
    ArrayVal[(rowPnt * rowWidth) + zS] = AngVal[zS];
    // send value to usb port
    msgTx = "ST" + str(Ch[zS]) + "." +"SA" + str(AngVal[zS]) + ".";
    if (Link[zS] > 0) {moveLinkedVal(zS);}
    else {setSliderX(); usbPortWrite(msgTx); rateFlag = true; drawFlag = 1;}
    if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
  }
}

// ----------------------------------------------------------------------

public void insertBalance() {
  // insert a Balance command either after or before the current row
  // default value is 1 until a flag is adjusted
  int zRP = rowPnt * rowWidth;
  newRow();  // create a new line, then overwrite it
  zRP = rowPnt * rowWidth;

  // now overwrite with the command data
  ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
  ArrayVal[zRP + 1] = 5;  // set the cmd = 4
  ArrayVal[zRP + 2] = 1;  // set the default 1 = ON, 0 = OFF value
  ArrayVal[zRP + 3] = 0;  // clear FL1
  ArrayVal[zRP + 4] = 0;  // clear FL2
  ArrayVal[zRP + 5] = 0;  // clear FL3
  ArrayVal[zRP + 6] = 0;  // clear RR1
  ArrayVal[zRP + 7] = 0;  // clear RR2
  ArrayVal[zRP + 8] = 0;  // clear RR3
  ArrayVal[zRP + 9] = 0;  // clear RL1
  ArrayVal[zRP + 10] = 0;  // clear RL2
  ArrayVal[zRP + 11] = 0;  // clear RL3
  ArrayVal[zRP + 12] = 0;  // clear steps
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void insertBlank() {
  // insert a 'Blank' command either after or before the current row
  int zRP = rowPnt * rowWidth;
  newRow();  // create a new line, then overwrite it
  zRP = rowPnt * rowWidth;
  
  // now overwrite with the command data
  ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
  ArrayVal[zRP + 1] = 0;  // set the Blank cmd = 0
  ArrayVal[zRP + 2] = 0;  // clear FR3
  ArrayVal[zRP + 3] = 0;  // clear FL1
  ArrayVal[zRP + 4] = 0;  // clear FL2
  ArrayVal[zRP + 5] = 0;  // clear FL3
  ArrayVal[zRP + 6] = 0;  // clear RR1
  ArrayVal[zRP + 7] = 0;  // clear RR2
  ArrayVal[zRP + 8] = 0;  // clear RR3
  ArrayVal[zRP + 9] = 0;  // clear RL1
  ArrayVal[zRP + 10] = 0;  // clear RL2
  ArrayVal[zRP + 11] = 0;  // clear RL3
  ArrayVal[zRP + 12] = 0;  // clear steps
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void InsertExCode() {
  // insert an executable code either after or before the current row
  // the default value is For... Next command, one loop
  //
  // 6  - For... command
  // 7  - Next command
  int zRP = rowPnt * rowWidth;
  newRow();  // create a new line, then overwrite it
  zRP = rowPnt * rowWidth;

  // now overwrite with the command data
  ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
  ArrayVal[zRP + 1] = 6;  // set cmd = 6, first excuteable command value, For...Next loop
  ArrayVal[zRP + 2] = 1;  // set the default 1 value
  ArrayVal[zRP + 3] = 0;  // clear FL1
  ArrayVal[zRP + 4] = 0;  // clear FL2
  ArrayVal[zRP + 5] = 0;  // clear FL3
  ArrayVal[zRP + 6] = 0;  // clear RR1
  ArrayVal[zRP + 7] = 0;  // clear RR2
  ArrayVal[zRP + 8] = 0;  // clear RR3
  ArrayVal[zRP + 9] = 0;  // clear RL1
  ArrayVal[zRP + 10] = 0;  // clear RL2
  ArrayVal[zRP + 11] = 0;  // clear RL3
  ArrayVal[zRP + 12] = 0;  // clear steps
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void InsertFlag() {
  // insert a Flag command either after or before the current row
  // default value is 1 until a flag is adjusted
  int zRP = rowPnt * rowWidth;
  newRow();  // create a new line, then overwrite it
  zRP = rowPnt * rowWidth;

  // now overwrite with the command data
  ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
  ArrayVal[zRP + 1] = 4;  // set the Flag cmd = 4
  ArrayVal[zRP + 2] = Flag;  // set the default Flag value
  ArrayVal[zRP + 3] = 0;  // clear FL1
  ArrayVal[zRP + 4] = 0;  // clear FL2
  ArrayVal[zRP + 5] = 0;  // clear FL3
  ArrayVal[zRP + 6] = 0;  // clear RR1
  ArrayVal[zRP + 7] = 0;  // clear RR2
  ArrayVal[zRP + 8] = 0;  // clear RR3
  ArrayVal[zRP + 9] = 0;  // clear RL1
  ArrayVal[zRP + 10] = 0;  // clear RL2
  ArrayVal[zRP + 11] = 0;  // clear RL3
  ArrayVal[zRP + 12] = 0;  // clear steps
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void insertGoto() {
  // insert a goto branch to a selected row, the 1st selected row in the list
  // find the selected line
  if (SelCnt < 1) {return;}   // abort if none selected. This should never happen
  usbEn = false;  // avoid any unintentional movement

  int zRP = rowPnt * rowWidth;
  if ((ArrayVal[zRP] < 0) && (ArrayVal[zRP + 1] == 1)) {
    // current line is a goto line, so simply update it
  } else {
    newRow();  // create a new line, then overwrite it
    zRP = rowPnt * rowWidth;
  }

  for (int zS = 0; zS < rowNum; zS++) {
    if (Select[zS]) {
      // 1st selected row, so set branch instruction in place of SR1
      ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
      ArrayVal[zRP + 1] = 1;  // set the goto cmd = 1
      ArrayVal[zRP + 2] = zS; // set the goto pointer
      ArrayVal[zRP + 3] = 0;  // clear FL1
      ArrayVal[zRP + 4] = 0;  // clear FL2
      ArrayVal[zRP + 5] = 0;  // clear FL3
      ArrayVal[zRP + 6] = 0;  // clear RR1
      ArrayVal[zRP + 7] = 0;  // clear RR2
      ArrayVal[zRP + 8] = 0;  // clear RR3
      ArrayVal[zRP + 9] = 0;  // clear RL1
      ArrayVal[zRP + 10] = 0;  // clear RL2
      ArrayVal[zRP + 11] = 0;  // clear RL3
      ArrayVal[zRP + 12] = 0;  // clear steps
      break;
    }
  } 
  ClearSelect();  // clear selsects as it may not be visible if at front of list
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void insertPause() {
  // insert a Pause command either after or before the current row
  // default value is 1 until a pause is adjusted

  int zRP = rowPnt * rowWidth;
  newRow();  // create a new line, then overwrite it
  zRP = rowPnt * rowWidth;

  // now overwrite with the command data
  ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
  ArrayVal[zRP + 1] = 3;  // set the pause cmd = 3
  ArrayVal[zRP + 2] = Pause; // set the current pause value
  ArrayVal[zRP + 3] = 0;  // clear FL1
  ArrayVal[zRP + 4] = 0;  // clear FL2
  ArrayVal[zRP + 5] = 0;  // clear FL3
  ArrayVal[zRP + 6] = 0;  // clear RR1
  ArrayVal[zRP + 7] = 0;  // clear RR2
  ArrayVal[zRP + 8] = 0;  // clear RR3
  ArrayVal[zRP + 9] = 0;  // clear RL1
  ArrayVal[zRP + 10] = 0;  // clear RL2
  ArrayVal[zRP + 11] = 0;  // clear RL3
  ArrayVal[zRP + 12] = 0;  // clear steps
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void insertReady() {
  // insert ready data either after or before the current row
  // if SHIFT key then only send data and execute
  if (keySHIFT) {
    msgTx  = "SP50.";
    msgTx += "ML0.130,70,95,130,70,85,130,70,95,130,70,85,640,MX0.";
    usbPortWrite(msgTx);
    msgTx = msgTx.substring(0,20) + "...";
  } else {
    int zRP = rowPnt * rowWidth;
    newRow();  // create a new line, then overwrite it
    zRP = rowPnt * rowWidth;
    // now overwrite with the command data
    ArrayVal[zRP] = 130;      // FR1
    ArrayVal[zRP + 1] = 70;  // FR2
    ArrayVal[zRP + 2] = 95;  // FR3
    
    ArrayVal[zRP + 3] = 130; // FL1
    ArrayVal[zRP + 4] = 70;  // FL2
    ArrayVal[zRP + 5] = 85;  // FL3
    
    ArrayVal[zRP + 6] = 130; // RR1
    ArrayVal[zRP + 7] = 70;  // RR2
    ArrayVal[zRP + 8] = 95;  // RR3
    
    ArrayVal[zRP + 9] = 130; // RL1
    ArrayVal[zRP + 10] = 70; // RL2
    ArrayVal[zRP + 11] = 85; // RL3
    
    ArrayVal[zRP + 12] = Steps + 600;
    sendRowData(rowPnt,rowPnt,1);   // send data
    setAngValFromRow();      // set servo values and sliders
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void insertRest() {
  // insert ready data either after or before the current row
  // if SHIFT key then only send data and execute
  if (keySHIFT) {
    // move to rest sitting, then turn power off
    msgTx  = "SP50.";
    msgTx += "ML0.85,60,95,85,60,85,11,0,95,11,0,85,640,";
    msgTx += "ML1.285,260,295,285,260,285,211,200,295,211,200,285,1,MO.";
    usbPortWrite(msgTx);
    msgTx = msgTx.substring(0,20) + "...";
  } else {
    int zRP = rowPnt * rowWidth;
    newRow();  // create a new line, then overwrite it
    zRP = rowPnt * rowWidth;
    // now overwrite with the command data
    ArrayVal[zRP] = 85;      // FR1
    ArrayVal[zRP + 1] = 60;  // FR2
    ArrayVal[zRP + 2] = 95;  // FR3
    
    ArrayVal[zRP + 3] = 85;  // FL1
    ArrayVal[zRP + 4] = 60;  // FL2
    ArrayVal[zRP + 5] = 85;  // FL3
    
    ArrayVal[zRP + 6] = 11;  // RR1
    ArrayVal[zRP + 7] = 0; // RR2
    ArrayVal[zRP + 8] = 95; // RR3
    
    ArrayVal[zRP + 9] = 11;  // RL1
    ArrayVal[zRP + 10] = 0; // RL2
    ArrayVal[zRP + 11] = 85; // RL3
    
    ArrayVal[zRP + 12] = Steps;
    sendRowData(rowPnt,rowPnt,1);   // send data
    setAngValFromRow();      // set servo values and sliders
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void insertSpeed() {
  // insert a speed command either after or before the current row
  // if current line is speed command then update

  int zRP = rowPnt * rowWidth;
  if ((ArrayVal[zRP] < 0) && (ArrayVal[zRP + 1] == 2)) {
    // current line is a speed line, so simply update it
  } else {
    newRow();  // create a new line, then overwrite it
    zRP = rowPnt * rowWidth;
  }
  // now overwrite with the command data
  ArrayVal[zRP] = -1;  // set the first value to -1 to indicate a command
  ArrayVal[zRP + 1] = 2;  // set the speed cmd = 2
  ArrayVal[zRP + 2] = Speed; // set the current speed value
  ArrayVal[zRP + 3] = 0;  // clear FL1
  ArrayVal[zRP + 4] = 0;  // clear FL2
  ArrayVal[zRP + 5] = 0;  // clear FL3
  ArrayVal[zRP + 6] = 0;  // clear RR1
  ArrayVal[zRP + 7] = 0;  // clear RR2
  ArrayVal[zRP + 8] = 0;  // clear RR3
  ArrayVal[zRP + 9] = 0;  // clear RL1
  ArrayVal[zRP + 10] = 0;  // clear RL2
  ArrayVal[zRP + 11] = 0;  // clear RL3
  ArrayVal[zRP + 12] = 0;  // clear steps
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void insertStand() {
  // insert squat data either after or before the current row
  // if SHIFT key then only send data and execute
  if (keySHIFT) {
    // move to upright stand position
    msgTx  = "SP50.";
    msgTx += "ML0.30,0,95,30,0,85,30,0,95,30,0,85,220,";
    msgTx += "ML1.160,90,95,160,90,85,160,90,95,160,90,85,460,";
    msgTx += "ML2.360,290,95,360,290,85,360,290,95,360,290,85, 5,";
    msgTx += "MO.";
    usbPortWrite(msgTx);
    msgTx = msgTx.substring(0,20) + "...";
  } else {
    int zRP = rowPnt * rowWidth;
    newRow();  // create a new line, then overwrite it
    zRP = rowPnt * rowWidth;
    // now overwrite with the command data
    ArrayVal[zRP] = 160;      // FR1
    ArrayVal[zRP + 1] = 90;  // FR2
    ArrayVal[zRP + 2] = 95; // FR3
    
    ArrayVal[zRP + 3] = 160;  // FL1
    ArrayVal[zRP + 4] = 90;  // FL2
    ArrayVal[zRP + 5] = 85; // FL3
    
    ArrayVal[zRP + 6] = 160;  // RR1
    ArrayVal[zRP + 7] = 90;  // RR2
    ArrayVal[zRP + 8] = 95; // RR3
    
    ArrayVal[zRP + 9] = 160;  // RL1
    ArrayVal[zRP + 10] = 90; // RL2
    ArrayVal[zRP + 11] = 85; // RL3
    
    ArrayVal[zRP + 12] = Steps + 400;
    sendRowData(rowPnt,rowPnt,1);   // send data
    setAngValFromRow();      // set servo values and sliders
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void insertSquat() {
  // insert squat data either after or before the current row
  // if SHIFT key then only send data and execute
  if (keySHIFT) {
    // move to squat and turn power OFF
    msgTx  = "SP50.";
    msgTx += "ML0.0,0,95,0,0,85,0,0,95,0,0,85,705,";
    msgTx += "ML1.0,0,95,0,0,85,0,0,95,0,0,85,  5,";
    msgTx += "ML2.200,200,295,200,200,285,200,200,295,200,200,285,5,MO.";
    usbPortWrite(msgTx);
    msgTx = msgTx.substring(0,20) + "...";
  } else {
    int zRP = rowPnt * rowWidth;
    newRow();  // create a new line, then overwrite it
    zRP = rowPnt * rowWidth;
    // now overwrite with the command data
    ArrayVal[zRP] = 0;      // FR1
    ArrayVal[zRP + 1] = 0;  // FR2
    ArrayVal[zRP + 2] = 95; // FR3
    
    ArrayVal[zRP + 3] = 0;  // FL1
    ArrayVal[zRP + 4] = 0;  // FL2
    ArrayVal[zRP + 5] = 85; // FL3
    
    ArrayVal[zRP + 6] = 0;  // RR1
    ArrayVal[zRP + 7] = 0;  // RR2
    ArrayVal[zRP + 8] = 95; // RR3
    
    ArrayVal[zRP + 9] = 0;  // RL1
    ArrayVal[zRP + 10] = 0; // RL2
    ArrayVal[zRP + 11] = 85; // RL3
    
    ArrayVal[zRP + 12] = Steps + 600;
    sendRowData(rowPnt,rowPnt,1);   // send data
    setAngValFromRow();      // set servo values and sliders
  } rateFlag = true; drawFlag = 1;
}


// ######################################################################


public void keyToggleRate() {
  // called by 'R' key pressed. Toggle angle rate flag if we are pointing
  // at a valid angle in the list.
  if (AngField < 0) {return;}  // not a valid field

  rateFlag = true; drawFlag = 1;
  int zRow = getRowMYD() + rowPnt;
  int zPnt = (zRow * rowWidth) + AngField;
  if (zRow == 0) {msgTx = "Error - first line"; return;}
  if ((zRow >= (rowNum - 1)) && (ArrayVal[zPnt] < 400)) {msgTx = "Error - last line"; return;}
  if (rowNum < 3) {msgTx = "Error - too few lines"; return;}
  if (ArrayVal[zRow * rowWidth] < 0) {msgTx = "Error - Command line"; return;}
  
  if (ArrayVal[zPnt] >= 400) {
    ArrayVal[zPnt] -= 400;   // remove the auto-rate +400 offset
    Rated = getRated();  // check to see if any more rate angles exist
  } else {
    ArrayVal[zPnt] += 400;  // add the auto-rate +400 offset
    Rated = true;
  }

  if (Rated) {calcRates();}  // update rate values
  
  if (zRow == rowPnt) {
    // changed a value in the current row
    AngVal[AngField] = ArrayVal[zPnt];
    // send updated value to usb port
    msgTx = "ST" + str(Ch[AngField]) + "." +"SA" + str(AngVal[AngField]) + ".";
    setAngValFromRow();
    usbPortWrite(msgTx);
    if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
  }
}


// ######################################################################


public void makeIntermediate(int zR) {
  // sets the value of row zR to intermediate angles with previous and next row
  if (zR == 0) {return;}             // invalid intermediate row
  if (zR == (rowNum - 1)) {return;}  // invalid intermediate row
  if (ArrayVal[zR * rowWidth] < 0) {return;}  // command line
  
  // the STEP value will not be modified
  zR = zR * rowWidth;
  for (int zP = 0; zP < (rowWidth - 1); zP++) {
    ArrayVal[zR + zP] = ArrayVal[zR + zP - rowWidth] + ((ArrayVal[zR + zP + rowWidth] - ArrayVal[zR + zP - rowWidth]) / 2);
  }
  msgTx = "Intermediate row created";
}

// ----------------------------------------------------------------------

public int mapInt(int x,int in_min,int in_max,int out_min,int out_max) {
  // integer version of Processing floating point map() function
  if ((in_max - in_min) == 0) {in_max += 1;}  // trap divide by zero error
  return ((x - in_min) * (out_max - out_min) / (in_max - in_min)) + out_min;
}

// ----------------------------------------------------------------------

public void moveLinkedVal(int zS) {
  // move the linked slider by the same amount if possible
  if (AngDif == 0) {return;}  // no change so abort function
  
  int zT0 = 0;
  if (Link[zS] < 2) {
    // single linked pair so adjust associated target
    switch(zS) {
      case 0: zT0 = FL1; break;  // FR1 linked to FL1
      case 1: zT0 = FL2; break;  // FR2 linked to FL2
      case 2: zT0 = FL3; break;  // FR3 linked to FL3
      
      case 3: zT0 = FR1; break;  // FL1 linked to FR1
      case 4: zT0 = FR2; break;  // FL2 linked to FR2
      case 5: zT0 = FR3; break;  // FL3 linked to FR3
  
      case 6: zT0 = RL1; break;  // RR1 linked to RL1
      case 7: zT0 = RL2; break;  // RR2 linked to RL2
      case 8: zT0 = RL3; break;  // RR3 linked to RL3
  
      case 9: zT0 = RR1; break;  // RL1 linked to RR1
      case 10: zT0 = RR2; break;  // RL2 linked to RR2
      case 11: zT0 = RR3; break;  // RL3 linked to RR3
    }
    adjustLinkVal(zT0);
  } else {
    // multi-linked values, so adjust three other values
    int zT1 = 0; int zT2 = 0;
    switch(zS) {
      case 0: zT0 = FL1; zT1 = RR1; zT2 = RL1; break;  // FR1 linked to FL1, RR1, RL1
      case 1: zT0 = FL2; zT1 = RR2; zT2 = RL2; break;  // FR2 linked to FL2, RR2, RL2
      case 2: zT0 = FL3; zT1 = RR3; zT2 = RL3; break;  // FR3 linked to FL3, RR3, RL3
      
      case 3: zT0 = FR1; zT1 = RR1; zT2 = RL1; break;  // FL1 linked to FR1, RR1, RL1
      case 4: zT0 = FR2; zT1 = RR2; zT2 = RL2; break;  // FL2 linked to FR2, RR2, RL2
      case 5: zT0 = FR3; zT1 = RR3; zT2 = RL3; break;  // FL3 linked to FR3, RR3, RL3
  
      case 6: zT0 = RL1; zT1 = FR1; zT2 = FL1; break;  // RR1 linked to RL1, FR1, FL1
      case 7: zT0 = RL2; zT1 = FR2; zT2 = FL2; break;  // RR2 linked to RL2, FR2, FL2
      case 8: zT0 = RL3; zT1 = FR3; zT2 = FL3; break;  // RR3 linked to RL3, FR3, FL3
  
      case 9: zT0 = RR1; zT1 = FR1; zT2 = FL1; break;  // RL1 linked to RR1, FR1, FL1
      case 10: zT0 = RR2; zT1 = FR2; zT2 = FL2; break;  // RL2 linked to RR2, FR2, FL2
      case 11: zT0 = RR3; zT1 = FR3; zT2 = FL3; break;  // RL3 linked to RR3, FR3, FL3
    }
    adjustLinkVal(zT0); adjustLinkVal(zT1); adjustLinkVal(zT2);
  }
  usbPortWrite(msgTx);
}

// ----------------------------------------------------------------------

public void moveRowBlockFT(int zF, int zT, int zN) {
  // moves row block from zF up to zT, block row size zN, then clears remainder
  copyRowBlockFT(zF,zT,zN);
  if (zT < zF) {
    // row block was copied up so clear region below copied area
    ClearRowBlockFT(zT + zN, zF + zN -1);
  } else {
    // row block was copied down so clear region above it
    ClearRowBlockFT(zF, zT -1);
  }
}

// ----------------------------------------------------------------------

public void moveSelArrayUp() {
  // called when a selected row has been deleted
  int zT = 0;
  for (int zS = 1; zS < arrayMax; zS++) {
    Select[zT] = Select[zS]; zT++;
  } Select[arrayMax - 1] = false;
}

// ----------------------------------------------------------------------

public void moveSelDwn() {
  // move selected rows towards rowNum
  if ((SelCnt < 1) || (rowNum < 2)) {return;}
  usbEn = false;  // avoid any unintentional movement
  
  cursor(WAIT);
  boolean zS = false;
  for (int zR = rowNum - 2; zR >= 0; zR--) {
    if (Select[zR] && !Select[zR + 1]) {
      // current line selected and next is not, so swap them
      int zF = zR * rowWidth; int zT = (zR + 1) * rowWidth; int zV;
      // swap the row angle and steps
      for (int zP = 0; zP < rowWidth; zP++) {
        zV = ArrayVal[zT + zP]; ArrayVal[zT + zP] = ArrayVal[zF + zP];
        ArrayVal[zF + zP] = zV; 
      } Select[zR + 1] = true; Select[zR] = false; zS = true;
    }
  } 
  if (zS && (rowPnt < (rowNum-1))) {rowPnt++;}
  rateFlag = true; drawFlag = 1;
  cursor(ARROW);
}

// ----------------------------------------------------------------------

public void moveSelUp() {
  // move selected rows towards row zero
  if ((SelCnt < 1) || (rowNum < 2)) {return;}
  usbEn = false;  // avoid any unintentional movement
  
  cursor(WAIT);
  boolean zS = false;
  for (int zR = 1; zR < rowNum; zR++) {
    if (Select[zR] && !Select[zR - 1]) {
      // current line selected and previous not, so swap them
      int zF = zR * rowWidth; int zT = (zR - 1) * rowWidth; int zV;
      // swap the row angle and steps
      for (int zP = 0; zP < rowWidth; zP++) {
        zV = ArrayVal[zT + zP]; ArrayVal[zT + zP] = ArrayVal[zF + zP];
        ArrayVal[zF + zP] = zV; 
      } Select[zR-1] = true; Select[zR] = false; zS = true;
    }
  } 
  if (zS && (rowPnt > 0)) {rowPnt--;}
  rateFlag = true; drawFlag = 1;
  cursor(ARROW);
}

// ----------------------------------------------------------------------

public void MultiDecX(int zS1,int zS2) {
  // multimove X decrement on servos zS1 and zS2
}

// ----------------------------------------------------------------------

public void MultiDecY(int zS1,int zS2) {
  // multimove Y decrement on servos zS1 and zS2
  int zA = getAngVal(AngVal[zS1]);
  boolean zDF = false;  // drawFlag
  if (zA > AngVLL[zS1]) {
    // work on tier 1 angles first
    AngLast = zA; AngVal[zS1] = zA - 1; AngDif = -1;
    // update value in memory array
    ArrayVal[(rowPnt * rowWidth) + zS1] = AngVal[zS1];
    // send value to usb port
    msgTx = "ST" + str(Ch[zS1]) + "." +"SA" + str(AngVal[zS1]) + ".";
    if (Link[zS1] > 0) {moveLinkedVal(zS1);}
    else {setSliderX(); usbPortWrite(msgTx); rateFlag = true; zDF = false;}
    if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
  }

  // now do tier 2 angles if tier 1 is divisible by 2, ie even
  if (((zA % 2) != 0) && (zA != 0)) {
    // tier 1 angle is divisible by 2 so adjust tier 2 angle
    // using this method tier 2 wiull be adjusted twice as fast as tier 2
    zA = getAngVal(AngVal[zS2]);
    if (zA > AngVLL[zS2]) {
      // work on tier 2 angles
      AngLast = zA; AngVal[zS2] = zA - 1; AngDif = -1;
      // update value in memory array
      ArrayVal[(rowPnt * rowWidth) + zS2] = AngVal[zS2];
      // send value to usb port
      msgTx = "ST" + str(Ch[zS2]) + "." +"SA" + str(AngVal[zS2]) + ".";
      if (Link[zS2] > 0) {moveLinkedVal(zS2);}
      else {setSliderX(); usbPortWrite(msgTx); rateFlag = true; zDF = false;}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
    }
  }
  if (zDF) {drawFlag = 1;}
}

// ----------------------------------------------------------------------

public void MultiIncX(int zS1,int zS2) {
  // multimove X increment on servos zS1 and zS2
}

// ----------------------------------------------------------------------

public void MultiIncY(int zS1,int zS2) {
  // multimove Y increment on servos zS1 and zS2
  int zA = getAngVal(AngVal[zS1]);
  boolean zDF = false;  // drawFlag
  if (zA < AngVUL[zS1]) {
    // work on tier 1 angles first
    AngLast = zA; AngVal[zS1] = zA + 1; AngDif = 1;
    // update value in memory array
    ArrayVal[(rowPnt * rowWidth) + zS1] = AngVal[zS1];
    // send value to usb port
    msgTx = "ST" + str(Ch[zS1]) + "." +"SA" + str(AngVal[zS1]) + ".";
    if (Link[zS1] > 0) {moveLinkedVal(zS1);}
    else {setSliderX(); usbPortWrite(msgTx);}
    if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
    rateFlag = true; zDF = false;
  }

  // now do tier 2 angles if tier 1 is divisible by 2, ie even
  // note that tier 2 has a max value of 90 to straighten the leg
  if (((zA % 2) != 0) && (zA != 0)) {
    // tier 1 angle is divisible by 2 so adjust tier 2 angle
    // using this method tier 2 wiull be adjusted twice as fast as tier 2
    zA = getAngVal(AngVal[zS2]);
    if (zA != 90) {
      // work on tier 2 angles
      if (zA > 90) {AngDif = -1;} else {AngDif = 1;}
      AngLast = AngVal[zS2]; AngVal[zS2] = zA + AngDif;
      // update value in memory array
      ArrayVal[(rowPnt * rowWidth) + zS2] = AngVal[zS2];
      // send value to usb port
      msgTx = "ST" + str(Ch[zS2]) + "." +"SA" + str(AngVal[zS2]) + ".";
      if (Link[zS2] > 0) {moveLinkedVal(zS2);}
      else {setSliderX(); usbPortWrite(msgTx);}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
      rateFlag = true; zDF = false;
    }
  }
  if (zDF) {drawFlag = 1;}
}

// ----------------------------------------------------------------------

public void MultiMoveXY(int zS1,int zS2) {
  // multimove in X/Y on servos zS1 and zS2
  // the movement is relative to the click mXD/mYD points
  // Temp[] holds the active row when the mouse was clicked
  
  int zA = getAngVal(Temp[zS1]);
  int zB = getAngVal(Temp[zS2]);
  int zD = 0;           // AngDif
  boolean zDF = false;  // drawFlag
  int zP = 0;  // new angle
  if (MultiPnt == 0) {AngLast = zA; AngLast0 = zB;}
  MultiPnt = 1;  Multi0 = zS1; Multi1 = zS2;  // prevent mouseDown() interaction
  // consider Y axis first
  if (mY > mYD) {
    // mouse is dragged down so increase angles
    zD = (mY - mYD)/3;  // mouse Y determines change
    zP = zA + zD;
    if ((zP >= AngVLL[zS1]) && (zP <= AngVUL[zS1])) {
      // work on tier 1 angles first
      AngVal[zS1] = zP;
      AngDif = zP - AngLast; AngLast = zP;
      // update value in memory array
      ArrayVal[(rowPnt * rowWidth) + zS1] = AngVal[zS1];
      // send value to usb port
      msgTx = "ST" + str(Ch[zS1]) + "." +"SA" + str(AngVal[zS1]) + ".";
      if (Link[zS1] > 0) {moveLinkedVal(zS1);}
      else {setSliderX(); usbPortWrite(msgTx);}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
      rateFlag = true; zDF = true;
    }
    // now do tier 2 angles at half the displacement
    // also add in the x-axis component
    zD /= 2; zD -= (mX - mXD)/5;
    zP = zB + zD;
    if ((zP >= AngVLL[zS2]) && (zP <= AngVUL[zS2])) {
      // work on tier 2 angles
      AngVal[zS2] = zP;
      AngDif = zP - AngLast0; AngLast0 = zP;
      // update value in memory array
      ArrayVal[(rowPnt * rowWidth) + zS2] = AngVal[zS2];
      // send value to usb port
      msgTx = "ST" + str(Ch[zS2]) + "." +"SA" + str(AngVal[zS2]) + ".";
      if (Link[zS2] > 0) {moveLinkedVal(zS2);}
      else {setSliderX(); usbPortWrite(msgTx);}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
      rateFlag = true; zDF = true;
    }
  } else {
    // mouse is dragged up so decrease angles
    zD = (mY - mYD)/3;  // mouse Y determines change
    zP = zA + zD;
    if ((zP >= AngVLL[zS1]) && (zP <= AngVUL[zS1])) {
      // work on tier 1 angles first
      AngVal[zS1] = zP;
      AngDif = zP - AngLast; AngLast = zP;
      // update value in memory array
      ArrayVal[(rowPnt * rowWidth) + zS1] = AngVal[zS1];
      // send value to usb port
      msgTx = "ST" + str(Ch[zS1]) + "." +"SA" + str(AngVal[zS1]) + ".";
      if (Link[zS1] > 0) {moveLinkedVal(zS1);}
      else {setSliderX(); usbPortWrite(msgTx);}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
      rateFlag = true; zDF = true;
    }
    // now do tier 2 angles at half the displacement
    zD /= 2; zD -= (mX - mXD)/5;
    zP = zB + zD;
    if ((zP >= AngVLL[zS2]) && (zP <= AngVUL[zS2])) {
      // work on tier 2 angles
      AngVal[zS2] = zP;
      AngDif = zP - AngLast0; AngLast0 = zP;
      // update value in memory array
      ArrayVal[(rowPnt * rowWidth) + zS2] = AngVal[zS2];
      // send value to usb port
      msgTx = "ST" + str(Ch[zS2]) + "." +"SA" + str(AngVal[zS2]) + ".";
      if (Link[zS2] > 0) {moveLinkedVal(zS2);}
      else {setSliderX(); usbPortWrite(msgTx);}
      if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
      rateFlag = true; zDF = true;
    }
  }
  if (zDF) {drawFlag = 1;}
}


// ######################################################################


public void newRow() {
  // include a new row either before or after the current line if memory permits
  // correct goto commands if they are any that are affected
  if (rowNum == rowDepth) {return;}  // array memory is exhausted
  
  if (mB == LEFT) {
    // add in a new row after the current one
    if (rowPnt != (rowNum-1)) {
      // current row is not the last one in the list so copy values down to create a new line
      for (int zR = rowNum; zR > rowPnt; zR--) {
        int zT = zR * rowWidth; int zF = (zR-1) * rowWidth;
        for (int zP = 0; zP < rowWidth; zP++) {
          ArrayVal[zT + zP] = ArrayVal[zF + zP];
        }
      }
    }
    // now create the new row by copying the current one
    // if that was a command line, create a new row
    if (ArrayVal[rowPnt * rowWidth] < 0) {
      // current row is a command line, so create a default row
      ClearRow(rowPnt + 1);
    } else {
      // duplicate line values
      int zT = (rowPnt + 1) * rowWidth; int zF = rowPnt * rowWidth;
      for (int zP = 0; zP < rowWidth; zP++) {
        ArrayVal[zT + zP] = ArrayVal[zF + zP];
      }
    }
    rowNum++;  // increase the number of rows in the list
    rowPnt++; 
    checkGotoCmdIns(rowPnt,1);  // check to correct Goto commands
  } else {
    // add in a new row before the current one by copying the current rows down one row
    // this process will naturally leave a copy from the original
    for (int zR = rowNum; zR > rowPnt; zR--) {
      int zT = zR * rowWidth; int zF = (zR-1) * rowWidth;
      for (int zP = 0; zP < rowWidth; zP++) {
        ArrayVal[zT + zP] = ArrayVal[zF + zP];
      }
    }
    if (ArrayVal[rowPnt * rowWidth] < 0) {
      // current row is a command line, so create a default row
      ClearRow(rowPnt);
    }
    rowNum++;  // increase the number of rows in the list
    checkGotoCmdIns(rowPnt,1);  // check to correct Goto commands
  }
  // if the SHIFT key is down then determine intermediate points
  if (keySHIFT) {makeIntermediate(rowPnt);}
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public String Num3Dig(int zV) {
  // returns a string number 3 characters wide
  String zNum = "";
  if (zV < 0) {zNum = " ";}  // command line
  else if (zV < 10) {zNum = "  ";}
  else if (zV < 100) {zNum = " ";}
  zNum = zNum + str(zV);
  return zNum;
}


// ######################################################################


public void pasteRows() {
  // paste the contents of the copy buffer
  usbEn = false;  // avoid any unintentional movement
  msgTx = "No copied data"; rateFlag = true; rateFlag = true; drawFlag = 1;
  if (CopyCnt < 1) {return;}
  
  cursor(WAIT);
  int zF = -1; int zR = 0; int zS = 0; int zT = -1;
  if (SelCnt > 0) {
    // one or more rows selected so create correct gap and overwrite them
    // if there is several selected regions, only the first is overwritten
    if (SelCnt == rowNum) {
      // all of the rows are selected so reset rows and copy in data
      ClearRowBlockFT(0, rowNum - 1); copyCopyTo(0);
      rowPnt = 0; rowNum = CopyCnt;
    } else {
      // not all rows are selected, so find start of selection and count them
      for (zR = 0; zR < rowNum; zR++) {
        if (Select[zR]) {
          // a row is selected, is it the 1st? 
          zS++; if (zF < 0) {zF = zR;}  // zF points to the 1st selected row
        } else {
          // row is not selected, were previous selected rows detected?
          // if so, zT points at 1st non-selected row after selection
          if (zF >= 0) {if (zT < 0) {zT = zR; break;}}  // yes, so at end of selection
        }
      }
      // zF points to the 1st selected row
      // zT points to the 1st non-selected row after the selection
      // zS is the number of selected lines to be replaced
      if (zT >= 0) {
        // if zT != -1 then some data after selection
        // now create suitable gap size
        if (zS > CopyCnt) {
          // more selected rows than copy can fill, so close the gap
          moveRowBlockFT(zT, zF + CopyCnt, rowNum - zT);
        } else if (zS < CopyCnt) {
          // fewer selected rows than copy, so increase the gap
          moveRowBlockFT(zT, zT + CopyCnt - zS, rowNum - zT);
        }
      }
      // gap now matches copy count, so copy into it, starting at row zF
      zT = CopyCnt; if ((zT + zF) >= rowDepth) {zT = rowDepth - zF - 1;}
      copyCopyTo(zF);  // copy in starting at row zF
      if (zS > CopyCnt) {checkGotoCmdDel(zT,zS-CopyCnt);}
      if (zS < CopyCnt) {checkGotoCmdIns(zT,CopyCnt-zS);}
      rowNum = rowNum - SelCnt + CopyCnt;
    }
    ClearSelect();
  } else {
    // no selected rows, so create a gap after current row and copy in
    // ensure there are enough rows for the copy data
    int zCnt = CopyCnt;
    if (rowPnt == (rowNum - 1)) {
      // last line so simply copy onto end
      copyCopyTo(rowPnt + 1); rowNum = rowNum + CopyCnt;
      if (rowNum > rowDepth) {rowNum = rowDepth;}
      println("Add to end");
    } else if ((rowDepth - rowNum) <= CopyCnt) {
      // copy memory is larger than rows left so limit copy and simply
      // overwrite data area
      zCnt = rowDepth - rowNum;
      copyCopyTo(rowPnt + 1); rowNum = rowDepth;
      println("Too big!");
    } else {
      // copied data will fit so move data down to create a gap
      moveRowBlockFT(rowPnt + 1, rowPnt + CopyCnt +1, rowNum - rowPnt);
      copyCopyTo(rowPnt + 1); rowNum = rowNum + CopyCnt;
      if (rowNum > rowDepth) {rowNum = rowDepth;}
      checkGotoCmdIns(rowPnt,CopyCnt);
      println("It fits!");
    }
  }
  msgTx = "Copied data was pasted";
  if (rowNum == rowDepth) {msgTx = "Memory is consumed!";}
  cursor(ARROW);
}

// ----------------------------------------------------------------------

public void playConst() {
  // send contents of memory and then execute it
  int zF = 0;           // first row of data
  int zL = rowNum - 1;  // last row of data
  drawFlag = 1;
  if (usbEn) {
    msgTx = "MC."; usbPortWrite(msgTx);
    sendSpeed();
    // check to see if rows are selected
    if (SelCnt > 0) {
      // rows are selected so find first selected block
      // if multiple blocks or rows are selected we only play the first
      zF = -1;
      for (int zR = 0; zR < rowNum; zR++) {
        if (Select[zR]) {
          if (zF < 0) {zF = zR;}  // first selected row
          zL = zR;                // track last selected row
        } else {if (zF >= 0) {break;}}  // exit at end of selected block
      }
    }
    // now export the data to RoboDog
    for (int zR = zF; zR <= zL; zR++) {
      sendRowData(zR,zR,0);
    }
    msgTx = "MF" + str(zF) + "."; usbPortWrite(msgTx);
    msgTx = "MR."; usbPortWrite(msgTx); playMode = 2;
    msgTx = "Play repeatedly";
  } else if (Scrn == 2) {
    // simulate the walking process by moving down through the list
    Walk_Init();
    WalkSim = 2;  // set Walk to continuous mode
    playMode = 2;
  }
}

// ----------------------------------------------------------------------

public void playOnce() {
  // if usbEn then send contents of memory and execute it
  int zF = 0;           // first row of data
  int zL = rowNum - 1;  // last row of data
  drawFlag = 1;
  if (usbEn) {
    msgTx = "MC."; usbPortWrite(msgTx);
    sendSpeed();
    // check to see if rows are selected
    if (SelCnt > 0) {
      // rows are selected so find first selected block
      // if multiple blocks or rows are selected we only play the first
      zF = -1;
      for (int zR = 0; zR < rowNum; zR++) {
        if (Select[zR]) {
          if (zF < 0) {zF = zR;}  // first selected row
          zL = zR;                // track last selected row
        } else {if (zF >= 0) {break;}}  // exit at end of selected block
      }
    }
    // now export the data to RoboDog
    for (int zR = zF; zR <= zL; zR++) {
      sendRowData(zR,zR,0);
    }
    msgTx = "MF" + str(zF) + "."; usbPortWrite(msgTx);
    msgTx = "MO."; usbPortWrite(msgTx); playMode = 1;
    msgTx = "Play once";
  } else if (Scrn == 2) {
    // simulate the walking process by moving down through the list
    Walk_Init();
    WalkSim = 1;  // set Walk to once mode
    playMode = 1;
  }
}

// ----------------------------------------------------------------------

public void playStop() {
  // send a command to stop any movement
  drawFlag = 1;
  if (usbEn) {
    msgTx = "MS.";
    usbPortWrite(msgTx);
    rowPnt = 0;  // return to the start row in the list
    sendRowData(rowPnt,rowPnt,1);  // move to first row angles
  } else if (Scrn == 2) {
    // stop Walk simulator
    WalkSim = 0; msgTx = "Sim Stopped";
  } playMode = 0;
}

// ----------------------------------------------------------------------

public void primeTemp() {
  // prime the Temp[] array with the default angles
  Temp[ 0] =  0;  // FR1
  Temp[ 1] =  0;  // FR2
  Temp[ 2] = 95;  // FR3
  Temp[ 3] =  0;  // FL1
  Temp[ 4] =  0;  // FL2
  Temp[ 5] = 85;  // FL3
  Temp[ 6] =  0;  // RR1
  Temp[ 7] =  0;  // RR2
  Temp[ 8] = 95;  // RR3
  Temp[ 9] =  0;  // RL1
  Temp[10] =  0;  // RL2
  Temp[11] = 85;  // RL3
  Temp[12] = 30;  // Stp
}


// ######################################################################


public void readRowSlider() {
  // determine row slider position and display accordingly
}

// ----------------------------------------------------------------------

public void readXSlider(int zS) {
  // determine X-slider position and send value
  servoTgt = zS;        // record which slider was clicked
  AngLast = AngVal[zS]; // remember the current value
  if (sliderDwn < 1) {getXSliderOffset(zS);}  // just been clicked
  JoyXval[zS] = mX + mXOff; JoyXval[zS] = max(JoyXval[zS], JoyXLL[zS]);
  JoyXval[zS] = min(JoyXval[zS], JoyXUL[zS]);

  // remove servo rate & power offset settings
  AngVal[zS] = getAngVal(AngVal[zS]);
  switch(zS) {
    case 0:  // FR1 
      AngVal[zS] = AngVLL[zS] + (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 1: // FR2
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 2: // FR3
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 3: //FL1
      AngVal[zS] = AngVLL[zS] + (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 4: // FL2
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 5: // FL3
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 6: //RR1
      AngVal[zS] = AngVLL[zS] + (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 7: // RR2
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 8: // RR3
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 9: // RL1
      AngVal[zS] = AngVLL[zS] + (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 10: // RL2
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
    case 11: // RL3
      AngVal[zS] = AngVUL[zS] - (((JoyXval[zS] - JoyXLL[zS]) * (AngVUL[zS] - AngVLL[zS])) / (JoyXUL[zS] - JoyXLL[zS])); break;
  } 
  
  if (AngLast != AngVal[zS]) {
    AngDif = AngVal[zS] - AngLast;  // note difference in case of linked sliders
    AngLast = AngVal[zS];
    // update value in memory array
    ArrayVal[(rowPnt * rowWidth) + zS] = AngVal[zS];
    // send value to usb port
    msgTx = "ST" + str(Ch[zS]) + "." +"SA" + str(AngVal[zS]) + ".";
    usbPortWrite(msgTx); rateFlag = true; drawFlag = 1;
    if (Link[zS] > 0) {moveLinkedVal(zS);}
  }
  sliderDwn = 1; // set slider down flag to prevent inc/dec from false triggering
}

// ----------------------------------------------------------------------

public void recallM0() {
  // recall data stored in M0 if any
  msgTx = "Data store M0 is empty"; rateFlag = true; drawFlag = 1;
  if (M0 < 1) {return;}  // no data stored so exit
  
  rowNum = M0; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    ArrayVal[zP] = Mem0[zP];
  } 
  rowPnt = 0; usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  ClearSelect();
  msgTx = "Data loaded from M0";
}

// ----------------------------------------------------------------------

public void recallM1() {
  // recall data stored in M1 if any
  msgTx = "Data store M1 is empty"; rateFlag = true; drawFlag = 1;
  if (M1 < 1) {return;}  // no data stored so exit

  rowNum = M1; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    ArrayVal[zP] = Mem1[zP];
  } 
  rowPnt = 0; usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  ClearSelect(); drawFlag = 1;
  msgTx = "Data loaded from M1";
}

// ----------------------------------------------------------------------

public void recallM2() {
  // recall data stored in M2 if any
  msgTx = "Data store M2 is empty"; rateFlag = true; drawFlag = 1;
  if (M2 < 1) {return;}  // no data stored so exit

  rowNum = M2; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    ArrayVal[zP] = Mem2[zP];
  } 
  rowPnt = 0; usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  ClearSelect(); drawFlag = 1;
  msgTx = "Data loaded from M2";
}

// ----------------------------------------------------------------------

public void recallM3() {
  // recall data stored in M0 if any
  msgTx = "Data store M3 is empty"; rateFlag = true; drawFlag = 1;
  if (M3 < 1) {return;}  // no data stored so exit

  rowNum = M3; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    ArrayVal[zP] = Mem3[zP];
  } 
  rowPnt = 0; usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  ClearSelect(); drawFlag = 1;
  msgTx = "Data loaded from M3";
}

// ----------------------------------------------------------------------

public void resetDegSteps() {
  // degrements the DegStep value
  DegStep = 3; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void resetLftFlds() {
  // set the values in the left fields to their defaults
  T1maxAngLft = 140;  // Walk Trainer T1 maxAng left side
  T2maxAngLft = 90;   // Walk Trainer T2 maxAng left side
  T2midAngLft = 70;   // Walk Trainer T2 midAng left side
  T2minAngLft = 50;   // Walk Trainer T2 minAng left side
}

// ----------------------------------------------------------------------

public void resetRhtFlds() {
  // set the values in the rightt fields to their defaults
  T1maxAngRht = 140;  // Walk Trainer T1 maxAng right side
  T2maxAngRht = 90;   // Walk Trainer T2 maxAng right side
  T2midAngRht = 70;   // Walk Trainer T2 midAng right side
  T2minAngRht = 50;   // Walk Trainer T2 minAng right side
}

// ----------------------------------------------------------------------

public void resetSpeed() {
  // set the Speed to the default value
  Speed = SpeedDef; drawFlag = 1; sendSpeed();
}

// ----------------------------------------------------------------------

public void resetStepRates() {
  // reset all start rate profiles
  for (int zR = 0; zR < rowNum; zR++) {
    int zP = (zR * rowWidth) + 12;
    int zS = ArrayVal[zP];
    // strip off rate values
    if (zS > 600) {zS = zS - 600;}
    else if (zS > 400) {zS = zS - 400;}
    else if (zS > 200) {zS = zS - 200;}
    ArrayVal[zP] = zS;
  } rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void RetRowFromTemp() {
  // return the angle values in Temp[] store to the current row
  int zRP = rowPnt * rowWidth;
  for (int zP = 0; zP < (rowWidth - 1); zP++) {
    ArrayVal[zP + zRP] = Temp[zP];
  }
}

// ----------------------------------------------------------------------

public void reverseSelRows() {
  // reverse the order of the selected rows
  // first find the range of selected rows, which need to be contiguous
  if ((SelCnt < 1) || (rowNum < 2)) {return;}
  usbEn = false;  // avoid any unintentional movement

  int zF = -1; int zL = -1; int zS;
  for (zS = 0; zS < rowNum; zS++) {
    if (Select[zS]) {
      // record the 1st selected row  
      if (zF < 0) {zF = zS;}
      // collect selected rows until break or end
      else {zL = zS;}
    } else {
      // exit if we find a n unselected row after finding one or more
      if (zF >= 0) {break;}
    }
  }
  rateFlag = true;drawFlag = 1;
  // did we only find one selected row?
  if (zL < 0) {msgTx = "Error. Only 1 row selected."; return;}
  
  // we only need to perform the operation on half the number of rows
  int zN = zL - zF + 1;
  msgTx = str(zN) + " rows reversed.";
  zN = zF + (zN/2);
  for (zS = zF; zS < zN; zS++) {
    //println(str(zS) + "   " + str(zL));
    int zRF = zS * rowWidth; int zRL = zL * rowWidth;
    for (int zP = 0; zP < rowWidth; zP++) {
      Temp[zP] = ArrayVal[zRF + zP];
      ArrayVal[zRF + zP] = ArrayVal[zRL + zP];
      ArrayVal[zRL + zP] = Temp[zP];
    }
    // progressively move the last row pointer towards the middle
    zL--;
  }
}


// ######################################################################


public void sendButton(String zS) {
  // called when a button is clicked on the scrn
  msgTx = zS; usbPortWrite(msgTx);
  drawFlag = 1;
}

// ----------------------------------------------------------------------

public void sendEcho() {
  // we send Ech status regardless of usbEn status
  if (Echo) {
    msgTx = "SR1.";
  } else {
    msgTx = "SR0.";
  }
  if (usbEn) {usbPortWrite(msgTx);}
  else if (comFlag > 0) {usbPort.write(msgTx);}
}

// ----------------------------------------------------------------------

public void sendRowData(int zR, int zA, int zEx) {
  // send the contents of a row as a group of individual angles, as:
  //
  // MLnn.n,n,. - start a memory load sequence, values separate by a comma, terminated by a period '.'
  //
  // called on demand or when clearing the array or switching rows
  if (zEx == 1) {sendSpeed();}
  msgTx = "ML" + str(zA) + ".";  // set RoboDog memory pointer
  int zP = zR * rowWidth;
  for (int zC = 0; zC < rowWidth; zC++){
    msgTx = msgTx + str(ArrayVal[zP + zC]) + ",";
  }
  if (zEx == 1) {
    msgTx += "MX" + str(zA) + ".";  // send execute command
  }
  usbPortWrite(msgTx);
  msgTx = msgTx.substring(0,20) + "...";
  drawFlag = 1;
}

// ----------------------------------------------------------------------

public void sendSpeed() {
  // called to send Speed value to arduinio
  msgTx = "SP" + str(Speed) + "."; usbPortWrite(msgTx);
}

// ----------------------------------------------------------------------

public void setAngValFromRow() {
  // called when a row changes or memory is cleared, this uses memory row
  // data to update the angle values and sliders
  int zP = rowPnt * rowWidth;
  if (ArrayVal[zP] < 0) {return;}  // command row so ignore
  
  AngVal[FR1] = ArrayVal[zP]; zP++;
  AngVal[FR2] = ArrayVal[zP]; zP++;
  AngVal[FR3] = ArrayVal[zP]; zP++;
  AngVal[FL1] = ArrayVal[zP]; zP++;
  AngVal[FL2] = ArrayVal[zP]; zP++;
  AngVal[FL3] = ArrayVal[zP]; zP++;
  AngVal[RR1] = ArrayVal[zP]; zP++;
  AngVal[RR2] = ArrayVal[zP]; zP++;
  AngVal[RR3] = ArrayVal[zP]; zP++;
  AngVal[RL1] = ArrayVal[zP]; zP++;
  AngVal[RL2] = ArrayVal[zP]; zP++;
  AngVal[RL3] = ArrayVal[zP];
  setSliderX();
}

// ----------------------------------------------------------------------

public void setGndY() {
  // set the GndY ground value based on the latest GndYM
  GndY = GndYM + 12;
}

// ----------------------------------------------------------------------

public void setRowData(int zRow,int zFR1,int zFR2,int zFR3,int zFL1,int zFL2,int zFL3,int zRR1,int zRR2,int zRR3,int zRL1,int zRL2,int zRL3,int zStp) {
  // called to load data directly into row zRow
  zRow = zRow * rowWidth;
  ArrayVal[zRow] = zFR1; zRow++;
  ArrayVal[zRow] = zFR2; zRow++;
  ArrayVal[zRow] = zFR3; zRow++;
  ArrayVal[zRow] = zFL1; zRow++;
  ArrayVal[zRow] = zFL2; zRow++;
  ArrayVal[zRow] = zFL3; zRow++;
  ArrayVal[zRow] = zRR1; zRow++;
  ArrayVal[zRow] = zRR2; zRow++;
  ArrayVal[zRow] = zRR3; zRow++;
  ArrayVal[zRow] = zRL1; zRow++;
  ArrayVal[zRow] = zRL2; zRow++;
  ArrayVal[zRow] = zRL3; zRow++;
  ArrayVal[zRow] = zStp;
}

// ----------------------------------------------------------------------

public void setSliderX() {
  // set the X-slider position for current AngVal[] values
  // note that some sliders work in reverse, as their corresponding servo is reversed
  // angles can also contain Pwr status, which has to be ignored
  // FR1
  JoyXval[0] = JoyXLL[0] + ((getAngVal(AngVal[0]) - AngVLL[0]) * (JoyXUL[0] - JoyXLL[0]))/(AngVUL[0] - AngVLL[0]);
  // FR2
  JoyXval[1] = JoyXUL[1] - ((getAngVal(AngVal[1]) - AngVLL[1]) * (JoyXUL[1] - JoyXLL[1]))/(AngVUL[1] - AngVLL[1]);
  // FR3
  JoyXval[2] = JoyXUL[2] - ((getAngVal(AngVal[2]) - AngVLL[2]) * (JoyXUL[2] - JoyXLL[2]))/(AngVUL[2] - AngVLL[2]);
  // FL1
  JoyXval[3] = JoyXLL[3] + ((getAngVal(AngVal[3]) - AngVLL[3]) * (JoyXUL[3] - JoyXLL[3]))/(AngVUL[3] - AngVLL[3]);
  // FL2
  JoyXval[4] = JoyXUL[4] - ((getAngVal(AngVal[4]) - AngVLL[4]) * (JoyXUL[4] - JoyXLL[4]))/(AngVUL[4] - AngVLL[4]);
  // FL3
  JoyXval[5] = JoyXUL[5] - ((getAngVal(AngVal[5]) - AngVLL[5]) * (JoyXUL[5] - JoyXLL[5]))/(AngVUL[5] - AngVLL[5]);
  // RR1
  JoyXval[6] = JoyXLL[6] + ((getAngVal(AngVal[6]) - AngVLL[6]) * (JoyXUL[6] - JoyXLL[6]))/(AngVUL[6] - AngVLL[6]);
  // RR2
  JoyXval[7] = JoyXUL[7] - ((getAngVal(AngVal[7]) - AngVLL[7]) * (JoyXUL[7] - JoyXLL[7]))/(AngVUL[7] - AngVLL[7]);
  // RR3
  JoyXval[8] = JoyXUL[8] - ((getAngVal(AngVal[8]) - AngVLL[8]) * (JoyXUL[8] - JoyXLL[8]))/(AngVUL[8] - AngVLL[8]);
  // RL1
  JoyXval[9] = JoyXLL[9] + ((getAngVal(AngVal[9]) - AngVLL[9]) * (JoyXUL[9] - JoyXLL[9]))/(AngVUL[9] - AngVLL[9]);
  // RL2
  JoyXval[10] = JoyXUL[10] - ((getAngVal(AngVal[10]) - AngVLL[10]) * (JoyXUL[10] - JoyXLL[10]))/(AngVUL[10] - AngVLL[10]);
  // RL3
  JoyXval[11] = JoyXUL[11] - ((getAngVal(AngVal[11]) - AngVLL[11]) * (JoyXUL[11] - JoyXLL[11]))/(AngVUL[11] - AngVLL[11]);
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void storeM0() {
  // transfer row data to memory store M0
  cursor(WAIT);
  M0 = rowNum; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    Mem0[zP] = ArrayVal[zP];
  } drawFlag = 1;
  msgTx = "Data stored in M0";
}

// ----------------------------------------------------------------------

public void storeM1() {
  // transfer row data to memory store M1
  cursor(WAIT);
  M1 = rowNum; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    Mem1[zP] = ArrayVal[zP];
  } drawFlag = 1;
  msgTx = "Data stored in M1";
}

// ----------------------------------------------------------------------

public void storeM2() {
  // transfer row data to memory store M2
  cursor(WAIT);
  M2 = rowNum; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    Mem2[zP] = ArrayVal[zP];
  } drawFlag = 1;
  msgTx = "Data stored in M2";
}

// ----------------------------------------------------------------------

public void storeM3() {
  // transfer row data to memory store M0
  cursor(WAIT);
  M3 = rowNum; int zM = rowNum * rowWidth;
  for (int zP = 0; zP < zM; zP++) {
    Mem3[zP] = ArrayVal[zP];
  } drawFlag = 1;
  msgTx = "Data stored in M3";
}


// ######################################################################


public void toggleEcho() {
  // toggles the comms echo function
  Echo = !Echo; sendEcho();
  if (!Echo) {msgRx = "Arduino is muted";}
  drawFlag = 1;
}

// ----------------------------------------------------------------------

public void toggleLink(int zA,int zB) {
  // toggles the link between servo zA and servo zB
  if (Link[zA] == 0) {
    // set the links even if the values aren't the same
    Link[zA] = 1; Link[zB] = 1;
  } else {Link[zA] = 0; Link[zB] = 0;}
  drawFlag = 1;
}

// ----------------------------------------------------------------------

public void toggleMultiLink(int zA) {
  // toggles multiple links in servo rows 1, 2 and 3
  drawFlag = 1;
  if (zA == FR1) {
    if (Link[FR1] < 2) {
      // multilink is not active so set all four links by adding 2
      Link[FR1] += 2; Link[FL1] += 2; Link[RR1] += 2; Link[RL1] += 2;
    } else {
      // multilink is active so set all four links by subtracting 2
      Link[FR1] -= 2; Link[FL1] -= 2; Link[RR1] -= 2; Link[RL1] -= 2;
    } return;
  }
  if (zA == FR2) {
    if (Link[FR2] < 2) {
      // multilink is not active so set all four links by adding 2
      Link[FR2] += 2; Link[FL2] += 2; Link[RR2] += 2; Link[RL2] += 2;
    } else {
      // multilink is active so set all four links by subtracting 2
      Link[FR2] -= 2; Link[FL2] -= 2; Link[RR2] -= 2; Link[RL2] -= 2;
    } return;
  }
  if (zA == FR3) {
    if (Link[FR3] < 2) {
      // multilink is not active so set all four links by adding 2
      Link[FR3] += 2; Link[FL3] += 2; Link[RR3] += 2; Link[RL3] += 2;
    } else {
      // multilink is active so set all four links by subtracting 2
      Link[FR3] -= 2; Link[FL3] -= 2; Link[RR3] -= 2; Link[RL3] -= 2;
    } return;
  }
}

// ----------------------------------------------------------------------

public void togglePlayPause() {
  // send command to toggle play pause
  drawFlag = 1;
  if (usbEn) {
    msgTx = "MP."; usbPortWrite(msgTx);
  } else if (Scrn == 2) {
    // pause/restart Walk simulator
    if (WalkSim != 0) {
      WalkSim = -WalkSim;  // toggle Walk mode +/-
      if (WalkSim < 0) {msgTx = "Sim Pause";}
    }
  }
}

// ----------------------------------------------------------------------

public void togglePwr(int zS) {
  // toggle the servo PWM pulses ON/OFF by modifying the angle value
  // all angles are within 0 - 180, so we can add 200 for the OFF state
  // if a rate offset of 400 is applied we need to sustain this
  int zOff = 0; if (AngVal[zS] >= 400) {zOff = 400; AngVal[zS] -= 400;}
  
  if (AngVal[zS] >= 200) {
    AngVal[zS] -= 200;  // remove the Pwr offset
    togPwr = 1;         // set all mode to sustain power
  } else {
    AngVal[zS] += 200;  // add the Pwr offset
    togPwr = 2;         // set all mode to remove power
  }
  AngVal[zS] += zOff;
  ArrayVal[(rowPnt * rowWidth) + zS] = AngVal[zS];
  // send updated value to usb port
  msgTx = "ST" + str(Ch[zS]) + "." +"SA" + str(AngVal[zS]) + ".";
  setAngValFromRow();
  usbPortWrite(msgTx); drawFlag = 1;
  if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
}

// ----------------------------------------------------------------------

public void togglePwrAll() {
  // sets the state of all servos to the current selection when the mouse
  // button is held down. togPwr indicates the target state
  if (togPwr > 1) {togPwr = 200;} else {togPwr = 0;}
  AngVal[FR1] = getAngVal(AngVal[FR1]) + togPwr; if (AngRate) {AngVal[FR1] += 400;}
  ArrayVal[(rowPnt * rowWidth) + FR1] = AngVal[FR1];
  
  AngVal[FR2] = getAngVal(AngVal[FR2]) + togPwr; if (AngRate) {AngVal[FR2] += 400;}
  ArrayVal[(rowPnt * rowWidth) + FR2] = AngVal[FR2];
  
  AngVal[FR3] = getAngVal(AngVal[FR3]) + togPwr; if (AngRate) {AngVal[FR3] += 400;}
  ArrayVal[(rowPnt * rowWidth) + FR3] = AngVal[FR3];
  
  AngVal[FL1] = getAngVal(AngVal[FL1]) + togPwr; if (AngRate) {AngVal[FL1] += 400;}
  ArrayVal[(rowPnt * rowWidth) + FL1] = AngVal[FL1];
  
  AngVal[FL2] = getAngVal(AngVal[FL2]) + togPwr; if (AngRate) {AngVal[FL2] += 400;}
  ArrayVal[(rowPnt * rowWidth) + FL2] = AngVal[FL2];
  
  AngVal[FL3] = getAngVal(AngVal[FL3]) + togPwr; if (AngRate) {AngVal[FL3] += 400;}
  ArrayVal[(rowPnt * rowWidth) + FL3] = AngVal[FL3];
  
  AngVal[RR1] = getAngVal(AngVal[RR1]) + togPwr; if (AngRate) {AngVal[RR1] += 400;}
  ArrayVal[(rowPnt * rowWidth) + RR1] = AngVal[RR1];
  
  AngVal[RR2] = getAngVal(AngVal[RR2]) + togPwr; if (AngRate) {AngVal[RR2] += 400;}
  ArrayVal[(rowPnt * rowWidth) + RR2] = AngVal[RR2];
  
  AngVal[RR3] = getAngVal(AngVal[RR3]) + togPwr; if (AngRate) {AngVal[RR3] += 400;}
  ArrayVal[(rowPnt * rowWidth) + RR3] = AngVal[RR3];
  
  AngVal[RL1] = getAngVal(AngVal[RL1]) + togPwr; if (AngRate) {AngVal[RL1] += 400;}
  ArrayVal[(rowPnt * rowWidth) + RL1] = AngVal[RL1];
  
  AngVal[RL2] = getAngVal(AngVal[RL2]) + togPwr; if (AngRate) {AngVal[RL2] += 400;}
  ArrayVal[(rowPnt * rowWidth) + RL2] = AngVal[RL2];
  
  AngVal[RL3] = getAngVal(AngVal[RL3]) + togPwr; if (AngRate) {AngVal[RL3] += 400;}
  ArrayVal[(rowPnt * rowWidth) + RL3] = AngVal[RL3];
  
  setAngValFromRow();
  sendRowData(rowPnt,rowPnt,1);
  mDwn = 0;        // clear mouse down timer
  drawFlag = 1;
}

// ----------------------------------------------------------------------

public void toggleRate(int zS) {
  // toggle the servo 'auto-rate' mode ON/OFF by modifying the angle value
  // all angles are within 0 - 180, +200 for Pwr OFF, so we can add 400 for the rate ON state
  rateFlag = true; drawFlag = 1;
  if (rowPnt == 0) {msgTx = "Error - first line"; return;}
  if ((rowPnt >= (rowNum - 1)) && (AngVal[zS] < 400)) {msgTx = "Error - last line"; return;}
  if (rowNum < 3) {msgTx = "Error - too few lines"; return;}
  
  if (AngVal[zS] >= 400) {
    AngVal[zS] -= 400;   // remove the auto-rate +400 offset
    Rated = getRated();  // check to see if any more rate angles exist
  } else {
    AngVal[zS] += 400;  // add the auto-rate +400 offset
    Rated = true;
  }
  ArrayVal[(rowPnt * rowWidth) + zS] = AngVal[zS];
  if (Rated) {calcRates();}  // update rate values
  
  // send updated value to usb port
  msgTx = "ST" + str(Ch[zS]) + "." +"SA" + str(AngVal[zS]) + ".";
  setAngValFromRow();
  usbPortWrite(msgTx);
  if (msgTx.length() > 28) {msgTx = msgTx.substring(0,24) + "...";}
}

// ----------------------------------------------------------------------

public void toggleRowSelect() {
  // toggle the selection of a given row
  int zF; int zT;
  int zR = getRowMYD();
  zR = rowPnt + zR;
  if ((zR >= 0) && (zR < rowNum)) {
    // valid row pointer
    if (SelLast < 0) {
      // first time selection
      Select[zR] = true; SelCnt++;
    } else {
      if (zR > SelLast) {zF = SelLast; zT = zR;} else {zF = zR; zT = SelLast;}
      if (Select[zR]) {
        if (keySHIFT) {
          // multi-line deselect
          for (zR = zF; zR < zT + 1; zR++) {
            if (Select[zR]) {Select[zR] = false; SelCnt--;}
          }
        } else {
          // single line deselect
          Select[zR] = false; SelCnt--;
        }
      } else {
        if (keySHIFT) {
          // multi-line select
          for (zR = zF; zR < zT + 1; zR++) {
            if (!Select[zR]) {Select[zR] = true; SelCnt++;}
          }
        } else {
          // single line select
          Select[zR] = true; SelCnt++;
        }
      }
    } SelLast = zR; drawFlag = 1;
  }
}

// ----------------------------------------------------------------------

public void toggleStepRate() {
  // user has clicked on step rate profile column
  boolean zC = false;    // something has changed flag
  int zM;                // rate mode flag
  int zS;                // current step value
  int zR = getRowMYD();
  zR = rowPnt + zR;
  if ((zR < 0) || (zR >= rowNum)) {return;}  // row is before/after list
  if (ArrayVal[zR * rowWidth] < 0) {return;}  // command line so exit
  
  // valid row pointer
  zR = (zR * rowWidth) + 12;
  zS = ArrayVal[zR];
  if (zS > 600) {zS -= 600; zM = 3;}
  else if (zS > 400) {zS -= 400; zM = 2;}
  else if (zS > 200) {zS -= 200; zM = 1;}
  else {zM = 0;}
    
  if (keySHIFT) {
    // special 'quick' copy/paste step rate function
    if (mB == LEFT) {
      // paste quick rate value, which has previously been set
      if (QuickRate >= 0) {zM = QuickRate; zC = true;}
    } else {
      // record 'quick' step rate value
      QuickRate = zM; zC = true;
      msgTx = "QuickRate = " + str(zM); drawFlag = 1; return;
    }
  } else {
    // toggle current rate value
    if (mB == LEFT) {zM++; if (zM > 3) {zM = 0;}}
    else {zM--; if (zM < 0) {zM = 3;}}
    zC = true;
  }
  if (zM > 2) {zS += 600;}
  else if (zM > 1) {zS += 400;}
  else if (zM > 0) {zS += 200;}
  ArrayVal[zR] = zS;
  if (zC) {drawFlag = 1;}
}

// ----------------------------------------------------------------------

public void toggleUsbEn() {
  // toggles the usbEn flag to control messages sent to the Arduino
  //if (SelCnt > 0) {ClearSelect();}
  if (comFlag < 1) {return;}  // no COM port connection
  
  // if enabled and we are playiong a move, then stop it first
  if ((usbEn) && (playMode > 1)) {playStop();}
  playMode = 0;  // reset the flag in case of playing once
  // now toggle the flag
  usbEn = !usbEn; drawFlag = 1;
  
  // if now enabled then send the current line data
  if (usbEn) {
    sendRowData(rowPnt,rowPnt,1);
  }
}

// ----------------------------------------------------------------------

public void transSel() {
  // transpose angles for selected lines
  msgTx = "No rows selected"; rateFlag = true; drawFlag = 1;  
  if (SelCnt < 1) {return;}
  usbEn = false;  // avoid any unintentional movement
  
  cursor(WAIT);
  int zP = 0; int zRP = 0; boolean zSR0 = false;
  for (int zR = 0; zR < rowNum; zR++) {
    // work through each selected row
    if (Select[zR]) {
      // row is selected
      zRP = zR * rowWidth;
      // check to see if this is a command row?
      if (ArrayVal[zRP] > 0) {
        // it is not Cmd so proceed
        // if current row then set a flag to update sliders at the end
        if (zR == rowPnt) {zSR0 = true;}
        // first grab row values in Temp[] store
        for (zP = 0; zP < (rowWidth - 1); zP++) {
          Temp[zP] = ArrayVal[zP + zRP];
        }
        // now perform transposition of angles
        // it is the tier 3 servos which need to be corrected
        ArrayVal[zRP]     = Temp[3];  // set FR1
        ArrayVal[zRP + 1] = Temp[4];  // set FR2
        ArrayVal[zRP + 2] = 180 - getAngVal(Temp[5]);  // set FR3
        if (AngRate) {ArrayVal[zRP + 2] += 400;}
        if (!AngPwr) {ArrayVal[zRP + 2] += 200;}
        
        ArrayVal[zRP + 3] = Temp[0];  // set FL1
        ArrayVal[zRP + 4] = Temp[1];  // set FL2
        ArrayVal[zRP + 5] = 180 - getAngVal(Temp[2]);  // set FL3
        if (AngRate) {ArrayVal[zRP + 5] += 400;}
        if (!AngPwr) {ArrayVal[zRP + 5] += 200;}
        
        ArrayVal[zRP + 6] = Temp[9];  // set RR1
        ArrayVal[zRP + 7] = Temp[10]; // set RR2
        ArrayVal[zRP + 8] = 180 - getAngVal(Temp[11]); // set RR3
        if (AngRate) {ArrayVal[zRP + 8] += 400;}
        if (!AngPwr) {ArrayVal[zRP + 8] += 200;}
        
        ArrayVal[zRP + 9] = Temp[6];  // set RL1
        ArrayVal[zRP + 10] = Temp[7]; // set RL1
        ArrayVal[zRP + 11] = 180 - getAngVal(Temp[8]); // set RL1
        if (AngRate) {ArrayVal[zRP + 11] += 400;}
        if (!AngPwr) {ArrayVal[zRP + 11] += 200;}
      }
    }
  }
  usbEn =false;
  if (zSR0) {setAngValFromRow();}  // set servo values and sliders
  msgTx = "Selected data transposed";
}

// ----------------------------------------------------------------------

public void transFR1FL1() {
  // transpose the FR1 and FL1 angles
  int zP =  rowPnt * rowWidth;
  Temp[0] = ArrayVal[zP];  // copy FR1
  ArrayVal[zP] =  ArrayVal[zP + 3];  // get FR1 from FL1
  ArrayVal[zP + 3] = Temp[0];        // replace FL1
  usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  msgTx = "FR1 and FL1 transposed";
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void transFR2FL2() {
  // transpose the FR2 and FL2 angles
  int zP =  rowPnt * rowWidth;
  Temp[0] = ArrayVal[zP + 1];  // copy FR2
  ArrayVal[zP + 1] =  ArrayVal[zP + 4];  // get FR2 from FL2
  ArrayVal[zP + 4] = Temp[0];            // replace FL2
  usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  msgTx = "FR2 and FL2 transposed";
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void transFR3FL3() {
  // transpose the FR3 and FL3 angles
  int zP =  rowPnt * rowWidth;
  Temp[0] = ArrayVal[zP + 2];  // copy FR3
  ArrayVal[zP + 2] = 180 - getAngVal(ArrayVal[zP + 5]);  // get FR3 from FL3
  if (AngRate) {ArrayVal[zP + 2] += 400;}
  if (!AngPwr) {ArrayVal[zP + 2] += 200;}
  ArrayVal[zP + 5] = 180 - getAngVal(Temp[0]);  // replace FL3
  if (AngRate) {ArrayVal[zP + 5] += 400;}
  if (!AngPwr) {ArrayVal[zP + 5] += 200;}
  usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  msgTx = "FR3 and FL3 transposed";
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void transRR1RL1() {
  // transpose the RR1 and RL1 angles
  int zP =  rowPnt * rowWidth;
  Temp[0] = ArrayVal[zP + 6];  // copy RR1
  ArrayVal[zP + 6] = ArrayVal[zP + 9];  // get RR1 from RL1
  ArrayVal[zP + 9] = Temp[0];           // replace RL1
  usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  msgTx = "RR1 and RL1 transposed";
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void transRR2RL2() {
  // transpose the RR2 and RL2 angles
  int zP =  rowPnt * rowWidth;
  Temp[0] = ArrayVal[zP + 7];  // copy RR2
  ArrayVal[zP + 7] = ArrayVal[zP + 10];  // get RR2 from RL2
  ArrayVal[zP + 10] = Temp[0];           // replace RL2
  usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  msgTx = "RR2 and RL2 transposed";
  rateFlag = true; drawFlag = 1;
}

// ----------------------------------------------------------------------

public void transRR3RL3() {
  // transpose the RR3 and RL3 angles
  int zP =  rowPnt * rowWidth;
  Temp[0] = ArrayVal[zP + 8];  // copy RR3
  ArrayVal[zP + 8] = 180 - getAngVal(ArrayVal[zP + 11]);  // get RR3 from RL3
  if (AngRate) {ArrayVal[zP + 8] += 400;}
  if (!AngPwr) {ArrayVal[zP + 8] += 200;}
  ArrayVal[zP + 11] = 180 - getAngVal(Temp[0]);            // replace RL3
  if (AngRate) {ArrayVal[zP + 11] += 400;}
  if (!AngPwr) {ArrayVal[zP + 11] += 200;}
  usbEn =false;
  setAngValFromRow();  // set servo values and sliders
  msgTx = "RR3 and RL3 transposed";
  rateFlag = true; drawFlag = 1;
}


// ######################################################################


public void usbPortOp() {
  // called from a timer to empty the usb buffer
  usbLen = usbBuff.length(); drawUsbLamp();
  if (usbLen < 1) {return;}  // nothing in buffer
  
  // we are going to send something from usbBuff
  if (usbLen <= WiFiPW) {
    // it should only take 2.8ms to send 32 bytes
    usbPort.write(usbBuff); Tx_t0 = millis();  // send small packet
    usbBuff = ""; usbLen = 0; // send small packet and clear buffer
  } else {
    usbPort.write(usbBuff.substring(0,WiFiPW)); // send WiFiPW length packet
    Tx_t0 = millis();
    usbBuff = usbBuff.substring(WiFiPW,usbBuff.length());  // shift buffer by 16 char
    usbLen = usbBuff.length();
  } Ping = PingMax;  // no need to ping if sending data, so reset timer
}

// ----------------------------------------------------------------------

public void usbPortWrite(String zmsg) {
  // called to send text over the serial port
  // nothing is sent if comFlag or usbEn are false
  if (!zmsg.equals("#")) {
    usbMsg = zmsg;  // keep a record or what was last sent or intended, ignoring pings
    if ((comFlag > 0) && usbEn) {
      usbBuff = usbBuff + zmsg;  // add new message to buffer
      //println(zmsg);
    } else {usbBuff = "";}  // flush the buffer if not sending
  } else {
    // this is a 'ping' # which is always sent if comm port is open
    // the droids eyes should always be red when this app is active
    if (comFlag > 0) {usbBuff = usbBuff + zmsg;}  // add new message to buffer
  }
}


// ######################################################################


public void Walk_Init() {
  // initiallises the 'Walk' simulator from the first row, which will be
  // drawn. A delay is set based on the row step value, but if the row
  // contains a Command then the timer is set to re-trigger immediately
  // first check for selected lines
  sim1st = 0;           // first row of data
  simEnd = rowNum - 1;  // last row of data
  if (SelCnt > 0) {
    // rows are selected so find first selected block
    // if multiple blocks or rows are selected we only play the first
    sim1st = -1;
    for (int zR = 0; zR < rowNum; zR++) {
      if (Select[zR]) {
        if (sim1st < 0) {sim1st = zR;}    // first selected row
        simEnd = zR;                      // track last selected row
      } else {if (sim1st >= 0) {break;}}  // exit at end of selected block
    }
  }
  
  rowPnt = sim1st;     // define the row to be drawn
  simPnt = sim1st;     // simulator start pointer
  CmdForNum = 0;  // For... Next loop count
  CmdForRow = 0;  // row pointer for For... Next loops
  WalkNext = millis();  // set the timer to trigger immediately
  msgTx = "Sim " + str(rowPnt);
  drawFlag = 1;
}

// ----------------------------------------------------------------------

public void Walk_Sim() {
  // called from draw() function when enabled and timer reached
  // still within the current list
  msgTx = "Sim " + str(simPnt) + "   " + str(t1);
  int zCD = 1;       // command step delay, default
  int zDF = 0;       // temp drawFlag
  int zP = simPnt * rowWidth;
  int zCmdCode = 0;  // command code type
  int zCmdVal0 = 0;  // command value
  if (ArrayVal[zP] < 0) {
    // command line so check for Goto, Speed, Pause, etc
    zCmdCode = ArrayVal[zP + 1];
    zCmdVal0 = ArrayVal[zP + 2];
    // a blank line is simply ignored
    if (zCmdCode == 1) {
      // this is a Goto command so go to it if in continuous mode
      if (WalkSim == 2) {simPnt = ArrayVal[zP + 2] - 1;}
      // if not then stop the process
      else {WalkSim = 0; msgTx = "Sim Terminated";}
    }
    else if(zCmdCode == 2) {
      // this is a speed change command
      Speed = zCmdVal0;
    } 
    else if(zCmdCode == 3) {
      // this is a Pause command, so determine pause delay
      zCD = 1000/Speed; zCD = max(zCD,4);
      zCD = zCD * zCmdVal0;
    } 
    else if(zCmdCode == 4) {
      // this is a Flag command, so simply ignore it
    } 
    else if(zCmdCode == 5) {
      // this is a Balance command, so simply ignore it
    } 
    else if(zCmdCode == 6) {
      // this is a For... Next loop command, so set pointers
      CmdForNum = zCmdVal0; CmdForRow = simPnt;
    } 
    else if(zCmdCode == 7) {
      // this is a Next command in a For... Next loop
      // loop back to previous row if there is more to do
      if (CmdForNum > 1) {CmdForNum--; simPnt = CmdForRow;}
      else {CmdForNum = 0; CmdForRow = 0;}
    } 
    // Command lines do not consume step time so simply add 1ms default delay
    // unless a Pause command has been encountered
     WalkNext = millis() + zCD;
  } else {
    // not a command line so load the step delay
    rowPnt = simPnt;  // set the drawing pointer after the simulated delay
    getWalkDelay(); zDF = 1;
  }
  
  simPnt++;  // point at the next line
  if (simPnt > simEnd ){
    // at the last line in the list so either stop or go back to the beginning
    if (WalkSim == 2) {
      // in continuous mode so loop back to the beginning
      Walk_Init();
    } else {
      // switch OFF the Walk simulator
      WalkSim = 0;  // stop the simulator
      msgTx = "Sim End";
    }
  }
  drawFlag = zDF;
}

// ----------------------------------------------------------------------

// Called from a mouse move this presents help messages for areas under
// the mouse pointer.

public void Help_Message() {
  if (HelpOFF > 0) {return;}  // if >0 then help system is disabled

  Help$ = "-";         // clear current help message
  HelpRowPnt = -1;     // clear global row pointer
  int zAF = AngField;  // note previous value to trap update
  AngField = -1;       // reset angle field highlights
  rowColPnt = -1;      // reset column field highlights

  int zrowPnt = (getRowMYD() + rowPnt); zrowPnt = max(zrowPnt,0); zrowPnt = min(zrowPnt,rowNum-1);
  int zCP = zrowPnt * rowWidth;
  int CmdCode = ArrayVal[zCP + 1] ;  // get command type code
  int CmdVal0 = ArrayVal[zCP + 2] ;  // get command 1st value
  boolean CmdRowHlp = (ArrayVal[zCP] < 0);
  CmdRow = (ArrayVal[rowPnt * rowWidth] < 0);

  if ((mX >= 377) && (mX <= 446) && (mY < 36)) {Help$ = "Change App modes.";}

  if ((mY >= 13) && (mY <= 47)) {
    // top function buttons 1st row
         if ((mX >= 601) && (mX < 671)) {Help$ = "Insert 'REST' data.   L-click after. R-click before.  SHIFT-click send only.";}
    else if ((mX >= 671) && (mX < 741)) {Help$ = "Insert 'READY' data.   L-click after. R-click before.  SHIFT-click send only.";}
    else if ((mX >= 741) && (mX <= 810)) {Help$ = "Insert 'STAND' data.   L-click after. R-click before.  SHIFT-click send only.";}
  }
  else if ((mY >= 48) && (mY <= 82)) {
    // top function buttons 2nd row
         if ((mX >= 601) && (mX < 671)) {Help$ = "";}
    else if ((mX >= 671) && (mX < 741)) {Help$ = "";}
    else if ((mX >= 741) && (mX <= 810)) {Help$ = "Insert 'SQUAT' data.   L-click after. R-click before.  SHIFT-click send only.";}
  }

  //#####################################################################
  //
  // Screen specific help functions
  //
  //#####################################################################
  if (Scrn == 0) {Help_Message_0();}
  else if (Scrn == 1) {Help_Message_1();}
  else {Help_Message_2();}
  if (rowColPnt >= 0) {drawFlag = 1;}
  //Help$ = str(rowColPnt);

  //#####################################################################
  //
  // All lower screen help functions
  //
  //#####################################################################
  if ((mY >= 547) && (mY <= 686)) {
    // Memory lines, sliders and buttons
    if ((SelCnt > 0) && (mX <= 105)) {
      // left-hand side buttons and arrows
      if (mY <= 580) {
        // left region top row
        if ((mX >= 72) && (mX <= 102)) {Help$ = "Move " + str(SelCnt) + " selected rows up.";}
      } else if (mY <= 616) {
        // left region 2nd row
        if (mX < 39) {Help$ = "Copy and cut " + str(SelCnt) + " selected rows.";}
        if ((mX >= 39) && (mX < 72)) {Help$ = "Copy " + str(SelCnt) + " selected rows.";}
      } else if (mY <= 650) {
        // left region 3rd row
        if ((mX >= 9) && (mX < 39)) {
          if ((CmdRowHlp) && (CmdType == 1)) {
            Help$ = "Overwrite a GOTO branch, to a selected row.";
          } else {
            Help$ = "Insert a GOTO branch, to a selected row.";
          }
        }
             if ((mX >= 39) && (mX < 71)) {Help$ = "Calculate steps for " + str(SelCnt) + " selected rows.  SHIFT-click to reset selected.";}
        else if ((mX >= 71) && (mX <= 102)) {Help$ = "Transpose " + str(SelCnt) + " selected row angles.";}
      } else {
        // left region 4th row
        if ((mX >= 39) && (mX < 71)) {
          if (SelCnt < 2) {
            Help$ = "Reverse selected row order. INACTIVE - too few selected.";
          } else {
            Help$ = "Reverse the order of " + str(SelCnt) + " selected rows.";
          }
        }
        if ((mX >= 71) && (mX <= 102)) {Help$ = "Move " + str(SelCnt) + " selected rows down.";}
      }
    }
    if (CopyCnt > 0) {  
      if ((mY > 575) && (mY < 614)) {  
        if ((mX >= 71) && (mX <= 102)) {Help$ = "Paste " + str(CopyCnt) + " copied rows.";}
      }  
    }
    if ((mX >= 585) && (mX <= 611)) {
      // vertical slider region
           if (mY < 566) {Help$ = "Decrement current row pointer.";}
      else if ((mY >= 566) && (mY < 589)) {Help$ = "Go to first row.";}
      else if ((mY >= 642) && (mY < 665)) {Help$ = "Go to last row.";}
      else if (mY >= 665) {Help$ = "Increment current row pointer.";}
    }
    if ((mX >= 614) && (mX <= 643)) {
      //R-hand insert buttons Blank,Delete,Pause,Balance
           if (mY < 576) {Help$ = "Insert a blank row. L-click after. R-click before.";}
      else if ((mY >= 578) && (mY < 613)) {
        if (SelCnt < 1) {Help$ = "Delete current row.";}
        else {Help$ = "Delete all selected rows.";}
      }
      else if ((mY >= 616) && (mY < 651)) {Help$ = "Insert a Pause. L-click after. R-click before.";}
      else if ((mY >= 653) && (mY < 689)) {Help$ = "Insert a Balance. L-click after. R-click before.";}
    }
    if ((mX >= 646) && (mX <= 675)) {
      //R-hand insert buttons Speed,New,Flag,TBD
           if (mY < 576) {
             if ((Cmd) && (CmdType == 2)) {
               Help$ = "Click to change 'Speed' command.";
             } else {
               Help$ = "Insert a 'Speed' command. L-click after. R-click before.";
             }
           }
      else if ((mY >= 587) && (mY <= 614)) {Help$ = "Add duplicate row. L-click after. R-click before. SHIFT-click intermediate.";}
      else if ((mY >= 616) && (mY <= 651)) {Help$ = "Insert a 'Flag' value. L-click after. R-click before.";}
      else if ((mY >= 653) && (mY <= 689)) {Help$ = "Insert an excuteable code. L-click after. R-click before.";}
    }
    if ((mY >= 559) && (mY <= 595)) {
      // Memory buttons top row
           if ((mX >= 691) && (mX < 719)) {Help$ = "M0 store. Click to recall. Click-hold to store data.";}
      else if ((mX >= 719) && (mX < 749)) {Help$ = "M1 store. Click to recall. Click-hold to store data.";}
      else if ((mX >= 749) && (mX < 778)) {Help$ = "M2 store. Click to recall. Click-hold to store data.";}
      else if ((mX >= 778) && (mX <= 807)) {Help$ = "M3 store. Click to recall. Click-hold to store data.";}
    }
    if ((mY >= 598) && (mY <= 633)) {
      // Play buttons 2nd row
      if (usbEn || (Scrn == 2)) {
             if ((mX >= 691) && (mX < 719)) {Help$ = "PLAY memory contents once.";}
        else if ((mX >= 719) && (mX < 749)) {Help$ = "PLAY memory contents repeatedly.";}
        else if ((mX >= 749) && (mX < 778)) {Help$ = "PAUSE playing.";}
        else if ((mX >= 778) && (mX <= 807)) {Help$ = "STOP playing.";}
      }  
    }
    if ((mY >= 636) && (mY <= 671)) {
      // 3rd row buttons - Clear,Copy,Paste
           if ((mX >= 691) && (mX < 749)) {Help$ = "Click-hold to clear memory contents.";}
      else if ((mX >= 749) && (mX < 778)) {Help$ = "Copy memory contents to clipboard.";}
      else if ((mX >= 778) && (mX < 807)) {Help$ = "Click-hold to paste clipboard contents into memory.";}
    }
  }

  //#####################################################################
  //
  // list lines region
  //
  //#####################################################################
  int zmY = 547; if (Scrn == 1) {zmY = 41;}
  if ((mY >= zmY) && (mY <= 686)) {
    if ((mX >= 105) && (mX < 582)) {
      // memory lines region, list window
      int zR = getRowMYD() + rowPnt;
      if ((zR >= 0) && (zR < rowNum)) {
        HelpRowPnt = zR; drawFlag = 1; // global pointer for current mouse row
        if ((mX >= 105) && (mX < 138)) {Help$ = "Select this row.  SHIFT+click to select a range.";}
        if (!CmdRowHlp) {
          if ((mX >= 154) && (mX < 494)) {
                AngField = getAngleMXD();
                if (zAF != AngField) {drawFlag = 1;}  // only redraw if a change has occured
                Help$ = "Servo angles. Click-L increase / Click-R decrease.";
            if ((mY >= 604) && (mY <= 626)) {
                // centre row
                //AngField = getAngleMXD(); HelpLast$ = "";
                Help$ = "Current memory row.";
                if (usbEn) {Help$ += " SHIFT L-click to send data.";}
                else {Help$ += " SHIFT+L/R-click to copy/paste data.";}
            }
          }
          if ((mX >= 512) && (mX < 537)) {
            switch(getProfile(getRowMYD() + rowPnt)) {
              case 0: Help$ = "Constant"; break;
              case 1: Help$ = "Accelerate"; break;
              case 2: Help$ = "Decelerate"; break;
              case 3: Help$ = "Switch A/D"; break;
            }
            Help$ += " rate profile. Click L/R to change. Click+hold to reset ALL.";
          }
          if ((mX >= 543) && (mX < 582)) {Help$ = "Steps value. L-click increase. R-click decrease. Shift-click calculate.";}
        } else {
          // command line
          if (CmdCode < 6) {Help$ = "Command: ";} else {Help$ = "Execute: ";}
               if (CmdCode == 0) {Help$ = "Blank line";}
          else if (CmdCode == 1) {Help$ += "Goto row " + str(CmdVal0);}
          else if (CmdCode == 2) {Help$ += "Set speed to " + str(CmdVal0) + ". L-click increase. R-click decrease.";}
          else if (CmdCode == 3) {Help$ += "Set pause to " + str(CmdVal0) + ". L-click increase. R-click decrease.";}
          else if (CmdCode == 4) {Help$ += "Set Flag to " + str(CmdVal0) + ". L-click increase. R-click decrease.";}
          else if (CmdCode == 5) {
            Help$ += "Set Balance to ";
            if (CmdVal0 == 0) {Help$ +=  "OFF";}
            else {Help$ +=  "ON";}
            Help$ +=  ". Click to toggle ON/OFF.";
          }
          else if (CmdCode == 6) {Help$ += "For...1 to ..." + str(CmdVal0) + " loop.  L-click increase. R-click decrease.";}
          else if (CmdCode == 7) {Help$ += "Next after a For... loop.  L-click increase. R-click decrease.";}
        }
      }
    }
  }
  
  if ((mY >= 703) && (mY < 741)) {
    // Processing logo
    if ((mX >= 12) && (mX <= 49)) {Help$ = "App written in 'Processing'.   Visit: https://processing.org/";}
  }
  if ((mY >= 690) && (mY < 716)) {
    // Com, RX, Tx fields
    if ((mX >= 105) && (mX <= 180)) {
      if (comFlag < 1) {Help$ = "USB port is disconnected. Left-click to connect.";}
      else {Help$ = "USB port is connected. Right-click to disconnect.";}
    }
    if ((mX >= 217) && (mX <= 440)) {Help$ = "Messages received from USB port.";}
    if ((mX >= 478) && (mX <= 682)) {
      Help$ = "Messages sent to the USB port - ";  
      if (usbEn) {Help$ += "ENABLED";} else {Help$ += "DISABLED";}
    }
    if ((mX >= 694) && (mX <= 712)) {Help$ = "USB activity data lamp";}
  }
  if ((mY >= 686) && (mY <= 721)) {
    // Echo button
    if ((mX >= 749) && (mX <= 807)) {Help$ = "Echo function. Click to toggle Arduino responses ON/OFF.";}
  }
  if ((mY >= 724) && (mY <= 748)) {
    // Help field & TechKnowTone
         if ((mX >= 106) && (mX <= 682)) {Help$ = "Hello";}
    else if ((mX >= 712) && (mX <= 807)) {Help$ = "Author: TechKnowTone     Released: " + Released$;}
  }
  
  if (!Help$.equals(HelpLast$)) {
    // string is new so display it and remember it
    HelpLast$ = Help$; drawFlag = 1;
  }
  
}

// ----------------------------------------------------------------------

public void Help_Message_0() {
  // display help messages for Scrn0
  if (CmdRow) {
    // command so make top part of form inactive
  } else {
    // not a command line so display help messages

    if ((mY >= 122) && (mY < 148)) {
        // Level 3 text fields & links
             if ((mX >= 148) && (mX < 160)) {Help$ = "Increment shoulder FR3 value.";}
        else if ((mX >= 160) && (mX < 186)) {Help$ = "Shoulder FR3 value. Click-n-drag to change.";}
        else if ((mX >= 186) && (mX < 198)) {Help$ = "Decrement shoulder FR3 value.";}
  
        else if ((mX >= 198) && (mX < 215)) {Help$ = "Click to set FR3 as FL3.";}
        else if ((mX >= 215) && (mX < 233)) {Help$ = "Link changes to FR3 and FL3 values.";}
        else if ((mX >= 233) && (mX < 250)) {Help$ = "Click to set FL3 as FR3.";}
  
        else if ((mX >= 250) && (mX < 262)) {Help$ = "Increment shoulder FL3 value.";}
        else if ((mX >= 262) && (mX <= 287)) {Help$ = "Shoulder FL3 value. Click-n-drag to change.";}
        else if ((mX > 287) && (mX <= 299)) {Help$ = "Decrement shoulder FL3 value.";}
  
        else if ((mX >= 376) && (mX < 388)) {Help$ = "Increment shoulder RR3 value.";}
        else if ((mX >= 388) && (mX <= 413)) {Help$ = "Shoulder RR3 value. Click-n-drag to change.";}
        else if ((mX > 413) && (mX <= 425)) {Help$ = "Decrement shoulder RR3 value.";}
  
        else if ((mX >= 425) && (mX < 441)) {Help$ = "Set RR3 as RL3.";}
        else if ((mX >= 441) && (mX < 459)) {Help$ = "Link changes to RR3 and RL3 values.";}
        else if ((mX >= 459) && (mX < 476)) {Help$ = "Set RL3 as RR3.";}
  
        else if ((mX >= 476) && (mX < 488)) {Help$ = "Increment shoulder RL3 value.";}
        else if ((mX >= 488) && (mX <= 514)) {Help$ = "Shoulder RL3 value. Click-n-drag to change.";}
        else if ((mX > 514) && (mX <= 526)) {Help$ = "Decrement shoulder RL3 value.";}
    
             if ((mX >= 148) && (mX < 198)) {rowColPnt = 2;}  // FR3
        else if ((mX >= 250) && (mX < 299)) {rowColPnt = 5;}  // FL3
        else if ((mX >= 376) && (mX < 425)) {rowColPnt = 8;}  // RR3
        else if ((mX >= 476) && (mX < 526)) {rowColPnt = 11;} // RL3
        
    } else if ((mY >= 148) && (mY <= 168)) {
        // Level 3 transposes & Pwr buttons
        if ((mX >= 165) && (mX <= 182)) {Help$ = "Toggle FR3 servo power. Hold to toggle all.";}
        else if ((mX >= 183) && (mX <= 198)) {Help$ = "Toggle FR3 auto-rate mode.";}
        else if ((mX >= 213) && (mX < 234)) {Help$ = "Transpose FR3/Ch5 with FL3/Ch4 angles";}
        else if ((mX >= 266) && (mX <= 284)) {Help$ = "Toggle FL3 servo power. Hold to toggle all.";}
        else if ((mX >= 285) && (mX <= 299)) {Help$ = "Toggle FL3 auto-rate mode.";}
        else if ((mX >= 391) && (mX <= 409)) {Help$ = "Toggle RR3 servo power. Hold to toggle all.";}
        else if ((mX >= 410) && (mX <= 425)) {Help$ = "Toggle RR3 auto-rate mode.";}
        else if ((mX >= 439) && (mX < 461)) {Help$ = "Transpose RR3/Ch10 with RL3/Ch11 angles";}
        else if ((mX >= 492) && (mX <= 510)) {Help$ = "Toggle RL3 servo power. Hold to toggle all.";}
        else if ((mX >= 510) && (mX <= 526)) {Help$ = "Toggle RL3 auto-rate mode.";}

    
    } else if ((mY >= 260) && (mY < 286)) {
        // Level 2 text fields & links
        if ((mX >= 174) && (mX < 186)) {Help$ = "Increment upper leg FR2 value.";}
        else if ((mX >= 186) && (mX <= 211)) {Help$ = "Upper leg FR2 value. Click-n-drag to change.";}
        else if ((mX > 211) && (mX <= 223)) {Help$ = "Decrement upper leg FR2 value.";}
  
        else if ((mX >= 223) && (mX < 240)) {Help$ = "Set FR2 as FL2.";}
        else if ((mX >= 240) && (mX < 258)) {Help$ = "Link changes to FR2 and FL2 values.";}
        else if ((mX >= 258) && (mX < 275)) {Help$ = "Set FL2 as FR2.";}
  
        else if ((mX >= 275) && (mX < 287)) {Help$ = "Increment upper leg FL2 value.";}
        else if ((mX >= 287) && (mX <= 312)) {Help$ = "Upper leg FL2 value. Click-n-drag to change.";}
        else if ((mX > 312) && (mX <= 324)) {Help$ = "Decrement upper leg FL2 value.";}
  
        else if ((mX >= 400) && (mX < 412)) {Help$ = "Increment upper leg RR2 value.";}
        else if ((mX >= 412) && (mX <= 438)) {Help$ = "Upper leg RR2 value. Click-n-drag to change.";}
        else if ((mX > 438) && (mX <= 450)) {Help$ = "Decrement upper leg RR2 value.";}
  
        else if ((mX >= 450) && (mX < 467)) {Help$ = "Set RR2 as RL2.";}
        else if ((mX >= 467) && (mX < 485)) {Help$ = "Link changes to RR2 and RL2 values.";}
        else if ((mX >= 485) && (mX < 501)) {Help$ = "Set RL2 as RR2.";}
  
        else if ((mX >= 501) && (mX < 513)) {Help$ = "Increment upper leg RL2 value.";}
        else if ((mX >= 513) && (mX <= 539)) {Help$ = "Upper leg RL2 value. Click-n-drag to change.";}
        else if ((mX > 539) && (mX <= 551)) {Help$ = "Decrement upper leg RL2 value.";}
    
             if ((mX >= 174) && (mX < 223)) {rowColPnt = 1;}  // FR2
        else if ((mX >= 275) && (mX < 324)) {rowColPnt = 4;}  // FL2
        else if ((mX >= 400) && (mX < 450)) {rowColPnt = 7;}  // RR2
        else if ((mX >= 501) && (mX < 551)) {rowColPnt = 10;} // RL2
    
     } else if ((mY >= 286) && (mY <= 305)) {
        // Level 2 transposes & Pwr buttons
        if ((mX >= 174) && (mX <= 188)) {Help$ = "Toggle FR2 auto-rate mode.";}
        else if ((mX >= 189) && (mX <= 206)) {Help$ = "Toggle FR2 servo power. Hold to toggle all.";}
        else if ((mX >= 238) && (mX < 258)) {Help$ = "Transpose FR2/Ch3 with FL2/Ch2 angles";}
        else if ((mX >= 275) && (mX <= 289)) {Help$ = "Toggle FL2 auto-rate mode.";}
        else if ((mX >= 290) && (mX <= 308)) {Help$ = "Toggle FL2 servo power. Hold to toggle all.";}
        else if ((mX >= 400) && (mX <= 415)) {Help$ = "Toggle RR2 auto-rate mode.";}
        else if ((mX >= 416) && (mX <= 434)) {Help$ = "Toggle RR2 servo power. Hold to toggle all.";}
        else if ((mX >= 465) && (mX < 485)) {Help$ = "Transpose RR2/Ch12 with RL2/Ch13 angles";}
        else if ((mX >= 500) && (mX <= 516)) {Help$ = "Toggle RL2 auto-rate mode.";}
        else if ((mX >= 517) && (mX <= 535)) {Help$ = "Toggle RL2 servo power. Hold to toggle all.";}
   
     } else if ((mY >= 385) && (mY < 411)) {
        // Level 1 text fields & links
        if ((mX >= 286) && (mX < 298)) {Help$ = "Decrement upper leg FR1 value.";}
        else if ((mX >= 298) && (mX <= 323)) {Help$ = "Lower leg FR1 value. Click-n-drag to change.";}
        else if ((mX > 323) && (mX <= 335)) {Help$ = "Increment upper leg FR1 value.";}
  
        else if ((mX >= 335) && (mX < 352)) {Help$ = "Set FR1 as FL1.";}
        else if ((mX >= 352) && (mX < 370)) {Help$ = "Link changes to FR1 and FL1 values.";}
        else if ((mX >= 370) && (mX < 385)) {Help$ = "Set FL1 as FR1.";}
  
        else if ((mX >= 385) && (mX < 397)) {Help$ = "Decrement upper leg FL1 value.";}
        else if ((mX >= 397) && (mX <= 422)) {Help$ = "Lower leg FL1 value. Click-n-drag to change.";}
        else if ((mX > 422) && (mX <= 434)) {Help$ = "Increment upper leg FL1 value.";}
  
        else if ((mX >= 512) && (mX < 524)) {Help$ = "Decrement upper leg RR1 value.";}
        else if ((mX >= 524) && (mX <= 549)) {Help$ = "Lower leg RR1 value. Click-n-drag to change.";}
        else if ((mX > 549) && (mX <= 561)) {Help$ = "Increment upper leg RR1 value.";}
  
        else if ((mX >= 561) && (mX < 578)) {Help$ = "Set RR1 as RL1.";}
        else if ((mX >= 578) && (mX < 596)) {Help$ = "Link changes to RR1 and RL1 values.";}
        else if ((mX >= 596) && (mX < 612)) {Help$ = "Set RL1 as RR1.";}
  
        else if ((mX >= 612) && (mX < 624)) {Help$ = "Decrement upper leg RL1 value.";}
        else if ((mX >= 624) && (mX <= 649)) {Help$ = "Lower leg RL1 value. Click-n-drag to change.";}
        else if ((mX > 649) && (mX <= 661)) {Help$ = "Increment upper leg RL1 value.";}
    
             if ((mX >= 286) && (mX < 335)) {rowColPnt = 0;}  // FR1
        else if ((mX >= 385) && (mX < 434)) {rowColPnt = 3;}  // FL1
        else if ((mX >= 512) && (mX < 561)) {rowColPnt = 6;}  // RR1
        else if ((mX >= 612) && (mX < 661)) {rowColPnt = 9;} // RL1
   
    } else if ((mY >= 411) && (mY <= 430)) {
        // Level 1 transposes & Pwr buttons
        if ((mX >= 286) && (mX <= 300)) {Help$ = "Toggle FR1 auto-rate mode.";}
        else if ((mX >= 301) && (mX <= 320)) {Help$ = "Toggle FR1 servo power. Hold to toggle all.";}
        else if ((mX >= 350) && (mX < 370)) {Help$ = "Transpose FR1/Ch1 with FL1/Ch0 angles";}
        else if ((mX >= 385) && (mX <= 400)) {Help$ = "Toggle FL1 auto-rate mode.";}
        else if ((mX >= 401) && (mX <= 419)) {Help$ = "Toggle FL1 servo power. Hold to toggle all.";}
        else if ((mX >= 511) && (mX <= 526)) {Help$ = "Toggle RR1 auto-rate mode.";}
        else if ((mX >= 527) && (mX <= 546)) {Help$ = "Click to toggle RR1 servo power. Hold to toggle all.";}
        else if ((mX >= 576) && (mX < 596)) {Help$ = "Transpose RR1/Ch14 with RL1/Ch15 angles";}
        else if ((mX >= 612) && (mX <= 627)) {Help$ = "Toggle RL1 auto-rate mode.";}
        else if ((mX >= 628) && (mX <= 646)) {Help$ = "Toggle RL1 servo power. Hold to toggle all.";}
     }
   }
  
  if ((mY >= 488) && (mY <= 545)) {
    // Multimove controls
         if ((mX >= 222) && (mX < 277)) {Help$ = "FR Multi-move handle control.";}
    else if ((mX >= 324) && (mX < 380)) {Help$ = "FL Multi-move handle control.";}
    else if ((mX >= 450) && (mX < 504)) {Help$ = "RR Multi-move handle control.";}
    else if ((mX >= 551) && (mX < 607)) {Help$ = "RL Multi-move handle control.";}
  }
  
  if ((mY >= 515) && (mY <= 539)) {
    // Degrees/step field
         if ((mX >= 12) && (mX < 31)) {Help$ = "Decrement Degrees/Step value.";}
    else if ((mX >= 31) && (mX < 81)) {Help$ = "Degrees/Step value used in step calcs. Click to reset.";}
    else if ((mX >= 81) && (mX <= 99)) {Help$ = "Increment Degrees/Step value.";}
    // Speed buttons and field
         if ((mX >= 722) && (mX < 740)) {Help$ = "Decrement Speed value.";}
    else if ((mX >= 740) && (mX < 791)) {Help$ = "Speed value (steps/sec). Click to reset to default.";}
    else if ((mX >= 791) && (mX <= 809)) {Help$ = "Increment Speed value.";}
  }
}

// ----------------------------------------------------------------------

public void Help_Message_1() {
  // display help messages for Scrn1
  if (!CmdRow) {
    // not a command so do this
    if ((mY >= 184) && (mY <= 436)) {
      // RoboDog image servo identification.
      // Maps image srvos to current row columns
      if (mY <= 216) {
        // Tier 3 servos
        if ((mX >= 599) && (mX <= 639)) {rowColPnt =  2; return;}  // FR3
        if ((mX >= 641) && (mX <= 677)) {rowColPnt =  8; return;}  // RR3
        if ((mX >= 706) && (mX <= 742)) {rowColPnt =  5; return;}  // FL3
        if ((mX >= 746) && (mX <= 784)) {rowColPnt = 11; return;}  // RL3
      }
      else if ((mY >= 233) && (mY < 289)) {
        // Tier 2 servos
        if ((mX >= 599) && (mX <= 639)) {rowColPnt =  1; return;}  // FR2
        if ((mX >= 641) && (mX <= 677)) {rowColPnt =  7; return;}  // RR2
        if ((mX >= 706) && (mX <= 742)) {rowColPnt =  4; return;}  // FL2
        if ((mX >= 746) && (mX <= 784)) {rowColPnt = 10; return;}  // RL2
      }
      else if ((mY >= 289) && (mY <= 335)) {
        // Tier 1 servos
        if ((mX >= 599) && (mX <= 639)) {rowColPnt =  0; return;}  // FR1
        if ((mX >= 641) && (mX <= 677)) {rowColPnt =  6; return;}  // RR1
        if ((mX >= 706) && (mX <= 742)) {rowColPnt =  3; return;}  // FL1
        if ((mX >= 746) && (mX <= 784)) {rowColPnt =  9; return;}  // RL1
      }
    }
  }
  
  if ((mY >= 515) && (mY <= 539)) {
    // Degrees/step field
         if ((mX >= 12) && (mX < 31)) {Help$ = "Decrement Degrees/Step value."; return;}
    else if ((mX >= 31) && (mX < 81)) {Help$ = "Degrees/Step value used in step calcs. Click to reset."; return;}
    else if ((mX >= 81) && (mX <= 99)) {Help$ = "Increment Degrees/Step value."; return;}
    // Speed buttons and field
         if ((mX >= 722) && (mX < 740)) {Help$ = "Decrement Speed value."; return;}
    else if ((mX >= 740) && (mX < 791)) {Help$ = "Speed value (steps/sec). Click to reset to default."; return;}
    else if ((mX >= 791) && (mX <= 809)) {Help$ = "Increment Speed value."; return;}
  }
}

// ----------------------------------------------------------------------

public void Help_Message_2() {
  // display help messages for Scrn2
  if (!CmdRow) {
    // not a command so do this
    if ((mY >= 99) && (mY <= 124)) {
      // RoboDog simulator image servo identification.
      // Maps image srvos to current row columns
      // Tier 3 servos
      if ((mX >= 272) && (mX <= 298)) {rowColPnt =  2; return;}  // FR3
      if ((mX >= 463) && (mX <= 489)) {rowColPnt =  8; return;}  // RR3
      if ((mX >= 325) && (mX <= 350)) {rowColPnt =  5; return;}  // FL3
      if ((mX >= 516) && (mX <= 542)) {rowColPnt = 11; return;}  // RL3
    }
    else if ((mY >= 171) && (mY < 196)) {
      // Tier 2 servos
      if ((mX >= 272) && (mX <= 298)) {rowColPnt =  1; return;}  // FR2
      if ((mX >= 463) && (mX <= 489)) {rowColPnt =  7; return;}  // RR2
      if ((mX >= 325) && (mX <= 350)) {rowColPnt =  4; return;}  // FL2
      if ((mX >= 516) && (mX <= 542)) {rowColPnt = 10; return;}  // RL2
    }
    // Tier 1 servos, areas move with leg drawing process
    if ((mY >= (HY0 - 11)) && (mY <= (HY0 + 11))) {  // FR1
      if ((mX >= (HX0 - 11)) && (mX <= (HX0 + 11))) {rowColPnt =  0; return;}
    }
    if ((mY >= (HY1 - 11)) && (mY <= (HY1 + 11))) {  // RR1
      if ((mX >= (HX1 - 11)) && (mX <= (HX1 + 11))) {rowColPnt =  3; return;}
    }
    if ((mY >= (HY2 - 11)) && (mY <= (HY2 + 11))) {  // FL1
      if ((mX >= (HX2 - 11)) && (mX <= (HX2 + 11))) {rowColPnt =  6; return;}
    }
    if ((mY >= (HY3 - 11)) && (mY <= (HY3 + 11))) {  // RL1
      if ((mX >= (HX3 - 11)) && (mX <= (HX3 + 11))) {rowColPnt =  9; return;}
    }
    drawFlag = 1;  // removes highlights when mouse moves off areas
  }
  
  // Calculation fields and buttons
  if ((mY >= 373) && (mY <= 532)) {
    if ((mX >= 103) && (mX <= 153)) {
      // 1st field column
           if ((mY >= 404) && (mY <= 431)) {Help$ = "Left legs max upper T2 servo angle.";}
      else if ((mY >= 436) && (mY <= 463)) {Help$ = "Calculated left steps value.";}
      else if ((mY >= 470) && (mY <= 497)) {Help$ = "Left legs max lower T1 servo angle.";}
      else if ((mY >= 503) && (mY <= 528)) {Help$ = "Calculated left side height.";}
    }
    if ((mX >= 272) && (mX <= 322)) {
      // 2nd field column
           if ((mY >= 404) && (mY <= 431)) {Help$ = "Left legs min upper T2 servo angle.";}
      else if ((mY >= 436) && (mY <= 463)) {Help$ = "Calculated upper T2 mid-point angle value.";}
      else if ((mY >= 470) && (mY <= 497)) {Help$ = "Left legs lower T1 servo lift angle.";}
    }
    
    if ((mX >= 352) && (mX <= 473)) {
      // centre controls - Link,transpose,Speed,Calc
      if (mY <= 407) {
        // Link button region
             if (mX < 377) {Help$ = "Copy right leg values to left leg.";}
        else if ((mX >= 377) && (mX <= 446)) {Help$ = "Link left and right leg values.";}
        else if (mX > 446) {Help$ = "Copy left leg values to right leg.";}
      }
      else if (mY <= 433) {
        // Transpose button
        if ((mX >= 399) && (mX <= 424)) {Help$ = "Transpose right and left leg settings.";}
      }
      if ((mY >= 454) && (mY <= 480)) {
        // Speed settings
             if ((mX >= 369) && (mX < 386)) {Help$ = "Decrement Speed value.";}
        else if ((mX >= 386) && (mX <= 438)) {Help$ = "Speed value.";}
        else if ((mX > 438) && (mX <= 455)) {Help$ = "Increment Speed value.";}
      }
      if ((mY >= 498) && (mY <= 532)) {
        // Calculate button
        if ((mX >= 377) && (mX < 447)) {Help$ = "Calculate a move table using current settings.";}
      }
    }
    
    if ((mX >= 567) && (mX <= 616)) {
     // 4th field column
          if ((mY >= 404) && (mY <= 431)) {Help$ = "Right legs max upper T2 servo angle.";}
     else if ((mY >= 436) && (mY <= 463)) {Help$ = "Calculated right steps value.";}
     else if ((mY >= 470) && (mY <= 497)) {Help$ = "Right legs max lower T1 servo angle.";}
     else if ((mY >= 503) && (mY <= 528)) {Help$ = "Calculated right side height.";}
   }
    if ((mX >= 737) && (mX <= 785)) {
      // 5th field column
           if ((mY >= 404) && (mY <= 431)) {Help$ = "Right legs min upper T2 servo angle.";}
      else if ((mY >= 436) && (mY <= 463)) {Help$ = "Calculated upper T2 mid-point angle value.";}
      else if ((mY >= 470) && (mY <= 497)) {Help$ = "Calculated right legs lower T1 servo lift angle.";}
    }
  }
}

// ----------------------------------------------------------------------
  public void settings() {  size(824,760); }
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "Proc_Move_Trainer_02" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
